/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author Alex Hogen (alex@edt.com)
 * @copyright Copyright (c) 2023 Engineering Design Team, Inc.
 */

#define pr_fmt(fmt) "%s:%s(): " fmt, KBUILD_MODNAME, __func__

#include "chardev_intf.h"

#include <linux/types.h>
#include <linux/kconfig.h>
#if IS_ENABLED(CONFIG_COMPAT)
#include <linux/compat.h>
#endif

#define CDEV_MINOR_BASE         0
#define CDEV_UNIT_NAME_LEN      24
#define CDEV_UNIT_FILE_NAME_LEN (2 * CDEV_UNIT_NAME_LEN)

#if defined(PCD)
static const char CDEV_PREFIX[] = "pcd";
#elif defined(PDV)
static const char CDEV_PREFIX[] = "pdv";
#else
#error "EDT driver type undefined."
#endif /* PCD / PDV */

/** Character device class, used by all chardevs created by this module.
 * @see edt_chardev_init() edt_chardev_cleanup()
 */
static struct class* edt_cdev_class_ = NULL;

static DEFINE_MUTEX(unit_mutex);
static LIST_HEAD(unit_list);

/*!
 * @brief Character class structure for each unit created.
 */
struct edt_cdev_unit {
  /*! Linux linked-list pointer. */
  struct list_head list_entry;

  /*! Private data pointer. _Probably_ the channel 0  @ref Edt_Dev this cdev is
   * intended to interface with. */
  void* private_data;

  /*! Character device class created. */
  struct cdev* cdev;

  /*! Cdev unit name */
  char name[CDEV_UNIT_NAME_LEN];

  /*! (Dynamic) major allocated for this class. */
  int major;

  /*! First/lowest minor allocated for this class. */
  int lowest_minor;

  /*! Number of nodes created. Typically to represent different DMA channels of
   * an EDT PCI device.
   */
  int num_nodes;
};

/*!
 * @brief A UEVENT callback to set permissions on character devices via UDEV
 * variables.
 *
 * @param dev
 *     Device to act on.
 * @param env
 *     Kernel uevent environment object.
 * @return
 *     Zero if successful.
 */
static int edt_cdev_uevent(const struct device* dev, struct kobj_uevent_env* env);

/* cdev fops open callback */
static int _cdev_open(struct inode* inode, struct file* filp);
/* cdev fops release callback */
static int _cdev_release(struct inode* inode, struct file* filp);
/* cdev fops read callback */
static ssize_t _cdev_read(struct file* filp, char* user_buf, size_t count, loff_t* seek);
/* cdev fops write callback */
static ssize_t _cdev_write(struct file* filp, const char* user_buf, size_t count, loff_t* offset);
/* cdev fops ioctl callback */
static long _cdev_unlocked_ioctl(struct file* filp, unsigned int cmd, unsigned long arg);
/* cdev fops mmap callback */
static int _cdev_mmap(struct file* filp, struct vm_area_struct* vma);
#if IS_ENABLED(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL) && HAVE_COMPAT_IOCTL
/* cdev fops ioctl32 callback */
static long _cdev_ioctl32(struct file* filep, unsigned int cmd, unsigned long arg);
#endif /* HAVE_COMPAT_IOCTL */

static int _cdev_find_unit_by_inode(struct inode* inode, Edt_Dev** private_data, int* index);

/*-----------------------------------------------------------------------------
 * Module init/cleanup
 *---------------------------------------------------------------------------*/

int edt_chardev_init(void)
{
  if (NULL == edt_cdev_class_)
  {
    pr_info("Creating chardev class\n");

// Linux v6.4 (commit dcfbb67e48) made this a real function and removed the
// first param: "owner". Backported kernels, especially RHEL, may use the
// newer function in older (e.g. 5.14) kernel versions.
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0)) || (!defined(class_create))
    edt_cdev_class_ = class_create(KBUILD_MODNAME);
#else
    edt_cdev_class_ = class_create(THIS_MODULE, KBUILD_MODNAME);
#endif
    if (IS_ERR(edt_cdev_class_))
    {
      pr_warn("Failed to create device class\n");
      return -1;
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0) && __GNUC__
#pragma GCC diagnostic push
/* RHEL-8 and 9 seems to have backported kernel 6.2 changes into 5.x or even 4.x
 * kernels. Lower the potential GCC error to a warning for kernels earlier than
 * 6.2.
 *
 * https://www.rpmfind.net/linux/RPM/centos/8-stream/baseos/x86_64/Packages/kernel-debug-core-4.18.0-522.el8.x86_64.html
 * https://bugzilla.redhat.com/show_bug.cgi?id=2178301
 * https://access.redhat.com/errata/RHSA-2023:7077
 */
#if __GNUC__ >= 5
#pragma GCC diagnostic warning "-Wincompatible-pointer-types"
#else
#pragma GCC diagnostic warning "-w"
#endif
#endif /* linux < v6.2.0 */

    edt_cdev_class_->dev_uevent = edt_cdev_uevent;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0) && __GNUC__
#pragma GCC diagnostic pop
#endif
  }

  return 0;
}

static int edt_cdev_uevent(const struct device* dev, struct kobj_uevent_env* env)
{
  (void)dev;

  // mode: 0666
  // Userspace apps need write permissions to be able to call mmap.
  return add_uevent_var(env,
                        "DEVMODE=%#o",
                        (S_IRUSR | S_IWUSR) | (S_IRGRP | S_IWGRP) | (S_IROTH | S_IWOTH));
}

int edt_chardev_cleanup(void)
{
  if (edt_cdev_class_)
  {
    pr_debug("Removing chardev class\n");
    class_destroy(edt_cdev_class_);
    edt_cdev_class_ = NULL;
  }

  return 0;
}

/*-----------------------------------------------------------------------------
 * Create/destroy cdev interfaces for each PCI device
 *---------------------------------------------------------------------------*/

/*
 * File operations (fops) for each character class.
 */
static const struct file_operations EDT_CDEV_FOPS = {.owner = THIS_MODULE,
                                                     .read = _cdev_read,
                                                     .write = _cdev_write,
                                                     .unlocked_ioctl = _cdev_unlocked_ioctl,
#if IS_ENABLED(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL) && HAVE_COMPAT_IOCTL
                                                     .compat_ioctl = _cdev_ioctl32,
#endif
                                                     .open = _cdev_open,
                                                     .release = _cdev_release,
                                                     .mmap = _cdev_mmap};

int edt_chardev_create_interfaces(const Edt_Dev* const edt_p)
{
  __label__ unroll_device_create;
  __label__ free_cdev_region;
  __label__ fail_obtain;
  int retval = 0;
  int node_idx = 0;
  struct edt_cdev_unit* unit = NULL;
  dev_t devt = 0;

  pci_dbg(edt_p->os.m_pci_dev, "Creating interfaces\n");

  unit = kzalloc(sizeof(*unit), GFP_KERNEL);
  if (!unit)
    return -ENOMEM;

  memset(unit, 0, sizeof(*unit));

  mutex_lock(&unit_mutex);

  snprintf(unit->name, CDEV_UNIT_NAME_LEN, "%s%d", CDEV_PREFIX, edt_p->m_Unitno);

  /* Allocate major */
  if (edt_p->m_chancnt < 0 || edt_p->m_chancnt >= MAX_MINORS)
  {
    pci_err(edt_p->os.m_pci_dev,
            "Channel count (%d) out of range. Cannot create interfaces.\n",
            edt_p->m_chancnt);
    retval = -ERANGE;
    goto fail_obtain;
  }

  retval = alloc_chrdev_region(&devt, CDEV_MINOR_BASE, edt_p->m_chancnt, unit->name);
  if (retval)
  {
    pci_warn(edt_p->os.m_pci_dev, "Failed to allocate major/minors\n");
    goto fail_obtain;
  }

  unit->major = MAJOR(devt);
  unit->lowest_minor = MINOR(devt);
  unit->num_nodes = edt_p->m_chancnt;
  unit->private_data = (void*)edt_p;

  unit->cdev = cdev_alloc();
  if (!unit->cdev)
  {
    pci_err(edt_p->os.m_pci_dev, "Could not alloc cdev\n");
    retval = -ENOMEM;
    goto free_cdev_region;
  }

  cdev_init(unit->cdev, &EDT_CDEV_FOPS);

  pci_dbg(edt_p->os.m_pci_dev, "cdev add\n");
  retval = cdev_add(unit->cdev, MKDEV(unit->major, unit->lowest_minor), unit->num_nodes);
  if (retval)
  {
    pci_err(edt_p->os.m_pci_dev, "Failed to add cdev\n");
    /* kobject_put() is normally done by cdev_del() */
    kobject_put(&unit->cdev->kobj);
    goto free_cdev_region;
  }

  /* Create "/dev" files */
  for (node_idx = 0; node_idx < unit->num_nodes; node_idx++)
  {
    char dev_fname[CDEV_UNIT_FILE_NAME_LEN] = "";
    struct device* device;

    snprintf(dev_fname, sizeof(dev_fname), "%s_%d", unit->name, (node_idx + unit->lowest_minor));

    device = device_create(edt_cdev_class_,
                           NULL,
                           MKDEV(unit->major, (node_idx + unit->lowest_minor)),
                           NULL,
                           "%s",
                           dev_fname);
    if (IS_ERR(device))
    {
      pci_err(edt_p->os.m_pci_dev, "Failed to create device '%s'. Aborting.\n", dev_fname);
      retval = -ENODEV;
      goto unroll_device_create;
    }
  }

  list_add_tail(&unit->list_entry, &unit_list);
  pci_info(edt_p->os.m_pci_dev, "Created %d interface devices.\n", unit->num_nodes);

  mutex_unlock(&unit_mutex);

  return retval;

unroll_device_create:
  for (node_idx--; node_idx >= 0; node_idx--)
  {
    device_destroy(edt_cdev_class_, MKDEV(unit->major, (node_idx + unit->lowest_minor)));
  }
  cdev_del(unit->cdev);

free_cdev_region:
  unregister_chrdev_region(MKDEV(unit->major, unit->lowest_minor), unit->num_nodes);

fail_obtain:
  mutex_unlock(&unit_mutex);

  kfree(unit);

  return retval;
}

int edt_chardev_destroy_interfaces(const Edt_Dev* const edt_p)
{
  struct edt_cdev_unit* unit = NULL;
  bool found = false;
  int minor = 0;

  mutex_lock(&unit_mutex);

  list_for_each_entry(unit, &unit_list, list_entry)
  {
    if (unit->private_data == edt_p)
    {
      found = true;
      break;
    }
  }

  if (!found)
  {
    pci_err(edt_p->os.m_pci_dev, "Could not find cdev unit to destroy.\n");
    mutex_unlock(&unit_mutex);
    return -ENODEV;
  }

  for (minor = unit->lowest_minor; minor < (unit->lowest_minor + unit->num_nodes); minor++)
  {
    device_destroy(edt_cdev_class_, MKDEV(unit->major, minor));
  }

  cdev_del(unit->cdev);
  unit->cdev = NULL;

  unregister_chrdev_region(MKDEV(unit->major, unit->lowest_minor), unit->num_nodes);

  pci_dbg(edt_p->os.m_pci_dev, "Removed %d interface devices.\n", unit->num_nodes);

  list_del(&unit->list_entry);
  kfree(unit);
  unit = NULL;

  mutex_unlock(&unit_mutex);
  return 0;
}

/*-----------------------------------------------------------------------------
 * File operations
 *---------------------------------------------------------------------------*/

static int _cdev_find_unit_by_inode(struct inode* inode, Edt_Dev** private_data, int* index)
{
  const int major = imajor(inode);
  const int minor = iminor(inode);
  struct edt_cdev_unit* unit = NULL;
  bool found = false;

  mutex_lock(&unit_mutex);

  list_for_each_entry(unit, &unit_list, list_entry)
  {
    if ((unit->major == major) && (minor >= unit->lowest_minor) &&
        (minor < (unit->lowest_minor + unit->num_nodes)))
    {
      found = true;
      break;
    }
  }

  if (!found)
  {
    mutex_unlock(&unit_mutex);
    return -ENODEV;
  }

  *private_data = unit->private_data;
  *index = (minor - unit->lowest_minor);

  mutex_unlock(&unit_mutex);
  return 0;
}

static int _cdev_open(struct inode* inode, struct file* file)
{
  int channel_num = -1;
  Edt_Dev* edt_p = NULL;
  Edt_Dev* edt_ch_p = NULL;
  int retval;

  retval = _cdev_find_unit_by_inode(inode, &edt_p, &channel_num);
  if (retval)
    return retval;

  if (!edt_p)
    return -ENODEV;

  edt_ch_p = edt_p->chan_p[channel_num];

  pr_debug("Opening cdev intf: %s unit %d channel %d  (mode: 0x%X)\n",
           CDEV_PREFIX,
           edt_ch_p->m_Unitno,
           edt_ch_p->m_DmaChannel,
           file->f_mode);

  /* use filp->private_data to point to the device data */
  file->private_data = edt_ch_p;

  edt_ch_p->m_IsOpen++;

  return 0;
}

static int _cdev_release(struct inode* inode, struct file* file)
{
  bool do_unmap = false;
  bool do_unlock = false;
  Edt_Dev* edt_p = file->private_data;
  int retval;

  if (!edt_p)
  {
    int channel_num = -1;
    pr_err("NULL data pointer!\n");

    retval = _cdev_find_unit_by_inode(inode, &edt_p, &channel_num);
    if (retval)
      return retval;
    edt_p = edt_p->chan_p[channel_num];
  }

  pr_debug("Release cdev %s unit %d channel %d\n",
           CDEV_PREFIX,
           edt_p->m_Unitno,
           edt_p->m_DmaChannel);

  pci_dbg(edt_p->os.m_pci_dev,
          "current pid: %d, dma pid: %d, numbufs: %d\n",
          edt_get_current_pid(),
          edt_p->dma_pid,
          edt_p->m_Numbufs);

  /* if we did the last DMA, shut it down */
  do_unmap = (edt_get_current_pid() == edt_p->dma_pid);
  do_unlock = (edt_get_current_pid() == edt_p->os.register_locker);

  if (do_unmap)
    edt_clear_dma(edt_p);

  if (do_unlock)
  {
    pci_dbg(edt_p->os.m_pci_dev, "Unlocking register lock\n");

    edt_unlock_registers(edt_p);
  }

  pci_dbg(edt_p->os.m_pci_dev,
          "abort: %d, done: %d, todo: %d, isopen: %d\n",
          edt_p->m_Abort,
          edt_p->m_Done,
          edt_p->m_Todo,
          --edt_p->m_IsOpen);

  return 0;
}

static ssize_t _cdev_read(struct file* file,
                          char* UNUSED_PARAM(user_buf),
                          size_t UNUSED_PARAM(count),
                          loff_t* UNUSED_PARAM(seek))
{
  if (!file->private_data)
    return -ENODEV;
  return -ENOSYS; /* Function not implemented */
}

static ssize_t _cdev_write(struct file* file,
                           const char* UNUSED_PARAM(user_buf),
                           size_t UNUSED_PARAM(count),
                           loff_t* UNUSED_PARAM(offset))
{
  if (!file->private_data)
    return -ENODEV;

  return -ENOSYS; /* Function not implemented */
}

static long _cdev_unlocked_ioctl(struct file* file, unsigned int cmd, unsigned long arg)
{
  if (!file->private_data)
    return -ENODEV;

  return edt_private_ioctl(file->private_data, cmd, arg, 0);
}

#if IS_ENABLED(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL) && HAVE_COMPAT_IOCTL
static long _cdev_ioctl32(struct file* filep, unsigned int cmd, unsigned long arg)
{
  void* edt_p = filep->private_data;

  return edt_private_ioctl(edt_p, cmd, (unsigned long)compat_ptr(arg), 1);
}
#endif /* HAVE_COMPAT_IOCTL */

static int _cdev_mmap(struct file* file, struct vm_area_struct* vma)
{
  Edt_Dev* edt_p = file->private_data;

  if (!edt_p)
    return -ENODEV;

  pci_dbg(edt_p->os.m_pci_dev,
          "mmap() -- unit %d, ch %d, pid %d\n",
          edt_p->m_Unitno,
          edt_p->m_DmaChannel,
          edt_get_current_pid());

  return edt_private_mmap(edt_p, vma);
}
