/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @file
 * @author Alex Hogen (alex@edt.com)
 * @copyright Copyright (c) 2023 Engineering Design Team, Inc.
 * @brief Functions to setup and manage character device (chardev) interfaces
 * associated with a PCI(e) device.
 */

#ifndef CHARDEV_INTF_H
#define CHARDEV_INTF_H

#include "edtdrvlnx.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <linux/device.h>
#include <linux/cdev.h>

/*!
 * @brief Create chardev region and device class, used by all char device
 * interfaces to this PCI device.
 *
 * @note Should be called before @ref edt_pcidev_create_interfaces()
 *
 * @returns
 *     Zero if successful, or negative errno code.
 */
int edt_chardev_init(void);

/*!
 * @brief Cleanup after edt_chardev_init().
 *
 * @pre @ref edt_pcidev_destroy_interfaces() should be called on each interface
 *     first, before calling this, so the cdev class remains valid.
 *
 * @returns
 *     Zero if successful, or negative errno code.
 */
int edt_chardev_cleanup(void);

/*!
 * @brief Create character device interfaces to this PCI device.
 *
 * If this is a PCD driver, this will create device files like:
 *
 *     /dev/pcd0_0
 *     /dev/pcd0_1
 *     /dev/pcd0_2
 *
 * And likewise for PDV:
 *
 *     /dev/pdv0_0
 *     /dev/pdv0_1
 *     /dev/pdv0_2
 *
 * ...where the first number is the EDT device's "unit" number, and the second
 * is the the DMA channel index.
 *
 * @param edt_p
 *     EDT device structure. All channel structs must be initialized before
 *     calling this function.
 * @return
 *     Zero if successful, or negative errno code.
 */
int edt_chardev_create_interfaces(const Edt_Dev* edt_p);

/*!
 * @brief Remove all character device files previously created by
 *     @ref edt_chardev_create_interfaces()
 *
 * @param edt_p
 *     EDT device structure.
 * @return
 *     Zero if successful, or negative errno code.
 */
int edt_chardev_destroy_interfaces(const Edt_Dev* edt_p);

#ifdef __cplusplus
}
#endif

#endif /* CHARDEV_INTF_H */
