/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 * @brief Local include files and declarations for Edt_Dev driver.
 *
 * @see edt_common.h
 */

#include "edt_common.h"
#include "edt_trace.h"

#ifdef __linux__
#include <linux/bug.h>
#endif

/*!
 * Initial value for number of rbufs
 */
int edt_default_rbufs_from_type(int id)
{
#if defined(PCD)
  if (id == PE4CDA16_ID)
  {
    return 64;
  }
#else
  (void)id;
#endif

  return 256;
}

int edt_safe_bufnum(Edt_Dev* edt_p, int bufnum)
{
  if (edt_p->m_Numbufs)
    return bufnum % edt_p->m_Numbufs;
  else
    return 0;
}

int edt_test_intr(Edt_Dev* edt_p)
{
  u_int cfg = edt_get_intcfg(edt_p);
  u_int stat = edt_get(edt_p, EDT_DMA_STATUS);

  // TODO: Add pci_enabled_msi(edt_p->os.m_pci_dev) check somewhere, since MSI
  // irqs are not shared between devices like legacy pin interrupts. Not sure
  // where to check though.

  if ((cfg & EDT_ICFG_PCI_EN_INTR) == 0)
  {
    EDTPRT(edt_p, DBG_ISR, TRL_VERBOSE, ("Reject cfg = %08x\n", cfg));
    return 0;
  }

  if ((stat & (EDT_DMA_INTR | EDT_RMT_INTR | EDT_PCI_INTR)) == 0)
  {
    EDTPRT(edt_p, DBG_ISR, TRL_VERBOSE, ("Reject stat = %08x\n", stat));
    return 0;
  }
  edt_p->m_Ints++;
  edt_p->m_savestat = stat;

#ifdef DISABLE_INTS_IN_ISR
  edt_disable_interrupts(edt_p);
#endif

  return (1);
}

static void edt_set_rbuf_endtime(Edt_Dev* edt_p, rbuf* p)
{
  EdtTraceTimeStamp(edt_p, &p->m_end_timeval, 1);
}

static void edt_increment_done(Edt_Dev* edt_p)
{
  rbuf* p;
  int foundbuf;
  int foundend;
  int foundsize;
  int another_done = 1;
  int another_loop = 0;

  while (another_done)
  {
    p = &edt_p->m_Rbufs[edt_p->m_Curbuf];

    edt_p->m_Done++;

    p->m_Active = 0;

    edt_p->m_hadcnt = 0;

    if (edt_p->m_DoEndDma)
    {
      EDTPRT(edt_p,
             DBG_DPC,
             TRL_INFO,
             ("DPC: doing end dma %x %x", edt_p->m_EndDmaDesc, edt_p->m_EndDmaVal));
      edt_set(edt_p, edt_p->m_EndDmaDesc, edt_p->m_EndDmaVal);
      EdtTrace(edt_p, 0xc0ff0003);
      EdtTrace(edt_p, edt_p->m_EndDmaDesc);
      EdtTrace(edt_p, edt_p->m_EndDmaVal);
      if (edt_p->m_EndAction == EDT_ACT_ONCE)
        edt_p->m_DoEndDma = 0;
    }

    EdtTrace(edt_p, TR_DONE);
    EdtTrace(edt_p, edt_p->m_Done);

    /* clearing here will end up with clear even if others loaded */
    /* chet changed while at par */
    /* edt_p->m_DmaActive = 0; */

    p = &edt_p->m_Rbufs[edt_p->m_Curbuf];

    edt_set_rbuf_endtime(edt_p, p);

#if defined(PDV)
    if (EDT_DEVID_IS_DVCL(edt_p->devid))
    {
      p->m_LinesXferred = edt_get(edt_p, PDV_CL_LINESPERFRAME);
      EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("DPC: C-Link Read lines as %d\n", p->m_LinesXferred));
      another_done = 0;
    }
    else
#endif
    {
      /* 0 implies all lines in image xferred */
      p->m_LinesXferred = 0;
    }

    if (edt_p->m_MultiDone)
    {
      foundbuf = TransferCountDebug(edt_p, edt_p->m_Curbuf, &foundend, &foundsize);
      if (foundbuf >= 0 && foundbuf == (int)edt_p->m_Curbuf)
        another_done = 0;

      if (++edt_p->m_Curbuf >= edt_p->m_Numbufs)
        edt_p->m_Curbuf = 0;

      if (another_done && foundbuf >= 0 && foundbuf == (int)edt_p->m_Curbuf && !foundend)
        another_done = 0;

      if (foundbuf < 0 || another_loop++ > (int)edt_p->m_Numbufs)
        another_done = 0;
    }
    else
    {
      if (++edt_p->m_Curbuf >= edt_p->m_Numbufs)
        edt_p->m_Curbuf = 0;

      another_done = 0;
    }

    EdtTrace(edt_p, 0xbeadbead);
    EdtTraceTimeStamp(edt_p, NULL, 1);

    edt_broadcast_event(edt_p, EDT_EODMA_EVENT);

    edt_p->m_needs_wakeup = 1;

    EDTPRT(edt_p,
           DBG_DPC,
           TRL_INFO,
           ("Setting %p m_needs_wakeup -> %d\n", edt_p, edt_p->m_needs_wakeup));

    edt_p->m_xfers++;
  }
}

/*
 * on NT this is done as deferred procedure return nonzero if need wakeup
 */

static int edtdrv_check_for_devint(Edt_Dev* edt_p, int* had_fv_p)
{
  int HadDevInt = 0;

  u_int stat = edt_p->m_savestat;
  u_int have_devint = stat & (EDT_RMT_INTR | EDT_TEMP_INTR);

  if (have_devint)
  {
    u_int cfg = edt_get(edt_p, EDT_DMA_CFG);

    if ((cfg & X_DONE) == 0)
    {
      EDTPRT(edt_p,
             DBG_DPC,
             TRL_ERROR,
             ("DPC: RMT_INT when we don't want it; disabling remote interrupt\n"));
      edt_refresh_intcfg(edt_p);
      edt_disable_interrupts(edt_p);
      edt_disable_remote_interrupt(edt_p);
      EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("Disable from check_devint\n"));

      return -1;
    }
  }

  if (have_devint)
  {
#if defined(PDV)

    int channel = 0;
    Edt_Dev* chan0_p = edt_p->chan_p[0];

    for (channel = 0; channel < chan0_p->m_chancnt; channel++)
    {
      edt_p = chan0_p->chan_p[channel];
      {
        HadDevInt = edt_DeviceInt(edt_p);
        if (HadDevInt)
        {
          if (edt_p->m_HasFvalDone)
          {
            *had_fv_p = channel;
          }
          EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("DPC: got devint\n"));
          break;
        }
      }
    }

#elif defined(PCD)
    (void)had_fv_p; /* UNUSED */

    HadDevInt = edt_DeviceInt(edt_p);

#endif /* PDV / PCD */
  }

  return HadDevInt;
}

Edt_Dev* edt_DpcForIsr(Edt_Dev* edt_p)
{
  Edt_Dev* chan0_p = NULL;

  int channel = 0;
  u_int cmd = 0;
  u_int command = 0;

  int HadDevInt = 0;
  int had_abort = 0;
  int haddone = 0;
  int had_fval_done = -1;

  if (NULL == edt_p)
  {
    return NULL;
  }

  edt_p->m_needs_wakeup = 0;

  HadDevInt = edtdrv_check_for_devint(edt_p, &had_fval_done);

  if (HadDevInt < 0)
  {
    return NULL;
  }

  chan0_p = edt_p->chan_p[0];

  for (channel = 0; channel < chan0_p->m_chancnt; channel++)
  {
    edt_p = chan0_p->chan_p[channel];
    cmd = edt_get(edt_p, EDT_SG_NXT_CNT);
    edt_p->m_LastCmd = cmd;

    command = cmd & (EDT_EN_MN_DONE | EDT_DMA_MEM_RD | EDT_BURST_EN | 0xffff);

    if ((cmd & EDT_PG_INT) || ((cmd & EDT_DMA_RDY) && (cmd & EDT_EN_RDY)) ||
        (cmd & EDT_EN_SG_DONE) || (edt_p->m_Abort))
    {
      if (edt_p->m_Abort)
      {
        edt_p->m_Abort = 0;
        command &= ~(EDT_EN_MN_DONE | EDT_DMA_ABORT);
        EDTPRT(edt_p,
               DBG_DPC,
               TRL_INFO,
               ("DPC: abort chan %d clearing MN_DONE %x in %x\n",
                edt_p->m_DmaChannel,
                command,
                EDT_EN_MN_DONE));
      }

      edt_set_sg_cnt(edt_p, command);

      break;
    }
  }

  if (channel == chan0_p->m_chancnt)
  {
    if (HadDevInt)
    {
      /* set edt_p to correct channel for fval */
      if (had_fval_done >= 0)
        edt_p = chan0_p->chan_p[had_fval_done];
      else
        edt_p = chan0_p;
    }
    else
    {
      EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("DPC: Nothing to do\n"));
      return NULL;
    }
  }

  edt_p->m_needs_wakeup = 0;

  if (edt_p->m_Abort)
  {
    EDTPRT(edt_p,
           DBG_DPC,
           TRL_INFO,
           ("DPC: Abort %d HadCnt %d HadDev %d\n", edt_p->m_Abort, edt_p->m_hadcnt, HadDevInt));
  }

  cmd = edt_p->m_LastCmd;

  if ((cmd & EDT_PG_INT) || edt_p->m_Abort || edt_p->m_hadcnt || edt_p->m_HasFvalDone)
  {
    cmd &= (EDT_DMA_MEM_RD | EDT_BURST_EN | 0xffff);

    edt_set_sg_cnt(edt_p, cmd);

    if (edt_p->m_Abort)
    {
      EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("DPC: aborting buf %d", edt_p->m_Curbuf));
      had_abort = 1;
      edt_p->m_Abort = 0;
    }
    else
    {
      cmd |= EDT_EN_MN_DONE;
      edt_set_sg_cnt(edt_p, cmd);
    }

    if (edt_p->m_HasFvalDone)
    {
      u_int intcfg;

      u_int donedbg;

      EdtTrace(edt_p, 0xbacabaca);
      donedbg = ((edt_p->m_Done & 0xff) << 24) | ((edt_p->m_Loaded & 0xff) << 16) |
                ((edt_p->m_Todo & 0xff) << 8) | (edt_p->m_Curbuf & 0xff);

      EdtTrace(edt_p, donedbg);

      EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("FV: Done calling abortdma\n"));

      edt_abortdma(edt_p);
      intcfg = edt_get(edt_p, EDT_DMA_INTCFG);
      intcfg &= ~EDT_RFIFO_ENB;
      edt_set(edt_p, EDT_DMA_INTCFG_ONLY, (intcfg >> 8) & 0xff);
      /* why do we need to do this again? */
      edt_set_sg_cnt(edt_p, cmd | EDT_DMA_ABORT);
      edt_set_sg_cnt(edt_p, cmd);

      edt_p->m_HasFvalDone = 0;
      had_fval_done = 1;

      intcfg = edt_get(edt_p, PDV_CFG);
      intcfg &= ~PDV_FIFO_RESET;

      edt_set(edt_p, PDV_CFG, intcfg | PDV_FIFO_RESET);
      /* will this screw up other channel ? */
      edt_set(edt_p, PDV_CMD, PDV_RESET_INTFC);

      edt_set(edt_p, PDV_CFG, intcfg);

      intcfg = edt_get(edt_p, EDT_DMA_INTCFG);
      intcfg |= EDT_RFIFO_ENB;

      edt_set(edt_p, EDT_DMA_INTCFG_ONLY, (intcfg >> 8) & 0xff);

      // always reset nxt, loaded, and curbuf
      if (edt_p->m_Loaded != edt_p->m_Done + 1)
      {
        edt_p->m_Loaded = edt_p->m_Done + 1;
        edt_p->m_Nxtbuf = edt_safe_bufnum(edt_p, edt_p->m_Loaded);
        edt_p->m_Curbuf = edt_safe_bufnum(edt_p, edt_p->m_Done);
      }

      edt_p->m_CurSg = 0;
      EDTPRT(edt_p,
             DBG_DPC,
             TRL_INFO,
             ("FVal Done Done = %d Loaded = %d Todo = %d\n",
              edt_p->m_Done,
              edt_p->m_Loaded,
              edt_p->m_Todo));

      edt_clear(edt_p, PDV_UTIL2, PDV_INT_ENFV);

      donedbg = ((edt_p->m_Done & 0xff) << 24) | ((edt_p->m_Loaded & 0xff) << 16) |
                ((edt_p->m_Todo & 0xff) << 8) | (edt_p->m_Curbuf & 0xff);

      EdtTrace(edt_p, donedbg);
      edt_set_sg_cnt(edt_p, cmd | EDT_DMA_ABORT);
      edt_set_sg_cnt(edt_p, cmd);
      edt_p->m_Abort = 0;
    }
    else
    {
      EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("DPC: Apparent Done\n"));
    }

    edt_increment_done(edt_p);
    haddone = 1;
  }

  /* removed if haddone here because it caused problems */
  if (!had_abort)
  {
    EdtTrace(edt_p, 0xeeeaaa00);
    EdtTrace(edt_p, haddone);
    if (edt_p->m_Continuous || (edt_p->m_Done < edt_p->m_Todo) || edt_p->m_Freerun)
    {
      edt_SetupDma(edt_p);
    }

    /* if (edt_p->m_Loaded < edt_p->m_Done) */
    {
      /* SetupRdy will leave RDY enabled if sglist not done also will reset EN_MN_DONE */
      edt_SetupRdy(edt_p);
    }
  }

  EDTPRT(edt_p,
         DBG_DPC,
         TRL_INFO,
         ("DPC: %d done %d todo %d loaded \n", edt_p->m_Done, edt_p->m_Todo, edt_p->m_Loaded));

  if (edt_p->m_Abort)
  {
    if (edt_get_active_channels(edt_p))
    {
      EDTPRT(edt_p, DBG_DPC, TRL_ERROR, ("DPC: abort buf still othter active\n"));
#ifdef DISABLE_INTS_IN_ISR
      edt_enable_interrupts(edt_p);
#endif

      edt_p->m_DmaActive = 0;
    }
    edt_p->m_Abort = 0;
  }
  else
  {
#ifdef DISABLE_INTS_IN_ISR
    edt_enable_interrupts(edt_p);
#endif
  }

  return (edt_p);
}

void EdtTrace(Edt_Dev* edt_p, u_int val)
{
  unsigned int index = edt_p->m_TraceIdx;
  int entries = edt_p->m_TraceEntrys;

  if ((NULL == edt_p->m_TraceBuf) || (edt_p->m_TraceBufSize < 2))
  {
#ifdef __linux__
    WARN_ONCE(1, "Invalid TraceBuf");
#endif
    return;
  }

  if (index >= edt_p->m_TraceBufSize - 2)
    index = 2;

  edt_p->m_TraceIdx += 2;
  edt_p->m_TraceEntrys += 2;

  if (edt_p->m_TraceIdx >= edt_p->m_TraceBufSize)
    edt_p->m_TraceIdx = 2;

  edt_p->m_TraceBuf[index] = (entries & 0xfffffff) | (edt_p->m_DmaChannel << 28);

  edt_p->m_TraceBuf[index + 1] = val;
}

#ifdef TIME_LOCKS

uint64_t edt_drv_timestamp()
{
  edt_timespec_t tv;

  edt_get_hres_time(NULL, &tv);

  return (uint64_t)tv.tv_sec * 1000000 + (uint64_t)tv.tv_nsec / 1000;
}

#endif

void EdtTraceTimeStamp(Edt_Dev* edt_p, edt_timespec_t* pTimeSpec, u_int UNUSED_PARAM(level))
{
  edt_timespec_t* pt = pTimeSpec;
  edt_timespec_t tv;

  if (!pt)
    pt = &tv;

  EdtTrace(edt_p, TR_TMSTAMP);

  edt_get_hres_time(edt_p, pt);
  EdtTrace(edt_p, pt->tv_sec);
  EdtTrace(edt_p, pt->tv_nsec);
}

void edt_set_tracebuf_size(Edt_Dev* edt_p, u_int size)
{
  if (edt_p->m_TraceBuf && (size == edt_p->m_TraceBufSize))
  {
    return;
  }

  // size must be >= EDT_TRACESIZE
  if (size < EDT_TRACESIZE)
  {
    size = EDT_TRACESIZE;
  }

  EDTPRT(edt_p, DBG_ALLOC, TRL_WARN, ("setting tracebuf to %d\n", size));

  edt_start_critical_src(edt_p, 26);

  if (edt_p->m_TraceBuf)
  {
    edt_free_tracebuf(edt_p);
  }
  edt_p->m_TraceBuf = (u_int*)edt_vmalloc(size * 4, "tracebuf");
  edt_p->m_TraceBufSize = size;

  edt_end_critical_src(edt_p, 26);

  edt_clear_tracebuf(edt_p);
}

void edt_free_tracebuf(Edt_Dev* edt_p)
{
  edt_clear_tracebuf(edt_p);

  if (edt_p->m_TraceBuf)
  {
    edt_vfree(edt_p->m_TraceBuf, edt_p->m_TraceBufSize * 4, "tracebuf");
    edt_p->m_TraceBufSize = 0;
    edt_p->m_TraceBuf = NULL;
  }
}

void edt_clear_tracebuf(Edt_Dev* edt_p)
{
  edt_p->m_TraceIdx = 0;
  edt_p->m_TraceEntrys = 0;
}

/* Set the debug mask - this sets bits in the mask array up to the
level specified, and clears those bits in all levels > level */
void edt_set_debug_mask(Edt_Dev* edt_p, u_int level, u_int mask)
{
  u_int clearval = ~mask;
  u_int i;

  EDTPRT(edt_p, DBG_ALL, TRL_VERBOSE, ("--> Debug mask set %d : %08x\n", level, mask));

  if (level > EDT_DEBUG_LEVELS)
    level = EDT_DEBUG_LEVELS;

  for (i = 0; i <= level; i++)
  {
    edt_p->DebugMask[i] |= mask;
    EDTPRT(edt_p, DBG_ALL, TRL_VERBOSE, ("%d : %02x\n", i, edt_p->DebugMask[i]));
  }

  for (i = level + 1; i < EDT_DEBUG_LEVELS; i++)
  {
    edt_p->DebugMask[i] &= clearval;
    EDTPRT(edt_p, DBG_ALL, TRL_VERBOSE, ("%d : %02x\n", i, edt_p->DebugMask[i]));
  }
}

u_int edt_get_debug_mask(Edt_Dev* edt_p, u_int dbglevel)
{
  if (dbglevel <= EDT_DEBUG_LEVELS)
    return edt_p->DebugMask[dbglevel];

  return 0xffffffff;
}

void edt_set_current_buffer(Edt_Dev* edt_p, u_int bufnum)
{
  edt_start_critical_src(edt_p, 16);
  EdtTrace(edt_p, TR_SETBUF);
  EdtTrace(edt_p, bufnum);
  edt_p->m_Freerun = 0;
  edt_p->m_Abort = 0;

#ifdef _WINDK_DRIVER_
  edt_p->m_WaitIrp = NULL;
  edt_p->ClearEvent(EDT_EVENT_BUF);
#else
  edt_p->m_WaitIrp = 0;
#endif

  edt_p->m_ContState = 0;
  edt_SetVars(edt_p, bufnum);

  edt_end_critical_src(edt_p, 16);
}
