/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 */

#include "edtdrv.h"
#include "edt_trace.h"

int edt_set_sg_cnt(Edt_Dev* edt_p, u_int command)
{
  u_int readback;
  u_int count = 0;
  u_char done = 0;

  while (!done)
  {
    edt_set(edt_p, EDT_SG_NXT_CNT, command);
    readback = edt_get(edt_p, EDT_SG_NXT_CNT) & 0x9b00ffff;

    if ((++count == 300))
    {
      EDTPRT(edt_p, DBG_DMA, TRL_ERROR, ("Setting EDT_SG_NXT_CNT failed\n"));
      return -1;
    }
    if (count > 10)
      EDTPRT(edt_p, DBG_ALL, TRL_INFO, ("Setting EDT_SG_NXT_CNT count %u\n", count));

    if (readback == (command & 0x9b00ffff))
      done = 1;
  }

  return 0;
}

int edt_set_sg_addr(Edt_Dev* edt_p, u_int command)
{
  u_int readback;
  u_int count = 0;
  u_char done = 0;

  while (!done)
  {
    edt_set(edt_p, EDT_SG_NXT_ADDR, command);
    readback = edt_get(edt_p, EDT_SG_NXT_ADDR);

    if ((++count == 300))
    {
      EDTPRT(edt_p, DBG_DMA, TRL_ERROR, ("Setting EDT_SG_NXT_ADDR failed\n"));
      return -1;
    }
    if (count > 10)
      EDTPRT(edt_p, DBG_ALL, TRL_INFO, ("Setting EDT_SG_NXT_ADDR count %d\n", count));

    if (readback == command)
      done = 1;
  }

  return 0;
}

/****************************************
 * Go through all channels counting which ones are active
 **************************************/

int edt_get_active_channels(Edt_Dev* edt_p)
{
  Edt_Dev* base_p = NULL;
  int i;
  int active = 0;
  u_int j;

  if (NULL == edt_p)
    return -EINVAL;

  base_p = edt_p->m_chan0_p;

  if (!base_p)
    return 0;

  for (i = 0; i < base_p->m_chancnt; i++)
  {
    // GCC 13 believes this pointer is never NULL
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Waddress"
    if (base_p->chan_p /* && base_p->chan_p[i]->m_DmaActive */)
#pragma GCC diagnostic pop
    {
      Edt_Dev* chan_p = base_p->chan_p[i];

      if (base_p->chan_p[i]->m_DmaActive)
      {
        active = 1;
        EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("edt_get_active_channel: chan %d open for dma\n", i));
      }
      for (j = 0; j < chan_p->m_Numbufs; j++)
      {
        if (chan_p->m_Rbufs[j].m_Active)
        {
          EDTPRT(edt_p,
                 DBG_DMA,
                 TRL_INFO,
                 ("edt_get_active_channel: chan %d buf %d active\n", i, j));
          active++;
        }
      }
    }
  }

  return active;
}

/*!
 * Count stats for all channels on a device:
 */
int edt_dma_chan_info(Edt_Dev* edt_p, edt_dma_info_t* dma_p)
{
  Edt_Dev* base_p = edt_p;
  int i;
  int active_chans = 0;
  int chan_active;
  int chan;

  base_p = edt_p->m_chan0_p;

  if (!base_p)
    return 0;
  dma_p->used_dma = 0;
  dma_p->alloc_dma = 0;
  dma_p->active_dma = 0;

// GCC 13 believes this pointer is never NULL
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Waddress"
  if (base_p->chan_p)
#pragma GCC diagnostic pop
  {
    for (chan = 0; chan < base_p->m_chancnt; chan++)
    {
      Edt_Dev* channel_p = base_p->chan_p[chan];
      chan_active = 0;
      if (channel_p->m_DmaActive)
      {
        dma_p->active_dma |= (1 << chan);
        chan_active = 1;
      }
      if (channel_p->m_ReportedNumbufs >= 1)
      {
        dma_p->alloc_dma |= (1 << chan);
        chan_active = 1;
      }
      if (channel_p->dma_pid)
      {
        dma_p->used_dma |= (1 << chan);
        chan_active = 1;
      }

      for (i = 0; i < 256; i++)
      {
        dma_p->direct_reads[i] += channel_p->m_direct_reads[i];
        dma_p->indirect_reads[i] += channel_p->m_indirect_reads[i];
        dma_p->direct_writes[i] += channel_p->m_direct_writes[i];
        dma_p->indirect_writes[i] += channel_p->m_indirect_writes[i];
      }
      for (i = 0; i < 8; i++)
      {
        dma_p->dma_reads[i] += channel_p->m_dma_reads[i];
        dma_p->dma_writes[i] += channel_p->m_dma_writes[i];
      }
      if (chan_active)
        active_chans++;
    }
  }

  dma_p->lock_time = base_p->m_total_lock_time;
  dma_p->wait_time = base_p->m_total_wait_time;
  dma_p->locks = base_p->m_nLocks;
  dma_p->interrupts = base_p->m_nInterrupts;
  memcpy(dma_p->lock_array, base_p->m_lock_array, sizeof(dma_p->lock_array));

  return (active_chans);
}

void edt_abortdma(Edt_Dev* edt_p)
{
  u_int cmd;
  int j;

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("start of abort %d done %d started pid %d cur %d tmout %d\n",
          edt_p->m_Done,
          edt_p->m_Started,
          edt_p->dma_pid,
          edt_get_current_pid(),
          edt_p->m_had_timeout));

  if (!edt_p->m_had_timeout && edt_p->dma_pid != edt_get_current_pid())
  {
    return;
  }

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("abortdma: channel chn %d  todo %d done %d loaded %d free %d started %d\n",
          edt_p->m_DmaChannel,
          edt_p->m_Todo,
          edt_p->m_Done,
          edt_p->m_Loaded,
          edt_p->m_Freerun,
          edt_p->m_Started));

  cmd = edt_get(edt_p, EDT_SG_NXT_CNT);
  cmd &= (EDT_DMA_MEM_RD | EDT_BURST_EN | 0xffff);

  edt_p->m_Freerun = 0;
  edt_p->m_CurSg = 0;
  edt_p->m_DmaActive = 0;

  for (j = 0; j < (int)edt_p->m_Numbufs; j++)
    edt_p->m_Rbufs[j].m_Active = 0;

  edt_p->m_DmaActive = 0;

  edt_lock_abort(edt_p);

  edt_p->m_Abort = 1;
  cmd |= EDT_DMA_ABORT;

  edt_set_sg_cnt(edt_p, cmd);

  edt_unlock_abort(edt_p);

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("abortdma: after abort cmd %x  started %d done %d loaded %d\n",
          edt_get(edt_p, EDT_SG_NXT_CNT),
          edt_p->m_Started,
          edt_p->m_Done,
          edt_p->m_Loaded));
}

static void edt_flush_for_dma(Edt_Dev* edt_p)
{
#ifdef PDV
  /* changed debug CKB */
  u_int intcfg;
  u_int cmd;
#endif

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_WARN, ("==> flush_for_dma\n"));

  EdtTrace(edt_p, TR_FLUSH);
  EdtTrace(edt_p, edt_p->m_Done);

  edt_set_rfifo(edt_p, 0);

  /* todo - add dep_flush */

#ifdef PDV

  intcfg = edt_get(edt_p, PDV_CFG);
  intcfg &= ~PDV_FIFO_RESET;
  edt_set(edt_p, PDV_CFG, (u_char)(intcfg | PDV_FIFO_RESET));
  edt_set(edt_p, PDV_CFG, intcfg);

  cmd = edt_get(edt_p, EDT_SG_NXT_CNT);
  cmd &= (EDT_DMA_MEM_RD | EDT_BURST_EN | 0xffff);

  edt_set_sg_cnt(edt_p, cmd | EDT_DMA_ABORT);  // DMA_ABORT self clears and reads 0

#endif

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_WARN, ("<== flush_for_dma\n"));
}

static void edt_do_start_action(Edt_Dev* edt_p)
{
  EdtTrace(edt_p, 0xaaaa0001);

#ifdef PDV
  /*
   * if m_Continuous - set continuous if doing > 1 image
   * set clear continuous  if we have finished next to last image
   */
  if (edt_p->m_Continuous)
    pdvdrv_step_continuous(edt_p);
  else
#endif /* PDV */
      if (edt_p->m_StartAction == EDT_ACT_ONCE || edt_p->m_StartAction == EDT_ACT_ALWAYS ||
          edt_p->m_StartAction == EDT_ACT_CYCLE)
  {
    if (edt_p->m_DoStartDma)
    {
      if (edt_p->m_Started == edt_p->m_Done)
      {
        edt_set(edt_p, edt_p->m_StartDmaDesc, edt_p->m_StartDmaVal);

        if (edt_p->m_DoneOnFval)
        {
          edt_or(edt_p, PDV_UTIL2, PDV_INT_ENFV);
        }

        edt_p->m_Started++;

        EdtTrace(edt_p, TR_DOGRAB);
        EdtTrace(edt_p, edt_p->m_Started);

        if (edt_p->m_StartAction == EDT_ACT_ONCE)
          edt_p->m_DoStartDma = 0;
      }
    }
  }

  EdtTrace(edt_p, 0xaaaa0002);
}

u_int edt_SetupDma(Edt_Dev* edt_p)
{
  u_int rfifo;
  rbuf* p;

  EdtTrace(edt_p, TR_SETUP_DMA);

  rfifo = edt_get_rfifo(edt_p);

  /* chet par */
  if (edt_p->m_Abort)
  {
    EDTPRT(edt_p, DBG_DMA_SETUP, TRL_ERROR, ("SetupDma: called with abort\n"));
    return (1);
  }

  p = &edt_p->m_Rbufs[edt_p->m_Nxtbuf];

  /* set up DMA - flush, set rfifo if needed */
  if ((edt_p->m_Done == edt_p->m_Loaded) && (edt_p->m_Done < edt_p->m_Todo || edt_p->m_Freerun))
  {
    if (edt_p->m_FirstFlush == EDT_ACT_ONCE || edt_p->m_FirstFlush == EDT_ACT_ALWAYS)
    {
      if (edt_p->m_DoFlush)
      {
        edt_flush_for_dma(edt_p);

        if (edt_p->m_FirstFlush == EDT_ACT_ONCE)
          edt_p->m_DoFlush = 0;
      }
    }

    /* set fifo direction */
    if (edt_p->m_AutoDir)
    {
      rfifo = !p->m_WriteFlag;
    }

    EDTPRT(
        edt_p,
        DBG_DMA_SETUP,
        TRL_INFO,
        ("Autodir %d writeflag %d setting rfifo to %d\n", edt_p->m_AutoDir, p->m_WriteFlag, rfifo));

    edt_set_rfifo(edt_p, rfifo);
  }

  edt_SetupRdy(edt_p);

  EdtTrace(edt_p, 0xaaaa0000);

  edt_do_start_action(edt_p);

  EdtTrace(edt_p, 0xaaaa0003);

  EdtTrace(edt_p, TR_SETUP_DMA_DONE);
  EdtTrace(edt_p, 0xaaaa0005);

  return 0;
}

u_int edt_SetupRdy(Edt_Dev* edt_p)
{
  u_int ret = 0;
  u_int idx = 0;
  rbuf* p;
  u_long command;

  EdtTrace(edt_p, TR_SETUP_RDY);
  EdtTrace(edt_p, edt_p->m_DmaChannel);

  command = edt_get(edt_p, EDT_SG_NXT_CNT);

  /* chet later */
  if (edt_p->m_Abort)
  {
    EDTPRT(edt_p, DBG_DMA, TRL_WARN, ("SETRDY called with abort\n"));
    return (0);
  }
  if (command & EDT_DMA_ABORT)
  {
    EDTPRT(edt_p, DBG_DMA, TRL_WARN, ("SetupRdy: sees abort\n"));
  }

  /* always turn off abort and en_sg_done bits */
  command &= ~(EDT_DMA_ABORT | EDT_EN_SG_DONE);

  p = &edt_p->m_Rbufs[edt_p->m_Nxtbuf];

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("SetupRdy: Entry command = %lx EDT_DMA_RDY %s\n",
          command,
          ((command & EDT_DMA_RDY) ? "TRUE" : "FALSE")));

  if (edt_p->m_Done >= edt_p->m_Todo && !edt_p->m_Freerun)
  {
    EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("SetupRdy: Return because Done >= Todo \n"));
    edt_p->m_DmaActive = 0;
    edt_p->m_CurSg = 0;
    return 0;
  }

  if (edt_p->m_Loaded >= edt_p->m_Todo && !edt_p->m_Freerun)
  {
    EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("SetupRdy: Return because Loaded >= Todo \n"));
    return 0;
  }

  if (command & EDT_DMA_RDY)
  {
    if (!edt_p->m_CurSg)
    {
      edt_p->m_CurSg = p;
      edt_p->m_CurSg->m_Idx = 0;
      edt_p->m_DmaActive = 1;
      EdtTrace(edt_p, 0xabcd0000);
    }
    else
      EdtTrace(edt_p, 0xabcd0000 | edt_p->m_CurSg->m_Idx);

    command &= (EDT_DMA_MEM_RD | EDT_BURST_EN | 0xffff);

    idx = edt_p->m_CurSg->m_Idx;

    EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("SetupRdy: Idx = %d cmd = %08lx\n", idx, command));

    EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("SetupRdy: chunks %d \n", edt_p->m_CurSg->m_nChunks));

#ifdef DO_STOP
    edt_set(edt_p, EDT_DMA_STOP, 1);
#endif
    if (idx < edt_p->m_CurSg->m_nChunks)
    {
      EdtPCISg* psgmem = &edt_p->m_CurSg->SgChunksList[idx];
      bus_addr32_t workaddr = psgmem->busaddr;

      EDTPRT(edt_p,
             DBG_DMA,
             TRL_INFO,
             ("Setting NXT_ADDR to %08x size 0x%x\n", workaddr, (u_short)psgmem->count));

      edt_set_sg_addr(edt_p, workaddr);

      EdtTrace(edt_p, TR_LOAD_SG);
      EdtTrace(edt_p, workaddr);
      EdtTrace(edt_p, psgmem->count);

      if (p->m_WriteFlag)
      {
        command |= EDT_DMA_MEM_RD;
      }
      else
      {
        command &= ~EDT_DMA_MEM_RD;
      }

      command &= ~0xffff;
      command |= psgmem->count;
      command |= EDT_DMA_START;

      idx++;
      edt_p->m_CurSg->m_Idx = idx;
    }

    if (idx >= edt_p->m_CurSg->m_nChunks)
    {
      edt_p->m_Loaded++;
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_INFO,
             ("SetupRdy: NXT_CNT=%08x ADDR=%08x\n",
              edt_get(edt_p, EDT_SG_NXT_CNT),
              edt_get(edt_p, EDT_SG_NXT_ADDR)));
      p->m_Active = 1;
      edt_p->m_DmaActive = 1;

      EdtTrace(edt_p, TR_LOADED);
      EdtTrace(edt_p, edt_p->m_Loaded);

      EdtTraceTimeStamp(edt_p, NULL, 0);

      if (++edt_p->m_Nxtbuf >= edt_p->m_Numbufs)
        edt_p->m_Nxtbuf = 0;

      edt_p->m_CurSg->m_Idx = 0;
      edt_p->m_CurSg = 0;
      ret = 1;
    }

  } /* is ready */

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("SetupRdy: Loaded %d Todo %d\n", edt_p->m_Loaded, edt_p->m_Todo));

  if (edt_p->m_Loaded < edt_p->m_Todo || edt_p->m_Freerun)
  {
    if (edt_p->m_Loaded > edt_p->m_Done + 1)
    {
      /* try to cut down on unneeded ints  */
    }
    else
      command |= EDT_EN_RDY;
  }

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("SetupRdy %s EDT_EN_RDY\n", (command & EDT_EN_RDY) ? "enable" : "disable"));

  if (edt_p->m_Done < edt_p->m_Todo || edt_p->m_Freerun)
    command |= EDT_EN_MN_DONE;

  if (edt_p->m_UseBurstEn)
    command |= EDT_BURST_EN;
  else
  {
    command &= ~EDT_BURST_EN;
  }

  EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("SetupRdy: command = %08lx\n", command));

  edt_set_sg_cnt(edt_p, command);
  EdtTraceTimeStamp(edt_p, &p->m_start_timeval, 1);

#ifdef DO_STOP
  edt_set(edt_p, EDT_DMA_STOP, 0);
#endif
  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("SetupRdy: Leaving NXT_CNT=%08x ADDR=%08x command=%08lx\n",
          edt_get(edt_p, EDT_SG_NXT_CNT),
          edt_get(edt_p, EDT_SG_NXT_ADDR),
          command));

  EdtTrace(edt_p, TR_SETUP_RDY_DONE);

  return (ret);
}

int edt_fifocount(Edt_Dev* edt_p)
{
  u_int cfg;
  int fifocnt;

  cfg = edt_get(edt_p, EDT_DMA_CFG);
  if (cfg & EDT_FIFO_FLG)
  {
    fifocnt = ((cfg & EDT_FIFO_CNT) >> EDT_FIFO_SHIFT) + 1;
    fifocnt *= 4;
  }
  else
    fifocnt = 0;

  return fifocnt;
}

int edt_InitializeBoard(Edt_Dev* edt_p)
{
  EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("Edt Initialize Board\n"));
  edt_set_sg_cnt(edt_p, EDT_BURST_EN);

  if (edt_p->m_DmaChannel == 0)
  {
    EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("DIsable from InitializeBoard\n"));
    edt_disable_interrupts(edt_p);
  }

  edt_set_sg_cnt(edt_p, EDT_BURST_EN | EDT_DMA_ABORT);
  edt_set_sg_cnt(edt_p, EDT_BURST_EN);

  edt_dep_init(edt_p);

  return 0;
}

void StartNext(Edt_Dev* edt_p)
{
  EdtTrace(edt_p, TR_START_BUF);
  EdtTrace(edt_p, edt_p->m_Todo);
  edt_p->m_Abort = 0; /* added at par */
  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("StartNext: loaded %d todo %d started %d done %d\n",
          edt_p->m_Loaded,
          edt_p->m_Todo,
          edt_p->m_Started,
          edt_p->m_Done));

#ifdef __linux__
  edt_prepare_dma_buffers(edt_p);
#endif

  if (edt_p->m_Continuous || (edt_p->m_Loaded < edt_p->m_Todo || edt_p->m_Freerun))
  {
    EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("Startnext: calls SetupDma %d\n", edt_p->m_Todo));

    edt_SetupDma(edt_p);

    edt_enable_dma_interrupt(edt_p);
  }
}

int edt_start_buffers(Edt_Dev* edt_p, unsigned int numtostart)
{
  edt_p->m_Abort = 0;
  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("--> edt_start_buffers calling StartNext todo %u\n", numtostart));
  edt_start_critical_src(edt_p, 7);

  if (numtostart == 0)
  {
    edt_p->m_Todo = 0;
    edt_p->m_Freerun = 1;
    EDTPRT(edt_p, DBG_DMA, TRL_WARN, ("Starting m_Freerun\n"));
  }
  else
  {
    if (edt_p->m_Freerun)
      edt_p->m_Todo = edt_p->m_Done + numtostart;
    else
      edt_p->m_Todo += numtostart;
    edt_p->m_Freerun = 0;
  }

  StartNext(edt_p);

  EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("<-- edt_start_buffers\n"));

  edt_end_critical_src(edt_p, 7);

  return STATUS_SUCCESS;
}
