/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 */

#include "edtdrv.h"
#include "edt_trace.h"
#include "edt_pci_devices.h"

#ifdef PDV
#include "pdv.h"
#endif

#ifdef PCD
#include "pcd.h"
#endif

/*!
 * Initialize board variables
 * Assume that DmaChannel and Devid are set coming in
 */
int edt_initialize_vars(Edt_Dev* edt_p)
{
  rbuf* p;

  /*
   * init vars
   */

  EDTPRT(edt_p, DBG_INIT, TRL_INFO, ("Initialize vars for channel %d\n", edt_p->m_DmaChannel));

#ifdef PCD
  edt_p->m_edt_dep_init = pcd_dep_init;
  edt_p->m_edt_DeviceInt = pcd_DeviceInt;
  edt_p->m_edt_device_ioctl = pcd_device_ioctl;
  edt_p->m_edt_hwctl = pcd_hwctl;
#endif
#ifdef PDV
  edt_p->m_edt_dep_init = pdv_dep_init;
  edt_p->m_edt_DeviceInt = pdv_DeviceInt;
  edt_p->m_edt_device_ioctl = pdv_device_ioctl;
  edt_p->m_edt_hwctl = pdv_hwctl;
#endif

  edt_SetVars(edt_p, 0);

  edt_p->m_Numbufs = 1;
  edt_p->m_ReportedNumbufs = 1;
  edt_p->m_TraceIdx = 2;
  edt_p->m_TraceEntrys = 0;
  edt_p->m_timeout = 0;
  edt_p->m_hadcnt = 0;
  edt_p->m_Freerun = 0;
  edt_p->m_Abort = 0;
  edt_p->m_Overflow = 0;
  edt_p->m_DoStartDma = 0;
  edt_p->m_DoEndDma = 0;
  edt_p->m_DmaIntMask = 0;

  edt_p->m_CurSg = 0;
  edt_p->m_Curbuf = 0; /* so transfer count will work  */
  edt_p->m_UseBurstEn = 1;

  edt_p->m_TimeoutAction = EDT_TIMEOUT_NULL;
  edt_p->m_TimeoutGoodBits = 32;
  edt_p->m_MultiDone = 0;

  /* Extended register access; see comment at start of edtreg.h.  -mcm 4/09 */
  edt_p->m_IndRegBase = 0x60;

#if defined(PCD)
  edt_p->mezz.id = MEZZ_ID_UNKNOWN;
#endif

#ifdef PDV
  edt_p->m_FirstFlush = EDT_ACT_ONCE;
  edt_p->m_StartAction = EDT_ACT_ALWAYS;
  edt_p->m_EndAction = EDT_ACT_ALWAYS;
  edt_p->m_DoFlush = 1;
  edt_p->m_AutoDir = 1;
#endif

#if defined(PCD)
  /* For PCD types.  Mark April '04 */
  edt_p->m_FirstFlush = EDT_ACT_NEVER;
  edt_p->m_StartAction = EDT_ACT_NEVER;
  edt_p->m_EndAction = EDT_ACT_NEVER;
  edt_p->m_DoFlush = 0;
#else
  edt_p->m_FirstFlush = EDT_ACT_ONCE;
  edt_p->m_StartAction = EDT_ACT_ALWAYS;
  edt_p->m_EndAction = EDT_ACT_ALWAYS;
  edt_p->m_DoFlush = 1;
#endif

  edt_p->m_AutoDir = 1;

  edt_p->m_Ints = 0;
  edt_p->m_WaitIrp = 0;
#if defined(PDV)
  if (edt_p->devid == PE1DVVL_ID || edt_p->devid == PE4DVVL_ID || edt_p->devid == PE8VLCLS_ID)
  {
    edt_p->m_ReadTimeout = 5000;
    edt_p->m_WriteTimeout = 5000;
  }
  else
#endif /* PDV */
  {
    edt_p->m_ReadTimeout = 0;
    edt_p->m_WriteTimeout = 0;
  }

  edt_p->m_serial = edt_vmalloc(sizeof(edt_serial), "m_serial");
  memset(edt_p->m_serial, 0, sizeof(edt_serial));

  EDTPRT(edt_p,
         DBG_INIT,
         TRL_INFO,
         ("init rbuf %p for max %x\n", edt_p->m_Rbufs, edt_p->m_MaxRbufs));

  if (edt_p->m_Rbufs == 0)
  {
    EDTPRT(edt_p, DBG_INIT, TRL_INFO, ("skip init with rbuf 0\n"));
  }
  else
    for (p = edt_p->m_Rbufs; p < &edt_p->m_Rbufs[edt_p->m_MaxRbufs]; p++)
    {
      p->m_SglistSize = 0;
      p->m_SgUsed = 0;

      p->SgChunksList = 0;
      p->is_adjusted = 0;
    }

  edt_set_tracebuf_size(edt_p, EDT_INITIAL_TRACESIZE);

  /*
   * reset device (including disabling interrupts)
   */

  if (edt_p->bar0_mmap)
    (void)edt_hwctl(INIT_DEVICE, NULL, edt_p);

  return 0;
}

int edt_init_board_specific_hw(Edt_Dev* edt_p)
{
  /*
   * reset device (including disabling interrupts)
   */
  EDTPRT(edt_p,
         DBG_INIT,
         TRL_INFO,
         ("--> edt_init_board_specific_hw MemBase %p hwctl %p\n",
          edt_p->bar0_mmap,
          edt_p->m_edt_hwctl));

  if (edt_p->bar0_mmap)
    (void)edt_p->m_edt_hwctl(INIT_DEVICE, NULL, edt_p);
  else if (edt_p->m_DmaChannel == 0)
  {
    EDTPRT(edt_p,
           DBG_INIT,
           TRL_ERROR,
           ("FATAL:  edt_p->MemBase %p; skipping edt_hwctl(INIT_DEVICE)\n", edt_p->bar0_mmap));
    return -ENOMEM;
  }

  EDTPRT(edt_p, DBG_INIT, TRL_INFO, ("<-- edt_init_board_specific_hw\n"));

  return 0;
}
