/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @file
 * @internal
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 */
#ifndef EDT_IOCTL_H
#define EDT_IOCTL_H

#include "edtdef.h"
#include "edt_trace.h"  // EDT_TRACESIZE

#if defined(__linux__) && defined(_KERNEL)
#include <linux/types.h>
#else
#include <stdint.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define EDT_IOCTL_ACTION_Shift 0
/* Action is low 10 bits */
#define EDT_IOCTL_ACTION_Mask (0x3FFU << EDT_IOCTL_ACTION_Shift)

#define EDT_IOCTL_SIZE_Shift 10
/* Size is bits [23:10] */
#define EDT_IOCTL_SIZE_Mask (0x3FFFU << EDT_IOCTL_SIZE_Shift)

#define EDT_IOCTL_TYPE_Shift 24
#define EDT_IOCTL_GET_Mask   (0x1U << EDT_IOCTL_TYPE_Shift)
#define EDT_IOCTL_SET_Mask   (0x2U << EDT_IOCTL_TYPE_Shift)

#define EIO_DECODE_ACTION(code) (code & EDT_IOCTL_ACTION_Mask)
/* Shorthand, looks better in case statements. */
#define EIODA(code) EIO_DECODE_ACTION(code)

#define EIO_DECODE_SIZE(code) ((code & EDT_IOCTL_SIZE_Mask) >> EDT_IOCTL_SIZE_Shift)
#define EIO_DECODE_SET(code)  ((code & EDT_IOCTL_SET_Mask) != 0)
#define EIO_DECODE_GET(code)  ((code & EDT_IOCTL_GET_Mask) != 0)

#if defined(_WIN32)
#if defined(_KERNEL)
#include <ntddk.h>
#else
#include <winioctl.h>
#endif /* _WIN32 and _KERNEL */

#define EDT_NT_IOCTL 0xf000f000
#endif /* _WIN32 */

#if defined(__linux__)
#include <linux/ioctl.h>

#define EDT_IOC_MAGIC 'k'

#define EDT_NT_IOCTL   _IOWR(EDT_IOC_MAGIC, 1, edt_ioctl_struct)
#define EDT_NT_IOCTL32 _IOWR(EDT_IOC_MAGIC, 1, edt_ioctl_struct32)

#define EDT_IOC_MAXNR 2

#endif /* __linux__ */

#if defined(__linux__) || (defined(_WIN32) && !defined(_KERNEL))
/*
 * This structure must correspond to the next edt_ioctl_struct
 * in order for 32-bit programs to run under 64-bit Linux.
 * This workaround was done for ioctl backwards compatibility purposes.
 * Mark Mason - Aug 2007
 */
typedef struct {
#ifdef _WIN32
  HANDLE device;
#else
  int device;
#endif
  uint32_t controlCode;
  uint32_t inSize;
  uint32_t outSize;
  uint32_t bytesReturned;
  uint32_t inBuffer;  /* userspace address */
  uint32_t outBuffer; /* userspace address */
} edt_ioctl_struct32;

/*
 * This structure must correspond to the above edt_ioctl_struct32
 * in order for 32-bit programs to run under 64-bit Linux.
 * Mark Mason - Aug 2007
 */
typedef struct {
#ifdef _WIN32
  HANDLE device;
#else
  int device;
#endif
  uint32_t controlCode;
  uint32_t inSize;
  uint32_t outSize;
  uint32_t bytesReturned;

  void* inBuffer;
  void* outBuffer;
} edt_ioctl_struct;
#endif /* Linux OR (Windows AND NOT driver) */

/* For passing 2 args. */
typedef struct {
  /* Must be multiple of 64 bits for amd64. */
  uint64_t value;
  uint32_t desc;
  uint32_t flags;
} edt_buf;

/* For serial read and write. */
#define EDT_SERBUF_SIZE  2048
#define EDT_SERBUF_OVRHD 16 /* Size of all but buffer. */

/** Data argument used with ES_SERIAL ioctl. */
typedef struct {
  uint32_t size;
  uint32_t misc;
  uint32_t flags;
  char buf[EDT_SERBUF_SIZE];
} ser_buf;

/* For describing ring buffer. */
typedef struct {
  uintptr_t addr;
  uint32_t index;
  uint32_t size;
  uint32_t writeflag;
} buf_args;

/* Dependent / PdvDependent struct size in bytes */
#define EDT_DEPSIZE 4096

#define SIZED_DATASIZE (EDT_DEPSIZE - (uint32_t)sizeof(uint32_t))

typedef struct {
  uint32_t size;
  uint32_t data[SIZED_DATASIZE / 4];
} edt_sized_buffer;

#define EDT_DEVICE_TYPE 0x8000

/* 'METHOD_BUFFERED' is defined in Windows KMDF "ntddk.h", and Windows SDK "winioctl.h". */
#if defined(_WIN32) && defined(METHOD_BUFFERED)
#define EDT_MAKE_IOCTL(t, c) (uint32_t) CTL_CODE((t), 0x800 + (c), METHOD_BUFFERED, FILE_ANY_ACCESS)
#elif defined(__linux__)
#define EDT_MAKE_IOCTL(t, c) (uint32_t)(c)
#endif

#define EIOC(action, type, size)                                                         \
  (((uint32_t)type) | (((uint32_t)size << EDT_IOCTL_SIZE_Shift) & EDT_IOCTL_SIZE_Mask) | \
   (((uint32_t)action << EDT_IOCTL_ACTION_Shift) & EDT_IOCTL_ACTION_Mask))

#define EDTS_DEBUG EIOC(10, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_DEBUG EIOC(11, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_REG   EIOC(14, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTG_REG   EIOC(15, EDT_IOCTL_GET_Mask | EDT_IOCTL_SET_Mask, sizeof(edt_buf))

/* TODO: Remove? edt_projects/general/legacy/pci_driver#26 */
#define EDTS_PROG_READBACK EIOC(21, EDT_IOCTL_SET_Mask, sizeof(uint32_t))

#define EDTS_DEBUG_MASK    EIOC(25, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_DEBUG_MASK    EIOC(26, EDT_IOCTL_GET_Mask | EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_ALLOC_KBUFFER EIOC(27, EDT_IOCTL_GET_Mask | EDT_IOCTL_SET_Mask, sizeof(buf_args))

#define EDTS_SERIAL    EIOC(29, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_SERIAL    EIOC(30, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_DEPENDENT EIOC(31, EDT_IOCTL_SET_Mask, EDT_DEPSIZE)
#define EDTG_DEPENDENT EIOC(32, EDT_IOCTL_GET_Mask, EDT_DEPSIZE)
#define EDTG_DEVID     EIOC(33, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_RTIMEOUT  EIOC(34, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_WTIMEOUT  EIOC(35, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_BUFDONE   EIOC(36, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_NUMBUFS   EIOC(37, EDT_IOCTL_SET_Mask, sizeof(int))
#define EDTS_BUF       EIOC(38, EDT_IOCTL_SET_Mask, sizeof(buf_args))
#define EDTS_STARTBUF  EIOC(39, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#if defined(__linux__)
#define EDTS_WAITBUF EIOC(40, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#endif
#define EDTS_FREEBUF  EIOC(41, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_STOPBUF  EIOC(42, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_SETBUF   EIOC(45, EDT_IOCTL_SET_Mask, sizeof(int))
#define EDTG_TIMEOUTS EIOC(47, EDT_IOCTL_GET_Mask, sizeof(int))
#define EDTG_TRACEBUF EIOC(48, EDT_IOCTL_GET_Mask, (EDT_TRACESIZE * sizeof(int)))
#define EDTS_STARTDMA EIOC(49, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTS_ENDDMA   EIOC(50, EDT_IOCTL_SET_Mask, sizeof(edt_buf))

// SERIAL_FIFO IOCTLs removed in MR !200, not fully working before, now handled in user mode.
// #define EDTS_SERIAL_FIFO      EIOC(51, EDT_IOCTL_SET_Mask, sizeof(int))
// #define EDTG_SERIAL_FIFO      EIOC(52, EDT_IOCTL_GET_Mask, sizeof(int))

#define EDTG_RTIMEOUT         EIOC(55, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTG_WTIMEOUT         EIOC(56, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_SERIALWAIT       EIOC(58, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(edt_buf))
#define EDTG_OVERFLOW         EIOC(60, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_AUTODIR          EIOC(61, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_FIRSTFLUSH       EIOC(62, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_CONFIG_COPY      EIOC(63, EDT_IOCTL_GET_Mask | EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTG_CONFIG           EIOC(64, EDT_IOCTL_GET_Mask | EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTS_CONFIG           EIOC(65, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTG_SGTODO           EIOC(69, EDT_IOCTL_GET_Mask, (EDT_TRACESIZE * 4))
#define EDTG_SGLIST           EIOC(70, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(buf_args))
#define EDTS_SGLIST           EIOC(71, EDT_IOCTL_SET_Mask, sizeof(buf_args))
#define EDTG_SGINFO           EIOC(72, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(edt_buf))
#define EDTG_TIMECOUNT        EIOC(73, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_STARTACT         EIOC(77, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_ENDACT           EIOC(78, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_RESETCOUNT       EIOC(79, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_RESETSERIAL      EIOC(80, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_CLR_EVENT        EIOC(81, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_ADD_EVENT_FUNC   EIOC(82, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_DEL_EVENT_FUNC   EIOC(83, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_WAIT_EVENT       EIOC(85, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_TMSTAMP          EIOC(87, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(int64_t) * 3)
#define EDTS_TIMEOUT_ACTION   EIOC(88, EDT_IOCTL_SET_Mask, sizeof(uint32_t)) /* SSDIO specific */
#define EDTG_TIMEOUT_GOODBITS EIOC(89, EDT_IOCTL_GET_Mask, sizeof(uint32_t)) /* SSDIO specific */
#define EDTS_REG_OR           EIOC(93, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(edt_buf))
#define EDTS_REG_AND          EIOC(94, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(edt_buf))
#define EDTG_GOODBITS         EIOC(95, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_BURST_EN         EIOC(96, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_BURST_EN         EIOC(97, EDT_IOCTL_GET_Mask, sizeof(uint32_t))

#define EDTG_BUFBYTECOUNT EIOC(102, EDT_IOCTL_GET_Mask, sizeof(uint32_t) * 2)
#define EDTS_DOTIMEOUT    EIOC(103, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_REFTMSTAMP   EIOC(104, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#if defined(PDV) && (PDV != 0)
#define EDTS_PDVCONT  EIOC(105, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_PDVDPATH EIOC(106, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#endif
#define EDTS_RESET_EVENT_COUNTER EIOC(107, EDT_IOCTL_SET_Mask, sizeof(uint32_t))

#define EDTG_TODO        EIOC(109, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_RESUME      EIOC(110, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_EVENT_HNDL  EIOC(112, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTS_MAX_BUFFERS EIOC(113, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_MAX_BUFFERS EIOC(114, EDT_IOCTL_GET_Mask, sizeof(uint32_t))

/* TODO: Remove? edt_projects/general/legacy/pci_driver#26 */
#define EDTS_PROG_XILINX EIOC(116, EDT_IOCTL_SET_Mask, sizeof(edt_sized_buffer))
#ifdef _WIN32
#define EDTS_MAPMEM   EIOC(117, EDT_IOCTL_GET_Mask | EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTS_UNMAPMEM EIOC(118, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#endif

#define EDTS_DRV_BUFFER EIOC(122, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_BITPATH    EIOC(140, EDT_IOCTL_SET_Mask, sizeof(edt_bitpath))
#define EDTG_BITPATH    EIOC(141, EDT_IOCTL_GET_Mask, sizeof(edt_bitpath))
#define EDTG_VERSION    EIOC(142, EDT_IOCTL_GET_Mask, sizeof(edt_version_string))
#define EDTG_BUILDID    EIOC(143, EDT_IOCTL_GET_Mask, sizeof(edt_version_string))

// WAITCHAR IOCTL removed in MR !200, now handled in user mode.
// #define EDTS_WAITCHAR EIOC(144, EDT_IOCTL_SET_Mask, sizeof(edt_buf))

#define EDTG_MEMSIZE EIOC(146, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#if defined(PCD)
#define EDTS_DIRECTION EIOC(147, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#endif

#if defined(PCD)
#define EDTS_ABORTDMA_ONINTR EIOC(151, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#endif

#if defined(PDV)
#define EDTS_FVAL_DONE     EIOC(152, EDT_IOCTL_SET_Mask, sizeof(uint8_t))
#define EDTG_FVAL_DONE     EIOC(153, EDT_IOCTL_GET_Mask, sizeof(uint8_t))
#define EDTG_LINES_XFERRED EIOC(154, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#endif
#define EDTS_PROCESS_ISR EIOC(155, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_CLEAR_DMAID EIOC(156, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#if defined(PDV)
#define EDTG_SERIAL_WRITE_AVAIL EIOC(158, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#endif

#define EDTG_WAIT_STATUS      EIOC(161, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_WAIT_STATUS      EIOC(162, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_TIMEOUT_OK       EIOC(163, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_TIMEOUT_OK       EIOC(164, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_MULTI_DONE       EIOC(165, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTG_MULTI_DONE       EIOC(166, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_TEST_LOCK_ON     EIOC(167, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_BUF_MMAP         EIOC(169, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(buf_args))
#define EDTS_MEZZ_BITPATH     EIOC(170, EDT_IOCTL_SET_Mask, sizeof(edt_bitpath))
#define EDTG_MEZZ_BITPATH     EIOC(171, EDT_IOCTL_GET_Mask, sizeof(edt_bitpath))
#define EDTG_DMA_INFO         EIOC(172, EDT_IOCTL_GET_Mask, sizeof(edt_dma_info_t))
#define EDTS_TRACE_REG        EIOC(179, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTS_REG_BIT_CLEARSET EIOC(181, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTS_REG_BIT_SETCLEAR EIOC(182, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#if defined(PCD)
#define EDTS_MEZZ_ID EIOC(184, EDT_IOCTL_SET_Mask, sizeof(edt_buf))
#define EDTG_MEZZ_ID EIOC(185, EDT_IOCTL_SET_Mask | EDT_IOCTL_GET_Mask, sizeof(edt_buf))
#endif
#define EDTG_NUMBUFS EIOC(186, EDT_IOCTL_GET_Mask, sizeof(int))

#define EDTS_INDIRECT_REG_BASE EIOC(195, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_INDIRECT_REG_BASE EIOC(196, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTG_MEM2SIZE          EIOC(208, EDT_IOCTL_GET_Mask, sizeof(uint32_t))
#define EDTS_INTR_MASK         EIOC(210, EDT_IOCTL_SET_Mask, sizeof(uint32_t))
#define EDTG_INTR_MASK         EIOC(211, EDT_IOCTL_GET_Mask, sizeof(uint32_t))

/* Defines for get DMA status. */
#define EDT_DMA_IDLE    0
#define EDT_DMA_ACTIVE  1
#define EDT_DMA_TIMEOUT 2
#define EDT_DMA_ABORTED 3

/* Defines for SG ioctls. */
#define EDT_SGLIST_SIZE     1
#define EDT_SGLIST_VIRTUAL  2
#define EDT_SGLIST_PHYSICAL 3
#define EDT_SGTODO_SIZE     4
#define EDT_SGTODO_VIRTUAL  5
#define EDT_SGTODO_FIRST_SG 6
#define EDT_SG_LOADSIZE     7
#define EDT_SG_ALLOCSIZE    8
#define EDT_SG_TOTALUSED    9

#ifdef __cplusplus
}
#endif

#endif /* EDT_IOCTL_H */
