/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#define pr_fmt(fmt) "%s:%s(): " fmt, KBUILD_MODNAME, __func__

#include <linux/version.h>
#include <linux/preempt.h>
#include <linux/printk.h>
#include <linux/pci.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)) && \
    (LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0)) && (!defined(PCIE_LINK_STATE_L0S))
#include <linux/pci-aspm.h>
#endif

#include "edtdrv.h"
#include "edt_common.h"

#include "edt_lnx_alloc.h"

int edt_initialize_vars(Edt_Dev* edt_p);

/* Linux `HZ` define likely set to the value of `CONFIG_HZ`. */
#define MSEC_TICKS(x) (((x)*HZ) / 1000)

#ifdef TEST_SG_PAGEN

void test_sg_pagen()
{
  int pagen;

  for (pagen = 0; pagen < 1024; pagen += 64)

    edt_printk("get_sg_address_for_page: %d PAGE_SG_PAGE = %d PAGE_SG_OFFSET = %d\n",
               pagen,
               PAGE_SG_PAGE(pagen),
               PAGE_SG_OFFSET(pagen));
}

#endif

static void edt_copy_ring_buffer_to_user(Edt_Dev* edt_p, int waitbuf)
{
  int whichbuf;

  rbuf* p;

  whichbuf = edt_safe_bufnum(edt_p, waitbuf);

  EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("copy_ring_buffer_to_user %d\n", whichbuf));

  p = edt_p->m_Rbufs + whichbuf;

  if (p->os.map.page_addresses)
  {
    edt_copy_pages_to_user((void*)p->m_user_address, p->os.map.page_addresses, p->m_XferSize);
  }
}

static void edt_clear_sgtodo(Edt_Dev* edt_p, rbuf* p)
{
  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, (" ==> clear_sgtodo \n"));

  if (p->AllocatedSgList)
  {
    edt_free_kernel_mem(p->AllocatedSgList,
                        p->AllocatedEntries * sizeof(EdtMemRange),
                        "AllocatedSgList");
    p->AllocatedSgList = NULL;
  }

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("<== clear_sgtodo \n"));
}

static void edt_clear_rbuf(Edt_Dev* edt_p, rbuf* p)
{
  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("edt_clear_rbuf p %p\n", p));

  if (p->os.map.page_list_ptrs != NULL)
  {
    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("set_max_rbuf: calling mem_map_release for buf %p\n", p));
    edt_mem_map_release(&(p->os.map));
  }

  edt_clear_sg_entry(edt_p, p);

  edt_clear_sgtodo(edt_p, p);

  edt_mem_map_clear_sg(&(p->os.map));
}

// asmlinkage void* edt_get_dev_from_minor(int minor_id)
// {
//   Edt_Dev* dev = NULL;

//   /* check for validiity */

//   if (minor_id < 0 || minor_id >= edt_num_minor_ids)
//   {
//     EDTPRT(dev, DBG_ALL, TRL_FATAL, ("get_dev_from_minor: Invalid minor id %d\n", minor_id));

//     return NULL;
//   }

//   return edt_devices[minor_id];
// }

int edt_set_max_rbufs(Edt_Dev* edt_p, int newmax)
{
  rbuf* old_rbufs = edt_p->m_Rbufs;
  int oldmax = edt_p->m_MaxRbufs;

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, (" ==> set_max_rbufs newmax = %d\n", newmax));

  if (newmax > 0)
  {
    int numtocopy;

    int size = newmax * sizeof(rbuf);

    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("Set Max rbufs: Calling vmalloc for %d rbufs %d bytes\n", newmax, size));

    edt_p->m_Rbufs = edt_vmalloc(size, "m_Rbufs");

    edt_memset(edt_p->m_Rbufs, 0, newmax * sizeof(rbuf));

    if (old_rbufs)
    {
      numtocopy = oldmax;
      if (newmax < oldmax)
        numtocopy = newmax;
      edt_memcpy(edt_p->m_Rbufs, old_rbufs, numtocopy * sizeof(rbuf));
    }
  }
  else
    edt_p->m_Rbufs = NULL;

  edt_p->m_MaxRbufs = newmax;

  if (old_rbufs)
  {
    int i;
    int size = oldmax * sizeof(rbuf);

    for (i = newmax; i < oldmax; i++)
    {
      edt_clear_rbuf(edt_p, old_rbufs + i);
    }

    /* free old rbufs */

    edt_vfree(old_rbufs, size, "m_Rbufs");
  }

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("<== set_max_rbufs\n"));

  return 0;
}

void edt_clear_dma(Edt_Dev* edt_p)
{
  unsigned int j;

  edt_abortdma(edt_p);

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("edt_clear_dma current = %x dma_pid = %x\n", edt_get_current_pid(), edt_p->dma_pid));

  for (j = 0; j < edt_p->m_Numbufs; j++)
  {
    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("Release calling unmap_buffer for unit %d chan %d %d\n",
            edt_p->m_Unitno,
            edt_p->m_DmaChannel,
            j));
    edt_unmap_buffer(edt_p, j);
  }

  edt_p->dma_pid = 0;

  edt_p->m_Curbuf = 0;
  edt_p->m_Nxtbuf = 0;
  edt_p->m_Wait = 0;
  edt_p->m_Done = 0;
  edt_p->m_Started = 0;
  edt_p->m_Loaded = 0;
  edt_p->m_Todo = 0;
  edt_p->m_Freerun = 0;
  edt_p->m_Prepared = 0;
  edt_p->m_ReportedNumbufs = 0; /* added 5/15/2008 rwh */
}

/*
 * Data management: read and write
 */

static asmlinkage void edt_check_wakeup_dma(void* dev_id)
{
  Edt_Dev* wakeup_p = (Edt_Dev*)dev_id;

  if (wakeup_p && wakeup_p->m_needs_wakeup)
  {
    EdtTrace(wakeup_p, TR_WAKEUP);
    edt_wakeup_dma(wakeup_p);
  }
}

asmlinkage irqreturn_t edt_intr(int UNUSED_PARAM(irq), void* data)
{
  Edt_Dev* edt_p = (Edt_Dev*)data;

  if (NULL == edt_p)
    return IRQ_NONE;

  if (edt_test_intr(edt_p))
  {
    Edt_Dev* wakeup_p = NULL;

    EdtTrace(edt_p, TR_ISR_START);
    EdtTrace(edt_p, edt_p->m_nInterrupts);
    edt_p->m_nInterrupts++;

    edt_start_critical_src(edt_p, 32);
    edt_disable_interrupts(edt_p);
    wakeup_p = edt_DpcForIsr(edt_p);
    edt_enable_interrupts(edt_p);

    if (wakeup_p)
    {
      edt_check_wakeup_dma(wakeup_p);
      EdtTrace(edt_p, wakeup_p->m_DmaChannel);
    }
    else
      EdtTrace(edt_p, 0xffffffff);

    EdtTraceTimeStamp(edt_p, NULL, 0);
    edt_end_critical_src(edt_p, 32);

    EdtTrace(edt_p, TR_ISR_END);
    return IRQ_HANDLED;
  }
  else
  {
    EDTPRT(edt_p, DBG_ISR, TRL_VERBOSE, ("%s: Looks like it's not our interrupt\n", __func__));
    return IRQ_NONE;
  }
}

/*!
 * called by edt_DpcForIsr
 */
int edt_wakeup_dma(Edt_Dev* edt_p)
{
  /* Race condition prevention. October 2001 MCM */
  edt_p->os.m_wait_done = 1;

  edt_wake_up_interruptible(edt_p->os.wait);

  return 0;
}

asmlinkage int edt_private_mmap(void* dev_ptr, void* vma)
{
  Edt_Dev* edt_p = (Edt_Dev*)dev_ptr;

  return edt_mmap_vmap_address(edt_p, vma);
}

int edt_waitdma(Edt_Dev* edt_p)
{
  int rc = 0;
  int timeoutval;
  u_int rawtimeoutval = 0;

  int whichbuf;

  rbuf* p;

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_VERBOSE,
         ("edt_waitdma: waitirp %x m_Done %d m_Wait %d m_Todo %d Started %d Ints %d Copy %d\n",
          edt_p->m_WaitIrp,
          edt_p->m_Done,
          edt_p->m_Wait,
          edt_p->m_Todo,
          edt_p->m_Started,
          edt_p->m_Ints,
          edt_p->m_use_buffer_copy));

  edt_start_critical_src(edt_p, 17);

  EdtTrace(edt_p, TR_WAIT_START);
  EdtTrace(edt_p, edt_p->m_Wait);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  EdtTrace(edt_p, edt_p->m_Done);
  EdtTrace(edt_p, edt_p->m_Todo);
  EdtTrace(edt_p, 0x12345678);

  edt_p->m_WaitStatus = EDT_WAIT_OK;

  if ((edt_p->m_WaitIrp && edt_p->m_Done >= edt_p->m_Wait) ||
      (!edt_p->m_WaitIrp && edt_p->m_Done >= edt_p->m_Todo))

  {
    EdtTrace(edt_p, 0xbbbb0007);
    EDTPRT(edt_p, DBG_DMA, TRL_VERBOSE, ("edt_waitdma: already done\n"));
    EDTPRT(edt_p,
           DBG_DMA,
           TRL_VERBOSE,
           ("edt_waitdma: -- waitirp %x m_Done %d m_Wait %d m_Todo %d\n",
            edt_p->m_WaitIrp,
            edt_p->m_Done,
            edt_p->m_Wait,
            edt_p->m_Todo));

    EdtTrace(edt_p, TR_WAIT_END);
    EdtTrace(edt_p, rc);
    EdtTraceTimeStamp(edt_p, NULL, 0);
    EdtTrace(edt_p, edt_p->m_Done);
    EdtTrace(edt_p, edt_p->m_Todo);

    whichbuf = edt_safe_bufnum(edt_p, edt_p->m_Wait - 1);

    p = edt_p->m_Rbufs + whichbuf;

#ifdef PCD
    if (edt_p->m_AbortDma_OnIntr_errno == ECONNRESET)
    {
      EdtTrace(edt_p, 0xbbbb0035);
      edt_p->m_AbortDma_OnIntr_errno = 0;

      if (edt_p->m_AbortDma_OnIntr == 2)
      {
        /*
         * Flush PCI and intfc fifos and clear DNR bit for TI -- Mark
         * 5/28/03
         */
        EdtTrace(edt_p, 0xbbbb0004);

        /* Turn off the PCI fifo */
        edt_clear(edt_p, EDT_DMA_INTCFG_ONLY, (EDT_RFIFO_ENB >> 8) & 0xff);

        /* reset the interface fifos */
        edt_clear(edt_p, PCD_CMD, PCD_ENABLE);
        edt_get(edt_p, PCD_CMD);
        edt_or(edt_p, PCD_CMD, PCD_ENABLE);

        /* Turn on the PCI fifos, which flushes them */
        edt_or(edt_p, EDT_DMA_INTCFG_ONLY, (EDT_RFIFO_ENB >> 8) & 0xff);

        /* Turn off the DNR bit */
        edt_clear(edt_p, PCD_CONFIG, 0x20);
      }

      rc = -ECONNRESET;
      EdtTrace(edt_p, 0xbbbb0005);
    }
#endif

    edt_end_critical_src(edt_p, 17);

#ifdef PCD
    if (edt_p->m_AbortDma_OnIntr_errno != ECONNRESET)
#endif
    {
      if (edt_p->m_use_buffer_copy)
      {
        if (!p->m_WriteFlag)
          edt_copy_ring_buffer_to_user(edt_p, edt_p->m_Wait);
      }
    }

    return rc;
  }

  /* start timeout */

  if (edt_p->m_WriteFlag)
  {
    timeoutval = edt_p->m_WriteTimeout;
  }
  else
  {
    timeoutval = edt_p->m_ReadTimeout;
  }

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("edt_waitdma: Waiting w/ %s timeout %d\n",
          (edt_p->m_WriteFlag) ? "Write" : "Read",
          timeoutval));

  /* Race condition prevention.  October 2001 MCM */
  edt_p->os.m_wait_done = 0;

  edt_end_critical_src(edt_p, 17);

  if (timeoutval <= 0)
  {
    rc = edt_wait_event_interruptible(edt_p->os.wait, &edt_p->os.m_wait_done, 0);
  }
  else
  {
    rawtimeoutval = MSEC_TICKS(timeoutval);

    rc = edt_wait_event_interruptible_timeout(edt_p->os.wait,
                                              &edt_p->os.m_wait_done,
                                              0,
                                              rawtimeoutval);
  }

  if (rc != 0 && rc != -EINTR)
  {
    EdtTrace(edt_p, 0xbbbb0006);
    if (timeoutval)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_INFO,
             ("edt_waitdma: Apparent timeout rc = %d \nm_Done %d m_Wait %d m_Todo %d rawtimeoutval "
              "= %d timeoutval = %d\n",
              rc,
              edt_p->m_Done,
              edt_p->m_Wait,
              edt_p->m_Todo,
              rawtimeoutval,
              timeoutval));
      if (!edt_p->m_TimeoutOK)
      {
        edt_timeout(edt_p);
        edt_p->m_WaitStatus = EDT_WAIT_TIMEOUT;
      }
      else
      {
        edt_p->m_WaitStatus = EDT_WAIT_OK_TIMEOUT;
      }
    }
  }
  else
  {
    whichbuf = edt_safe_bufnum(edt_p, edt_p->m_Wait - 1);

    p = edt_p->m_Rbufs + whichbuf;

    if (edt_p->m_use_buffer_copy)
    {
      if (!edt_p->m_WriteFlag)
        edt_copy_ring_buffer_to_user(edt_p, edt_p->m_Wait);
    }

    edt_p->m_WaitStatus = EDT_WAIT_OK;
  }

#ifdef PCD
  if (edt_p->m_AbortDma_OnIntr_errno == ECONNRESET)
  {
    edt_p->m_AbortDma_OnIntr_errno = 0;

    if (edt_p->m_AbortDma_OnIntr == 2)
    {
      /*
       * Flush PCI and intfc fifos and clear DNR bit for TI -- Mark
       * 5/28/03
       */

      edt_start_critical_src(edt_p, 18);

      /* Turn off the PCI fifo */
      edt_clear(edt_p, EDT_DMA_INTCFG_ONLY, (EDT_RFIFO_ENB >> 8) & 0xff);

      /* reset the interface fifos */
      edt_clear(edt_p, PCD_CMD, PCD_ENABLE);
      edt_get(edt_p, PCD_CMD);
      edt_or(edt_p, PCD_CMD, PCD_ENABLE);

      /* Turn on the PCI fifos, which flushes them */
      edt_or(edt_p, EDT_DMA_INTCFG_ONLY, (EDT_RFIFO_ENB >> 8) & 0xff);

      /* Turn off the DNR bit */
      edt_clear(edt_p, PCD_CONFIG, 0x20);
      EdtTrace(edt_p, edt_get(edt_p, PCD_CONFIG));

      edt_end_critical_src(edt_p, 18);
    }

    rc = -ECONNRESET;
  }
  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("edt_waitdma: rc %d OnIntr %d\n", rc, edt_p->m_AbortDma_OnIntr));
#endif /* PCD */

  EdtTrace(edt_p, TR_WAIT_END);
  EdtTrace(edt_p, rc);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  EdtTrace(edt_p, edt_p->m_Done);
  EdtTrace(edt_p, edt_p->m_Todo);

  return (rc);
}

/*
 * The ioctl() implementation
 */
asmlinkage int edt_private_ioctl(void* dev_ptr,
                                 unsigned int cmd,
                                 unsigned long arg,
                                 int compat32_flag)
{
  edt_ioctl_struct eis;
  edt_ioctl_struct32* eis_from32 = NULL;
  Edt_Dev* edt_p = (Edt_Dev*)dev_ptr;
  u_char* argbuf;
  int use_useradd = 0;
  int retval = 0;
  uint64_t local_argbuf[8];

  /*
   * extract the type and number bitfields, and don't decode wrong cmds:
   * return EINVAL before verify_area()
   */

  if (_IOC_TYPE(cmd) != EDT_IOC_MAGIC)
  {
    EDTPRT(edt_p, DBG_IOCTLS, TRL_ERROR, ("IOCTL %08x not IOC_MAGIC\n", cmd));

    return -EINVAL;
  }
  if (_IOC_NR(cmd) > EDT_IOC_MAXNR)
  {
    EDTPRT(edt_p, DBG_IOCTLS, TRL_ERROR, ("IOCTL %08x nr > IOC_MAXNR\n", cmd));

    return -EINVAL;
  }

  /* EDT_NT_IOCTL32 is defined differently depending on the base platform,
   * but the action should be the same.  Part of the 64/32 compatibility
   * changes.  Mark  Jan'08.
   */
  if (cmd == EDT_NT_IOCTL32)
    cmd = EDT_NT_IOCTL;

  switch (cmd)
  {
  case EDT_NT_IOCTL: /* Set: arg points to the value */

    if (compat32_flag)
    {
      eis_from32 = (edt_ioctl_struct32*)arg;
      eis.device = eis_from32->device;
      eis.controlCode = eis_from32->controlCode;
      eis.inSize = eis_from32->inSize;
      eis.outSize = eis_from32->outSize;
      eis.inBuffer = (void*)(uintptr_t)eis_from32->inBuffer;
      eis.outBuffer = (void*)(uintptr_t)eis_from32->outBuffer;
    }
    else
    {
      edt_copy_from_user((void*)&eis, (void*)arg, sizeof(eis));
    }

    if ((eis.inSize <= sizeof(local_argbuf)) && (eis.outSize <= sizeof(local_argbuf)))
      argbuf = (u_char*)local_argbuf;
    else
      argbuf = (u_char*)edt_p->m_IoctlBuf;

    if (eis.inSize)
    {
      edt_copy_from_user(argbuf, eis.inBuffer, eis.inSize);
    }
    else if (eis.inBuffer)
    {
      argbuf = eis.inBuffer;
      use_useradd = 1;
    }

    retval = EdtIoctlCommon(edt_p,
                            eis.controlCode,
                            eis.inSize,
                            eis.outSize,
                            argbuf,
                            (u_int*)&eis.bytesReturned,
                            NULL);

    if (eis.bytesReturned && !use_useradd)
    {
      edt_copy_to_user(eis.outBuffer, argbuf, eis.bytesReturned);
    }

    if (eis.outSize)
    {
      if (compat32_flag)
      {
        eis_from32->bytesReturned = eis.bytesReturned;
      }
      else
      {
        edt_copy_to_user((void*)arg, (void*)&eis, sizeof(eis));
      }
    }

    break;

  default: /* redundant, as cmd was checked against
            * MAXNR */
    return -EINVAL;
  }

  return retval;
}

/*!
 * @brief Enable PCI device, memory-map BAR addresses to (kernel) virtual memory.
 *
 * Added support for BAR1 (Mem2) register access.  -mcm June 2011
 *
 * @param edt_p
 *     EDT device pointer. Must already have `os.m_pci_dev` field initialized.
 * @return
 *     Zero if successful. Negative error code otherwise.
 */
static int edt_map_pci_board(Edt_Dev* edt_p)
{
  int retval = 0;
  int bar_mask = 0;
  struct pci_dev* pcidev = edt_p->os.m_pci_dev;

  edt_p->bar0_mmap = NULL;
  edt_p->bar0_size = 0;
  edt_p->bar1_mmap = NULL;
  edt_p->bar1_size = 0;

  if (pcim_enable_device(pcidev))
  {
    pci_err(pcidev, "pci_enable_device() failed\n");
    return -EIO;
  }

  pci_set_master(pcidev);

  /* Disable ASPM L0 standby if we can */
  pci_disable_link_state(pcidev, PCIE_LINK_STATE_L0S);

  if (!(pci_resource_flags(pcidev, 0) & IORESOURCE_MEM))
  {
    pci_err(pcidev, "Incorrect BAR0 configuration.\n");
    return -ENODEV;
  }
  bar_mask = BIT(0);

  if (!(pci_resource_flags(pcidev, 1) & IORESOURCE_MEM))
    pci_warn(pcidev, "Incorrect BAR1 configuration. Skipping.\n");
  else
    bar_mask |= BIT(1);

  /* Map BARs */
  retval = pcim_iomap_regions(pcidev, bar_mask, KBUILD_MODNAME);
  if (retval)
  {
    pci_err(pcidev, "pci_iomap_regions() failed\n");
    return -ENODEV;
  }

  edt_p->bar0_mmap = pcim_iomap_table(pcidev)[0]; /* BAR0 */
  if (NULL == edt_p->bar0_mmap)
  {
    pci_err(pcidev, "pci_iomap_table() BAR0 failed\n");
    return -ENOMEM;
  }
  edt_p->bar0_size = pci_resource_len(pcidev, 0);

  pci_info(pcidev, "Mapped BAR0 [size=0x%lx]\n", edt_p->bar0_size);

  /* Optionally get BAR1 */
  if (bar_mask & BIT(1))
  {
    edt_p->bar1_mmap = pcim_iomap_table(pcidev)[1];
    if (NULL == edt_p->bar1_mmap)
    {
      pci_err(pcidev, "pci_iomap_table() BAR1 failed\n");
      return -ENOMEM;
    }
    edt_p->bar1_size = pci_resource_len(pcidev, 1);

    pci_info(pcidev, "Mapped BAR1 [size=0x%lx]\n", edt_p->bar1_size);
  }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
  retval = dma_set_mask(&pcidev->dev, DMA_BIT_MASK(32));
  if (retval)
  {
    pci_err(pcidev, "DMA mask set failed\n");
    return -ENODEV;
  }
  retval = dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(32));
  if (retval)
  {
    pci_err(pcidev, "Consistent DMA mask set failed\n");
    return -ENODEV;
  }
#else
  retval = pci_set_dma_mask(pcidev, DMA_BIT_MASK(32));
  if (retval)
  {
    pci_err(pcidev, "DMA mask set failed\n");
    return -ENODEV;
  }
  retval = pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(32));
  if (retval)
  {
    pci_err(pcidev, "Consistent DMA mask set failed\n");
    return -ENODEV;
  }
#endif

  pci_dbg(pcidev, "DMA mask set succeeded\n");

  return retval;
}

static int edt_init_channel_struct(struct pci_dev* pci_dev,
                                   const struct pci_device_id* pci_id,
                                   Edt_Dev* edt_p,
                                   int channel,
                                   Edt_Dev* chan0_p,
                                   int unit)
{
  int result;
  int i;

  edt_p->os.m_pci_dev = pci_dev;
  edt_p->devid = pci_id->device;

  edt_p->m_Unitno = unit;
  edt_p->m_DmaChannel = channel;

  // abort_lock is unique to a channel
  edt_p->os.abort_lock = edt_create_spinlock(edt_p);
  if (!edt_p->os.abort_lock)
    return -ENOMEM;

  if (0 == channel)
  {
    edt_p->m_chan0_p = edt_p;
    edt_p->chan_p[0] = edt_p;

    result = edt_map_pci_board(edt_p);
    if (0 != result)
    {
      pr_warn("Problem setting up device memory ranges.\n");
      return result;
    }

    edt_p->os.crit_lock = edt_create_spinlock(edt_p);
    if (!edt_p->os.crit_lock)
      return -ENOMEM;

    edt_p->os.register_lock = edt_create_semaphore(edt_p);
    if (!edt_p->os.register_lock)
      return -ENOMEM;
  }
  else
  {
    edt_p->m_chan0_p = chan0_p;

    edt_p->chan_p[0] = chan0_p;

    edt_p->bar0_mmap = chan0_p->bar0_mmap;
    edt_p->bar1_mmap = chan0_p->bar1_mmap;
    edt_p->bar1_size = chan0_p->bar1_size;

    // crit_lock and register_lock are shared across all channels.
    edt_p->os.crit_lock = chan0_p->os.crit_lock;
    edt_p->os.register_lock = chan0_p->os.register_lock;
  }

  result = edt_initialize_vars(edt_p);
  if (0 != result)
  {
    pr_warn("Initialize vars failed on ch %d\n", edt_p->m_DmaChannel);
    return result;
  }

  edt_open_wait_handle(&edt_p->os.serial_wait);
  edt_open_wait_handle(&edt_p->os.wait);

  for (i = 0; i < EDT_MAX_EVENT_TYPES; i++)
    edt_open_wait_handle(&edt_p->os.event_wait[i]);

#if IS_ENABLED(CONFIG_TPE3)
  edt_p->m_kernel_buffers = 1;
#endif

  return 0;
}

Edt_Dev* edt_create_device_structures(struct pci_dev* pcidev,
                                      const struct pci_device_id* pci_id,
                                      unsigned int unitno)
{
  int nchannels;                    /* Total/max number of channels this EDT PCI device supports. */
  int ch_idx;                       /* (dma) channel index */
  Edt_Dev** device_channels = NULL; /* List of allocated channel structures */
  Edt_Dev* device_ch0 = NULL;       /* Pointer of first channel structure. */
  int retval = -1;

  nchannels = edt_pcidev_num_dma_channels(pci_id->device);
  if (nchannels < 0)
  {
    pci_err(pcidev, "Could not get DMA ch count for device type 0x%04x\n", pci_id->device);
    return NULL;
  }
  if (nchannels > MAX_MINORS)
  {
    pci_err(pcidev, "Driver supports %d ch max, device has %d\n", MAX_MINORS, nchannels);
    return NULL;
  }
  /* Create at least 1 channel. */
  if (nchannels == 0)
    nchannels = 1;

  /* Allocate array of pointers */
  device_channels = (Edt_Dev**)devm_kzalloc(&pcidev->dev, nchannels * sizeof(Edt_Dev*), GFP_KERNEL);
  if (!device_channels)
  {
    pr_err("Could not alloc %d channel array\n", nchannels);
    return NULL;
  }

  /* Allocate channel structs */
  for (ch_idx = 0; ch_idx < nchannels; ch_idx++)
  {
    device_channels[ch_idx] = (Edt_Dev*)devm_kzalloc(&pcidev->dev, sizeof(Edt_Dev), GFP_KERNEL);
    if (NULL == device_channels[ch_idx])
    {
      pr_err("Could not alloc channel %d struct\n", ch_idx);
      return NULL;
    }
    memset(device_channels[ch_idx], 0, sizeof(Edt_Dev));
  }
  device_ch0 = device_channels[0];

  pci_info(pcidev, "Setting up %d channels for unit %d\n", nchannels, unitno);

  for (ch_idx = 0; ch_idx < nchannels; ch_idx++)
  {
    pr_debug("Init ch %d struct\n", ch_idx);
    device_channels[ch_idx]->m_TraceLevel = 0;
    device_channels[ch_idx]->m_DebugLevel = TRL_WARN;

    // the debug mask is set from a lower debug level up to EDT_DEBUG_LEVELS - so
    // all debug levels >= current will display these errors
    // All masks should show at debug level TRL_ERROR
    edt_set_debug_mask(device_channels[ch_idx], TRL_ERROR, DBG_ALL);

    // we want info messages except for reg set / get
    edt_set_debug_mask(device_channels[ch_idx], TRL_WARN, DBG_DMA_SETUP);

    EDTPRT(device_channels[ch_idx],
           DBG_INIT,
           TRL_INFO,
           ("init_module: still Going to set max rbufs for %d\n",
            edt_default_rbufs_from_type(pci_id->device)));

    // This should always be set. Mark and Steve, June 2014
    edt_set_kernel_buffers(device_channels[ch_idx], EDT_MMAP_KBUFS);

    edt_set_max_rbufs(device_channels[ch_idx], edt_default_rbufs_from_type(pci_id->device));

    if (device_channels[ch_idx] != device_ch0)
    {
      device_channels[ch_idx]->m_serial = device_ch0->m_serial;
    }

    retval = edt_init_channel_struct(pcidev,
                                     pci_id,
                                     device_channels[ch_idx],
                                     ch_idx,
                                     device_ch0,
                                     unitno);
    if (0 != retval)
    {
      pr_err("Could not initialize ch %d struct\n", ch_idx);
      return NULL;
    }

    device_channels[ch_idx]->m_chancnt = nchannels;
  }

  /* cross-reference channel pointers */
  for (ch_idx = 0; ch_idx < nchannels; ch_idx++)
  {
    int ch_idx2;
    for (ch_idx2 = 0; ch_idx2 < nchannels; ch_idx2++)
    {
      device_channels[ch_idx]->chan_p[ch_idx2] = device_channels[ch_idx2];
    }
  }

  return device_ch0;
}

static int edt_find_buffer(Edt_Dev* edt_p, int buf, u_int sgaddr, u_int dmastat, int onlyactive)
{
  u_int sgfirst;
  u_int sglast;
  int found = 0;
  unsigned int idx;
  int dmadone;
  int sgdone;
  int sgempty;
  int startbuf;
  EdtPCISg* sgPtr;
  rbuf* p;
  int on_last_pg;

  dmadone = (dmastat & EDT_MN_DMA_DONE) ? 1 : 0;
  sgdone = (dmastat & EDT_DMA_DONE) ? 1 : 0;
  sgempty = (dmastat & EDT_DMA_RDY) ? 1 : 0;
  on_last_pg = (dmastat & (EDT_CURPG_INT | EDT_NXTPG_INT | EDT_PG_INT)) ? 1 : 0;
  startbuf = buf;
  for (idx = 0; idx < edt_p->m_Numbufs; idx++)
  {
    p = &edt_p->m_Rbufs[buf];
    if (!p->m_SglistSize)
      break;
    sgPtr = edt_get_sg_address_for_page(edt_p, p, 0);
    if (!sgPtr)
      break;
    EDTPRT(edt_p,
           DBG_DMA,
           TRL_INFO,
           ("find_buffer:  debug buf %d active %d used %x size %x ptr %p\n",
            buf,
            p->m_Active,
            p->m_SgUsed,
            p->m_SglistSize,
            sgPtr));

    if (!onlyactive || p->m_Active)
    {
      unsigned int i;

      EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("find_buffer:  entrys %d\n", p->AllocatedEntries));
      for (i = 0; i < p->AllocatedEntries; i++)
      {
        if (p->AllocatedSgList[i].size == 0)
          break;
        sgfirst = (u_int)(p->AllocatedSgList[i].physaddr);
        sglast = sgfirst + p->AllocatedSgList[i].size;
        if (!onlyactive)
          EDTPRT(edt_p,
                 DBG_DMA,
                 TRL_INFO,
                 ("find_buffer:  buf %d add %x first %x last %x\n", buf, sgaddr, sgfirst, sglast));
        if (sgaddr >= sgfirst && sgaddr <= sglast)
        {
          if (sgaddr == sglast)
          {
            found = 1;
            EdtTrace(edt_p, 0xdead1122);
            EdtTrace(edt_p, sgdone);
            if (!sgdone)
            {
              EDTPRT(edt_p,
                     DBG_DMA,
                     TRL_INFO,
                     ("find_buffer:  sgaddr == sglast but sgdone not set\n"));
            }
          }
          else
          {
            found = 1;
          }
          break;
        }
        /* sg may be on next buf with main dma on current */
        else
        {
          if (on_last_pg)
          {
            found = 1;
            break;
          }
        }
      }
      if (found)
        break;
    }

    buf++;
    if ((unsigned int)buf == edt_p->m_Numbufs)
      buf = 0;
  }
  if (!found)
    EDTPRT(edt_p,
           DBG_DMA,
           TRL_INFO,
           ("find_buffer:  startbuf %d buf %d found %d sgaddr %x done %d sgdone %d sgempty %d\n",
            startbuf,
            buf,
            found,
            sgaddr,
            dmadone,
            sgdone,
            sgempty));
  if (found)
    return (buf);
  else
    return (-1);
}

int TransferCountDebug(Edt_Dev* edt_p, int buf, int* endptr, int* sizeptr)
{
  u_int sgaddr = edt_get(edt_p, EDT_SG_CUR_ADDR);
  u_int dmastat = edt_get(edt_p, EDT_SG_NXT_CNT);

  *endptr = 0;
  *sizeptr = 0;
  return (edt_find_buffer(edt_p, buf, sgaddr, dmastat, 1));
}

#if defined(PCD)
static int edt_scan_sg_for_all_addr(Edt_Dev* edt_p, int buf, u_int dmaAddr)
{
  rbuf* p = &edt_p->m_Rbufs[buf];
  u_int dmaCount = 0;
  unsigned int i;
  EdtPCISg* sgPtr;
  int found = 0;
  u_int listAddr = 0;
  u_int listCount = 0;

  for (i = 0; i < p->m_SgUsed / 8; i++)
  {
    /* find page in sglist */
    if (!p->m_SglistSize)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_ERROR,
             ("edt_scan_sg_for_all_addr:  zero SgListSize with i %d used %d\n", i, p->m_SgUsed));
      return -1;
    }
    sgPtr = edt_get_sg_address_for_page(edt_p, p, i);

    if (!sgPtr)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_INFO,
             ("edt_scan_sg_for_all_addr:  null sgPtr with i %d used %d\n", i, p->m_SgUsed));
      return -1;
    }

    listAddr = BUS_DATA(sgPtr[0].busaddr);
    listCount = BUS_DATA(sgPtr[0].count) & 0xffff;

    if (dmaAddr >= listAddr && dmaAddr < listAddr + listCount)
    {
      dmaCount += (dmaAddr - listAddr);
      found = 1;
      EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("Found %08x in place at %x\n", dmaAddr, dmaCount));
    }
    else if (dmaAddr == (listAddr + listCount))
    {
      /* we are at end of this sg chunk */
      /* but without wrap */
      /* if don't catch will show full buf */
      dmaCount += listCount;
      found = 1;
      EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("Found %08x at end at %x\n", dmaAddr, dmaCount));
    }
    else
    {
      dmaCount += listCount;
    }
  }

  if (found)
    return dmaCount;
  else
    return -1;
}
#endif /* PCD */

#if defined(PCD)
static int edt_scan_sg_for_addr(Edt_Dev* edt_p,
                                int buf,
                                u_int dmaAddr,
                                u_int at_end,
                                u_int* found_at_end)
{
  rbuf* p = &edt_p->m_Rbufs[buf];
  u_int dmaCount = 0;
  unsigned int i;
  EdtPCISg* sgPtr;
  int found = 0;
  u_int listAddr = 0;
  u_int listCount = 0;

  for (i = 0; i < p->m_SgUsed / 8; i++)
  {
    /* find page in sglist */
    if (!p->m_SglistSize)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_ERROR,
             ("edt_scan_sg_for_addr:  zero SgListSize with i %d used %d\n", i, p->m_SgUsed));
      return -1;
    }
    sgPtr = edt_get_sg_address_for_page(edt_p, p, i);

    if (!sgPtr)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_ERROR,
             ("edt_scan_sg_for_addr:  null sgPtr with i %d used %d\n", i, p->m_SgUsed));
      return -1;
    }

    listAddr = BUS_DATA(sgPtr[0].busaddr);
    listCount = BUS_DATA(sgPtr[0].count) & 0xffff;

    if (!at_end && (dmaAddr >= listAddr && dmaAddr < listAddr + listCount))
    {
      dmaCount += (dmaAddr - listAddr);
      found = 1;
      break;
    }
    else if (at_end && (dmaAddr == listAddr + listCount))
    {
      /* we are at end of this sg chunk */
      /* but without wrap */
      /* if don't catch will show full buf */
      dmaCount += listCount;
      found = 1;
      *found_at_end = 1;
      break;
    }
    else
    {
      dmaCount += listCount;
    }
  }

  if (found)
    return dmaCount;
  else
    return -1;
}
#endif /* PCD */

#if defined(PCD)
static int last_offset = 0;
static int last_buffer = 0;
static u_int xfer_counter = 0;

static int TransferCountCommon_LX(Edt_Dev* edt_p,
                                  int buf,
                                  u_int* endptr,
                                  u_int* sizeptr,
                                  int UNUSED_PARAM(dsbldma),
                                  int UNUSED_PARAM(onlyactive))
{
  u_int dmaAddr;
  u_int dmaAddr2;
  u_int i;
  u_int startbuf;
  u_int testbuf;
  int offset;
  int count1;
  int count2;
  int at_end = 0;
  u_int found_at_end = 0;

  /* just read dmaaddr */

  xfer_counter++;
  count1 = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
  dmaAddr = edt_get(edt_p, EDT_DMA_CUR_ADDR);
  count2 = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
  dmaAddr2 = edt_get(edt_p, EDT_DMA_CUR_ADDR);

  offset = dmaAddr & 0xfff;

  /* Try and determine the state - are we at end or beginning of a page */

  if (count2 == 0 && offset == 0)
  {
    /* at end of an entry */
    at_end = 1;
  }
  else if (count1 == 0 && count2 != 0)
  {
    /* loaded new entry */
    dmaAddr = edt_get(edt_p, EDT_DMA_CUR_ADDR);
  }
  else if (count2 > count1 && offset == 0)
  {
    /* if dmaAddr2 is same page as dmaAddr not at end */
    if (dmaAddr2 - dmaAddr < 0x1000)
      at_end = 0;
    else
      at_end = 1;
  }

  startbuf = edt_p->m_Curbuf;
  found_at_end = 0;

  for (i = 0; i < edt_p->m_Numbufs; i++)
  {
    testbuf = (startbuf + i) % edt_p->m_Numbufs;

    offset = edt_scan_sg_for_addr(edt_p, testbuf, dmaAddr, at_end, &found_at_end);

    if (offset >= 0)
    {
      buf = testbuf;
      *sizeptr = offset;
      *endptr = 1;

      if (buf == last_buffer && offset < last_offset)
      {
        EDTPRT(edt_p,
               DBG_DMA,
               TRL_ERROR,
               ("apparent fail buf = %d addr %08x count1 %x count2 %x at_end %d\n",
                buf,
                dmaAddr,
                count1,
                count2,
                at_end));
        EDTPRT(edt_p,
               DBG_DMA,
               TRL_ERROR,
               ("%d: offset %x last %x found at end %d\n",
                xfer_counter,
                offset,
                last_offset,
                found_at_end));
        edt_scan_sg_for_all_addr(edt_p, buf, dmaAddr);
      }
      last_buffer = buf;
      last_offset = offset;
      return buf;
    }
  }

  EDTPRT(edt_p,
         DBG_DMA,
         TRL_ERROR,
         ("Unable to find addr %x -> %x count1 %x count2 %x at_end %d\n",
          dmaAddr,
          dmaAddr2,
          count1,
          count2,
          at_end));

  *sizeptr = 0;
  *endptr = 0;
  return -1;
}
#endif /* PCD */

int TransferCountCommon(Edt_Dev* edt_p,
                        int buf,
                        u_int* endptr,
                        u_int* sizeptr,
                        int dsbldma,
                        int onlyactive)
{
  u_int dmaCount;
  u_int dmaAddr;
  u_int dmaAddrChk;
  u_int listAddr = 0;
  u_int listCount = 0;
  u_int sgaddr;
  u_int i;
  int savebuf;
  int found;
  int foundend;
  int save_endbuf;
  u_int curcount;
  u_int curcountChk;
  u_int dmastat;
  EdtPCISg* sgPtr;
  rbuf* p;

  /* we can't read addr and cnt atomically - disable dma during read */
  /* not available on all boards and danger of interrupt after DMA_STOP */
  /* so only do when ioctl for bufbytecount */
  /* element to the next, we can just return with full count for that sg list */

#if defined(PCD)
  if (EDT_DEVID_IS_LX(edt_p->devid))
    return TransferCountCommon_LX(edt_p, buf, endptr, sizeptr, dsbldma, onlyactive);
#endif

  if (dsbldma)
  {
    int ii = 0;

    edt_set(edt_p, EDT_DMA_STOP, 1);
    curcount = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
    if (curcount == 0)
    {
      edt_set(edt_p, EDT_DMA_STOP, 0);
      while (ii < 10)
      {
        ii++;
        curcount = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
        if (curcount)
          break;
      }
      edt_set(edt_p, EDT_DMA_STOP, 1);
      curcount = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
    }
    dmaAddr = edt_get(edt_p, EDT_DMA_CUR_ADDR);
    sgaddr = edt_get(edt_p, EDT_SG_CUR_ADDR);
    dmastat = edt_get(edt_p, EDT_SG_NXT_CNT);
    edt_set(edt_p, EDT_DMA_STOP, 0);
  }
  else
  {
    dmaAddrChk = edt_get(edt_p, EDT_DMA_CUR_ADDR);
    curcountChk = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;

    dmaAddr = edt_get(edt_p, EDT_DMA_CUR_ADDR);
    curcount = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
    if (dmaAddr != dmaAddrChk || curcount != curcountChk)
    {
      dmaAddr = edt_get(edt_p, EDT_DMA_CUR_ADDR);
      curcount = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
    }
    if (listCount == curcount || curcount == 0)
    {
      dmaAddr = edt_get(edt_p, EDT_DMA_CUR_ADDR);
      curcount = edt_get(edt_p, EDT_DMA_CUR_CNT) & 0xffff;
    }
    sgaddr = edt_get(edt_p, EDT_SG_CUR_ADDR);
    dmastat = edt_get(edt_p, EDT_SG_NXT_CNT);
  }

  savebuf = buf;
  buf = edt_find_buffer(edt_p, buf, sgaddr, dmastat, onlyactive);

  if (buf < 0)
  {
    *endptr = 0;
    *sizeptr = 0;
    return buf;
  }
  else if (dmastat & EDT_PG_INT)
  {
    p = &edt_p->m_Rbufs[buf];
    /* since we are in critical section - int could be pending */
    *endptr = 1;
    *sizeptr = p->m_XferSize;
    EdtTrace(edt_p, 0xdead1122);
    EdtTrace(edt_p, buf);
    EdtTrace(edt_p, dmastat);
    EdtTrace(edt_p, p->m_XferSize);
    EdtTrace(edt_p, edt_get(edt_p, EDT_DMA_STATUS));
    return buf;
  }
  if (buf != savebuf)
  {
    EdtTrace(edt_p, 0xdead2222);
    EdtTrace(edt_p, buf);
    EdtTrace(edt_p, dmastat);
    /* this is tricky because the new sg addr could be loaded without */
    /* the sgdma happening yet - check if always sgdone == 0 here */
    /* since async anyway - assume no data in new buffer */
    *endptr = 0;
    *sizeptr = 0;
    return buf;
  }

  found = 0;
  foundend = 0;
  dmaCount = 0;
  save_endbuf = 0;

  p = &edt_p->m_Rbufs[buf];
  EDTPRT(edt_p,
         DBG_DMA,
         TRL_INFO,
         ("TransferCountCommon:  debug buf %d active %d used %x size %x\n",
          buf,
          p->m_Active,
          p->m_SgUsed,
          p->m_SglistSize));
  dmaCount = 0;
  for (i = 0; i < p->m_SgUsed / 8; i++)
  {
    /* find page in sglist */
    if (!p->m_SglistSize)
    {
      break;
    }
    sgPtr = edt_get_sg_address_for_page(edt_p, p, i);

    if (!sgPtr)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_INFO,
             ("TransferCountCommon:  null sgPtr with i %d used %d\n", i, p->m_SgUsed));
      break;
    }

    listAddr = BUS_DATA(sgPtr[0].busaddr);
    listCount = BUS_DATA(sgPtr[0].count) & 0xffff;
    if (curcount)
    {
      if (dmaAddr >= listAddr && dmaAddr < listAddr + listCount)
      {
        dmaCount += (dmaAddr - listAddr);
        found = 1;
        break;
      }
      else
      {
        dmaCount += listCount;
      }
    }
    else if (dmaAddr == listAddr + listCount)
    {
      /* we are at end of this sg chunk */
      /* but without wrap */
      /* if don't catch will show full buf */
      dmaCount += listCount;
      found = 1;
      break;
    }
    else
    {
      /* this takes care of wrap  */
      /* but also have problem when count is */
      /* reloaded but addr is not, or atleast */
      /* we read when both are being updated */
      /* problem - could also be earlier addr */
      if (dmaAddr == ((listAddr & 0xffff0000) | (((listAddr & 0xffff) + listCount) & 0xffff)))
      {
        /* check for wrap when done */
        /* if not done - would reload in a couple pci clocks */
        dmaCount += listCount;
        foundend = 1;
        save_endbuf = buf;
        /* this should set found */
        found = 1;
        break;
      }
      else
        dmaCount += listCount;
    }
  }
  EdtTrace(edt_p, 0xdead3333);
  EdtTrace(edt_p, buf);
  EdtTrace(edt_p, dmastat);
  EdtTrace(edt_p, found);
  EdtTrace(edt_p, dmaCount);
  EdtTrace(edt_p, edt_get(edt_p, EDT_DMA_STATUS));

  *endptr = foundend;
  *sizeptr = dmaCount;
  if (found)
  {
    return (buf);
  }
  else if (foundend)
  {
    return (save_endbuf);
  }
  else
  {
    return (-1);
  }
}

u_int TransferBufCount(Edt_Dev* edt_p, u_int* buf_p)
{
  u_int dmaCount;
  u_int buf;
  u_int endptr;

  buf = TransferCountCommon(edt_p, edt_p->m_Curbuf, &endptr, &dmaCount, 1, 0);
  *buf_p = buf;
  return (dmaCount);
}

u_int TransferCount(Edt_Dev* edt_p)
{
  u_int dmaCount;
  u_int endptr;

  TransferCountCommon(edt_p, edt_p->m_Curbuf, &endptr, &dmaCount, 0, 0);
  return (dmaCount);
}

int edt_wait_serial(Edt_Dev* edt_p, int msecs)
{
  int retval = 0;

  int waitval = MSEC_TICKS(msecs);

  retval = edt_wait_event_interruptible_timeout(edt_p->os.serial_wait,
                                                &edt_p->os.m_serial_wait_done,
                                                0,
                                                waitval);
  edt_start_critical_src(edt_p, 14);
  if (retval == 0)
    edt_p->os.m_serial_wait_done = 0;
  edt_end_critical_src(edt_p, 14);

  return retval;
}

void edt_wakeup_serial(Edt_Dev* edt_p)
{
  edt_p->os.m_serial_wait_done = 1;

  edt_wake_up(edt_p->os.serial_wait);
}

void edt_broadcast_event(Edt_Dev* edt_p, int nEventId)
{
  EdtTrace(edt_p, TR_SETEVENT);
  if (edt_p->m_EventRefs[nEventId])
  {
    edt_p->m_event_flags[nEventId] = 1;
    ++edt_p->m_EventCounters[nEventId];
    edt_wake_up(edt_p->os.event_wait[nEventId]);
  }
}

int edt_wait_on_event(Edt_Dev* edt_p, int nEventId)
{
  int retval = 0;

  edt_p->m_event_flags[nEventId] = 0;
  do
  {
    retval = edt_wait_event_interruptible(edt_p->os.event_wait[nEventId],
                                          &edt_p->m_event_flags[nEventId],
                                          0);

  } while (retval >= 0 && edt_p->m_event_flags[nEventId] == 0 &&
           edt_p->m_event_disable[nEventId] == 0);

  edt_p->m_event_flags[nEventId] = 0;

  return retval;
}

// #define TIME_LOCKS

void edt_start_critical_src(Edt_Dev* chan_p, int src)
{
  Edt_Dev* edt_p = chan_p->m_chan0_p;
#ifdef TIME_LOCKS
  uint64_t t;
#endif

  preempt_disable(); /* added 03/2015 RWH */

  if (edt_p->m_TraceLevel & 2)
  {
    EdtTrace(edt_p, TR_LOCK);
    EdtTrace(edt_p, src);
  }

#ifdef TIME_LOCKS
  edt_p->m_wait_start_time = edt_drv_timestamp();
#endif

  edt_start_intr_critical(edt_p->os.crit_lock, &edt_p->os.cpu_flags);

#ifdef TIME_LOCKS
  edt_p->m_lock_start_time = edt_drv_timestamp();
  t = edt_p->m_lock_start_time - edt_p->m_wait_start_time;
  if (t < 1000000)
    edt_p->m_total_wait_time += t;
  else
  {
    EDTPRT(edt_p, DBG_LOCKS, TRL_INFO, ("Time > 1\n", t));
  }
#endif

  edt_p->m_nLocks++;

  edt_p->m_lock_array[src]++;

  /* NOTE: removed code here re. m_InCritical; see git comment for 5.4.8.4 RWH 03/2015*/
}

void edt_end_critical_src(Edt_Dev* chan_p, int src)
{
  Edt_Dev* edt_p = chan_p->m_chan0_p;

  /* NOTE: removed code here re. m_InCritical; see git comment for 5.4.8.4 RWH 03/2015*/

#ifdef TIME_LOCKS
  edt_p->m_total_lock_time += edt_drv_timestamp() - edt_p->m_lock_start_time;
#endif

  edt_end_intr_critical(edt_p->os.crit_lock, edt_p->os.cpu_flags);

  if (edt_p->m_TraceLevel & 2)
  {
    EdtTrace(edt_p, TR_UNLOCK);
    EdtTrace(edt_p, src);
  }

  preempt_enable(); /* added 03/2015 RWH */
}

/* Exported serial routines ******************/

void edt_get_hres_time(Edt_Dev* UNUSED_PARAM(edt_p), edt_timespec_t* ts)
{
// Reason For Version Specification: https://lwn.net/Articles/740063/
//   supplemental: https://lore.kernel.org/linux-fsdevel/43741269.9cZ5YESnMi@kreacher/
//   Chose 5.3.0: Found via error compiling for Fedora 33 (k: 5.8.6) [gcc 10]; gcc 10 is Fedora 31+
//   (k: 5.3).
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 3, 0)
  struct timeval tval;
#else
  struct __kernel_old_timeval tval;
#endif

  edt_do_gettimeofday(&tval);

  ts->tv_sec = tval.tv_sec;
  ts->tv_nsec = tval.tv_usec * 1000;  // convert to nanoseconds
}

void edt_lock_abort(Edt_Dev* edt_p)
{
  edt_start_intr_critical(edt_p->os.abort_lock, &edt_p->os.cpu_flags);
}

void edt_unlock_abort(Edt_Dev* edt_p)
{
  edt_end_intr_critical(edt_p->os.abort_lock, edt_p->os.cpu_flags);
}

int edt_trylock_registers(Edt_Dev* edt_p)
{
  int rc = edt_trylock_semaphore(edt_p->os.register_lock);
  if (rc == 0)
    edt_p->os.register_locker = edt_get_current_pid();
  return rc;
}

int edt_lock_registers(Edt_Dev* edt_p)
{
  int ret = edt_lock_semaphore(edt_p->os.register_lock);
  edt_p->os.register_locker = edt_get_current_pid();
  return ret;
}

int edt_unlock_registers(Edt_Dev* edt_p)
{
  edt_unlock_semaphore(edt_p->os.register_lock);
  edt_p->os.register_locker = 0;
  return 0;
}
