/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#include "edtdrv.h"

#include "edt_lnx_alloc.h"

// #define MAX_SG_SIZE 0x1000

// This should normally be 8 - 8 pages per full sg block, 4K sg elements per block
// #define MAX_SG_PAGES (MAX_SG_SIZE / PAGE_SIZE)

// Convert a page number to its place in AllocatedList array
// #define SG_PAGE_SHIFT (PAGE_SHIFT - 3)

// #define PAGE_SG_PAGE(pn)   (pn >> SG_PAGE_SHIFT)
// #define MAX_SG_MASK        ((1 << SG_PAGE_SHIFT) - 1)
// #define PAGE_SG_OFFSET(pn) ((pn & MAX_SG_MASK))

void dump_mem_mapping(char* name, EdtMemMapping* map)
{
  edt_printk("**********  MAP  ***************\n");
  edt_printk("%-16s: %p\n", name, map);
  if (!map)
    return;
  edt_printk("%-20s: %08lx\n", "size", map->size);

  edt_printk("%-20s: %p\n", "page_addresses", map->page_addresses);
  edt_printk("%-20s: %p\n", "vma", map->vma);
  edt_printk("%-20s: %04x\n", "npages", map->npages);
  edt_printk("%-20s: %04x\n", "offset", map->offset);
  edt_printk("%-20s: %04x\n", "sg_entries_needed", map->sg_entries_needed);
  edt_printk("%-20s: %02x\n", "user_mem", map->user_mem);
  edt_printk("%-20s: %02x\n", "own_pages", map->own_pages);
  edt_printk("%-20s: %02x\n", "is_mapped", map->is_mapped);
  edt_printk("%-20s: %02x\n", "rw", map->rw);
  edt_printk("%-20s: %02x\n", "is_dma_mapped", map->is_dma_mapped);
  edt_printk("%-20s: %p\n", "os_sglist", map->os_sglist);
  edt_printk("%-20s: %d\n", "sglist_count", map->sglist_count);
  edt_printk("%-20s: %d\n", "pgsize", map->pgsize);
  edt_printk("%-20s: %d\n", "sgsize", map->sgsize);
  edt_printk("%-20s: %p\n", "pci_dev", map->pci_dev);
  edt_printk("%-20s: %p\n", "sg_list_ptrs", map->sg_list_ptrs);
  edt_printk("%-20s: %p\n", "chunks", map->chunks);
  edt_printk("%-20s: %p\n", "sg_list_ptrs", map->sg_list_ptrs);
  edt_printk("%-20s: %d\n", "nalloc_chunks", map->nalloc_chunks);
  edt_printk("%-20s: %d\n", "nchunks", map->nchunks);
  edt_printk("*********************************\n");
}

/*!
 * @brief Kernel-print the status of a ring buffer.
 *
 * @param name
 *     cstr name of this ring buffer
 * @param p
 *     The ring buffer pointer to print.
 */
static void dump_rbuf(const char* const name, rbuf* p)
{
  edt_printk("%-16s: %p\n", name, p);
  if (!p)
    return;

  dump_mem_mapping("map", &p->os.map);

  edt_printk("%-20s: %08x \n", "m_XferSize", p->m_XferSize);
  edt_printk("%-20s: %08x \n", "SgEntries", p->SgEntries);
  edt_printk("%-20s: %p \n", "SgChunksList", p->SgChunksList);
  edt_printk("%-20s: %08x \n", "m_nChunks", p->m_nChunks);
  edt_printk("%-20s: %08x \n", "InitialChunks", p->InitialChunks);
  edt_printk("%-20s: %08x \n", "AllocatedEntries", p->AllocatedEntries);
  edt_printk("%-20s: %p \n", "AllocatedSgList", p->AllocatedSgList);
  edt_printk("%-20s: %08x \n", "m_SglistSize", p->m_SglistSize);
  edt_printk("%-20s: %08x \n", "m_SgUsed", p->m_SgUsed);
  edt_printk("%-20s: %08x \n", "m_Idx", p->m_Idx);

  edt_printk("%-20s: %p \n", "m_user_address", p->m_user_address);

  edt_printk("%-20s: %08x \n", "m_WriteFlag", p->m_WriteFlag);
  edt_printk("%-20s: %08x \n", "m_FirstAddr", p->m_FirstAddr);
  edt_printk("%-20s: %08x \n", "m_LinesXferred", p->m_LinesXferred);
  edt_printk("%-20s: %p \n", "m_SgVirtAddr", p->m_SgVirtAddr);
  edt_printk("%-20s: %08x \n", "is_adjusted", p->is_adjusted);

  edt_printk("%-20s: %08x \n", "m_SgAdjustedEntry", p->m_SgAdjustedEntry);
  edt_printk("%-20s: %08x \n", "m_SgAdjustedSize", p->m_SgAdjustedSize);
  edt_printk("%-20s: %08x \n", "m_Active", p->m_Active);
  edt_printk("%-20s: %08x \n", "SgArrayAdjustedChunk", p->SgArrayAdjustedChunk);
  edt_printk("%-20s: %08x \n", "SgArrayAdjustedSize", p->SgArrayAdjustedSize);
  edt_printk("%-20s: %08x \n", "AllocatedChunks", p->AllocatedChunks);
  edt_printk("%-20s: %08x \n", "SgAllocSize", p->SgAllocSize);
}

/*!
 * @brief Kernel-print the scatter-gather lists in a ring buffer.
 *
 * @param edt_p
 *     EDT device pointer.
 * @param p
 *     Ring buffer pointer to dump.
 */

static void dump_sglists(Edt_Dev* edt_p, rbuf* p)
{
  unsigned int i;

  EdtPCISg* sgp;

  for (i = 0; i < p->AllocatedEntries; i++)
  {
    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("Chunk %3d: %08x %d\n", i, p->SgChunksList[i].busaddr, p->SgChunksList[i].count));
  }

  for (i = 0; i < p->AllocatedEntries; i++)
  {
    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("MemRange %3d: %p %u %d \n",
            i,
            p->AllocatedSgList[i].addr,
            p->AllocatedSgList[i].physaddr,
            p->AllocatedSgList[i].size));
  }

  for (i = 0; i < p->SgEntries; i++)
  {
    sgp = edt_get_sg_address_for_page(edt_p, p, i);

    if (sgp)
      EDTPRT(edt_p,
             DBG_DMA_SETUP,
             TRL_INFO,
             ("%4d: %p-> %08x %08x\n", i, sgp, sgp->busaddr, sgp->count));
    else
      EDTPRT(edt_p,
             DBG_DMA_SETUP,
             TRL_INFO,
             ("return from edt_get_sg_address_for_page NULL !!! \n"));
  }
}

void edt_clear_sg_entry(Edt_Dev* edt_p __attribute__((__unused__)), rbuf* p)
{
  p->m_SglistSize = 0;
}

EdtPCISg* edt_get_sg_address_for_page(Edt_Dev* UNUSED_PARAM(edt_p), rbuf* p, int find_entry)
{
  EdtPCISg* sg = NULL;

  unsigned int chunk = 0;
  unsigned int entry = find_entry;

  while (p->AllocatedSgList[chunk].size / 8 < entry && (chunk < p->m_nChunks))
  {
    entry -= p->AllocatedSgList[chunk].size / 8;
    chunk++;
  }

  if (chunk < p->m_nChunks)
    sg = (((EdtPCISg*)p->AllocatedSgList[chunk].addr) + (entry));

  return sg;
}

void edt_prepare_dma_buffers(Edt_Dev* edt_p)
{
  unsigned int count;
  int whichbuf;
  rbuf* p;

  if (edt_p->m_use_buffer_copy && !edt_p->m_Freerun)
  {
    for (count = edt_p->m_Prepared; count < edt_p->m_Todo; count++)
    {
      whichbuf = edt_safe_bufnum(edt_p, count);

      p = edt_p->m_Rbufs + whichbuf;

      if (p->m_WriteFlag)
      {
        if (p->os.map.page_addresses && p->m_WriteFlag)
          edt_copy_pages_from_user(p->os.map.page_addresses,
                                   (void*)p->m_user_address,
                                   p->m_XferSize);

        EDTPRT(edt_p,
               DBG_DMA_SETUP,
               TRL_INFO,
               ("prepare_dma_buffers:  %d useraddr = %p user_size = %x\n",
                whichbuf,
                p->m_user_address,
                p->m_XferSize));
      }
    }
    edt_p->m_Prepared = edt_p->m_Todo;
  }
}

int edt_unmap_buffer(Edt_Dev* edt_p, u_int bufnum)
{
  rbuf* p;

  /* get memory for sg list */
  p = &edt_p->m_Rbufs[bufnum];

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("edt_unmap_buffer %d, is_dma_mapped %u\n", bufnum, p->os.map.is_dma_mapped));

  edt_clear_sg_entry(edt_p, p);

  if (p->os.map.is_dma_mapped)
    edt_pci_unmap_sg(edt_p->os.m_pci_dev, &p->os.map);

  p->m_SglistSize = 0;

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("edt_unmap_buffer: calling mem_map_release for buf %d\n", bufnum));

  /* give up memory */

  edt_mem_map_release(&p->os.map);

  edt_mem_map_clear_sg(&p->os.map);

  return 0;
}

/*********************************************
 *
 * Allocate rbuf sg elements
 * Assumes  os.map already set up,
 * so rbuf just needs pointers or copies of os.map
 *
 ********************************************/

int edt_allocate_sglist(Edt_Dev* edt_p, u_int iolength, int bufnum)
{
  rbuf* p = &edt_p->m_Rbufs[bufnum];

  unsigned int i;

  EdtPCISg* last_sg;

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("==> edt_allocate_sglist size %d\n", iolength));

  p->SgChunksList = p->os.map.chunks;

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("Allocated %d nalloc_chunks %d\n", p->AllocatedEntries, p->os.map.nalloc_chunks));

  if (p->AllocatedEntries < p->os.map.nalloc_chunks)
  {
    if (p->AllocatedEntries && p->AllocatedSgList)
    {
      edt_free_kernel_mem(p->AllocatedSgList,
                          p->AllocatedEntries * sizeof(EdtMemRange),
                          "AllocatedSgList");
      p->AllocatedSgList = NULL;
    }

    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("edt_allocate_sglist %d -> %d\n", p->AllocatedEntries, p->os.map.nalloc_chunks));

    p->AllocatedSgList =
        (EdtMemRange*)edt_get_kernel_mem(p->os.map.nalloc_chunks * sizeof(EdtMemRange),
                                         "AllocatedSgList");
    if (p->AllocatedSgList == NULL)
    {
      edt_printk("Failed to allocate memory for scatter gather list: %s\n", __func__);
      return -1;
    }

    edt_memset(p->AllocatedSgList, 0, p->os.map.nalloc_chunks * sizeof(EdtMemRange));

    p->AllocatedEntries = p->os.map.nalloc_chunks;

    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("check_sg_size: Allocating %d in SgChunksList\n", p->os.map.nalloc_chunks));
  }

  for (i = 0; i < p->os.map.nalloc_chunks; i++)
  {
    p->AllocatedSgList[i].addr = p->os.map.sg_list_ptrs[i]->vaddr;
    p->AllocatedSgList[i].size = (p->os.map.chunks[i].count) & 0xffff;
    p->AllocatedSgList[i].physaddr = p->os.map.chunks[i].busaddr;
  }

  /* update rbuf variables */

  p->m_XferSize = iolength;

  p->m_SglistSize = p->os.map.sg_entries_needed * 8;
  p->m_SgUsed = p->m_SglistSize;
  p->SgEntries = p->os.map.sg_entries_needed;

  p->m_nChunks = p->os.map.nchunks;

  p->InitialChunks = p->m_nChunks;

  p->m_Idx = 0;

  p->m_SgVirtAddr = (u_int*)p->AllocatedSgList->addr;

  p->m_FirstAddr = p->m_SgVirtAddr[0];

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("nchunks=%d p->SgEntried %d\n", p->m_nChunks, p->SgEntries));

  // mark last entry w/ interrupt bit

  last_sg = edt_get_sg_address_for_page(edt_p, p, p->SgEntries - 1);
  if (last_sg)
  {
    last_sg->count |= BUS_DATA(EDT_LIST_PAGEINT);
  }

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("<== edt_allocate_sglist\n"));

  return (0);
}

static int edt_BuildSgList(Edt_Dev* edt_p,
                           u_int bufnum,
                           user_addr_t UNUSED_PARAM(uaddr),
                           u_int xfersize)
{
  int rc = 0;

  rbuf* p = &edt_p->m_Rbufs[bufnum];

  p->os.map.pci_dev = edt_p->os.m_pci_dev;

  rc = edt_mem_map_setup(&p->os.map, xfersize);
  if (rc != 0)
  {
    edt_printk("Failed memory map setup: %s\n", __func__);
    return rc;
  }

  rc = edt_allocate_sglist(edt_p, xfersize, bufnum);
  if (CHECK_DEBUG_MASK(edt_p, DBG_DMA_SETUP, TRL_INFO))
    dump_rbuf("edt_BuildSgList", p);

  if (CHECK_DEBUG_MASK(edt_p, DBG_DMA_SETUP, TRL_INFO))
    dump_sglists(edt_p, p);

  return rc;
}

int edt_map_buffer(Edt_Dev* edt_p,
                   int bufnum,
                   user_addr_t buf,
                   u_int size,
                   u_int direction,
                   void* UNUSED_PARAM(pFile))
{
  rbuf* m;
  int ret;

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         (" ==> map_buffer:  %3d addr %p length %x\n", bufnum, buf, size));

  if (edt_p->dma_pid && (edt_p->dma_pid != edt_get_current_pid()))
  {
    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_ERROR,
           ("map_buffer:  DMA ID already set to something different - UhOh!!\n"));

    return -EINVAL;
  }

  if (size == 0)
  {
    EDTPRT(edt_p, DBG_DMA_SETUP, TRL_ERROR, ("Can't map zero-length buffer\n"));
    return -EINVAL;
  }

  m = &(edt_p->m_Rbufs[bufnum]);

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("map_buffer:  Setting m->m_WriteFlag to %d\n", direction));

  m->m_WriteFlag = direction;
  m->os.map.rw = direction;
  edt_p->m_WriteFlag = direction;

  edt_map_buffer_mutex_lock();

  ret = edt_BuildSgList(edt_p, bufnum, buf, size);

  edt_map_buffer_mutex_unlock();

  if (ret)
    return (ret);

  if (edt_p->dma_pid == 0)
    EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("Setting dma pid = %x\n", edt_get_current_pid()));

  edt_p->dma_pid = edt_get_current_pid();

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("map_buffer:  Map chan %u buffer %d  %p  pid %x\n",
          edt_p->m_DmaChannel,
          bufnum,
          m->os.map.page_list_ptrs,
          edt_p->dma_pid));

  if (bufnum == 0 && CHECK_DEBUG_MASK(edt_p, DBG_DMA_SETUP, TRL_INFO))
    dump_rbuf("After Map", m);

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, (" <== map_buffer\n"));

  return 0;
}

asmlinkage void* edt_get_dmamem_page(void* devp, u_int pagen)
{
  Edt_Dev* edt_p = (Edt_Dev*)devp;
  u_int curpage = 0;
  u_int lastpage;
  rbuf* p;
  unsigned int bufnum = 0;

  p = &edt_p->m_Rbufs[0];

  EDTPRT(edt_p, DBG_MEMMAP, TRL_VERBOSE, ("edt_get_dmamem_page dev %p page %d\n", devp, pagen));

  if (!edt_p->m_use_kernel_buffers)
  {
    EDTPRT(edt_p, DBG_MEMMAP, TRL_ERROR, ("get_dmamem_page called w/o kernel buffers\n"));
    return NULL;
  }

  for (bufnum = 0; bufnum < edt_p->m_Numbufs; bufnum++, p++)
  {
    p = &edt_p->m_Rbufs[bufnum];

    lastpage = curpage + p->os.map.npages;

    if (pagen < lastpage)
    {
      if (p->os.map.page_list_ptrs)
      {
        EDTPRT(edt_p,
               DBG_MEMMAP,
               TRL_VERBOSE,
               ("Found page buf %d page %d %p\n",
                bufnum,
                (pagen - curpage),
                p->os.map.page_list_ptrs[pagen - curpage]->page));

        return p->os.map.page_list_ptrs[pagen - curpage]->page;
      }
      else
      {
        EDTPRT(edt_p, DBG_MEMMAP, TRL_ERROR, ("edt_get_dmamem_page: Error pages not mapped\n"));
        return NULL;
      }
    }

    curpage = lastpage;
  }

  return NULL;
}

asmlinkage void* edt_get_sglistmem_page(void* devp, u_int pageno)
{
  Edt_Dev* edt_p = (Edt_Dev*)devp;
  u_int curpage = 0;
  u_int lastpage;
  rbuf* p;
  unsigned int bufnum = 0;

  for (bufnum = 0; bufnum < edt_p->m_Numbufs; bufnum++, p++)
  {
    p = &edt_p->m_Rbufs[bufnum];

    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("edt_get_sglistmem_page dev %p pageno %d entries %d\n",
            devp,
            pageno,
            p->AllocatedEntries));

    lastpage = curpage + p->AllocatedEntries;

    EDTPRT(edt_p,
           DBG_DMA_SETUP,
           TRL_INFO,
           ("curpage %d lastpage %d pageno - curpage %d\n", curpage, lastpage, pageno - curpage));

    if (pageno < lastpage)
    {
      EDTPRT(edt_p,
             DBG_DMA_SETUP,
             TRL_INFO,
             ("page %p\n", p->AllocatedSgList[pageno - curpage].addr));

      if (p->AllocatedSgList[pageno - curpage].addr != NULL)
      {
        EDTPRT(edt_p,
               DBG_DMA_SETUP,
               TRL_INFO,
               ("Found page buf %d page %d %p\n",
                bufnum,
                (pageno - curpage),
                p->AllocatedSgList[pageno - curpage].addr));

        return edt_virt_to_page(p->AllocatedSgList[pageno - curpage].addr);
      }
      else
      {
        EDTPRT(edt_p,
               DBG_DMA_SETUP,
               TRL_INFO,
               ("edt_get_sglistmem_page error: pages not mapped\n"));
        return NULL;
      }
    }

    curpage = lastpage;
  }

  return NULL;
}

static int edt_clear_kernel_buffers(Edt_Dev* edt_p)
{
  unsigned int j;

  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("clear_kernel_buffers for %d bufs \n", edt_p->m_Numbufs));

  for (j = 0; j < edt_p->m_Numbufs; j++)
  {
    rbuf* p;
    p = edt_p->m_Rbufs + j;

    // this change is in SCs fast driver but applies to all? RWH 12/3/2015
    edt_mem_map_release(&p->os.map);
  }

  return 0;
}

u_int edt_set_kernel_buffers(Edt_Dev* edt_p, u_int newstate)
{
  int oldval = -1;

  oldval = ((edt_p->m_use_buffer_copy) ? EDT_COPY_KBUFS : 0) |
           ((edt_p->m_mmap_buffers) ? EDT_MMAP_KBUFS : 0);

  EDTPRT(edt_p,
         DBG_DMA_SETUP,
         TRL_INFO,
         ("Use buffer oldval = %04x copy = %d mmap buffers %d kernel bufs %d set to %04x\n",
          oldval,
          edt_p->m_use_buffer_copy,
          edt_p->m_mmap_buffers,
          edt_p->m_use_kernel_buffers,
          newstate));

  if (newstate == (u_int)-1) {}
  else
  {
    int old_use;

    edt_p->m_use_buffer_copy = ((newstate & EDT_COPY_KBUFS) != 0);
    edt_p->m_mmap_buffers = (newstate & EDT_MMAP_KBUFS) != 0;

    old_use = edt_p->m_use_kernel_buffers;
    edt_p->m_use_kernel_buffers = edt_p->m_use_buffer_copy | edt_p->m_mmap_buffers;

    if (old_use && !edt_p->m_use_kernel_buffers)
      edt_clear_kernel_buffers(edt_p);
  }

  return oldval;
}
