/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#define pr_fmt(fmt) "%s:%s(): " fmt, KBUILD_MODNAME, __func__

#include "edtdef.h"
#include "edt_driver_version.h"
#include "edt_lnx_kernel.h"
#include "edt_mem_tag.h"

#include "edtreg.h"
#include "edt_pci_devices.h"
#include "chardev_intf.h"

#include <linux/kconfig.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>
#include <linux/gfp.h>
#include <linux/idr.h> /* id allocation */

#if IS_ENABLED(CONFIG_SYSCTL)
#include <linux/sysctl.h>
#endif

#include <linux/spinlock_types.h>

#include <linux/pid.h>
#include <linux/pid_namespace.h>

// For 3.7 kernels
#ifndef VM_RESERVED
#define VM_RESERVED VM_IO
#endif

/* Turn this on to dump all alloc/free calls */
// #define PRINT_MEMALLOC1

/* Note that a SYSFS file of the same name can be used to update this at runtime. */
static int reserve_pages = 0;
module_param(reserve_pages, int, 0644);

#define DO_MUTEX          1
#define PRINT_MUTEX_STATS 1

#ifdef PRINT_MUTEX_STATS
static uint64_t start_mutex_tstamp = 0;
static uint64_t elapsed_wait_time = 0;
static uint64_t elapsed_hold_time = 0;
static int mutex_calls = 0;
#endif

/*!
 * The ID of a board. Referred to as the EDT "unit" number in userspace
 * applications. Used as the first number in /dev file names.
 */
DEFINE_IDA(unitnum_ida);

static DEFINE_MUTEX(map_buffer_mutex);  // Serialize access to buffer mapping

#if defined(DO_MUTEX) && defined(PRINT_MUTEX_STATS)
static uint64_t edt_lnx_timestamp(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0)

  struct timeval tv;

  do_gettimeofday(&tv);
  return (uint64_t)tv.tv_sec * 1000000 + (uint64_t)tv.tv_usec;

#else

  struct timespec64 ts;

  ktime_get_real_ts64(&ts);
  return (uint64_t)ts.tv_sec * 1000000 + (uint64_t)(ts.tv_nsec / 1000);

#endif
}

asmlinkage void edt_print_mutex_stats(void)
{
  int wait_fraction = elapsed_wait_time % 1000000;
  int hold_fraction = elapsed_hold_time % 1000000;

  uint64_t mean = elapsed_hold_time / mutex_calls;
  int mean_fraction = mean % 1000000;

  edt_printk("Mutex calls %d wait %lld.%06d hold %lld.%06d mean hold %lld.%06d\n",
             mutex_calls,
             elapsed_wait_time / 1000000,
             wait_fraction,
             elapsed_hold_time / 1000000,
             hold_fraction,
             mean / 1000000,
             mean_fraction);
}

#endif

asmlinkage void edt_map_buffer_mutex_lock(void)
{
#ifdef DO_MUTEX

#ifdef PRINT_MUTEX_STATS
  uint64_t start_wait_tstamp = edt_lnx_timestamp();
  mutex_calls++;
#endif

  mutex_lock(&map_buffer_mutex);

#ifdef PRINT_MUTEX_STATS
  start_mutex_tstamp = edt_lnx_timestamp();
  elapsed_wait_time += start_mutex_tstamp - start_wait_tstamp;
#endif

#endif
}

asmlinkage void edt_map_buffer_mutex_unlock(void)
{
#ifdef DO_MUTEX

#ifdef PRINT_MUTEX_STATS
  elapsed_hold_time += edt_lnx_timestamp() - start_mutex_tstamp;

  if ((mutex_calls & 0xff) == 0)
  {
    edt_print_mutex_stats();
  }
#endif /* PRINT_MUTEX_STATS */
  mutex_unlock(&map_buffer_mutex);
#endif /* DO_MUTEX */
}

asmlinkage EdtSemaphore_t edt_create_semaphore(const Edt_Dev* const edt_p)
{
  EdtSemaphore_t m;

  m = (EdtSemaphore_t)devm_kzalloc(&(edt_p->os.m_pci_dev->dev),
                                   sizeof(*(EdtSemaphore_t){NULL}),
                                   GFP_KERNEL);
  if (!m)
  {
    pr_err("Semaphore alloc failed\n");
    return NULL;
  }
  sema_init(m, 1);

  return m;
}

asmlinkage int edt_lock_semaphore(EdtSemaphore_t m)
{
  if (down_interruptible(m))
  {
    pr_debug("Unable to lock semaphore\n");
    return -EINTR;
  }
  return 0;
}

asmlinkage void edt_unlock_semaphore(EdtSemaphore_t m)
{
  up(m);
}

asmlinkage int edt_trylock_semaphore(EdtSemaphore_t m)
{
  return down_trylock(m);
}

asmlinkage void edt_printk(const char* const fmt, ...)
{
  static char prtbuf[512];
  static const char prepend[] = KBUILD_MODNAME;

  va_list ap;

  va_start(ap, fmt);
  vsnprintf(prtbuf, (sizeof(prtbuf) - 1), fmt, ap);
  prtbuf[sizeof(prtbuf) - 1] = '\0';
  va_end(ap);

  printk(KERN_DEBUG "%s: %s", prepend, prtbuf);
}

asmlinkage void edt_abort_dma_sleep(void)
{
  //  This must be long enough to allow pending PCIe operations in firmware to
  //  complete; otherwise intermittant crashes.
  udelay(400);

  // TODO: Warning:  this code is not preventing kernel crashes.
}

/* Actually returns tgid, which should be the same for all
threads in a process */
asmlinkage pid_t edt_get_current_pid(void)
{
  if (current)
  {
    return current->tgid;
  }
  else
    return 0;
}

asmlinkage unsigned long edt_copy_to_user(void* to, const void* from, unsigned long n)
{
  return copy_to_user(to, from, n);
}

asmlinkage unsigned long edt_copy_from_user(void* to, const void* from, unsigned long n)
{
  return copy_from_user(to, from, n);
}

asmlinkage unsigned long edt_copy_pages_to_user(void* to, void** sg, unsigned long n)
{
  int npages;
  int i;
  int rc = -1;
  unsigned long extra = PAGE_MASK;
  unsigned char* to_p = to;

  npages = n >> PAGE_SHIFT;

  for (i = 0; i < npages; i++)
  {
    rc = copy_to_user(to_p, sg[i], PAGE_SIZE);
    to_p += PAGE_SIZE;
  }
  if (rc != 0)
    return rc;

  extra = n & ~PAGE_MASK;

  if (extra)
    rc = copy_to_user(to_p, sg[npages], extra);

  return rc;
}

asmlinkage unsigned long edt_copy_pages_from_user(void** sg, void* from, unsigned long n)
{
  int npages;
  int i;
  unsigned long extra = PAGE_MASK;
  unsigned char* from_p = from;
  int rc = -1;

  npages = n >> PAGE_SHIFT;

  for (i = 0; i < npages; i++)
  {
    rc = copy_from_user(sg[i], from_p, PAGE_SIZE);
    from_p += PAGE_SIZE;
  }

  extra = n & ~PAGE_MASK;

  if (extra)
    rc = copy_from_user(sg[npages], from_p, extra);

  return rc;
}

asmlinkage void edt_wake_up(EdtWaitHandle h)
{
  wait_queue_head_t* p = (wait_queue_head_t*)h;

  wake_up(p);
}

/*!
 * @internal
 * @brief Thin wrapper around Linux pci_irq_vector(), providing support for
 * kernels older than 4.8.0.
 *
 * @param dev
 *   PCI device to operate on.
 * @param nr
 *   Device-relative interrupt vector index (0-based).
 *
 * @return Linux IRQ number of a device vector.
 */
static int edt_pci_irq_vector_(struct pci_dev* dev, unsigned int nr)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)
  return pci_irq_vector(dev, nr);
#else
  /* Backport of 5.15 implementation. */
  if (WARN_ON_ONCE(nr > 0))
    return -EINVAL;
  return dev->irq;
#endif
}

asmlinkage void edt_wake_up_interruptible(EdtWaitHandle h)
{
  wait_queue_head_t* p = (wait_queue_head_t*)h;
  wake_up_interruptible(p);
}

asmlinkage void edt_start_intr_critical(EdtSpinLock_t* lock, unsigned long* flags)
{
  spin_lock_irqsave((spinlock_t*)lock, *flags);
}

asmlinkage void edt_end_intr_critical(EdtSpinLock_t* lock, unsigned long flags)
{
  spin_unlock_irqrestore((spinlock_t*)lock, flags);
}

asmlinkage EdtSpinLock_t* edt_create_spinlock(const Edt_Dev* const edt_p)
{
  EdtSpinLock_t* lockp;

  lockp =
      (EdtSpinLock_t*)devm_kzalloc(&(edt_p->os.m_pci_dev->dev), sizeof(EdtSpinLock_t), GFP_KERNEL);
  if (!lockp)
  {
    pr_err("Spinlock alloc failed\n");
    return NULL;
  }

  spin_lock_init(lockp);

  return lockp;
}

#ifndef DEFINE_RAW_SPINLOCK
#define DEFINE_RAW_SPINLOCK DEFINE_SPINLOCK
#define raw_spinlock_t      spinlock_t
#endif

asmlinkage void edt_open_wait_handle(EdtWaitHandle* h)
{
  wait_queue_head_t** p = (wait_queue_head_t**)h;

  *p = (wait_queue_head_t*)edt_kmalloc(sizeof(wait_queue_head_t), "wait_handle");
  init_waitqueue_head(*p);
}

asmlinkage void edt_close_wait_handle(EdtWaitHandle* h)
{
  wait_queue_head_t** p = (wait_queue_head_t**)h;

  edt_kfree(*p, "wait_handle");

  *p = NULL;
}

asmlinkage void* edt_memcpy(void* dest, const void* src, size_t length)
{
  return memcpy(dest, src, length);
}

asmlinkage void* edt_memset(void* dest, int value, size_t length)
{
  return memset(dest, value, length);
}

asmlinkage void
// Reason For Version Specification: https://lwn.net/Articles/740063/
//   supplemental:
//   https://lore.kernel.org/linux-fsdevel/43741269.9cZ5YESnMi@kreacher/
//   Chose 5.3.0: Found via error compiling for Fedora 33 (k: 5.8.6) [gcc 10];
//   gcc 10 is Fedora 31+ (k: 5.3).
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 3, 0)
edt_do_gettimeofday(struct timeval* tv)
#else
edt_do_gettimeofday(struct __kernel_old_timeval* tv)
#endif

{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0)

  do_gettimeofday(tv);

#else

  struct timespec64 ts;

  ktime_get_real_ts64(&ts);
  tv->tv_sec = ts.tv_sec;
  tv->tv_usec = ts.tv_nsec / 1000;

#endif
}

asmlinkage int edt_wait_event(EdtWaitHandle* handle,
                              volatile unsigned short* arg1,
                              unsigned short arg2)
{
  wait_queue_head_t* p = (wait_queue_head_t*)handle;

  wait_event((*p), *arg1 != arg2);

  return 0;
}

asmlinkage int edt_wait_event_timeout(EdtWaitHandle* handle,
                                      volatile unsigned short* arg1,
                                      unsigned short arg2,
                                      signed long timeout)
{
  wait_queue_head_t* p = (wait_queue_head_t*)handle;

  int rc;

  rc = wait_event_timeout((*p), *arg1 != arg2, timeout);

  if (rc == 0)
    rc = -ETIMEDOUT;
  else if (rc > 0)
    rc = 0;

  return rc;
}

asmlinkage int edt_wait_event_interruptible(EdtWaitHandle* handle,
                                            volatile unsigned short* arg1,
                                            unsigned short arg2)
{
  wait_queue_head_t* p = (wait_queue_head_t*)handle;
  int rc = 0;

  rc = wait_event_interruptible((*p), *arg1 != arg2);

  return rc;
}

asmlinkage int edt_wait_event_interruptible_timeout(EdtWaitHandle* handle,
                                                    const volatile unsigned short* const arg1,
                                                    const unsigned short arg2,
                                                    signed long timeout)
{
  wait_queue_head_t* p = (wait_queue_head_t*)handle;

  int rc;

  rc = wait_event_interruptible_timeout((*p), *arg1 != arg2, timeout);

  if (rc == 0)
    rc = -ETIMEDOUT;
  else if (rc > 0)
    rc = 0;

  return rc;
}

asmlinkage int edt_pci_disable_device(struct pci_dev* p_pcidev)
{
  edt_printk("PCI disable device %p\n", p_pcidev);
  pci_disable_device(p_pcidev);
  return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0)
static int edt_dmamem_fault(struct vm_area_struct* vma, struct vm_fault* vmf)
{
  unsigned long pageno;
  void* dev = vma->vm_private_data;
  struct page* page = NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)
#ifdef PRINT_MEMALLOC1
  edt_printk("edt:fault() for address %08x\n", (unsigned long)vmf->virtual_address);
#endif
  pageno = ((unsigned long)vmf->virtual_address - vma->vm_start) >> PAGE_SHIFT;
#else
  pageno = ((unsigned long)vmf->address - vma->vm_start) >> PAGE_SHIFT;
#endif
  page = (struct page*)edt_get_dmamem_page(dev, pageno);

  if (page == NULL)
  {
    edt_printk("edt:fault(): NULL PAGE (VM_FAULT_SIGBUS) pageno : 0x%lx\n", pageno);
    return VM_FAULT_SIGBUS;
  }

#ifdef PRINT_MEMALLOC1
  edt_printk("edt:fault(): pageno >> PAGE_SHIFT: 0x%lx\n", pageno);
#endif

  get_page(page);
  vmf->page = page;

  return 0;
}

static int edt_sglistmem_fault(struct vm_area_struct* vma, struct vm_fault* vmf)
{
  unsigned long pageno;
  void* dev = vma->vm_private_data;
  struct page* page = NULL;

  // Test for kernel version beyond which "virtual_address" changes to "address".
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)
  pageno = ((unsigned long)vmf->virtual_address - vma->vm_start) >> PAGE_SHIFT;
#else
  pageno = ((unsigned long)vmf->address - vma->vm_start) >> PAGE_SHIFT;
#endif

  page = (struct page*)edt_get_sglistmem_page(dev, pageno);

  if (page == NULL)
  {
    edt_printk(
        "edt:edt_sglistmem_fault(): NULL PAGE (VM_FAULT_SIGBUS) pageno >> PAGE_SHIFT: 0x%lx\n",
        pageno);
    return VM_FAULT_SIGBUS;
  }

#ifdef PRINT_MEMALLOC1
  edt_printk("edt:fault(): offset >> PAGE_SHIFT: 0x%lx\n", offset);
#endif

  get_page(page);
  vmf->page = page;

  return 0;
}

#else

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
static vm_fault_t edt_dmamem_fault(struct vm_fault* vmf)
#else
static int edt_dmamem_fault(struct vm_fault* vmf)
#endif
{
  unsigned long pageno;
  void* dev = vmf->vma->vm_private_data;
  struct page* page = NULL;

  pageno = ((unsigned long)vmf->address - vmf->vma->vm_start) >> PAGE_SHIFT;
  page = (struct page*)edt_get_dmamem_page(dev, pageno);

  if (page == NULL)
  {
    edt_printk("edt:fault(): NULL PAGE (VM_FAULT_SIGBUS) pageno : 0x%lx\n", pageno);
    return VM_FAULT_SIGBUS;
  }

  get_page(page);
  vmf->page = page;

  return 0;
}

// map an allocated sglist mem page from driver

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 17, 0)
static vm_fault_t edt_sglistmem_fault(struct vm_fault* vmf)
#else
static int edt_sglistmem_fault(struct vm_fault* vmf)
#endif
{
  unsigned long pageno;
  void* dev = vmf->vma->vm_private_data;
  struct page* page = NULL;

  pageno = ((unsigned long)vmf->address - vmf->vma->vm_start) >> PAGE_SHIFT;

  page = (struct page*)edt_get_sglistmem_page(dev, pageno);

  if (page == NULL)
  {
    edt_printk(
        "edt:edt_sglistmem_fault(): NULL PAGE (VM_FAULT_SIGBUS) pageno >> PAGE_SHIFT: 0x%lx\n",
        pageno);
    return VM_FAULT_SIGBUS;
  }

#ifdef PRINT_MEMALLOC1
  edt_printk("edt:fault(): offset >> PAGE_SHIFT: 0x%lx\n", offset);
#endif

  get_page(page);
  vmf->page = page;

  return 0;
}
#endif  // VERSION 4.11.0

static void edt_dmamem_open_stub(struct vm_area_struct* UNUSED_PARAM(vma))
{}

static void edt_dmamem_close_stub(struct vm_area_struct* UNUSED_PARAM(vma))
{}

static const struct vm_operations_struct EDT_DMAMEM_OPS = {.open = edt_dmamem_open_stub,
                                                           .close = edt_dmamem_close_stub,
                                                           .fault = edt_dmamem_fault};

static const struct vm_operations_struct EDT_SGLIST_MEM_OPS = {.open = edt_dmamem_open_stub,
                                                               .close = edt_dmamem_close_stub,
                                                               .fault = edt_sglistmem_fault};

asmlinkage int edt_mmap_vmap_address(void* dev, void* vma_ptr)
{
  Edt_Dev* edt_p = (Edt_Dev*)dev;
  int rc = 0;
  struct vm_area_struct* vma = (struct vm_area_struct*)vma_ptr;
  const unsigned long vmsize = vma->vm_end - vma->vm_start;
  unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

  if (vma->vm_flags & VM_EXEC)
  {
    pci_err(edt_p->os.m_pci_dev, "mmap() PROT_EXEC not permitted.\n");
    return -EPERM;
  }

  /* Map either BAR0 or BAR1. */
  if (offset < PDMA_OFFSET)
  {
    int bar = -1;
    phys_addr_t bar_physical = 0;
    size_t bar_size = 0;
    phys_addr_t bar_end = 0;
    phys_addr_t start = 0;
    phys_addr_t end = 0;
    const unsigned long BAR_EXTRA_VM_FLAGS =
        (VM_IO | VM_DONTCOPY | VM_DONTEXPAND | VM_RESERVED | VM_DONTDUMP | VM_PFNMAP);

    if (offset < 0x10000)
    {
      bar = 0;
      bar_size = edt_p->bar0_size;
    }
    else if (offset < PDMA_OFFSET)
    {
      offset -= 0x10000;
      bar = 1;
      bar_size = edt_p->bar1_size;
    }

    bar_physical = pci_resource_start(edt_p->os.m_pci_dev, bar);
    bar_end = bar_physical + bar_size - 1;
    start = bar_physical + offset;
    end = start + vmsize - 1;

    pci_dbg(edt_p->os.m_pci_dev, "mmap() BAR%d, offset %lu, size %lu\n", bar, offset, vmsize);

    if ((start > bar_end) || (end > bar_end))
    {
      pci_err(edt_p->os.m_pci_dev, "Requested mmap not within BAR%d range\n", bar);
      return -EINVAL;
    }

    vma->vm_pgoff = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)
    vma->vm_page_prot = pgprot_device(vma->vm_page_prot);
#endif
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#if EDT_KCOMPAT_HAS_VM_FLAGS_SETTER
    vm_flags_set(vma, BAR_EXTRA_VM_FLAGS);
#else
    vma->vm_flags |= BAR_EXTRA_VM_FLAGS;
#endif

    rc = vm_iomap_memory(vma, start, vmsize);
    if (rc)
      pci_err(edt_p->os.m_pci_dev,
              "vm_iomap_memory() returned %d attempting to mmap BAR%d\n",
              rc,
              bar);
  }
  /* 16MB used to map SgList pages to user programs; used by Pdma - Mark 03/07 */
  else if (offset < 0x10000000)
  {
    pci_dbg(edt_p->os.m_pci_dev, "%s() sglist pages\n", __func__);
    offset -= PDMA_OFFSET;

#if EDT_KCOMPAT_HAS_VM_FLAGS_SETTER
    vm_flags_set(vma, VM_RESERVED);
#else
    vma->vm_flags |= VM_RESERVED;
#endif
    vma->vm_ops = &EDT_SGLIST_MEM_OPS;
    vma->vm_private_data = dev;
  }
  else
  {
    pci_info(edt_p->os.m_pci_dev, "%s() DMA buffers\n", __func__);
    /* map DMA buffer memory using nopage */
#if EDT_KCOMPAT_HAS_VM_FLAGS_SETTER
    vm_flags_set(vma, VM_RESERVED);
#else
    vma->vm_flags |= VM_RESERVED;
#endif
    vma->vm_ops = &EDT_DMAMEM_OPS;
    vma->vm_private_data = dev;
#ifdef PRINT_MEMALLOC1
    edt_printk("mapped dma mem at %x\n", offset);
#endif
  }

  return rc;
}

/*!
 * @brief Linux PCI API callback for adding PCI devices.
 *
 * Called by Linux:
 *
 *     ...for already existing devices or later if a new device gets inserted
 *     for all PCI devices which match the ID table and are not 'owned' by
 *     other drivers yet.
 *
 * @param pcidev
 *     PCI device structure to probe.
 * @param id
 *     PCI ID information of the device.
 * @returns
 *     Zero if this driver decides to take ownership of the device.
 *     Negative error code otherwise.
 */
static int edt_pcidrv_probe(struct pci_dev* pcidev, const struct pci_device_id* id)
{
  Edt_Dev* edt_p = NULL;
  int retval = 0;
  bool msi = false;
  int irq_flags;
  const char* board_name = "";
  int unit = -1;

  if (0 != edt_pcidev_str_from_pid(id->device, &board_name))
    board_name = "";

  pci_info(pcidev, "Probing PCI device %s [%04x:%04x]\n", board_name, id->vendor, id->device);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
  unit = ida_alloc(&unitnum_ida, GFP_KERNEL);
#else
  unit = ida_simple_get(&unitnum_ida, 0, 0, GFP_KERNEL);
#endif
  if (unit < 0)
  {
    pci_err(pcidev, "Could not assign a unit number\n");
    return unit;
  }

  // Create a custom structure for this device.
  edt_p = edt_create_device_structures(pcidev, id, unit);
  if (!edt_p)
  {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
    ida_free(&unitnum_ida, unit);
#else
    ida_simple_remove(&unitnum_ida, unit);
#endif
    return -ENOMEM;
  }

  // Attach our local structure to this device
  pci_set_drvdata(pcidev, edt_p);

  /* Set up a single MSI interrupt */
  edt_p->os.irq = 0;
  irq_flags = 0;
  msi = true;

#if EDT_KCOMPAT_HAS_PCI_IRQ_INTX
  retval = pci_alloc_irq_vectors(pcidev, 1, 1, PCI_IRQ_INTX | PCI_IRQ_MSI);
  if (retval != 1)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
  retval = pci_alloc_irq_vectors(pcidev, 1, 1, PCI_IRQ_LEGACY | PCI_IRQ_MSI);
  if (retval != 1)
#else
  retval = pci_enable_msi(pcidev);
  if (retval)
#endif
  {
    pci_info(pcidev, "MSI interrupts could not be enabled.\n");
    irq_flags = IRQF_SHARED;
    msi = false;
  }
  retval = devm_request_irq(&pcidev->dev,
                            edt_pci_irq_vector_(pcidev, 0),
                            edt_intr,
                            irq_flags,
                            KBUILD_MODNAME,
                            edt_p);
  if (retval)
  {
    pci_err(pcidev, "Failed to register interrupt handler. Aborting.\n");
    return -ENODEV;
  }

  {
    int i;
    edt_p->os.irq = edt_pci_irq_vector_(pcidev, 0);
    for (i = 0; i < edt_p->m_chancnt; i++)
    {
      edt_p->chan_p[i]->os.irq = edt_p->os.irq;
    }
  }

  if (msi)
    pci_info(pcidev, "MSI IRQ %d\n", edt_p->os.irq);
  else
    pci_info(pcidev, "IRQ %d\n", edt_p->os.irq);

  retval = edt_chardev_create_interfaces(edt_p);
  if (retval)
  {
    pci_err(edt_p->os.m_pci_dev, "(%d) Failed to create interface devices.\n", retval);
    return retval;
  }

  return 0;
}

/*!
 * @brief Linux PCI API callback for removing PCI devices.
 *
 * Called by Linux:
 *
 *     ...whenever a device being handled by this driver is removed (either
 *     during deregistration of the driver or when it's manually pulled out
 *     of a hot-pluggable slot)
 *
 * Also refer to:
 *     https://www.kernel.org/doc/html/latest/PCI/pci.html#pci-device-shutdown
 *
 * @param pcidev
 *     PCI device being removed.
 */
static void edt_pcidrv_remove(struct pci_dev* pcidev)
{
  size_t i = 0;
  Edt_Dev* edt_p = pci_get_drvdata(pcidev);
  pci_info(pcidev, "Removing\n");

  if (edt_p)
  {
    edt_disable_dma_interrupt(edt_p);
    edt_clear_dma(edt_p);

    edt_p->bar0_mmap = NULL;
    edt_p->bar0_size = 0;
    edt_p->bar1_mmap = NULL;
    edt_p->bar1_size = 0;
  }

  /* Clear bus-master bit. */
  pci_clear_master(pcidev);

  /* If there is more we can unroll with just the PCI dev, then remove the
   * 'return' here. Added now to avoid freeing memory that wasn't allocated. */
  if (!edt_p)
  {
    pr_warn("EDT struct is null\n");
    return;
  }

  if (0 != edt_chardev_destroy_interfaces(edt_p))
  {
    pci_warn(edt_p->os.m_pci_dev, "Device interfaces not removed cleanly.\n");
  }

  if (edt_p->m_serial)
  {
    EDTPRT(edt_p, DBG_CREATE_CLOSE, TRL_INFO, ("cleanup_module: Clearing serial_p\n"));
    edt_vfree(edt_p->m_serial, sizeof(edt_serial), "m_serial");
  }

  edt_close_wait_handle(&edt_p->os.serial_wait);
  edt_close_wait_handle(&edt_p->os.wait);
  for (i = 0; i < EDT_MAX_EVENT_TYPES; i++)
    edt_close_wait_handle(&edt_p->os.event_wait[i]);

  edt_free_tracebuf(edt_p);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
  ida_free(&unitnum_ida, edt_p->m_Unitno);
#else
  ida_simple_remove(&unitnum_ida, edt_p->m_Unitno);
#endif
}

/* Sysfs 'show' method of driver attribute 'reserve_pages'. */
static ssize_t reserve_pages_show(struct device_driver* UNUSED_PARAM(drv), char* buf)
{
  return scnprintf(buf, PAGE_SIZE, "%d", reserve_pages);
}

/* Sysfs 'store' method of driver attribute 'reserve_pages'. */
static ssize_t reserve_pages_store(struct device_driver* UNUSED_PARAM(drv),
                                   const char* buf,
                                   size_t count)
{
  int parsed_value = 0;

  if ((count > 0) && (1 == sscanf(buf, "%d", &parsed_value)))
  {
    int result = count;

    if (0 != edt_cache_create_free_list(parsed_value))
    {
      result = -EADDRINUSE;
    }

    // Update the module param value, in case someone reads it.
    reserve_pages = edt_cache_free_list_count();

    return result;
  }
  return -EINVAL;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 0, 0)
static DRIVER_ATTR_RW(reserve_pages);
#else
static DRIVER_ATTR(reserve_pages, (S_IRUGO | S_IWUSR), reserve_pages_show, reserve_pages_store);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
static struct attribute* edt_pci_driver_attrs[] = {&driver_attr_reserve_pages.attr, NULL};

/* Create "edt_pci_driver_groups" from the above "edt_pci_driver_attrs". Given
 * to the driver struct's `.group` member. */
ATTRIBUTE_GROUPS(edt_pci_driver);
#endif

/* List of PCI VID/PIDs supported by this driver. */
static const struct pci_device_id EDT_PCI_DEVICES[] = {
#define EXPAND_AS_PID(_defname, _pid, _dma_ch, _name, _desc) {PCI_DEVICE(PCI_VENDOR_ID_EDT, _pid)},
    EDT_PCI_DEVICES_LISTX(EXPAND_AS_PID)
#undef EXPAND_AS_PID
        {0} /* End of List */
};

MODULE_DEVICE_TABLE(pci, EDT_PCI_DEVICES);

/* PCI device IDs and driver callbacks  */
static struct pci_driver edt_pci_driver = {.name = KBUILD_MODNAME,
                                           .id_table = EDT_PCI_DEVICES,
                                           .probe = edt_pcidrv_probe,
                                           .remove = edt_pcidrv_remove,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
                                           .groups = edt_pci_driver_groups
#endif
};

static int __init edt_module_init(void)
{
  int ret = -1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
#error "This version of the EDT driver no longer supports kernels < 3.10.0"
#endif

  pr_info("Initializing: v%s (%s)\n", EDT_DRIVER_VERSION, EDT_DRIVER_BUILDID);

#ifdef DEBUG_LEAKS
  init_mem_pools();
#endif

  ret = edt_chardev_init();
  if (0 != ret)
  {
    pr_err("Module failed to init chardev.\n");
    return -1;
  }

  /* If module parameter was set at load time, attempt to reserve pages. */
  if (0 != reserve_pages)
  {
    pr_info("Reserve %d pages.\n", reserve_pages);
    if (0 != edt_cache_create_free_list(reserve_pages))
    {
      pr_warn("Failed to reserve %d pages.\n", reserve_pages);
      reserve_pages = edt_cache_free_list_count();
    }
  }

  ret = pci_register_driver(&edt_pci_driver);
  if (0 != ret)
    return ret;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0)
  /* Kernel < 4.14, need to create/remove driver attribute files ourselves, via
   * driver_create_file() and driver_remove_file(). Kernel >= 4.14, can use the
   * `.groups` driver struct member to create/remove sysfs attribute files
   * automatically with the driver.
   */
  ret = driver_create_file(&edt_pci_driver.driver, &driver_attr_reserve_pages);
  if (0 != ret)
    pr_warn("Driver attr sysfs file could not be created.");
#endif

  return ret;
}

static void __exit edt_module_exit(void)
{
  pr_info("Unloading\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0)
  driver_remove_file(&edt_pci_driver.driver, &driver_attr_reserve_pages);
#endif

  /* Call PCI unregister driver first, so that our "pci_driver.remove" function
   * will get called on all devices.
   *
   * Otherwise, calling cdev cleanup would destroy the character class before
   * all cdev devices have been deleted.
   */
  pci_unregister_driver(&edt_pci_driver);

  edt_chardev_cleanup();

  edt_release_cache_list();

#ifdef DEBUG_LEAKS
  dump_mem_pools();
  clear_mem_pools();
#endif
}

module_init(edt_module_init);
module_exit(edt_module_exit);

MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Engineering Design Team, Inc. <tech@edt.com>");
MODULE_DESCRIPTION("EDT device driver, v" EDT_MODULE_VERSION_LONG);
MODULE_VERSION(EDT_MODULE_VERSION);
