/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#ifndef _EDT_LNX_KERNEL_H
#define _EDT_LNX_KERNEL_H

#include "edt_mem_mapping.h"
#include "kcompat.h"

#include <linux/kconfig.h>

/* for asmlinkage */
#include <linux/linkage.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/version.h>
#include <linux/irqreturn.h>
#include <linux/semaphore.h>
#include <linux/pci.h>

#define UNUSED_PARAM(x) UNUSED_##x __attribute__((__unused__))

/**************************************************/
/* Edt wrapper functions around kernel calls */
/**************************************************/

/* typedefs for kernel structures - all code except edt_lnx_kernel doesn't
   know about these */

typedef void* EdtWaitHandle;

typedef spinlock_t EdtSpinLock_t;
typedef struct semaphore* EdtSemaphore_t;

asmlinkage void edt_printk(const char* const format, ...) __attribute__((format(printf, 1, 2)));

asmlinkage void* edt_memcpy(void* dest, const void* src, size_t length);
asmlinkage void* edt_memset(void* dest, int c, size_t length);

asmlinkage unsigned long edt_copy_to_user(void* to, const void* from, unsigned long n);

asmlinkage unsigned long edt_copy_from_user(void* to, const void* from, unsigned long n);

asmlinkage unsigned long edt_copy_pages_to_user(void* to, void** from, unsigned long n);

asmlinkage unsigned long edt_copy_pages_from_user(void** sg, void* from, unsigned long n);

asmlinkage void edt_wake_up(EdtWaitHandle p);

asmlinkage void edt_wake_up_interruptible(EdtWaitHandle p);

asmlinkage void edt_start_intr_critical(EdtSpinLock_t* lock, unsigned long* flags);

asmlinkage void edt_end_intr_critical(EdtSpinLock_t* lock, unsigned long flags);

asmlinkage pid_t edt_get_current_pid(void);

asmlinkage int edt_register_interrupt(uint32_t* const irq,
                                      void* const p_edtdev,
                                      void* const p_pci,
                                      const char* const name);

asmlinkage int edt_pci_disable_device(struct pci_dev* p_pcidev);

asmlinkage void edt_print_mutex_stats(void);

/********************************************/
/* Memory access functions                  */
/********************************************/

asmlinkage void* edt_kmalloc(size_t, char* tag);
asmlinkage void edt_kfree(const void* ptr, char* tag);

asmlinkage void* edt_get_free_pages(int size, char* tag);

asmlinkage int edt_cache_create_free_list(int size);

/*!
 * @brief Get number of pages in the cache page list.
 *
 * Pre-allocated cache page pool is setup by @ref edt_cache_create_free_list().
 *
 * @return Total number of pages in the cache list.
 */
unsigned int edt_cache_free_list_count(void);

asmlinkage void edt_free_pages(void* addr, int size, char* tag);

asmlinkage int edt_mem_map_alloc_kernel(EdtMemMapping* map, int size);

asmlinkage void edt_vfree(void* addr, u_int size, char* tag);

asmlinkage void* edt_vmalloc(unsigned int size, char* tag);

asmlinkage void edt_free_kernel_mem(void* addr, int size, char* tag);
asmlinkage void* edt_get_kernel_mem(int size, char* tag);

asmlinkage void edt_release_cache_list(void);

asmlinkage void edt_open_wait_handle(EdtWaitHandle* h);

asmlinkage void edt_close_wait_handle(EdtWaitHandle* h);

asmlinkage int edt_pci_unmap_sg(void* dev, EdtMemMapping* map);

asmlinkage void edt_map_buffer_mutex_lock(void);
asmlinkage void edt_map_buffer_mutex_unlock(void);

/*
 * function to avoid race conditions while sleeping per
 * Linux Device Drivers 2nd Edition.
 */
asmlinkage int edt_wait_event_interruptible_timeout(EdtWaitHandle* handle,
                                                    const volatile unsigned short* const arg1,
                                                    const unsigned short arg2,
                                                    signed long timeout);
asmlinkage int edt_wait_event_interruptible(EdtWaitHandle* handle,
                                            volatile unsigned short* arg1,
                                            unsigned short arg2);

asmlinkage int edt_wait_event_timeout(EdtWaitHandle* handle,
                                      volatile unsigned short* arg1,
                                      unsigned short arg2,
                                      signed long timeout);
asmlinkage int edt_wait_event(EdtWaitHandle* handle,
                              volatile unsigned short* arg1,
                              unsigned short arg2);

asmlinkage void edt_mem_map_release(EdtMemMapping* map);

asmlinkage int edt_mem_map_setup(EdtMemMapping* map, int bufsize);
asmlinkage int edt_mem_map_clear_sg(EdtMemMapping* map);

asmlinkage int edt_mmap_vmap_address(void* dev, void* vma_ptr);

asmlinkage void* edt_virt_to_page(void* vaddr);
asmlinkage void* edt_get_dmamem_page(void* dev, u_int offset);
asmlinkage void* edt_get_sglistmem_page(void* devp, u_int pagen);

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 3, 0)
asmlinkage void edt_do_gettimeofday(struct timeval*);
#else
asmlinkage void edt_do_gettimeofday(struct __kernel_old_timeval*);
#endif

asmlinkage int edt_private_ioctl(void* dev_ptr,
                                 unsigned int cmd,
                                 unsigned long arg,
                                 int compat32_flag);
asmlinkage int edt_private_mmap(void* dev_ptr, void* vma);

asmlinkage irqreturn_t edt_intr(int irq, void* data);

// asmlinkage void* edt_get_dev_from_minor(int minor);

asmlinkage void edt_abort_dma_sleep(void);

#endif /* _EDT_LNX_KERNEL_H */
