/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#include "edtdef.h"

#define pr_fmt(fmt) "%s:%s(): " fmt, KBUILD_MODNAME "_mm", __func__

#include "edt_lnx_kernel.h"
#include "edt_mem_tag.h"
#include "edt_lnx_mm.h"
#include "edt_lnx_alloc.h"
#include "edt_lnx_kernel.h"

#include <linux/vmalloc.h>
#include <linux/kconfig.h>
#include <linux/version.h>
#include <linux/pci.h>

#ifdef GFP_DMA32
#define GFP_FLAG (GFP_DMA32 | GFP_KERNEL)
#else
#define GFP_FLAG GFP_KERNEL
#endif

/* Turn this on to dump all alloc/free calls */
// #define PRINT_MEMALLOC
/* turn this on for SG allocation */
// #define PRINT_MEMALLOC1

#ifdef PRINT_MEMALLOC
#define print_memalloc(x) edt_printk x
#else
#define print_memalloc(x)
#endif

#ifdef PRINT_MEMALLOC1
#define print_memalloc1(x) edt_printk x
#else
#define print_memalloc1(x)
#endif

#define USE_CACHED_PAGES

#ifdef USE_CACHED_PAGES
DEFINE_SPINLOCK(rsv_pg_lock);

#define LOCK_RSV_PAGES

#define LIST_PTR_GRAN 512

typedef struct _edt_cache_block {
  struct _edt_cache_block* next;
  EdtMemList* pointers;
  int active_size;
} EdtCacheBlock;

static int using_cache_pages = 0;
static int uncached_pages = 0;

static EdtMemList* cache_free_list = NULL;
static EdtMemList* cache_active_list = NULL;
static u_int cache_list_pages = 0;
static u_int cache_free_size = 0;
static u_int cache_active_size = 0;
static EdtCacheBlock* block_list = NULL;

static asmlinkage int edt_merge_last_chunks(EdtMemMapping* map);

/*************************************
 *
 * Memory Allocation/deallocation
 *
 *************************************/

asmlinkage void* edt_kmalloc(size_t size, char* tag)
{
  void* p;
  p = kmalloc(size, GFP_KERNEL);

  add_mem_tag(kmalloc_tags, p, size, tag);

  return p;
}

asmlinkage void edt_kfree(const void* ptr, char* tag)
{
  clear_mem_tag(kmalloc_tags, ptr, tag);

  kfree(ptr);
}

static EdtCacheBlock* create_cache_block(void)
{
  EdtCacheBlock* block;

  block = (EdtCacheBlock*)vmalloc(sizeof(EdtCacheBlock));
  if (NULL == block)
    return NULL;

  block->active_size = LIST_PTR_GRAN;
  block->pointers = (EdtMemList*)vmalloc(block->active_size * sizeof(EdtMemList));
  if (NULL == block->pointers)
  {
    vfree(block);
    return NULL;
  }

  memset(block->pointers, 0, block->active_size * sizeof(EdtMemList));

  block->next = NULL;

  return block;
}

static void* edt_alloc_cache_page(const char* tag)
{
  void* p = alloc_page(GFP_FLAG);
  if (NULL == p)
  {
    pr_warn("Cache page alloc failed.\n");
    return NULL;
  }
  add_mem_tag(page_cache_tags, p, PAGE_SIZE, tag);

  return p;
}

static void edt_free_cache_page(void* p, const char* tag)
{
  clear_mem_tag(page_cache_tags, p, tag);

  __free_page(p);
}

/********************************
 * Move front element of from to front of to
 ********************************/

static EdtMemList* move_memlist(EdtMemList* lp, EdtMemList** from, EdtMemList** to)
{
  EdtMemList* dp;
  EdtMemList* fp;

  if (lp == NULL || from == NULL || to == NULL)
  {
    edt_printk("null call to move_list_element from %p to %p\n", from, to);
    return NULL;
  }

  // cut out lp

  fp = *from;
  dp = *to;

  if (fp)
  {
    if (lp->next != lp)
    {
      if (lp->prev)
        lp->prev->next = lp->next;
      if (lp->next)
        lp->next->prev = lp->prev;

      if (lp == fp)
        *from = lp->next;
    }
    else
    {
      if (lp == fp)
        *from = NULL;
    }

    lp->next = lp->prev = lp;
  }
  else
  {
    edt_printk("source in moev_memlist is empty\n");
    return NULL;
  }

  if (dp)
  {
    if (dp->prev)
      dp->prev->next = lp;
    lp->prev = dp->prev;
    dp->prev = lp;
    lp->next = dp;
  }
  else
  {
    lp->next = lp;
    lp->prev = lp;
  }
  *to = lp;

  return lp;
}

/* Returns 0 if successful. */
asmlinkage int edt_cache_create_free_list(int size)
{
  print_memalloc(("edt_cache_create_free_list %x\n", size));
  if (!using_cache_pages)
  {
    if (uncached_pages > 0)
    {
      edt_printk("Unable to create cache list, there are uncached pages allocated\n");
      return -1;
    }

    if (size)
      using_cache_pages = 1;
    else
      return -1;
  }

  if ((int)cache_list_pages > size)
  {
    if (cache_active_size > 0)
    {
      edt_printk("Can't shrink page cache if active pages\n");
      return -1;
    }

    print_memalloc(("Releasing pages to shrink cache list\n"));
    edt_release_cache_list();
  }

  if ((int)cache_list_pages < size)
  {
    while ((int)cache_list_pages < size)
    {
      int i;
      EdtCacheBlock* block = create_cache_block();
      if (NULL == block)
      {
        pr_warn("Cache block alloc failure.\n");
        return -1;
      }

      cache_list_pages += block->active_size;

      for (i = 0; i < block->active_size; i++)
      {
        EdtMemList* lp = block->pointers + i;

        // Break out of while loop on first failed alloc.
        lp->page = edt_alloc_cache_page("cache");
        if (NULL == lp->page)
        {
          pr_warn("Only allocated %d of %d pages.\n", cache_free_size, size);
          return -1;
        }

        lp->vaddr = page_address(lp->page);
        lp->mapped = 0;
        if (cache_free_list)
        {
          lp->next = cache_free_list;
          if (cache_free_list->prev)
            cache_free_list->prev->next = lp;
          lp->prev = cache_free_list->prev;
          cache_free_list->prev = lp;
        }
        else
        {
          lp->next = lp;
          lp->prev = lp;
        }

        cache_free_list = lp;
        cache_free_size++;
      }
      block->next = block_list;
      block_list = block;
    }
  }

  return 0;
}

unsigned int edt_cache_free_list_count(void)
{
  return cache_free_size + cache_active_size;
}

#endif

static void lock_reserved_pages(void)
{
#ifdef LOCK_RSV_PAGES
  spin_lock(&rsv_pg_lock);
#endif
}

static void unlock_reserved_pages(void)
{
#ifdef LOCK_RSV_PAGES
  spin_unlock(&rsv_pg_lock);
#endif
}

static asmlinkage EdtMemList* edt_alloc_kernel_page(char* tag)
{
  void* p;
  EdtMemList* lp;

  if (using_cache_pages)
  {
    // If we have pages on the free list, use them

    lock_reserved_pages();

    // if we don't have any allocated list pointers available, add to the cache

    if (cache_free_list == NULL || cache_free_size <= 0)
      edt_cache_create_free_list(cache_list_pages + 1);

    if (cache_free_list == NULL)
    {
      edt_printk("cache_list doesn't exist, failing\n");

      unlock_reserved_pages();
      return NULL;
    }

    lp = cache_free_list;

    move_memlist(lp, &cache_free_list, &cache_active_list);

    p = lp->page;

    cache_free_size--;
    cache_active_size++;

    unlock_reserved_pages();
  }
  else
  {
    // no caching
    lp = (EdtMemList*)edt_kmalloc(sizeof(EdtMemList), "edt_alloc_kernel_page-memlist");
    if (lp == NULL)
    {
      edt_printk("Failed to allocate kernel page memory list: %s\n", __func__);
      return lp;
    }

    p = (void*)edt_alloc_cache_page(tag);
    if (p == NULL)
    {
      edt_printk("Failed to allocate cache page: %s\n", __func__);
      return p;
    }

    lp->page = p;
    lp->vaddr = page_address(lp->page);

    uncached_pages++;
  }

  return lp;
}

static asmlinkage void edt_unalloc_mem_list_page(EdtMemList* lp, char* tag)
{
  if (using_cache_pages)
  {
    lock_reserved_pages();

    if (lp)
    {
      // cut out of active list

      lp = move_memlist(lp, &cache_active_list, &cache_free_list);

      cache_active_size--;
      cache_free_size++;
    }
    else
      edt_printk("NULL mem list\n");

    unlock_reserved_pages();
  }
  else
  {
    edt_free_cache_page(lp->page, tag);

    edt_kfree(lp, tag);

    uncached_pages--;
  }
}

asmlinkage void edt_release_cache_list(void)
{
  int i;

  if (using_cache_pages)
  {
    lock_reserved_pages();

    print_memalloc(("Releasing cache\n"));

    while (block_list != NULL)
    {
      EdtCacheBlock* block = block_list;
      block_list = block_list->next;

      for (i = 0; i < block->active_size; i++)
      {
        if (block->pointers[i].page)
          edt_free_cache_page(block->pointers[i].page, "cache");
        else
          edt_printk("edt_release_cache_list mem ptr is null\n");
      }

      vfree(block->pointers);
      vfree(block);
    }

    cache_list_pages = 0;

    unlock_reserved_pages();

    cache_free_list = NULL;
    cache_active_list = NULL;
    cache_active_size = 0;
  }
}

asmlinkage void* edt_virt_to_page(void* vaddr)
{
  return virt_to_page(vaddr);
}

asmlinkage void* edt_get_free_pages(int size, char* tag)
{
  int i, pow2 = 1;
  unsigned long p;

  for (i = 0;; i++)
  {
    if (pow2 * PAGE_SIZE >= (unsigned long)size)
      break;
    pow2 *= 2;
  }

  p = __get_free_pages(GFP_KERNEL | GFP_DMA32, i);

  add_mem_tag(page_tags, (void*)p, PAGE_SIZE * pow2, tag);

  return (void*)p;
}

asmlinkage void edt_free_pages(void* addr, int size, char* tag)
{
  int i, pow2 = 1;

  for (i = 0;; i++)
  {
    if (pow2 * PAGE_SIZE >= (unsigned long)size)
      break;
    pow2 *= 2;
  }

  clear_mem_tag(page_tags, addr, tag);

  free_pages((unsigned long)addr, i);
}

asmlinkage void edt_vfree(void* addr, u_int UNUSED_PARAM(size), char* tag)
{
  if (addr == NULL)
    edt_printk("warning attempt to vfree NULL tag:%s\n", tag);

  vfree(addr);

  clear_mem_tag(vmalloc_tags, addr, tag);
}

asmlinkage void* edt_vmalloc(unsigned int size, char* tag)
{
  void* p;

  p = vmalloc_32(size);

  add_mem_tag(vmalloc_tags, p, size, tag);

  return p;
}

asmlinkage void* edt_get_kernel_mem(int size, char* tag)
{
  void* p;

  p = (void*)edt_get_free_pages(size, tag);

  return p;
}

/* generic free - requires size */

asmlinkage void edt_free_kernel_mem(void* p, int size, char* tag)
{
  if (p == NULL || size == 0)
    return;

  edt_free_pages(p, size, tag);
}

asmlinkage int edt_pci_unmap_sg(void* UNUSED_PARAM(dev), EdtMemMapping* map)
{
  int retval = 0;

  if (map != NULL && map->is_dma_mapped)
  {
    if (map->os_sglist == NULL)
    {
      edt_printk("edt_pci_unmap_sg ERROR: pci_unmap_sg for list %p sglist_count %d for %d pages \n",
                 map->os_sglist,
                 map->sglist_count,
                 map->npages);

      dump_mem_mapping("edt_pci_unmap_sg", map);

      return retval;
    }
    else
    {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
      dma_unmap_sg(&(map->pci_dev->dev),
                   (struct scatterlist*)map->os_sglist,
                   map->npages,
                   (map->rw) ? DMA_TO_DEVICE : DMA_FROM_DEVICE);

#else
      pci_unmap_sg(map->pci_dev,
                   (struct scatterlist*)map->os_sglist,
                   map->npages,
                   (map->rw) ? PCI_DMA_TODEVICE : PCI_DMA_FROMDEVICE);
#endif
      map->is_dma_mapped = 0;
      retval = 1;
    }
  }

  return retval;
}

/*********************
 * Get the DMA memory for the sg list
 * already constructed
 * mlp gets the chunks
 *********************/

/**************************************
 * Routines for mapping sg lists from user space
 *
 *************************************/

/* generic allocation, using kmalloc for < page, get_free_pages for larger */

/***************************************
 *
 *  create a page list mapping
 * This allocates a page array to hold the mapped user space
 * pages. This only works if the physical addresses are < 4 GB
 *
 ***************************************/

//
// This function can be used as an escape during allocation or
// to clear the buffers when done
//

static void edt_mem_map_clear_kernel_buffers(EdtMemMapping* map, int ntoclear)
{
  int j;

  for (j = 0; j < ntoclear; j++)
  {
    if (map->page_list_ptrs[j])
    {
      edt_unalloc_mem_list_page(map->page_list_ptrs[j], "page_list_ptrs[i]");
      map->page_list_ptrs[j] = 0;
    }
  }

  if (map->page_addresses)
    edt_free_kernel_mem(map->page_addresses, map->pgsize, "page_addresses");
  if (map->page_list_ptrs)
    edt_free_kernel_mem(map->page_list_ptrs, map->pgsize, "page_list_ptrs");

  map->page_list_ptrs = NULL;
  map->page_addresses = NULL;

  return;
}

static void edt_mem_map_clear_sg_ptrs(EdtMemMapping* map, int ntoclear)
{
  int j;

  if (map->sg_list_ptrs)
  {
    for (j = 0; j < ntoclear; j++)
    {
      if (map->sg_list_ptrs[j])
      {
        map->sg_list_ptrs[j]->mapped = false;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
        dma_unmap_single(&(map->pci_dev->dev),
                         map->sg_list_ptrs[j]->busaddr,
                         PAGE_SIZE,
                         DMA_BIDIRECTIONAL);
#else
        pci_unmap_single(map->pci_dev, map->sg_list_ptrs[j]->busaddr, PAGE_SIZE, PCI_DMA_TODEVICE);
#endif
        map->sg_list_ptrs[j]->busaddr = (bus_addr32_t)0xFFFFFFFF;

        edt_unalloc_mem_list_page(map->sg_list_ptrs[j], "sg_list_ptrs[i]");
        map->sg_list_ptrs[j] = 0;
      }
    }

    edt_free_kernel_mem(map->sg_list_ptrs,
                        map->nalloc_chunks * sizeof(EdtMemList*),
                        "map->sg_list_ptrs");
  }

  if (map->chunks)
    edt_free_kernel_mem(map->chunks, map->nalloc_chunks * sizeof(EdtPCISg*), "map->chunks");

  map->sg_list_ptrs = 0;
  map->chunks = 0;
  map->nalloc_chunks = 0;
  map->nchunks = 0;

  return;
}

/*
 * @brief edt_mem_map_alloc_kernel
 *
 * @param[out] map
 * @param[in] size
 *
 * @returns
 *
 */
asmlinkage int edt_mem_map_alloc_kernel(EdtMemMapping* map, int size)
{
  int npages;
  int rc = 0;
  int pgsize;
  int i;

  npages = size / PAGE_SIZE;

  if (size & ~PAGE_MASK)
    npages++;

  print_memalloc(("==> mem_map_alloc_kernel with size %d pages %d\n", size, npages));

  pgsize = npages * sizeof(struct page*);

  // Everybody gets page_pointers and page_addresses

  map->pgsize = pgsize;

  map->page_addresses = (void**)edt_get_kernel_mem(pgsize, "page_addresses");

  if (map->page_addresses == NULL)
  {
    printk("edt_mem_map_alloc_kernel:  out of memory for kernel buffers\n");
    edt_free_kernel_mem(map->page_addresses, pgsize, "page_addresses");
    map->page_addresses = NULL;
    return -ENOMEM;
  }

  map->page_list_ptrs = (EdtMemList**)edt_get_kernel_mem(pgsize, "page_list_ptrs");

  if (map->page_list_ptrs == NULL)
  {
    printk("edt_mem_map_alloc_kernel:  out of memory for kernel buffers\n");
    edt_free_kernel_mem(map->page_list_ptrs, pgsize, "page_list_ptrs");
    map->page_list_ptrs = NULL;
    return -ENOMEM;
  }

  for (i = 0; i < npages; i++)
  {
    map->page_addresses[i] = 0;
    map->page_list_ptrs[i] = 0;
  }

  {
    for (i = 0; i < npages; i++)
    {
      print_memalloc(("mem_map_alloc_kernel[%d] at %p\n", i, map->page_list_ptrs + i));
      map->page_list_ptrs[i] = (EdtMemList*)edt_alloc_kernel_page("page_list_ptrs[i]");

      print_memalloc(("mem_map_alloc_kernel.page_addresses[%d] = %p\n", i, map->page_list_ptrs[i]));

      if (!map->page_list_ptrs[i])

      {
        edt_printk("mem_map_alloc_kernel:  Error allocating dma page\n");
        edt_mem_map_clear_kernel_buffers(map, i);

        return -ENOMEM;
      }
    }

    for (i = 0; i < npages; i++)
    {
      map->page_addresses[i] = (void*)map->page_list_ptrs[i]->vaddr;
      set_page_dirty(map->page_list_ptrs[i]->page);

      print_memalloc(("mem_map_alloc_kernel.page_addresses[%d] = %p\n", i, map->page_addresses[i]));
    }
  }

  map->size = size;

  map->npages = npages;
  map->own_pages = 1;
  map->is_mapped = 1;
  map->offset = 0;
  map->is_dma_mapped = 0;

  print_memalloc(("<== mem_map_alloc_kernel\n"));

  return rc;
}

/*
 * For multiple chunks, merge last two to spread out interrupts...
 */
asmlinkage int edt_merge_last_chunks(EdtMemMapping* map)
{
  const int chunk_cur = map->nchunks - 1;
  const int chunk_before = chunk_cur - 1;
  const int chunk_size = PAGE_SIZE / 8;

  if (map->nchunks <= 1)
  {
    return 0;
  }

  if (map->chunks[chunk_cur].count < (chunk_size / 2))
  {
    EdtPCISg* vchunk_from = map->sg_list_ptrs[chunk_before]->vaddr;
    EdtPCISg* vchunk = map->sg_list_ptrs[chunk_cur]->vaddr;

    unsigned int i = 0;
    int index = 0;
    int move_amount = (map->chunks[chunk_before].count + map->chunks[chunk_cur].count) / 2;
    int move_entries = move_amount / 8;
    int move_target = move_entries;
    int move_source = (map->chunks[chunk_before].count / 8) - move_entries;

    print_memalloc1(("Before merge chunk %d count = %x chunk %d count = %x\n",
                     chunk_before,
                     map->chunks[chunk_before].count,
                     chunk_cur,
                     map->chunks[chunk_cur].count));
    print_memalloc1(
        ("Adjusting last SG chunk to balance amount %d size %d\n", move_amount, move_entries));

    for (i = 0; i < map->chunks[chunk_before].count / 8; i++)
      print_memalloc1(
          ("%d:%3d %08x %08x\n", chunk_before, i, vchunk_from[i].busaddr, vchunk_from[i].count));

    for (i = 0; i < map->chunks[chunk_cur].count / 8; i++)
      print_memalloc1(("%d:%3d %08x %08x\n", chunk_cur, i, vchunk[i].busaddr, vchunk[i].count));

    print_memalloc1(("Moving vchunk to %d\n", move_target));

    for (index = 0; index < move_entries; index++, move_target++)
    {
      vchunk[move_target] = vchunk[index];
    }

    move_source = (map->chunks[chunk_before].count / 8) - move_entries;

    print_memalloc1(("Moving vchunk_from from %d\n", move_source));
    for (index = 0; index < move_entries; index++, move_source++)
    {
      vchunk[index] = vchunk_from[move_source];
    }

    move_amount = move_entries * 8;

    map->chunks[chunk_before].count -= move_amount;
    map->chunks[chunk_cur].count += move_amount;

    print_memalloc1(("After merge chunk %d count = %x chunk %d count = %x\n",
                     chunk_before,
                     map->chunks[chunk_before].count,
                     chunk_cur,
                     map->chunks[chunk_cur].count));

    for (i = 0; i < map->chunks[chunk_before].count / 8; i++)
      print_memalloc1(
          ("%d:%3d %08x %08x\n", chunk_before, i, vchunk_from[i].busaddr, vchunk_from[i].count));

    for (i = 0; i < map->chunks[chunk_cur].count / 8; i++)
      print_memalloc1(("%d:%3d %08x %08x\n", chunk_cur, i, vchunk[i].busaddr, vchunk[i].count));
  }

  return 0;
}

/*******************************************
 * edt_mem_map_setup
 * Top-level buffer setup, will fill in map
 * with list of pages and the sglist and sglist
 * broken up into chunks
 * Assumes map->pci_dev is set already
 *******************************************/

asmlinkage int edt_mem_map_setup(EdtMemMapping* map, int bufsize)
{
  // Check dma page allocation status
  int rc = 0;
  int npages = bufsize >> PAGE_SHIFT;
  u_char build_new = 1;
  struct scatterlist* sg = NULL;
  size_t total = 0;
  int offset = 0;
  int i = 0;
  int sgsize = 0;

  EdtMemList** pages;

  unsigned int cur_chunk = 0;
  unsigned int chunk_count = 0;

  EdtPCISg* vchunk = NULL;
  unsigned int chunk_size = 0;

  if (bufsize > (npages << PAGE_SHIFT))
    npages++;

  /* return mem-mapped address */

  print_memalloc1(("==> edt_mem_map_setup:   map %p %x old size %x size %x\n",
                   map,
                   map->page_addresses,
                   map->npages,
                   npages));

  if (map->page_addresses) /* already allocated something */
  {
    /* check if already ready */
    /* only keep old if npages is the same */
    if (npages && (npages == map->npages))
      build_new = 0;

    edt_mem_map_release(map);

    /* size of 0 frees buffer memory */
    if (npages == 0)  // we're done
    {
      return 0;
    }
  }

  // allocate dma pages if needed     -> page_list_ptrs

  if (build_new)
  {
    rc = edt_mem_map_alloc_kernel(map, bufsize);
    if (rc != 0)
    {
      edt_printk("Failed to allocate DMA pages: %s\n", __func__);
      return rc;
    }
  }

  // allocate os_sglist and map pages -> os_sglist
  /* linux sg structure */

  total = map->size;
  sgsize = map->npages * sizeof(struct scatterlist);

  sg = edt_get_kernel_mem(sgsize, "map->os_sglist");
  if (sg == NULL)
  {
    edt_printk("%s:  out of memory for sglist\n", __func__);
    return -ENOMEM;
  }

  memset(sg, 0, sgsize);

#if IS_ENABLED(CONFIG_DEBUG_SG)
  sg_init_table(sg, map->npages);
#endif

  pages = (EdtMemList**)map->page_list_ptrs;
  if (NULL == pages)
  {
    edt_printk("%s: (null ptr) kernel buffers not allocated\n", __func__);
    return -ENOMEM;
  }

  if (NULL == pages[0])
  {
    pr_err("(null ptr) Cannot set SG page. pages[%d] is cleared/unallocated.\n", 0);
    return -ENOMEM;
  }

  sg_set_page(&sg[0], pages[0]->page, PAGE_SIZE - offset, offset);

  if (sg[0].length > total)
    sg[0].length = total;

  total -= sg[0].length;

  for (i = 1; i < map->npages && total > 0; i++)
  {
    if (NULL == pages[i])
    {
      pr_err("(null ptr) Cannot set SG page. pages[%d] is cleared/unallocated.\n", i);
      return -ENOMEM;
    }

    sg_set_page(&sg[i], pages[i]->page, (total >= PAGE_SIZE) ? PAGE_SIZE : total, 0);
    total -= sg[i].length;
  }

  map->os_sglist = sg;
  map->sgsize = sgsize;

// map in dma memory using sg
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
  map->sglist_count = dma_map_sg(&(map->pci_dev->dev),
                                 sg,
                                 map->npages,
                                 (map->rw) ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
#else
  map->sglist_count =
      pci_map_sg(map->pci_dev, sg, map->npages, (map->rw) ? PCI_DMA_TODEVICE : PCI_DMA_FROMDEVICE);
#endif

  map->is_dma_mapped = 1;

  map->sg_entries_needed = 0;

  print_memalloc1(("<== mem_map_setup sglist_count = %d\n", map->sglist_count));

  // allocate sg chunks as we go
  // allocate DMA capable sg list     -> sg_list_ptrs, chunks, nalloc_chunks

  map->nalloc_chunks = (map->npages + 511) / 512;
  map->sg_list_ptrs = (EdtMemList**)edt_get_kernel_mem(map->nalloc_chunks * sizeof(EdtMemList*),
                                                       "map->sg_list_ptrs");
  if (map->sg_list_ptrs == NULL)
  {
    edt_printk("Failed to allocate sg_list_ptrs: %s\n", __func__);
    return -1;
  }

  /* chunks is an array of busaddr/count pairs for
      the sg pages */

  map->chunks = (EdtPCISg*)edt_get_kernel_mem(map->nalloc_chunks * sizeof(EdtPCISg), "map->chunks");
  if (map->chunks == NULL)
  {
    edt_printk("Failed to allocate map chunks: %s\n", __func__);
    return -1;
  }

  // allocate dma memory for SG lists

  for (i = 0; i < (int)map->nalloc_chunks; i++)
  {
    /* sg_list_ptrs is an array of allocated pages
       to hold DMA-able memory */
    map->sg_list_ptrs[i] = (EdtMemList*)edt_alloc_kernel_page("sg_list_ptrs[i]");
    if (map->sg_list_ptrs[i] == NULL)
    {
      edt_printk("Failed to allocate DMA memory for sg lists: %s\n", __func__);
      return -1;
    }

    /* get bus address of sg page */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
    map->sg_list_ptrs[i]->busaddr = dma_map_single(&(map->pci_dev->dev),
                                                   map->sg_list_ptrs[i]->vaddr,
                                                   PAGE_SIZE,
                                                   DMA_BIDIRECTIONAL);
#else
    map->sg_list_ptrs[i]->busaddr =
        pci_map_single(map->pci_dev, map->sg_list_ptrs[i]->vaddr, PAGE_SIZE, PCI_DMA_TODEVICE);
#endif

    map->chunks[i].busaddr = map->sg_list_ptrs[i]->busaddr;
    map->chunks[i].count = 0;

    print_memalloc1(
        ("Chunk %d address and count = %08x %08x\n", map->chunks[i].busaddr, map->chunks[i].count));
  }

  cur_chunk = 0;
  vchunk = (EdtPCISg*)map->sg_list_ptrs[0]->vaddr;
  chunk_count = 0;

  chunk_size = PAGE_SIZE / 8;

  print_memalloc1(("map->sglist_count = %d\n", map->sglist_count));

  for (i = 0; i < map->sglist_count; i++)
  {
    dma_addr_t addr = sg_dma_address(sg + i);

    dma_addr_t endaddr;
    int size;

    int tmpsize;

    size = sg_dma_len(sg + i);

    print_memalloc1(("Loop %d addr = %llx size = %08x\n", i, addr, size));

    endaddr = addr + size;

    /* deal with 64k-1 and 64k boundaries */
    /* to see how many actual entries we will need */

    while (addr < endaddr)
    {
      if (((addr + size) >> 16) != (addr >> 16))
      {
        u_int block_boundary = ((addr + size) >> 16) << 16;
        tmpsize = block_boundary - addr;
        if (tmpsize == 0x10000)
        {
          tmpsize = 0x8000;
        }
      }
      else
        tmpsize = size;

      map->sg_entries_needed++;

      // add to current chunk

      vchunk[chunk_count].busaddr = addr;
      vchunk[chunk_count].count = tmpsize;

      print_memalloc1(("&vchunk[chunk_count] = %p \n", vchunk + chunk_count));

      chunk_count++;
      map->chunks[cur_chunk].count += 8;

      print_memalloc1(("vaddr %p chunk_count %d count = %d addr %llx size %08x\n",
                       vchunk,
                       chunk_count,
                       map->chunks[cur_chunk].count,
                       addr,
                       tmpsize));

      size -= tmpsize;
      addr += tmpsize;

      if (chunk_count >= chunk_size)
      {
        map->chunks[cur_chunk].count = chunk_size * 8;
        chunk_count = 0;
        if (i < map->sglist_count - 1 && cur_chunk < map->nalloc_chunks - 1)
        {
          cur_chunk++;
          print_memalloc1(("Loop 1 Bumping cur_chunk\n"));
          vchunk = map->sg_list_ptrs[cur_chunk]->vaddr;
        }
      }
    }
  }

  print_memalloc1(("after loop cur_chunk = %d chunk_count = %d  count =%d \n",
                   cur_chunk,
                   chunk_count,
                   map->chunks[cur_chunk].count));
  if (map->chunks[cur_chunk].count == 0 && chunk_count)
  {
    map->chunks[cur_chunk].count = (chunk_count)*8;

    // FIX for TEGRA !!
    // cur_chunk++;

    print_memalloc1(("After loop Bumping cur_chunk\n"));
  }

  map->nchunks = cur_chunk + 1;

  // Deal with short last chunk
  edt_merge_last_chunks(map);

  for (cur_chunk = 0; cur_chunk < map->nchunks; cur_chunk++)
    print_memalloc1(("%d: map->chunk.addr = %08x map->chunk.count = %08x\n",
                     cur_chunk,
                     map->chunks[cur_chunk].busaddr,
                     map->chunks[cur_chunk].count));

  if (map->nchunks > map->nalloc_chunks)
    map->nchunks = map->nalloc_chunks;

  print_memalloc1(("nchunks = %d alloc_chunks %d\n", map->nchunks, map->nalloc_chunks));

  // at this point chunks has the busaddr of individual sg pages
  // plus counts

  return 0;
}

/*!
 * delete the sg array defined in edt_mem_map_setup_sg
 */
asmlinkage int edt_mem_map_clear_sg(EdtMemMapping* map)
{
  if (map->os_sglist)
  {
    edt_free_kernel_mem(map->os_sglist, map->sgsize, "map->os_sglist");

    map->os_sglist = NULL;
  }

  return 0;
}

asmlinkage void edt_mem_map_release(EdtMemMapping* map)
{
  if (!map)
  {
    edt_printk("Error null map variable\n");
    return;
  }

  if (map->is_dma_mapped && map->pci_dev)
  {
    edt_pci_unmap_sg(map->pci_dev, map);
  }

  if (map->page_list_ptrs && map->own_pages)
  {
    // kernel buffers
    edt_mem_map_clear_kernel_buffers(map, map->npages);
    edt_mem_map_clear_sg_ptrs(map, map->nalloc_chunks);
  }
  else
    return;

  map->is_mapped = 0;
}
