/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#include "edtdef.h"

#include "edt_lnx_kernel.h"
#include "edt_mem_tag.h"
#include "edt_lnx_mm.h"

#include <linux/kconfig.h>

#include <linux/sysctl.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/pci.h>

// Uncomment this to check for memory leaks - all allocations
// are tracked in one of 3 pools
// When the module is unloaded it will print the status of all
// allocations

// max entries 256k

EdtMemTagPool* kmalloc_tags = NULL;
EdtMemTagPool* vmalloc_tags = NULL;
EdtMemTagPool* page_tags = NULL;
EdtMemTagPool* page_cache_tags = NULL;

#ifdef DEBUG_LEAKS

#define MAX_MEMTAG_SIZE 64
#define MAX_MEM_TAGS    64

// struct to hold strings used in allocation functions
//
typedef struct _tag_pool {
  char tag[64];
  int adds;
  int refs;
} EdtTagPool;

static EdtTagPool tag_pool[MAX_MEM_TAGS];

static int ntags = 0;

int lookup_tag(const char* tag)
{
  int rc = -1;
  int i;

  if (ntags == 0)
    memset(&tag_pool, 0, sizeof(tag_pool));

  for (i = 0; i < ntags; i++)
  {
    if (!strncmp(tag, tag_pool[i].tag, MAX_MEMTAG_SIZE))
    {
      return i;
    }
  }

  return rc;
}

static int add_tag(const char* tag, int index)
{
  if (index == -1)
  {
    if (ntags < MAX_MEM_TAGS)
    {
      strncpy(tag_pool[ntags].tag, tag, MAX_MEMTAG_SIZE - 1);
      index = ntags;
      ntags++;
    }
    else
      return -1;
  }

  tag_pool[index].refs++;
  tag_pool[index].adds++;

  return index;
}

static int rm_tag(const char* tag, int index)
{
  if (index == -1 || index >= MAX_MEMTAG_SIZE)
  {
    return -1;
  }

  tag_pool[index].refs--;

  return index;
}

void add_mem_tag(EdtMemTagPool* pool, const void* addr, const int size, const char* tag)
{
  if (pool->entries < pool->total)
  {
    int e;
    int t;
    e = pool->first_empty;

    pool->tags[e].addr = (void*)addr;
    pool->tags[e].size = size;

    t = lookup_tag(tag);
    t = add_tag(tag, t);
    pool->tags[e].tagid = t;

    // update first_empty

    while (e < pool->total && (pool->tags[e].addr))
      e++;

    if (e < pool->total)
      pool->first_empty = e;

    pool->entries++;
    if (pool->entries > pool->max_entries)
      pool->max_entries = pool->entries;

    if (e > pool->last_empty)
      pool->last_empty = e;
  }
  else
    edt_printk("pool %s full\n", pool->name);

  pool->total_allocated += size;
  if (pool->total_allocated > pool->max_allocated)
    pool->max_allocated = pool->total_allocated;
}

void clear_mem_tag(EdtMemTagPool* pool, const void* addr, const char* tag)
{
  int e;
  int t;

  for (e = 0; e < pool->total; e++)
    if (pool->tags[e].addr == addr)
      break;

  if (e == pool->total)
    edt_printk("Addr in pool %s (%p) not found\n", pool->name, addr);
  else
  {
    pool->tags[e].addr = 0;
    t = lookup_tag(tag);
    if (t == -1)
      edt_printk("clear_mem_tag tag %s not found\n", tag);
    else
      rm_tag(tag, t);

    if (e < pool->first_empty)
      pool->first_empty = e;
    pool->entries--;
    pool->total_allocated -= pool->tags[e].size;
  }
}

void init_mem_pool(EdtMemTagPool** pool_ptr, char* name)
{
  EdtMemTagPool* pool = vmalloc(sizeof(EdtMemTagPool));

  memset(pool, 0, sizeof(*pool));

  strcpy(pool->name, name);

  pool->total = MEMTAG_SIZE;

  *pool_ptr = pool;
}

void init_mem_pools(void)
{
  init_mem_pool(&kmalloc_tags, "kmalloc");
  init_mem_pool(&vmalloc_tags, "vmalloc");
  init_mem_pool(&page_tags, "page_alloc");
  init_mem_pool(&page_cache_tags, "page_cache_alloc");
}

void clear_mem_pools(void)
{
  if (kmalloc_tags == NULL)
    return;

  vfree(kmalloc_tags);
  vfree(vmalloc_tags);
  vfree(page_tags);
  vfree(page_cache_tags);
}

void dump_mem_pool(EdtMemTagPool* pool)
{
  int e;

  int sum = 0;

  edt_printk("********** %12s **********\n", pool->name);

  for (e = 0; e < pool->total; e++)
  {
    if (pool->tags[e].addr)
    {
      if (pool->tags[e].tagid != -1)
        edt_printk("%4d: %p %d %d:%s\n",
                   e,
                   pool->tags[e].addr,
                   pool->tags[e].size,
                   pool->tags[e].tagid,
                   tag_pool[pool->tags[e].tagid].tag);
      else
        edt_printk("%4d: %p %d\n", e, pool->tags[e].addr, pool->tags[e].size);
      sum += pool->tags[e].size;
    }
  }

  edt_printk("Pool %12s Total size = %8d Max allocated = %8d\n",
             pool->name,
             sum,
             pool->max_allocated);
  edt_printk("**********************************\n", pool->name);
}
void dump_mem_pools(void)
{
  int i;

  for (i = 0; i < ntags; i++)
    edt_printk("%2d: %d (of %d adds) x %s\n",
               i,
               tag_pool[i].refs,
               tag_pool[i].adds,
               tag_pool[i].tag);

  dump_mem_pool(kmalloc_tags);
  dump_mem_pool(vmalloc_tags);
  dump_mem_pool(page_tags);
  dump_mem_pool(page_cache_tags);
}

#endif
