/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 * @brief Defines or constants for the device id defined in the PCI/PCIe
 * firmware and for the stored mezzanine id values for mezzanine cards.
 *
 * @details
 *
 * Some (not all) EDT PCI devices have human-readable names listed in the PCI ID
 * Repository. The `lspci` tool on Linux can parse the "pci.ids" file to show
 * the human-readable names in its output.
 *
 * See https://admin.pci-ids.ucw.cz/mods/PC/123d and `man update-pciids`.
 */

#ifndef EDT_PCI_DEVICES_H
#define EDT_PCI_DEVICES_H

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(_KERNEL)
#include "edtdef.h"
#endif

#if defined(PCD) && (PCD != 0)
#include "edt_pci_devices_pcd.h"
#elif defined(PDV) && (PDV != 0)
#include "edt_pci_devices_pdv.h"
#else
#error "Must define EDT driver type"
#endif

/*! Engineering Design Team's PCI Vendor ID. */
#define PCI_VENDOR_ID_EDT 0x123dU

/* Enum of EDT PCI device PIDs */
enum EDT_PCI_PID {
  /** Unknown PCI product. */
  EDT_PCI_UNKNOWN_ID = 0U,
  /** Dummy/fake device ID */
  DMY_ID = 0x0080U,

#define EXPAND_AS_DEFNAME_ENUM(_defname, _pid, _dma_ch, _name, _desc) _defname = _pid,

  EDT_PCI_DEVICES_LISTX(EXPAND_AS_DEFNAME_ENUM)

#undef EXPAND_AS_DEFNAME_ENUM
};

#define EDT_DEVID_IS_DUMMY(id) ((id == DMY_ID))

/* Allow user-space EDT libraries to include this header to get product VID/PID
 * constants, but functions here are kernel-only. Can be exposed to userspace
 * via IOCTLs. */
#if defined(_KERNEL) && (_KERNEL != 0)

#if defined(__linux__)
#include <linux/types.h>
#endif

/*!
 * @brief Get the friendly name of an EDT device, given its PCI PID value.
 *
 * @param[in] pid
 *     PCI Product ID value.
 * @param[out] cstr
 *     A pointer to static-const string is returned here if a match was found.
 *     May be returned as NULL otherise.
 * @return
 *     Zero if string for the given pid was found.
 */
int edt_pcidev_str_from_pid(enum EDT_PCI_PID pid, const char** const cstr);

/*!
 * @brief Get the (max) number of DMA channels an EDT PCI device is capable of.
 *
 * This is the number of channel devices that the EDT kernel driver is expected
 * to create during board discovery and enumeration. Some boards or variants
 * of EDT firmware may implement "dummy" DMA channels to fulfill the total DMA
 * channel requirement. In other words, depending on the given firmware running
 * on a given EDT card, not all DMA channels advertized may be useable.
 *
 * @param[in] pid
 *     PCI Product ID value.
 * @returns
 *     Negative error code if the given pid was not found.
 *     If pid was found, the number of DMA channels (>=0).
 */
int edt_pcidev_num_dma_channels(enum EDT_PCI_PID pid);

#endif /* _KERNEL */

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* EDT_PCI_DEVICES_H */
