/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 */

#include "edtdef.h"

#include "edtdrv.h"

#define TRACE_REGS
/* #define ACCUMULATE_REG_STATS */
#ifdef ACCUMULATE_REG_STATS

static u_int local_reg_sets[256];
static u_int local_reg_gets[256];

#endif

static int edt_reg_tracable(Edt_Dev* edt_p, u_int reg_def);
static void edt_set_recursive(Edt_Dev* edt_p, u_int reg_def, u_int val);

/* reg_trace_mask determines which registers to dump */

int edt_reg_tracable(Edt_Dev* edt_p, u_int reg_def)
{
  u_short offset = (u_short)EDT_REG_ADDR(reg_def);
  u_char type = (u_char)EDT_REG_TYPE(reg_def);

  if (type <= 3 && edt_p->reg_trace_mask[type] && offset < 0x100 &&
      edt_p->reg_trace_mask[type][offset])
    return 1;

  return 0;
}

void edt_reg_set_trace_mask(Edt_Dev* edt_p, u_int reg_def)
{
  u_short offset = (u_short)EDT_REG_ADDR(reg_def);
  u_char size = (u_char)EDT_REG_SIZE(reg_def) & 0x7f;
  u_char type = (u_char)EDT_REG_TYPE(reg_def);
  u_int i;
  int state = (reg_def & 0x80000000) ? 1 : 0;

  if (state)
    edt_p->m_TraceLevel = 1;

  if (type > 3)
  {
    EDTPRT(edt_p, DBG_REG, TRL_ERROR, ("Can only trace regs w/ type <= 3\n"));
    return;
  }

  if (edt_p->reg_trace_mask[type] == NULL)
  {
#if defined(__linux__)
    edt_p->reg_trace_mask[type] = (u_char*)edt_vmalloc(256, "reg_trace_mask");
#elif defined(_NT_DRIVER_)
    edt_p->reg_trace_mask[type] = (u_char*)edt_vmalloc(256, "reg_trace_mask");
#else
    /* not implemented */
    return;
#endif
  }

  /* special reg offset will turn on all
  registers */
  if (offset == 0xffff)
  {
    for (i = 0; i < 256; i++)
      edt_p->reg_trace_mask[type][i] = (u_char)state;

    /* exception  - REMOTE will show up too much, they
    must be turned on explicitly */

    if (type == 0)
    {
      edt_p->reg_trace_mask[type][EDT_REMOTE_OFFSET & 0xff] = 0;
      edt_p->reg_trace_mask[type][EDT_REMOTE_DATA & 0xff] = 0;
    }
  }
  else
  {
    for (i = 0; i < size; i++)
      edt_p->reg_trace_mask[type][offset + i] = (u_char)state;
  }
}

void edt_set_recursive(Edt_Dev* edt_p, u_int reg_def, u_int val)
{
  u_short offset = (u_short)EDT_REG_ADDR(reg_def);
  u_char size = (u_char)EDT_REG_SIZE(reg_def);
  u_char type = (u_char)EDT_REG_TYPE(reg_def);
  u_char register_access_class = EDT_REG_CLASS(reg_def);
  u_short startoffset = offset;
  u_int startval = val;
  volatile u_int tmpval = 0;
  (void)tmpval;

  EDTPRT(edt_p,
         DBG_REG,
         TRL_INFO,
         ("edt_set: mem: %p reg %08x val %08x\n", edt_p->bar0_mmap, reg_def, val));

  if (register_access_class) /* Extended register access; see comment at start of edtreg.h.  -mcm
                                4/09 */
  {
    switch (register_access_class)
    {
    case 1:
    {
      size &= 0x0f;

      EDTPRT(edt_p,
             DBG_REG,
             TRL_INFO,
             ("edt_set: ir base %x reg %08x size %d val %08x\n",
              edt_p->m_IndRegBase,
              reg_def,
              size,
              val));

      for (; size--; val >>= 8)
      {
        /* 24-bit address */
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase), reg_def & 0x000000ff);
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase + 1), (reg_def & 0x0000ff00) >> 8);
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase + 2), (reg_def & 0x00ff0000) >> 16);

        /* 8-bit data */
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase + 3), val & 0xff);

        ++reg_def;
      }
      break;
    }

    case 2:
    {
      /* 24-bit address */
      edt_set(edt_p, PCD_CMD | 4, reg_def & 0x000000ff);
      edt_set(edt_p, PCD_CMD | 3, (reg_def & 0x0000ff00) >> 8);
      edt_set(edt_p, PCD_CMD | 2, (reg_def & 0x00ff0000) >> 16);

      /* 32-bit data */
      edt_set(edt_p, PCD_CMD | 9, (val & 0xff000000) >> 24);
      edt_set(edt_p, PCD_CMD | 8, (val & 0x00ff0000) >> 16);
      edt_set(edt_p, PCD_CMD | 7, (val & 0x0000ff00) >> 8);
      edt_set(edt_p, PCD_CMD | 5, (val & 0x000000ff));

      break;
    }

    case 3:
    {
      u_int offset28 = reg_def & 0x0fffffff;

      EDT_PCI_MAPMEM(edt_p);

      PCI_MEM2_PUTL(edt_p, (offset28), val);

      /* readback to flush cache */
      /* turned on for all 6/1/07 jsc - then off again for performance 6/26/07 */

      // Read back from the same address instead of addr0.  Mark & Seb May 9, 2012
      tmpval = PCI_MEM2_GETL(edt_p, offset28);

      EDT_PCI_UNMAPMEM(edt_p);

      break;
    }

    case 7:
    {
      u_int offset28 = reg_def & 0x0fffffff;

      // Clear interface enable bit
      edt_set(edt_p, PCD_CMD | (0x4c), 0x00);

      /* 27-bit address */
      edt_set(edt_p, PCD_CMD | 0x04000000 | 0x44, offset28);

      /* 32-bit data */
      edt_set(edt_p, PCD_CMD | 0x04000000 | 0x48, val);

      // Set interface enable and direction bits
      edt_set(edt_p, PCD_CMD | (0x4c), 0x03);
      edt_set(edt_p, PCD_CMD | (0x4c), 0x00);

      break;
    }
    }
  }
  else  // Register access class 0
  {
    switch (type)
    {
    case LOCAL_DMA_TYPE:
      offset += edt_p->m_DmaChannel * EDT_CHAN_OFFSET;
#if defined(PCD)
      if (EDT_DEVID_USES_CHAN16_BASE(edt_p->devid))
      {
        offset += EDT_CHAN16_BASE;
      }
#endif
      /* FALLTHROUGH */

    case LOCAL_XILINX_TYPE:
    case REG_FILE_TYPE:
      EDT_PCI_MAPMEM(edt_p);
      switch (size)
      {
      case 1:
        PCI_PUTC(edt_p, (offset), (u_char)val);
        break;
      case 2:
        PCI_PUTS(edt_p, (offset), val);
        break;
      case 4:
        PCI_PUTL(edt_p, (offset), val);
        break;
      }

      /* readback to flush cache */
      /* turned on for all 6/1/07 jsc - then off again for performance 6/26/07 */
      tmpval = PCI_GETC(edt_p, 0);
      EDT_PCI_UNMAPMEM(edt_p);

      /* make sure shadow reg up to date */
      if (reg_def == EDT_DMA_CFG || reg_def == EDT_DMA_INTCFG)
      {
        if (edt_p->m_chan0_p)
          edt_p->m_chan0_p->m_DmaIntCfg = (val >> 8) & 0xff;
        else
          edt_p->m_DmaIntCfg = (val >> 8) & 0xff;
        EDTPRT(edt_p, DBG_INTR_EN, TRL_WARN, ("Setting intcfg to %02x\n", edt_p->m_DmaIntCfg));
      }
      else if (reg_def == EDT_DMA_INTCFG_ONLY)
      {
        if (edt_p->m_chan0_p)
          edt_p->m_chan0_p->m_DmaIntCfg = (u_char)val;
        else
          edt_p->m_DmaIntCfg = (u_char)val;
        EDTPRT(edt_p, DBG_INTR_EN, TRL_WARN, ("Setting intcfg to %02x\n", edt_p->m_DmaIntCfg));
      }

#ifdef ACCUMULATE_REG_STATS
      local_reg_sets[offset]++;
#endif
      break;

    case REMOTE_XILINX_TYPE:
    {
      /*
       * Set the register address: Get the contents of the cmd_config
       * register, then replace register address and auto increment (lower
       * 8 bits)
       */

      EDTPRT(edt_p, DBG_REG, TRL_INFO, ("Remote set %d	at %x\n", size, offset));

#if defined(PDV)
      if (edt_p->devid == PE1DVVL_ID || edt_p->devid == PE4DVVL_ID)
      {
        offset += edt_p->m_DmaChannel * EDT_CL_INTFC_OFFSET;
      }
      else if (edt_p->devid == PE8VLCLS_ID)
      {
        offset += edt_p->m_DmaChannel * 0x80;
      }
#endif /* PDV */

      /* This should deal with interrupts accessing the remote
      registers */

      /* Now write the data to the Xilinx port */
      for (; size--; val >>= 8)
      {
        edt_set(edt_p, EDT_REMOTE_OFFSET, offset & 0xff);

        edt_set(edt_p, EDT_REMOTE_DATA, val & 0xff);

        offset++;
      }
    }

    break;

    case MAC8100_TYPE:
      edt_set(edt_p, EDT_REMOTE_OFFSET, offset);
      /* to add time - give time for address to latch before data */
      tmpval = edt_get(edt_p, EDT_REMOTE_OFFSET);
      edt_set(edt_p, EDT_REMOTE_DATA, val);
      break;

    default:
      break;
    }
  }

#ifdef TRACE_REGS

  if ((edt_p->m_TraceLevel & 1) && edt_reg_tracable(edt_p, reg_def))
  {
    EdtTrace(edt_p, TR_REG_SET);
    EdtTrace(edt_p, (reg_def & 0xffff0000) + startoffset);
    EdtTrace(edt_p, startval);
  }

#endif
}

static u_int edt_get_recursive(Edt_Dev* edt_p, u_int reg_def)
{
  u_short offset = (u_short)EDT_REG_ADDR(reg_def);
  u_char size = (u_char)EDT_REG_SIZE(reg_def);
  u_char type = (u_char)EDT_REG_TYPE(reg_def);
  u_char register_access_class = (u_char)EDT_REG_CLASS(reg_def);
  u_char count = size;
  u_short startoffset = offset;
  int shift;
  volatile u_int tmpval;

  u_int val, retval = 0xffffffff;

  EDTPRT(edt_p,
         DBG_REG,
         TRL_INFO,
         ("edt_get mem:: %p at %x size %x type %d \n", edt_p->bar0_mmap, offset, size, type));

  if (register_access_class) /* Extended register access; see comment at start of edtreg.h.  -mcm
                                4/09 */
  {
    switch (register_access_class)
    {
    case 1:
    {
      size &= 0x0f;
      val = 0;
      shift = 0;
      EDTPRT(edt_p,
             DBG_REG,
             TRL_INFO,
             ("edt_get: ir base %x reg %08x size %d val %08x\n",
              edt_p->m_IndRegBase,
              reg_def,
              size,
              val));

      for (; size--; shift += 8)
      {
        /* 24-bit address */
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase), reg_def & 0x000000ff);
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase + 1), (reg_def & 0x0000ff00) >> 8);
        edt_set(edt_p, PCD_CMD | (edt_p->m_IndRegBase + 2), (reg_def & 0x00ff0000) >> 16);

        /* 8-bit data */
        val |= (edt_get(edt_p, PCD_CMD | (edt_p->m_IndRegBase + 3)) & 0xff) << shift;

        ++reg_def;
      }

      retval = val;
      break;
    }

    case 2:
    {
      /* 24-bit address */
      edt_set(edt_p, PCD_CMD | 4, reg_def & 0x000000ff);
      edt_set(edt_p, PCD_CMD | 3, (reg_def & 0x0000ff00) >> 8);
      edt_set(edt_p, PCD_CMD | 2, (reg_def & 0x00ff0000) >> 16);

      /* 32-bit data */

      /* trigger read */

      edt_set(edt_p, PCD_CMD | 6, 0);

      retval = 0;
      retval |= (edt_get(edt_p, PCD_CMD | 9) << 24);
      retval |= (edt_get(edt_p, PCD_CMD | 8) << 16);
      retval |= (edt_get(edt_p, PCD_CMD | 7) << 8);
      retval |= edt_get(edt_p, PCD_CMD | 5);
      break;
    }

    case 3:
    {
      u_int offset28 = reg_def & 0x0fffffff;

      EDT_PCI_MAPMEM(edt_p);

      retval = (u_int)PCI_MEM2_GETL(edt_p, (offset28));

      EDT_PCI_UNMAPMEM(edt_p);

      break;
    }

    case 7:
    {
      u_int regAddr = reg_def & 0x0fffffff;

      retval = 0;

      // Clear interface enable bit
      edt_set(edt_p, PCD_CMD | (0x4c), 0x00);

      /* 27-bit address */
      edt_set(edt_p, PCD_CMD | 0x04000000 | 0x44, regAddr);

      // Set interface enable and direction bits
      edt_set(edt_p, PCD_CMD | (0x4c), 0x01);

      // Wait for register data valid, but not too long
      while ((edt_get(edt_p, PCD_CMD | 0x4c) & 0x04) == 0)
        ;

      /* 32-bit data */
      retval = edt_get(edt_p, PCD_CMD | 0x04000000 | 0x48);

      edt_set(edt_p, PCD_CMD | (0x4c), 0x00);

      break;
    }
    }
  }
  else  // Register access class 0
  {
    switch (type)
    {
    case LOCAL_DMA_TYPE:
      offset += edt_p->m_DmaChannel * EDT_CHAN_OFFSET;
#if defined(PCD)
      if (EDT_DEVID_USES_CHAN16_BASE(edt_p->devid))
      {
        offset += EDT_CHAN16_BASE;
      }
#endif
      /* FALLTHROUGH */

    case LOCAL_XILINX_TYPE:
    case REG_FILE_TYPE:

      EDT_PCI_MAPMEM(edt_p);

      switch (size)
      {
      case 1:
        retval = (u_int)PCI_GETC(edt_p, (offset));
        break;
      case 2:
        retval = (u_int)PCI_GETS(edt_p, (offset));
        break;
      case 4:
        retval = (u_int)PCI_GETL(edt_p, offset);
        break;

      default:
        break;
      }
#ifdef ACCUMULATE_REG_STATS
      local_reg_gets[offset]++;
#endif

      if (reg_def == EDT_DMA_CFG || reg_def == EDT_DMA_INTCFG)
      {
        if (edt_p->m_chan0_p)
          edt_p->m_chan0_p->m_DmaIntCfg = (u_char)((retval >> 8) & 0xff);
        else
          edt_p->m_DmaIntCfg = (u_char)((retval >> 8) & 0xff);
      }
      else if (reg_def == EDT_DMA_INTCFG_ONLY)
      {
        if (edt_p->m_chan0_p)
          edt_p->m_chan0_p->m_DmaIntCfg = (u_char)retval;
        else
          edt_p->m_DmaIntCfg = (u_char)retval;
      }

      EDT_PCI_UNMAPMEM(edt_p);

      break;

    case REMOTE_XILINX_TYPE:
      /*
       * Set the register address: Get the contents of the cmd_config
       * register. Next replace register address and auto increment (lower
       * 8 bits).
       */
      {
        EDTPRT(edt_p, DBG_REG, TRL_INFO, ("Remote get %d at	%x\n", count, offset));

        /* Now read the	data from the Xilinx port */
        shift = 0;
        val = 0;

#if defined(PDV)
        if (edt_p->devid == PE1DVVL_ID || edt_p->devid == PE4DVVL_ID)
        {
          offset += edt_p->m_DmaChannel * EDT_CL_INTFC_OFFSET;
        }
        else if (edt_p->devid == PE8VLCLS_ID)
        {
          offset += edt_p->m_DmaChannel * 0x80;
        }
#endif /* PDV */

        for (val = 0; count--; shift += 8)
        {
          edt_set(edt_p, EDT_REMOTE_OFFSET, offset & 0xff);

          tmpval = edt_get(edt_p, EDT_REMOTE_DATA) & 0xff;

          offset++;

          val |= (tmpval << shift);
          retval = val;
        }
        EDTPRT(edt_p, DBG_REG, TRL_INFO, ("Remote value %x\n", retval));
      }

      break;

    case MAC8100_TYPE:
      edt_set(edt_p, EDT_REMOTE_OFFSET, offset);
      /* to add time - give time for address to latch before data */
      tmpval = edt_get(edt_p, EDT_REMOTE_OFFSET);
      retval = edt_get(edt_p, EDT_REMOTE_DATA) & 0xffff;
      break;

    default:
      break;
    }
  }

  EDTPRT(edt_p, DBG_REG, TRL_INFO, ("edt_get reg %08x val %08x\n", reg_def, retval));

#ifdef TRACE_REGS
  if ((edt_p->m_TraceLevel & 1) && edt_reg_tracable(edt_p, reg_def))
  {
    EdtTrace(edt_p, TR_REG_GET);
    EdtTrace(edt_p, (reg_def & 0xffff0000) + startoffset);
    EdtTrace(edt_p, retval);
  }
#endif

  return (retval);
}

/* Actual routine to get/set registers is in
edt_set/get_recursive. This outer call is to track register accesses
*/

void edt_set(Edt_Dev* edt_p, u_int reg_def, u_int val)
{
#ifdef LOG_REG_ACCESS
  u_short offset = (u_short)EDT_REG_ADDR(reg_def);
  u_char type = (u_char)EDT_REG_TYPE(reg_def);

  switch (type)
  {
  case LOCAL_XILINX_TYPE:
    if (offset < 256)
      edt_p->m_direct_writes[offset]++;
    break;
  case LOCAL_DMA_TYPE:
    if (offset < 32)
      edt_p->m_dma_writes[offset]++;
    break;
  case REMOTE_XILINX_TYPE:
    if (offset < 256)
      edt_p->m_indirect_writes[offset]++;
    break;
  }
#endif

  edt_set_recursive(edt_p, reg_def, val);
}

u_int edt_get(Edt_Dev* edt_p, u_int reg_def)
{
#ifdef LOG_REG_ACCESS
  u_short offset = (u_short)EDT_REG_ADDR(reg_def);
  u_char type = (u_char)EDT_REG_TYPE(reg_def);

  switch (type)
  {
  case LOCAL_XILINX_TYPE:
    if (offset < 256)
      edt_p->m_direct_reads[offset]++;
    break;
  case LOCAL_DMA_TYPE:
    if (offset < 32)
      edt_p->m_dma_reads[offset]++;
    break;
  case REMOTE_XILINX_TYPE:
    if (offset < 256)
      edt_p->m_indirect_reads[offset]++;
    break;
  }
#endif

  return edt_get_recursive(edt_p, reg_def);
}

/*!
 * @brief On PCD cards, sets DMA direction to read or write.
 *
 * Most users will not need to use this function, but instead can just
 * set the direction when calling @ref edt_configure_ring_buffers().
 *
 * @param edt_p
 *     Pointer to edt device structure returned by @ref edt_open() or
 *     @ref edt_open_channel()
 * @param direction
 *     Either EDT_READ or EDT_WRITE (not both)
 */
#if defined(PCD)
void edtdrv_set_direction(Edt_Dev* edt_p, int direction)
{
  u_char cmd;
  EDTPRT(edt_p, DBG_DMA_SETUP, TRL_INFO, ("--> edtdrv_set_direction %d\n", direction));

  cmd = (u_char)edt_get(edt_p, PCD_CMD);
  if (direction)
  {
    /* Set the direction register for DMA output */
    if (EDT_DEVID_HAS_PCD_DIR_BIT(edt_p->devid))
    {
      /* Enable board	and input direction */
      cmd &= ~PCD_DIR;
    }

    edt_p->m_AutoDir = 0;

    cmd |= PCD_ENABLE;
    edt_set(edt_p, PCD_CMD, cmd);
  }
  else
  {
    /* Set the direction register for DMA input */
    if (EDT_DEVID_HAS_PCD_DIR_BIT(edt_p->devid))
    {
      cmd |= PCD_DIR;
    }

    edt_p->m_AutoDir = 0;

    cmd |= PCD_ENABLE;
    edt_set(edt_p, PCD_CMD, cmd);
  }
}
#endif /* PCD */

u_int edt_and(Edt_Dev* edt_p, u_int type, u_int val)
{
  u_int tmp;

  tmp = edt_get(edt_p, type);
  tmp &= val;
  edt_set(edt_p, type, tmp);

  return tmp;
}

u_int edt_or(Edt_Dev* edt_p, u_int type, u_int val)
{
  u_int tmp;

  tmp = edt_get(edt_p, type);
  tmp |= val;
  edt_set(edt_p, type, tmp);

  return tmp;
}

u_int edt_clear(Edt_Dev* edt_p, u_int type, u_int val)
{
  u_int tmp;

  tmp = edt_get(edt_p, type);
  tmp &= ~val;
  edt_set(edt_p, type, tmp);

  return tmp;
}

/************************************
Somewhat optimized code for programming
the interface FPGA
***********************************/

#define X_CLOCK_ONE  (X_DATA | X_CCLK | X_INIT | X_PROG)
#define X_CLOCK_ZERO (X_CCLK | X_INIT | X_PROG)

static u_int edt_xilinx_lut[256][8];

static int edt_xilinx_lut_initialized = 0;

#define DMA_CFG_OFFSET (u_short) EDT_REG_ADDR(EDT_DMA_CFG)

static void edt_setget_prog(Edt_Dev* edt_p, u_int val)
{
  PCI_PUTL(edt_p, DMA_CFG_OFFSET, val);
  (void)PCI_GETL(edt_p, DMA_CFG_OFFSET);
}

static void edt_set_prog(Edt_Dev* edt_p, u_int val)
{
  PCI_PUTL(edt_p, DMA_CFG_OFFSET, val);
}

static void edt_xilinx_lut_init(void)
{
  int c;
  u_int* sets;

  for (c = 0; c < 256; c++)
  {
    sets = edt_xilinx_lut[c];

    sets[0] = (c & 0x80) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[1] = (c & 0x40) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[2] = (c & 0x20) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[3] = (c & 0x10) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[4] = (c & 0x8) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[5] = (c & 0x4) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[6] = (c & 0x2) ? X_CLOCK_ONE : X_CLOCK_ZERO;
    sets[7] = (c & 0x1) ? X_CLOCK_ONE : X_CLOCK_ZERO;
  }

  edt_xilinx_lut_initialized = 1;
}

void edtdrv_program_flash(Edt_Dev* edt_p, const u_char* bp, int size, int setget)
{
  int i;
  u_int* sets;
  u_char c;

  if (!edt_xilinx_lut_initialized)
    edt_xilinx_lut_init();

  if (setget)
  {
    for (i = 0; i < size; i++)
    {
      c = bp[i];

      sets = edt_xilinx_lut[c];

      edt_setget_prog(edt_p, sets[0]);
      edt_setget_prog(edt_p, sets[1]);
      edt_setget_prog(edt_p, sets[2]);
      edt_setget_prog(edt_p, sets[3]);
      edt_setget_prog(edt_p, sets[4]);
      edt_setget_prog(edt_p, sets[5]);
      edt_setget_prog(edt_p, sets[6]);
      edt_setget_prog(edt_p, sets[7]);
    }
  }
  else
  {
    for (i = 0; i < size; i++)
    {
      c = bp[i];

      sets = edt_xilinx_lut[c];

      edt_set_prog(edt_p, sets[0]);
      edt_set_prog(edt_p, sets[1]);
      edt_set_prog(edt_p, sets[2]);
      edt_set_prog(edt_p, sets[3]);
      edt_set_prog(edt_p, sets[4]);
      edt_set_prog(edt_p, sets[5]);
      edt_set_prog(edt_p, sets[6]);
      edt_set_prog(edt_p, sets[7]);
    }
  }
  return;
}

/* Functions to handle accesses to the DMA_CFG register, which is
used a lot, to minimize PCI accesses */

/* write value of m_DmaIntCfg to board*/

void edt_update_intcfg(Edt_Dev* edt_p)
{
  if (edt_p->m_DmaChannel)
  {
    edt_update_intcfg(edt_p->m_chan0_p);
    return;
  }

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("update intcfg %x\n", edt_p->m_DmaIntCfg));

  edt_set(edt_p, EDT_DMA_INTCFG_ONLY, edt_p->m_DmaIntCfg);
}

/* get value of EDT_DMA_INTCFG_ONLY  and
put in m_DmaIntCfg */

void edt_refresh_intcfg(Edt_Dev* edt_p)
{
  if (edt_p->m_DmaChannel)
  {
    edt_refresh_intcfg(edt_p->m_chan0_p);
    return;
  }

  edt_p->m_DmaIntCfg = (u_char)edt_get(edt_p, EDT_DMA_INTCFG_ONLY);
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("refresh intcfg %x\n", edt_p->m_DmaIntCfg));
}

void edt_set_intcfg(Edt_Dev* edt_p, u_int cfg)
{
  if (edt_p->m_DmaChannel)
  {
    edt_set_intcfg(edt_p->m_chan0_p, cfg);
    return;
  }

  edt_p->m_DmaIntCfg = (u_char)cfg;

  edt_set(edt_p, EDT_DMA_INTCFG_ONLY, cfg);

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("set_intcfg %x\n", cfg));
}

u_int edt_get_intcfg(Edt_Dev* edt_p)
{
  if (edt_p->m_DmaChannel)
    return edt_get_intcfg(edt_p->m_chan0_p);

  edt_p->m_DmaIntCfg = (u_char)edt_get(edt_p, EDT_DMA_INTCFG_ONLY);

  return edt_p->m_DmaIntCfg;
}

void edt_set_intcfg_bits(Edt_Dev* edt_p, u_int mask)
{
  if (edt_p->m_DmaChannel)
  {
    edt_set_intcfg_bits(edt_p->m_chan0_p, mask);
    return;
  }

#ifdef UPDATE_WHEN_NEEDED
  if ((edt_p->m_DmaIntCfg & mask) == mask)
    return;
#endif

  edt_p->m_DmaIntCfg |= mask;

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("set_intcfg_bits %x\n", mask));
  edt_update_intcfg(edt_p);
}

void edt_clear_intcfg_bits(Edt_Dev* edt_p, u_int mask)
{
  if (edt_p->m_DmaChannel)
  {
    edt_clear_intcfg_bits(edt_p->m_chan0_p, mask);
    return;
  }

#ifdef UPDATE_WHEN_NEEDED
  if ((edt_p->m_DmaIntCfg & mask) == 0)
    return;
#endif

  edt_p->m_DmaIntCfg &= ~mask;

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("clear_intcfg_bits %x\n", mask));
  edt_update_intcfg(edt_p);
}

void edt_disable_interrupts(Edt_Dev* edt_p)
{
  u_char cfg;
  if (edt_p->m_DmaChannel)
    edt_disable_interrupts(edt_p->m_chan0_p);

  cfg = edt_p->m_DmaIntMask;

  if (edt_p->m_TraceLevel & 4)
    EdtTrace(edt_p, TR_DISABLE_INT);

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("disable_interrupts %x\n", cfg));
  edt_clear_intcfg_bits(edt_p, cfg);

  cfg = edt_get(edt_p, EDT_DMA_CFG);
  EDTPRT(edt_p, DBG_ISR, TRL_INFO, ("<-- edt_disable_interrupts %08x\n", cfg));
}

void edt_enable_interrupts(Edt_Dev* edt_p)
{
  u_int cfg;

  if (edt_p->m_DmaChannel)
    edt_enable_interrupts(edt_p->m_chan0_p);
  cfg = edt_p->m_DmaIntMask;

  /* steve 12/21/10 removed enable TEMP interrupt */

  if (edt_p->m_TraceLevel & 4)
    EdtTrace(edt_p, TR_ENABLE_INT);
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("enable interrupts %x\n", cfg));
  edt_set_intcfg_bits(edt_p, cfg);

  cfg = edt_get(edt_p, EDT_DMA_CFG);

  EDTPRT(edt_p, DBG_ISR, TRL_INFO, ("<-- edt_enable_interrupts %08x\n", cfg));
}

u_int edt_get_rfifo(Edt_Dev* edt_p)
{
  if (edt_p->m_DmaChannel)
    return edt_get_rfifo(edt_p->m_chan0_p);
  return ((edt_p->m_DmaIntCfg & EDT_ICFG_RFIFO_ENB) == EDT_ICFG_RFIFO_ENB);
}

void edt_set_rfifo(Edt_Dev* edt_p, int val)
{
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("==> set_rfifo %d\n", val));

  if (edt_p->m_DmaChannel)
  {
    edt_set_rfifo(edt_p->m_chan0_p, val);
    return;
  }
  if (val)
    edt_p->m_DmaIntCfg |= EDT_ICFG_RFIFO_ENB;
  else
    edt_p->m_DmaIntCfg &= ~EDT_ICFG_RFIFO_ENB;

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("set rfifo %d\n", val));
  edt_update_intcfg(edt_p);

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("<== set_rfifo\n"));
}

void edt_set_interrupt_mask(Edt_Dev* edt_p, u_int val)
{
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("Set interrupt mask %x\n", val));
  if (edt_p->m_DmaChannel)
  {
    edt_set_interrupt_mask(edt_p->m_chan0_p, val);
    return;
  }

  edt_p->m_DmaIntMask = (u_char)val;

  edt_p->m_DmaIntCfg =
      (u_char)((edt_p->m_DmaIntCfg &
                ~(EDT_ICFG_RMT_EN_INTR | EDT_ICFG_PCI_EN_INTR | EDT_ICFG_TEMP_EN_INTR)) |
               val);

  edt_update_intcfg(edt_p);
}

u_int edt_get_interrupt_mask(Edt_Dev* edt_p)
{
  if (edt_p->m_DmaChannel)
  {
    return edt_get_interrupt_mask(edt_p->m_chan0_p);
  }
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("Get interrupt mask %x\n", edt_p->m_DmaIntMask));
  return (u_int)edt_p->m_DmaIntMask;
}

void edt_enable_remote_interrupt(Edt_Dev* edt_p)
{
  u_int mask = edt_get_interrupt_mask(edt_p) | (EDT_ICFG_RMT_EN_INTR | EDT_ICFG_PCI_EN_INTR);

  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("enable remote interrupt\n"));

  edt_set_interrupt_mask(edt_p, mask);
}

void edt_disable_remote_interrupt(Edt_Dev* edt_p)
{
  u_int mask = edt_get_interrupt_mask(edt_p) & ~EDT_ICFG_RMT_EN_INTR;
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("disable remote interrupt\n"));
  edt_set_interrupt_mask(edt_p, mask);
}

void edt_enable_dma_interrupt(Edt_Dev* edt_p)
{
  u_int mask = edt_get_interrupt_mask(edt_p) | EDT_ICFG_PCI_EN_INTR;
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("enable dma interrupt\n"));

  edt_set_interrupt_mask(edt_p, mask);
}

void edt_disable_dma_interrupt(Edt_Dev* edt_p)
{
  u_int mask = edt_get_interrupt_mask(edt_p) & ~EDT_ICFG_PCI_EN_INTR;
  EDTPRT(edt_p, DBG_INTR_EN, TRL_INFO, ("disable dma interrupt\n"));
  edt_set_interrupt_mask(edt_p, mask);
}

void edt_time_reg_access(Edt_Dev* edt_p)
{
  int i;
  volatile int x;

  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    x = PCI_GETL(edt_p, 0xc4);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    PCI_PUTL(edt_p, 0xc4, x);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    x = edt_get(edt_p, EDT_DMA_CFG);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    edt_set(edt_p, EDT_DMA_CFG, x);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    x = edt_get(edt_p, EDT_DMA_INTCFG_ONLY);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    edt_set(edt_p, EDT_DMA_INTCFG_ONLY, x);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    x = edt_get(edt_p, PCD_CMD);
  EdtTraceTimeStamp(edt_p, NULL, 0);
  for (i = 0; i < 1000000; i++)
    edt_set(edt_p, PCD_CMD, x);
  EdtTraceTimeStamp(edt_p, NULL, 0);
}

void edt_reg_bit_setclear(Edt_Dev* edt_p, u_int reg, u_int mask)
{
  volatile u_int val;
  u_int tmp;

  edt_start_critical_src(edt_p, 9);

  tmp = edt_get(edt_p, reg);

  /* Set the selected bits */
  val = tmp | mask;
  edt_set(edt_p, reg, val);

  /* Clear the selected bits */
  val = tmp & ~mask;
  edt_set(edt_p, reg, val);

  edt_end_critical_src(edt_p, 9);
}

void edt_reg_bit_clearset(Edt_Dev* edt_p, u_int reg, u_int mask)
{
  volatile u_int val;
  u_int tmp;

  edt_start_critical_src(edt_p, 9);

  tmp = edt_get(edt_p, reg);

  /* Clear the selected bits */
  val = tmp & ~mask;
  edt_set(edt_p, reg, val);

  /* Set the selected bits */
  val = tmp | mask;
  edt_set(edt_p, reg, val);

  edt_end_critical_src(edt_p, 9);
}

void edtdrv_set_startdma(Edt_Dev* edt_p, u_int desc, u_int value)
{
  EDTPRT(edt_p, DBG_IOCTLS, TRL_INFO, ("Set STARTDMA desc = %x Val = %x\n", desc, value));
  edt_start_critical_src(edt_p, 5);
  edt_p->m_StartDmaDesc = desc;
  edt_p->m_StartDmaVal = (u_int)value;
  edt_p->m_Started = edt_p->m_Done;
  edt_p->m_DoStartDma = 1;
  edt_end_critical_src(edt_p, 5);
}

void edtdrv_set_enddma(Edt_Dev* edt_p, u_int desc, u_int value)
{
  edt_start_critical_src(edt_p, 6);
  edt_p->m_EndDmaDesc = desc;
  edt_p->m_EndDmaVal = (u_int)value;
  edt_p->m_DoEndDma = 1;
  edt_end_critical_src(edt_p, 6);
}
