/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * Scatter/Gather List
 */
#include "edtdrv.h"

int edt_SetLogSgList(Edt_Dev* edt_p, int bufnum, u_int* log_list, u_int log_entrys)
{
  u_int tmpsize;
  u_int* log_p;
  u_int* new_p;
  u_int* cur_p;
  u_int new_entrys;
  u_int cur_entrys;
  u_int cur_idx = 0;
  u_int ii;
  rbuf* p = &edt_p->m_Rbufs[bufnum];
  u_int* sg_list;
  u_int cur_size = 0;
  int cur_total;
  u_int cur_addr = 0;
  u_int new_size;
  u_int log_size;
  int log_addr;
  u_int offset;
  u_int* new_list;
  int allocsize;

  sg_list = (u_int*)p->m_SgVirtAddr;
  cur_entrys = p->m_SgUsed / 8;

  if (log_entrys > cur_entrys)
    allocsize = (log_entrys + 1) * 8;
  else
    allocsize = (p->m_SgUsed + 1) * 8;

  new_list = (u_int*)edt_vmalloc(allocsize, "SetLogSgList");

  new_p = new_list;
  log_p = log_list;

  /* need to break up available physical chunks into logical request */
  new_entrys = 0;
  for (ii = 0; ii < log_entrys; ii++)
  {
    cur_p = sg_list;
    cur_total = 0;
    cur_idx = 0;
    log_addr = *log_p++;
    log_size = *log_p++;

    while (cur_total <= log_addr)
    {
      cur_addr = *cur_p++;
      cur_size = *cur_p++ & 0xffff;
      cur_total += cur_size;
      cur_idx++;
      if (cur_idx > cur_entrys)
      {
        EDTPRINTF(edt_p, 1, ("here0 cur_idx %x entrys %x\n", cur_idx, cur_entrys));
        break;
      }
    }
    if (cur_idx > cur_entrys)
      break;
    /* enter starting address in new list */
    offset = log_addr - (cur_total - cur_size);
    *new_p++ = cur_addr + offset;
    /* now add entrys needed to fulfill logical size */
    new_size = cur_size - offset;
    tmpsize = log_size;
    while (tmpsize)
    {
      new_entrys++;
      if (new_size >= tmpsize)
      {
        *new_p++ = tmpsize;
        break;
      }
      else
      {
        *new_p++ = new_size;
        tmpsize -= new_size;
        *new_p++ = *cur_p++;
        new_size = *cur_p++ & 0xffff;
        cur_idx++;
        if (cur_idx > cur_entrys)
        {
          EDTPRINTF(edt_p, 1, ("here1 cur_idx %x entrys %x\n", cur_idx, cur_entrys));
          break;
        }
      }
    }
  }
  /* back up for interrupt bit */
  *(new_p - 1) |= EDT_LIST_PAGEINT;

  EDTPRINTF(edt_p, 1, ("here2 new_entrys %d cur %d idx %d\n", new_entrys, cur_entrys, cur_idx));
  edt_set_sglist(edt_p, bufnum, new_list, new_entrys);

#if defined(__linux__)
  edt_vfree(new_list, allocsize, "SetLogicalSgList");
#elif defined(_NT_DRIVER_)
  ExFreePool(new_list);
#endif
  return (0);
}

uint64_t edt_get_sginfo(Edt_Dev* edt_p, u_int bufnum, u_int desc)
{
  rbuf* p;
  uint64_t value = 0;

  if (edt_p->m_Rbufs == NULL)
  {
    EDTPRT(edt_p, DBG_DMA, TRL_ERROR, ("Error - no rbufs allocated yet for sginfo\n"));
    return 0;
  }

  p = &edt_p->m_Rbufs[bufnum];
  edt_p->m_CurInfo = bufnum;

  if (!p->m_SglistSize)
  {
    return 0;
  }

  switch (desc)
  {
  case EDT_SGLIST_SIZE:
    value = (uint64_t)p->m_SgUsed;
    break;
  case EDT_SGLIST_VIRTUAL:
#if defined(__linux__)
    value = (uint64_t)(p->m_SgVirtAddr);
#else
    value = (uint64_t)p->m_SgVirtAddr;
#endif
    break;

  case EDT_SGTODO_SIZE:
    value = (uint64_t)p->m_nChunks;
    break;
  case EDT_SGTODO_VIRTUAL:
    value = (uint64_t)(p->SgChunksList);
    break;

  case EDT_SGTODO_FIRST_SG:
    /* New and used for PDMA (4/07): */
    value = (uint64_t)p->SgChunksList[0].busaddr;
    break;
  }

  return value;
}

int edt_read_sg_list(Edt_Dev* edt_p, u_int bufnum, u_int* target, u_int targetsize)
{
  EdtPCISg* cur_p;
  EdtPCISg* new_p;
  u_int ii;
  rbuf* p = &edt_p->m_Rbufs[bufnum];
  u_int maxentrys;
  u_int readentrys;

  maxentrys = p->m_SglistSize / 8;
  readentrys = targetsize / 8;

  new_p = (EdtPCISg*)target;

  if (readentrys > maxentrys)
  {
    readentrys = maxentrys;
  }
  for (ii = 0; ii < readentrys; ii++)
  {
    cur_p = edt_get_sg_address_for_page(edt_p, p, ii);

    edt_copy_to_user(new_p, cur_p, sizeof(EdtPCISg));
    new_p++;
  }

  return (0);
}

int edt_set_sglist(Edt_Dev* edt_p, u_int bufnum, u_int* source, u_int new_entrys)
{
  EdtPCISg* cur_p;
  EdtPCISg* new_p;
  u_int ii;
  rbuf* p = &edt_p->m_Rbufs[bufnum];
  u_int maxentrys;
  /* now copy to old if enough room */
  maxentrys = p->m_SglistSize / 8;

  new_p = (EdtPCISg*)source;
  if (new_entrys > maxentrys)
  {
    edt_allocate_sglist(edt_p, new_entrys * 4096, bufnum);
    maxentrys = p->m_SglistSize / 8;
  }
  for (ii = 0; ii < new_entrys; ii++)
  {
    cur_p = edt_get_sg_address_for_page(edt_p, p, ii);
    cur_p->busaddr = new_p->busaddr;
    cur_p->count = new_p->count;
    new_p++;
  }

  p->m_SgUsed = new_entrys * 8;

  return (0);
}

/*********************************************
 *
 *
 *
 ********************************************/

static void edt_clear_sglist_resize(Edt_Dev* edt_p, int bufnum)
{
  rbuf* p = &edt_p->m_Rbufs[bufnum];
  EdtPCISg* sgPtr = 0;

  if (p->is_adjusted)
  {
    sgPtr = edt_get_sg_address_for_page(edt_p, p, p->m_SgAdjustedEntry);

    /* reset */

    /* fix sgPtr back to original */
    sgPtr->count = p->m_SgAdjustedSize;

    p->m_SgAdjustedSize = 0;
    p->m_SgAdjustedEntry = 0;

    p->is_adjusted = 0;

    p->SgChunksList[p->SgArrayAdjustedChunk].count = p->SgArrayAdjustedSize;

    p->m_nChunks = p->InitialChunks;

    p->is_adjusted = 0;
  }
}

static void edt_set_adjusted_sg_size(Edt_Dev* UNUSED_edt_p,
                                     rbuf* p,
                                     int foundentry,
                                     EdtPCISg* sgPtr,
                                     int newpagesize)
{
  (void)UNUSED_edt_p;

  // Storecd
  p->m_SgAdjustedEntry = foundentry;
  p->m_SgAdjustedSize = (sgPtr->count & 0xffff);

  /* set the table entry */
  sgPtr->count = (newpagesize)&0xffff;
  sgPtr->count |= EDT_LIST_PAGEINT;

  p->is_adjusted = 1;
}

static void edt_set_adjusted_sg_array(Edt_Dev* UNUSED_edt_p, rbuf* p, int chunk, int newchunksize)
{
  (void)UNUSED_edt_p;

  // Store
  p->SgArrayAdjustedChunk = chunk;
  p->SgArrayAdjustedSize = p->SgChunksList[chunk].count;

  /* set the table entry */
  p->SgChunksList[chunk].count = newchunksize;

  p->m_nChunks = chunk + 1;

  p->is_adjusted = 1;
}

int edt_set_ringbuf_physical_addr(Edt_Dev* edt_p, u_int bufnum, uint64_t phyaddr)
{
  u_int entry;

  /* go through an existing SG list and set physaddr to a constant */

  rbuf* p = &edt_p->m_Rbufs[bufnum];

  for (entry = 0; entry < p->SgEntries; entry++)
  {
    EdtPCISg* sgPtr = edt_get_sg_address_for_page(edt_p, p, entry);

    sgPtr->busaddr = (unsigned int)phyaddr;
  }

  return 0;
}

int edt_resize_sglist(Edt_Dev* edt_p, u_int bufnum, u_int newsize, int writeflag)
{
  u_int size = 0;
  u_int entry = 0;
  u_int testSize = 0;
  int foundentry = -1;
  EdtPCISg* sgPtr = NULL;
  rbuf* p = NULL;

  if (bufnum >= edt_p->m_MaxRbufs)
  {
    EDTPRT(edt_p, DBG_DMA_SETUP, TRL_WARN, ("edt_resize_sglist: Bad buf index\n"));
    return EFAULT;
  }

  p = &edt_p->m_Rbufs[bufnum];

  if (p == NULL)
  {
    EDTPRT(edt_p, DBG_DMA_SETUP, TRL_WARN, ("edt_resize_sglist: NULL ptr\n"));
    return EFAULT;
  }

  /* go through an existing SG list */
  /* and change size */

  if (p->m_XferSize == newsize && (unsigned int)writeflag == p->m_WriteFlag)
  {
    return 0;
  }

  /* clear any existing resize */

  edt_clear_sglist_resize(edt_p, bufnum);

  /* reset the size loaded for SG list */

  for (entry = 0; entry < p->SgEntries; entry++)
  {
    sgPtr = edt_get_sg_address_for_page(edt_p, p, entry);

    testSize = sgPtr->count & 0xffff;

    if (size + testSize >= newsize)
    {
      foundentry = entry;

      edt_set_adjusted_sg_size(edt_p, p, foundentry, sgPtr, newsize - size);
      break;
    }

    size += testSize;
  }

  foundentry = (foundentry + 1) * 8;

  if (foundentry >= 0)
  {
    /* store existing table entries */
    u_int runningtotal = 0;
    u_int chunk;
    int foundchunk = -1;

    for (chunk = 0; chunk < p->m_nChunks; chunk++)
    {
      if (runningtotal + p->SgChunksList[chunk].count >= (u_int)foundentry)
      {
        foundchunk = chunk;

        edt_set_adjusted_sg_array(edt_p, p, foundchunk, foundentry - runningtotal);

        break;
      }

      runningtotal += p->SgChunksList[chunk].count;
    }

    if (foundchunk == -1)
    {
      EDTPRT(edt_p, DBG_DMA_SETUP, TRL_ERROR, ("edt_resize_sglist: ran off end of SgChunksList\n"));
    }
  }
  else
  {
    EDTPRT(edt_p, DBG_DMA_SETUP, TRL_ERROR, ("edt_resize_sglist: entry %x \n", bufnum));
  }

  return (-ENOMEM);
}
