/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @file
 * @author Alex Hogen (alex@edt.com)
 * @brief Datatypes and configuration macros shared between userspace libraries
 * and kernel driver source.
 *
 * @copyright Copyright (c) 2023 Engineering Design Team, Inc.
 *
 */

#ifndef EDT_SHARED_TYPES_H
#define EDT_SHARED_TYPES_H

#include "edtdef.h"

#if defined(__linux__) && defined(__KERNEL__)
#include <linux/types.h>
#elif defined(_NT_DRIVER_)
#include <stdint.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef char edt_version_string[128];

/**
 * Typedef for edt_bitpath to send and retrieve bitfile pathnames
 * from the driver.
 */
typedef char edt_bitpath[128];

#if defined(PCD)
#define EDT_MEZZ_MAX_EXTENDED_WORDS 32
typedef struct _EdtMezzDescriptor {
  int id;
  int n_extended_words;
  int extended_rev;
  uint32_t extended_data[EDT_MEZZ_MAX_EXTENDED_WORDS];
} EdtMezzDescriptor;
#endif /* PCD */

/** Structure holding a timestamp broken down into seconds and nanoseconds. */
typedef struct edt_timespec {
  /** Timestamp, integer seconds */
  int64_t tv_sec;
  /** Timestamp, nanoseconds */
  int64_t tv_nsec;
} edt_timespec_t;

/** Information about active DMA operations.
 *
 * @see @ref edt_get_dma_info().
 */
typedef struct edt_dma_info {
  uint32_t used_dma;   /**< Which channels have started DMA within current open/close. */
  uint32_t alloc_dma;  /**< Which channels have allocated > 1 ring buffer. */
  uint32_t active_dma; /**< Which channels have active DMA transfers. */
  uint32_t interrupts;
  uint32_t locks;
  uint64_t wait_time;
  uint64_t lock_time;
  uint32_t lock_array[MAX_LOCK_SRC + 1];
  uint32_t direct_reads[256];
  uint32_t direct_writes[256];
  uint32_t indirect_reads[256];
  uint32_t indirect_writes[256];
  uint32_t dma_reads[8];
  uint32_t dma_writes[8];
  uint32_t active_list_size;
  uint32_t free_list_size;
} edt_dma_info_t;

/* Defines for flush, start, end DMA action. */
#define EDT_ACT_NEVER            0
#define EDT_ACT_ONCE             1
#define EDT_ACT_ALWAYS           2
#define EDT_ACT_ONELEFT          3
#define EDT_ACT_CYCLE            4
#define EDT_ACT_KBS              5
#define EDT_ACT_ALWAYS_WRITEONLY 6

/* Defines for @ref edt_set_kernel_buffers(). */
#define EDT_USER_BUFS  (1U << 0) /**< @deprecated Do not use. */
#define EDT_COPY_KBUFS (1U << 1) /**< Use buffer copy. */
#define EDT_MMAP_KBUFS (1U << 2) /**< Use memory-mapped buffers. */

/* Defines for return from wait. */
typedef enum edt_dma_wait_status {
  /** Error: Unknown wait status. */
  EDT_WAIT_UNKNOWN = -1,
  EDT_WAIT_OK = 0,
  EDT_WAIT_TIMEOUT = 1,
  EDT_WAIT_OK_TIMEOUT = 2,
} EdtDmaWaitStatus;

/** Timeout action. */
typedef enum edt_dma_timeout_action {
  /** No extra action taken. */
  EDT_TIMEOUT_NULL,
  /** Flush any valid bits left in input circuits of SSDIO. */
  EDT_TIMEOUT_BIT_STROBE,
} EdtDmaTimeoutAction;

/*-----------------------------------------------------------------------------
 * Kernel events
 *---------------------------------------------------------------------------*/

#define EDT_MAX_KERNEL_EVENTS 20
#define EDT_BASE_EVENTS       1

/** Event flag bits, indicating interest in these events.
 *
 * @see edt_set_event_func()
 */
enum edt_kernel_event {
  EDT_EODMA_EVENT = EDT_BASE_EVENTS,
  EDT_EVENT_BUF,
  EDT_EVENT_STAT,

  /**
   * Image has been acquired, shutter has closed, subject can be moved if
   * necessary, DMA will now restart.
   *
   * <b>Applicable Boards:</b> VisionLink F1/F4
   */
  EDT_PDV_EVENT_ACQUIRE,

  /**
   * Interrupt occurred on Status 1 line.
   *
   * <b>Applicable Boards:</b> PCI CD
   */
  EDT_EVENT_PCD_STAT1,

  /**
   * Interrupt occurred on Status 2 line.
   *
   * <b>Applicable Boards:</b> PCI CD
   */
  EDT_EVENT_PCD_STAT2,

  /**
   * Interrupt occurred on Status 3 line.
   *
   * <b>Applicable Boards:</b> PCI CD
   */
  EDT_EVENT_PCD_STAT3,

  /**
   * Interrupt occurred on Status 4 line.
   *
   * <b>Applicable Boards:</b> PCI CD
   */
  EDT_EVENT_PCD_STAT4,
  EDT_PDV_STROBE_EVENT,
  /**
   * Frame Valid line is set.
   *
   * <b>Applicable Boards:</b> VisionLink F1/F4
   */
  EDT_PDV_EVENT_FVAL,
  EDT_PDV_EVENT_TRIGINT,
  EDT_EVENT_TEMP,
  /* Always the last in the enum. */
  EDT_MAX_EVENT_TYPES
};

#define EDT_EODMA_EVENT_NAME       "edt_eodma"
#define EDT_BUF_EVENT_NAME         "edt_buf"
#define EDT_STAT_EVENT_NAME        "edt_stat"
#define EDT_PDV_ACQUIRE_EVENT_NAME "edt_acquire"
#define EDT_EVENT_PCD_STAT1_NAME   "edt_pcd_stat1"
#define EDT_EVENT_PCD_STAT2_NAME   "edt_pcd_stat2"
#define EDT_EVENT_PCD_STAT3_NAME   "edt_pcd_stat3"
#define EDT_EVENT_PCD_STAT4_NAME   "edt_pcd_stat4"
#define EDT_PDV_STROBE_EVENT_NAME  "edt_pdv_strobe"
#define EDT_PDV_EVENT_FVAL_NAME    "edt_pdv_fval"
#define EDT_PDV_EVENT_TRIGINT_NAME "edt_pdv_trigint"
#define EDT_EVENT_TEMP_NAME        "edt_temp_intr"

/*
 * The EVENT_MODE controls the way the driver responds to events.
 * Mode 0 is similar to ACT_ONCE; mode 1 is similar to ACT_ALWAYS,
 * and mode 2 is similar to ACT_ALWAYS but disables the event
 * while user code is processing it to prevent OS lockups.
 */
#define EDT_EVENT_MODE_MASK       0xFF000000
#define EDT_EVENT_MODE_SHFT       24
#define EDT_EVENT_MODE_ONCE       0
#define EDT_EVENT_MODE_CONTINUOUS 1
#define EDT_EVENT_MODE_SERIALIZE  2

#ifdef __cplusplus
}
#endif

#endif /* EDT_SHARED_TYPES_H */
