/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 */

#ifndef _EDT_DRV_LINUX_H_
#define _EDT_DRV_LINUX_H_

#include "edtdef.h"

#include "edt_lnx_kernel.h"
#include "edt_lnx_mm.h"
#include "edt_pci_devices.h"
#include "edtreg.h"

#include "edt_shared_types.h"

#include <linux/types.h>  // TODO: consider removing (pci_driver#13)
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/pci.h>
#include <linux/version.h>

/* Backport dev print helpers added in Linux kernel 4.16 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 16, 0)
#define pci_printk(level, pdev, fmt, arg...) dev_printk(level, &(pdev)->dev, fmt, ##arg)

#define pci_emerg(pdev, fmt, arg...)  dev_emerg(&(pdev)->dev, fmt, ##arg)
#define pci_alert(pdev, fmt, arg...)  dev_alert(&(pdev)->dev, fmt, ##arg)
#define pci_crit(pdev, fmt, arg...)   dev_crit(&(pdev)->dev, fmt, ##arg)
#define pci_err(pdev, fmt, arg...)    dev_err(&(pdev)->dev, fmt, ##arg)
#define pci_warn(pdev, fmt, arg...)   dev_warn(&(pdev)->dev, fmt, ##arg)
#define pci_notice(pdev, fmt, arg...) dev_notice(&(pdev)->dev, fmt, ##arg)
#define pci_info(pdev, fmt, arg...)   dev_info(&(pdev)->dev, fmt, ##arg)
#define pci_dbg(pdev, fmt, arg...)    dev_dbg(&(pdev)->dev, fmt, ##arg)

#define pci_notice_ratelimited(pdev, fmt, arg...) dev_notice_ratelimited(&(pdev)->dev, fmt, ##arg)

#define pci_info_ratelimited(pdev, fmt, arg...) dev_info_ratelimited(&(pdev)->dev, fmt, ##arg)

#define pci_WARN(pdev, condition, fmt, arg...) \
  WARN(condition, "%s %s: " fmt, dev_driver_string(&(pdev)->dev), pci_name(pdev), ##arg)

#define pci_WARN_ONCE(pdev, condition, fmt, arg...) \
  WARN_ONCE(condition, "%s %s: " fmt, dev_driver_string(&(pdev)->dev), pci_name(pdev), ##arg)
#endif /* kernel 4.16.x pci print macros */

#define EDTPRINTF(edt_p, level, msg) \
  ((!edt_p || (edt_p && (edt_p->m_DebugLevel >= level))) ? (edt_printk msg) : 0)

#define SHOWCHANNEL

#ifdef SHOWCHANNEL
#define EDTPRT(edt_p, mask, level, msg)                              \
  (!(edt_p) ? (edt_printk msg)                                       \
            : ((((Edt_Dev*)(edt_p))->DebugMask[(level)] & (mask))    \
                   ? (edt_printk("[%d.%d]: ",                        \
                                 ((Edt_Dev*)(edt_p))->m_Unitno,      \
                                 ((Edt_Dev*)(edt_p))->m_DmaChannel), \
                      edt_printk msg)                                \
                   : 0))
#else
#define EDTPRT(edt_p, mask, level, msg) \
  (!(edt_p) ? (edt_printk msg)          \
            : ((((Edt_Dev*)(edt_p))->DebugMask[(level)] & (mask)) ? (edt_printk msg) : 0))

#endif

#define CHECK_DEBUG_MASK(edt_p, mask, level) \
  (!(edt_p) ? 0 : ((((Edt_Dev*)(edt_p))->DebugMask[(level)] & (mask)) ? 1 : 0))

typedef struct _Lnx_Rbuf {
  EdtMemMapping map;

} Lnx_Rbuf;

typedef struct _Lnx_Edt_Dev {
  void** data;

  EdtWaitHandle serial_wait;
  EdtWaitHandle wait;

  u_short m_wait_done;        /* Linux race condition prevention. Oct 2001 MCM */
  u_short m_serial_wait_done; /* Linux race condition prevention. Oct 2001 MCM */

  EdtWaitHandle event_wait[EDT_MAX_EVENT_TYPES];

  /* IRQ flags when locking/unlocking spinlocks in an ISR context. */
  unsigned long cpu_flags;
  /* The (first) IRQ number allocated to this device. Value is copied to all
   * 'channel' devices. */
  uint32_t irq;

  EdtSpinLock_t* crit_lock;

  EdtSpinLock_t* abort_lock;

  EdtSemaphore_t register_lock;

  struct pci_dev* m_pci_dev;

  pid_t register_locker;

} Lnx_Edt_Dev;

#define OS_SPECIFIC_RBUF Lnx_Rbuf

#define OS_SPECIFIC_DEV Lnx_Edt_Dev

#include "Edt_Dev.h"

#define EDT_PCI_MAPMEM(edt_p)
#define EDT_PCI_UNMAPMEM(edt_p)

/* rwh 11/18/2015 (v.5.5.1.3): added #ifdef for new vs old kernels
 * Changed PCI_PUT/GET macros from obsolete readw/writew to ioread/iowrite
 * Combined duplicated but #ifdef'd _ia64_ and 32-bit versions of these macros
 * Removed ppc-specific (#if defined BIG_ENDIAN), not supported
 */
#define BUS_DATA(data) (data)

/* When writing new FPGA firmware for a new device, sometimes it can be
 * helpful to see BAR0 and BAR1 register accesses from the driver's
 * perspective. Set this to 1 to enable printing ioread/iowrite accesses.
 * This is obviously very slow and should never be shipped enabled.
 */
#if 0

#include <linux/pci.h>

#define PCI_PUTC(edt_p, addr, value)                            \
  do                                                            \
  {                                                             \
    pci_warn(edt_p->os.m_pci_dev, "iowrite8(BAR0+0x%X)", addr); \
    iowrite8(value, edt_p->bar0_mmap + (addr));                 \
  } while (0)

#define PCI_PUTS(edt_p, addr, value)                             \
  do                                                             \
  {                                                              \
    pci_warn(edt_p->os.m_pci_dev, "iowrite16(BAR0+0x%X)", addr); \
    iowrite16(value, edt_p->bar0_mmap + (addr));                 \
  } while (0)

#define PCI_PUTL(edt_p, addr, value)                             \
  do                                                             \
  {                                                              \
    pci_warn(edt_p->os.m_pci_dev, "iowrite32(BAR0+0x%X)", addr); \
    iowrite32(value, edt_p->bar0_mmap + (addr));                 \
  } while (0)

#define PCI_GETC(edt_p, addr)                                                 \
  ({                                                                          \
    unsigned int val = ioread8(edt_p->bar0_mmap + (addr));                    \
    pci_warn(edt_p->os.m_pci_dev, "ioread8(BAR0+0x%X) -> 0x%02X", addr, val); \
    val;                                                                      \
  })

#define PCI_GETS(edt_p, addr)                                                  \
  ({                                                                           \
    unsigned int val = ioread16(edt_p->bar0_mmap + (addr));                    \
    pci_warn(edt_p->os.m_pci_dev, "ioread16(BAR0+0x%X) -> 0x%04X", addr, val); \
    val;                                                                       \
  })

#define PCI_GETL(edt_p, addr)                                                  \
  ({                                                                           \
    unsigned int val = ioread32(edt_p->bar0_mmap + (addr));                    \
    pci_warn(edt_p->os.m_pci_dev, "ioread32(BAR0+0x%X) -> 0x%08X", addr, val); \
    val;                                                                       \
  })

#define PCI_MEM2_PUTL(edt_p, addr, value)                        \
  do                                                             \
  {                                                              \
    pci_warn(edt_p->os.m_pci_dev, "iowrite32(BAR1+0x%X)", addr); \
    iowrite32(value, edt_p->bar1_mmap + (addr));                 \
  } while (0)

#define PCI_MEM2_GETL(edt_p, addr)                                             \
  ({                                                                           \
    unsigned int val = ioread32(edt_p->bar1_mmap + (addr));                    \
    pci_warn(edt_p->os.m_pci_dev, "ioread32(BAR1+0x%X) -> 0x%08X", addr, val); \
    val;                                                                       \
  })

#else
#define PCI_PUTC(edt_p, addr, value) iowrite8(value, edt_p->bar0_mmap + (addr))
#define PCI_PUTS(edt_p, addr, value) iowrite16(value, edt_p->bar0_mmap + (addr))
#define PCI_PUTL(edt_p, addr, value) iowrite32(value, edt_p->bar0_mmap + (addr))

#define PCI_GETC(edt_p, addr) ioread8(edt_p->bar0_mmap + (addr))
#define PCI_GETS(edt_p, addr) ioread16(edt_p->bar0_mmap + (addr))
#define PCI_GETL(edt_p, addr) ioread32(edt_p->bar0_mmap + (addr))

#define PCI_MEM2_PUTL(edt_p, addr, value) iowrite32(value, edt_p->bar1_mmap + (addr))
#define PCI_MEM2_GETL(edt_p, addr)        ioread32(edt_p->bar1_mmap + (addr))
#endif

#define DBG_V1           2
#define DBG_V2           3
#define DBG_V3           4
#define DBG_TEST         1
#define DBG_ERROR        0
#define DBG_INIT_MASK    2
#define DEF_DBG          2
#define DBG_DMA_MASK     2
#define DBG_SG_BLD_MASK  16
#define DBG_SG_PAGE_MASK 32
#define DBG_MAP_MASK     64
#define DBG_REG_MASK     128
#define DBG_DPC_MASK     2
#define DBG_TIMEOUT_MASK 512
#define DBG_FULL_MASK    0x1000 /* lowest level - full debug */

#define DBG_ALWAYS_MASK 0 /* always prints */

#if defined(PCD) && (PCD != 0)
#include "pcd.h"
#endif

#if defined(PDV) && (PDV != 0)
#include "pdv.h"
#endif

/*!
 * @brief Creates and initializes all DMA channel pointers.
 *
 * Should only be called from @ref edt_pcidrv_probe().
 *
 * Will call @ref edt_init_channel_struct() on each channel pointer it creates.
 *
 * @param pci_dev
 *     PCI device being added.
 * @param pci_id
 *     PCI device id structure to read information from.
 * @param unit_number
 *     Unit number to assign to this device and all it's channels.
 *
 * @return
 *     The channel 0 pointer if successful. NULL otherwise.
 */
Edt_Dev* edt_create_device_structures(struct pci_dev* pci_dev,
                                      const struct pci_device_id* pci_id,
                                      unsigned int unit_number);

u_int EdtIoctlCommon(Edt_Dev* edt_p,
                     u_int controlCode,
                     u_int inSize,
                     u_int outSize,
                     void* argBuf,
                     u_int* bytesReturned,
                     void* pIrp);

u_int edt_get(Edt_Dev* edt_p, u_int reg_def);
void edt_set(Edt_Dev* edt_p, u_int reg_def, u_int val);

asmlinkage EdtSpinLock_t* edt_create_spinlock(const Edt_Dev* const edt_p);

asmlinkage EdtSemaphore_t edt_create_semaphore(const Edt_Dev* const edt_p);
asmlinkage int edt_lock_semaphore(EdtSemaphore_t p) __attribute__((warn_unused_result));
asmlinkage void edt_unlock_semaphore(EdtSemaphore_t p);
asmlinkage int edt_trylock_semaphore(EdtSemaphore_t p);

#define edt_end_critical(edt_p)   edt_end_critical_src(edt_p, 0)
#define edt_start_critical(edt_p) edt_start_critical_src(edt_p, 0)

int edt_unmap_buffer(Edt_Dev* edt_p, u_int bufnum);

u_int edt_set_kernel_buffers(Edt_Dev* edt_p, u_int newstate);

EdtPCISg* edt_get_sg_address_for_page(Edt_Dev* edt_p, rbuf* p, int pagen);
uint64_t edt_get_sginfo(Edt_Dev* edt_p, u_int bufnum, u_int desc);

int edt_SetLogSgList(Edt_Dev* edt_p, int bufnum, u_int* log_list, u_int log_entrys);

int edt_read_sg_list(Edt_Dev* edt_p, u_int bufnum, u_int* target, u_int targetsize);

int edt_set_sglist(Edt_Dev* edt_p, u_int bufnum, u_int* source, u_int new_entrys);

int edt_resize_sglist(Edt_Dev* edt_p, u_int bufnum, u_int newsize, int writeflag);

void edt_clear_dma(Edt_Dev* edt_p);

/* Note this may not work on non PC systems */

#ifndef PAGE_SHIFT
#define PAGE_SHIFT (12)
#endif

#ifndef PAGE_SIZE
#define PAGE_SIZE (1UL << PAGE_SHIFT)
#endif

#ifndef PAGE_MASK
#define PAGE_MASK (PAGE_SIZE - 1)
#endif

#define userout(src, dst, size) edt_copy_to_user(dst, src, size)
#define userin(src, dst, size)  edt_copy_from_user(dst, src, size)

#endif
