/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2024 Engineering Design Team, Inc.
 */
#ifndef KCOMPAT_H
#define KCOMPAT_H

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33))
#include <linux/utsrelease.h>
#else
#include <generated/utsrelease.h>
#endif

/*************************************************************************************************/
/************************************ ALL LINUX DISTRIBUTIONS ************************************/
/*************************************************************************************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 3, 0)

/* Reason For Version Specification: https://lore.kernel.org/lkml/ZA7x9y60sfGOanHl@kroah.com/T/
 *   Chose 6.3.0: vma->vm_flags is now const
 *     (see: https://lore.kernel.org/lkml/20230126193752.297968-3-surenb@google.com/)
 */
#define EDT_KCOMPAT_HAS_VM_FLAGS_SETTER 1

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(6,3,0) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)

/* Reason For Version Specification:
 *   https://lore.kernel.org/linux-rdma/37fcc27a-7c6b-4aed-88be-92aadfaa67fe@intel.com/t/
 *   6.8.0: PCI_IRQ_LEGACY depreciated in favor of PCI_IRQ_INTX
 *     (see: https://lore.kernel.org/lkml/20240115225703.GA73226@bhelgaas/)
 *   6.10.0: PCI_IRQ_LEGACY removed
 *     (see: https://lore.kernel.org/lkml/20240520222943.GA7973@bhelgaas/)
 */
#define EDT_KCOMPAT_HAS_PCI_IRQ_INTX 1

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0) */

/*************************************************************************************************/
/*********************************** DEBIAN BASED DISTRIBUTION ***********************************/
/*************************************** (Debian & Ubuntu) ***************************************/
/*************************************************************************************************/

#if defined(UTS_UBUNTU_RELEASE_ABI)

#endif /* defined(UTS_UBUNTU_RELEASE_ABI) */

/*************************************************************************************************/
/************************************ RHEL BASED DISTRIBUTION ************************************/
/****************** (AlmaLinux, CentOs Linux, CentOs Stream, RHEL, Rocky Linux) ******************/
/*************************************************************************************************/

/* See: /usr/include/linux/version.h */
#if defined(RHEL_MAJOR) && defined(RHEL_MINOR)

/*
 * Is this RHEL version greater than or equal to maj,min?
 * (Use to accomodate RHEL backporting kernel changes.)
 */
#define RHEL_RELEASE_GT_OR_EQ(maj, min) \
  (RHEL_RELEASE_VERSION(RHEL_MAJOR, RHEL_MINOR) >= RHEL_RELEASE_VERSION(maj, min))

#if RHEL_RELEASE_GT_OR_EQ(9, 5)

/*
 * RHEL 9.5 (kernel 5.14.0-503) backported the kernel 6.3 const update.
 */
#define EDT_KCOMPAT_HAS_VM_FLAGS_SETTER 1

#endif /* RHEL_RELEASE_GT_OR_EQ(9,5) */

#if RHEL_RELEASE_GT_OR_EQ(9, 6)

/*
 * CentOs Stream 9.6 (kernel 5.14.0-529) backported the kernel 6.10 removal.
 */
#define EDT_KCOMPAT_HAS_PCI_IRQ_INTX 1

#endif /* RHEL_RELEASE_GT_OR_EQ(9, 6) */

#endif /* defined(RHEL_MAJOR) && defined(RHEL_MINOR) */

/*************************************************************************************************/
/******* All macros should be defined for every Linux version, set to 0 here if undefined. *******/
/*************************************************************************************************/

#ifndef EDT_KCOMPAT_HAS_VM_FLAGS_SETTER
#define EDT_KCOMPAT_HAS_VM_FLAGS_SETTER 0
#endif

#ifndef EDT_KCOMPAT_HAS_PCI_IRQ_INTX
#define EDT_KCOMPAT_HAS_PCI_IRQ_INTX 0
#endif

#endif /* KCOMPAT_H */
