/* SPDX-License-Identifier: BSD-3-Clause OR GPL-2.0-only */

/*!
 * @internal
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (C) 1997 Engineering Design Team, Inc.
 */

#include "edtdrv.h"
#include "pdv.h"

#include "edt_driver_ioctl.h"

/*
 * hooks to map driver dependencies into driver independent code.
 */
int pdv_device_ioctl(Edt_Dev* edt_p,
                     u_int controlCode,
                     u_int inSize,
                     u_int outSize,
                     void* argBuf,
                     u_int* bytesReturned,
                     void* UNUSED_pIrp)
{
  u_int retval = 0;
  u_char arg8 = 0;
  u_short arg16 = 0;
  u_int arg32 = 0;

  (void)UNUSED_pIrp;

  if (argBuf == 0)
    EDTPRT(edt_p, DBG_BOARD, TRL_ERROR, ("code %x in argBuf 0", controlCode));
  else
    switch (inSize)
    {
    case 1:
      arg8 = *(u_char*)argBuf;
      break;
    case 2:
      arg16 = *(u_short*)argBuf;
      break;
    case 4:
      arg32 = *(u_int*)argBuf;
      break;
    default:
      break;
    }

  *bytesReturned = outSize;

  /* check both incase ioctl called instead of edt_ioctl */
  switch (controlCode)
  {
  case ES_ADD_EVENT_FUNC:
  {
    edt_start_critical_src(edt_p, 40);
    switch (arg32)
    {
    case EDT_PDV_EVENT_ACQUIRE:
    {
      EDTPRT(edt_p, DBG_BOARD, TRL_INFO, ("ES_ADD_EVENT_FUNC acquire"));
      /* Enable ACQUIRE interrupt */
      edt_set(edt_p, PDV_CMD, PDV_AQ_CLR); /* first clr */
      edt_or(edt_p, PDV_CFG, PDV_INT_ENAQ);
      edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_DEV_INT);
      edt_enable_remote_interrupt(edt_p);

      break;
    }
    case EDT_PDV_EVENT_FVAL:
    {
      /* Enable FVAL interrupt */
      EDTPRT(edt_p, DBG_BOARD, TRL_INFO, ("ES_ADD_EVENT_FUNC fval"));
      edt_set(edt_p, PDV_CMD, PDV_FV_CLR);      /* first clr */
      edt_clear(edt_p, PDV_UTIL3, PDV_TRIGINT); /* make sure its not TRIGINT */
      edt_or(edt_p, PDV_UTIL2, PDV_INT_ENFV);
      edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_DEV_INT);
      edt_enable_remote_interrupt(edt_p);

      break;
    }
    case EDT_PDV_EVENT_TRIGINT:
    {
      /* Enable TRIGINT interrupt */
      EDTPRT(edt_p, DBG_BOARD, TRL_INFO, ("ES_ADD_EVENT_FUNC fval"));
      edt_set(edt_p, PDV_CMD, PDV_FV_CLR);   /* first clr */
      edt_or(edt_p, PDV_UTIL3, PDV_TRIGINT); /* change meaning of FV int */
      edt_or(edt_p, PDV_UTIL2, PDV_INT_ENFV);
      edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_DEV_INT);
      edt_enable_remote_interrupt(edt_p);

      break;
    }
    case EDT_PDV_STROBE_EVENT:
    {
      /* Enable STROBE interrupt; clear then set */
      EDTPRT(edt_p, DBG_BOARD, TRL_INFO, ("ES_ADD_EVENT_FUNC strobe"));
      edt_clear(edt_p, PDV_SERIAL_DATA_CNTL, LHS_INTEN);
      edt_or(edt_p, PDV_SERIAL_DATA_CNTL, LHS_INTEN);
      edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_DEV_INT);
      edt_enable_remote_interrupt(edt_p);

      break;
    }
    default:
      EDTPRT(edt_p, DBG_BOARD, TRL_VERBOSE, ("ES_ADD_EVENT_FUNC default"));
      break;
    }

    edt_end_critical_src(edt_p, 40);

    break;
  }

  case ES_DEL_EVENT_FUNC:
  {
    edt_start_critical_src(edt_p, 41);
    switch (arg32)
    {
    case EDT_PDV_EVENT_ACQUIRE:
    {
      u_char config_reg;

      /* Disable ACQUIRE interrupt */
      config_reg = (u_char)edt_get(edt_p, PDV_CFG);
      config_reg &= ~PDV_INT_ENAQ;
      edt_set(edt_p, PDV_CFG, config_reg);

      break;
    }
    case EDT_PDV_EVENT_FVAL:
    {
      u_char util2_reg;

      /* Disable FVAL interrupt */
      util2_reg = (u_char)edt_get(edt_p, PDV_UTIL2);
      util2_reg &= ~PDV_INT_ENFV;
      edt_set(edt_p, PDV_UTIL2, util2_reg);

      break;
    }
    case EDT_PDV_EVENT_TRIGINT:
    {
      u_char util2_reg;
      u_char util3_reg;

      /* Disable TRIGINT interrupt */
      util2_reg = (u_char)edt_get(edt_p, PDV_UTIL2);
      util2_reg &= ~PDV_INT_ENFV;
      edt_set(edt_p, PDV_UTIL2, util2_reg);

      /* And turn off TRIGINT bit so it can be used as FVAL if needed */
      util3_reg = (u_char)edt_get(edt_p, PDV_UTIL3);
      util3_reg &= ~PDV_TRIGINT;
      edt_set(edt_p, PDV_UTIL3, util3_reg);

      break;
    }
    case EDT_PDV_STROBE_EVENT:
    {
      /* Disable STROBE interrupt */
      edt_clear(edt_p, PDV_SERIAL_DATA_CNTL, LHS_INTEN);

      break;
    }
    default:
      EDTPRT(edt_p, DBG_BOARD, TRL_VERBOSE, ("ES_DEL_EVENT_FUNC default"));
      break;
    }

    edt_end_critical_src(edt_p, 41);

    break;
  }

  case ES_CLR_EVENT:
    break;

  default:
    EDTPRT(edt_p, DBG_BOARD, TRL_ERROR, ("pdv_device_ioctl:  unimplemented ioctl %x", controlCode));
    retval = 1;
    break;
  }

  if (retval)
    *bytesReturned = 0;
  else
  {
    if (argBuf == 0)
      EDTPRT(edt_p, DBG_BOARD, TRL_ERROR, ("code %x out argBuf 0", controlCode));
    else
      switch (outSize)
      {
      case 1:
        *(u_char*)argBuf = arg8;
        break;
      case 2:
        *(u_short*)argBuf = arg16;
        break;
      case 4:
        *(u_int*)argBuf = arg32;
        break;
      default:
        break;
      }
  }

  return (retval);
}

/*
 * edt_hwctl():  interface to device dependent hardware operations.
 */

int pdv_hwctl(int operation, void* context, Edt_Dev* edt_p)
{
  switch (operation)
  {
  case INIT_DEVICE:
    edt_InitializeBoard(edt_p);
    break;
  case LAST_CLOSE:
  {
    u_int cmd;

    if (edt_p->m_DebugLevel)
      EDTPRT(edt_p, DBG_BOARD, TRL_VERBOSE, ("pdv LAST CLOSE\n"));
    cmd = edt_get(edt_p, EDT_SG_NXT_CNT);
    cmd &= (EDT_DMA_MEM_RD | EDT_BURST_EN);
    edt_p->m_Done = edt_p->m_Todo = edt_p->m_Loaded;
    edt_p->m_DmaActive = 0;
    edt_p->m_Freerun = 0;
    cmd |= EDT_DMA_ABORT;

    edt_set_sg_cnt(edt_p, cmd);

    /* no break ; */
  }
  case IDLE_DEVICE:
  {
    break;
  }

  case ABORT_DMA:

    break;

  case DISABLE_INTR:

    break;

  case TEST_INTR:
    break;

  case ACK_INTR:
    break;

  case GET_DMA_RESID:
    break;

  case LOAD_REGISTERS:
    break;

  case DMA_COUNTDOWN:
    break;

  case EDT_DMA_ADDRESS:
    break;

  case MMAP_REGISTERS:
    break;

  case POSTIO:
    break;

  case STARTIO:
    break;

  case PRE_STARTIO_OP:
    break;

  case FIRST_OPEN:
  {
    break;
  }

  case FREE_SOFTWARE:
    break;

  default:
    EDTPRT(edt_p,
           DBG_BOARD,
           TRL_ERROR,
           ("Unknown edt_hwctl() operation 0x%x context 0x%p", operation, context));
    break;
  }

  return (0);
}

DRV_EXPORT void pdvdrv_serial_write(Edt_Dev* edt_p, char* buf, int size, int flags)
{
  int i, j;
  u_int dbgtrace;

  int dbgsize = 0;

  static char dbgbuf[20];
  char* tmp_p = (char*)&dbgtrace;
  edt_serial* ser_p = edt_p->m_serial;

  EdtTrace(edt_p, TR_SER_WRITE);
  EdtTrace(edt_p, size);

  for (j = 0; j < 4; j++)
  {
    for (i = 0; i < 4; i++)
    {
      if (dbgsize >= size)
      {
        tmp_p[i] = 0;
        dbgbuf[dbgsize] = 0;
      }
      else
      {
        tmp_p[i] = buf[dbgsize];
        dbgbuf[dbgsize] = buf[dbgsize];
        if (dbgbuf[dbgsize] == '\n' || dbgbuf[dbgsize] == '\r')
          dbgbuf[dbgsize] = 0;
      }
      dbgsize++;
    }
    EdtTrace(edt_p, dbgtrace);
  }

  EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("write %s flags 0x%x\n", dbgbuf, flags));

  edt_start_critical_src(edt_p, 12);

  for (i = 0; i < size; i++)
  {
    /*
     * Copy data into the output ring fifo.  Also keep track of how many
     * unwritten bytes are present.
     */
    ser_p->m_Wr_buf[ser_p->m_Wr_producer] = buf[i] & 0xff;

    if (++ser_p->m_Wr_producer == PDV_SSIZE)
      ser_p->m_Wr_producer = 0;

    if (ser_p->m_Wr_producer == ser_p->m_Wr_consumer)
    {
      /*
       * This is a ring-fifo overflow condition: count stays the same
       * and we lose the oldest character in the fifo.
       */
      if (++ser_p->m_Wr_consumer == PDV_SSIZE)
        ser_p->m_Wr_consumer = 0;
    }
    else
      ser_p->m_Wr_count++;
  }

  edt_or(edt_p,
         PDV_SERIAL_DATA_CNTL,
         /* edt_p->m_baudbits |*/
         PDV_EN_TX_INT | PDV_EN_TX | PDV_EN_RX | PDV_EN_RX_INT | PDV_EN_DEV_INT);

  edt_end_critical_src(edt_p, 12);

  edt_enable_remote_interrupt(edt_p);
}

DRV_EXPORT u_int pdvdrv_serial_read(Edt_Dev* edt_p, char* buf, int size)
{
  int n = 0;
  while ((n < size) && (edt_get(edt_p, PDV_SERIAL_DATA_STAT) & PDV_RECEIVE_RDY))
  {
    buf[n] = (char)edt_get(edt_p, PDV_SERIAL_DATA);
    edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_CLR_RX_INT);
    n++;
  }

  if (edt_get(edt_p, PDV_SERIAL_DATA_STAT) & PDV_RECEIVE_RDY)
    edt_wakeup_serial(edt_p);
  else
    edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_RX_INT);

  EDTPRT(edt_p,
         DBG_SERIAL,
         TRL_INFO,
         ("read channel %d size %d count %d\n", edt_p->m_DmaChannel, size, n));

  return n;
}

DRV_EXPORT int pdvdrv_serial_wait(Edt_Dev* edt_p, int mwait, int UNUSED_count)
{
  int stat;

  (void)UNUSED_count;

  edt_start_critical_src(edt_p, 14);
  stat = edt_get(edt_p, PDV_SERIAL_DATA_STAT);
  edt_end_critical_src(edt_p, 14);

  // not inside a critical section while waiting.
  if ((stat & PDV_RECEIVE_RDY) == 0)
    edt_wait_serial(edt_p, mwait);

  edt_start_critical_src(edt_p, 15);
  stat = edt_get(edt_p, PDV_SERIAL_DATA_STAT);
  edt_end_critical_src(edt_p, 15);
  if (stat & PDV_RECEIVE_RDY)
    return 1;
  else
    return 0;
}

static int pdv_DeviceIntPdv(Edt_Dev* edt_p)
{
  u_int pdvstat = edt_get(edt_p, PDV_STAT);
  u_int status = edt_get(edt_p, PDV_SERIAL_DATA_STAT);
  u_int util3 = edt_get(edt_p, PDV_UTIL3);

  unsigned int had_devint = 0;
  edt_serial* ser_p = edt_p->m_serial;

  // channel 2 interrupts are spurious (UNLESS PCIE FOX)
  if ((EDT_DEVID_IS_DVCL(edt_p->devid) && edt_p->m_DmaChannel == 2))
  {
    EDTPRT(edt_p, DBG_DPC, TRL_ERROR, ("DevInt at channel 2\n"));
    return 0;
  }

  EDTPRT(edt_p,
         DBG_SERIAL,
         TRL_INFO,
         ("DeviceInt pdvstat %x status %x util3 %x\n", pdvstat, status, util3));

  if (pdvstat & PDV_OVERRUN)
  {
    EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("overflow\n"));
    edt_p->m_Overflow++;
  }

  if (status & PDV_AQUIRE_INT)
  {
    EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("acquire interrupt status %x\n", status));
    edt_set(edt_p, PDV_CMD, PDV_AQ_CLR);

    /* EDT event func wakeup for pdv acquire interrupt -- Mark 01/99 */
    edt_broadcast_event(edt_p, EDT_PDV_EVENT_ACQUIRE);
    had_devint = 1;
  }

  if (status & PDV_FVAL_INT)
  {
    /* trigger interrupt appropriates frame valid bits */
    if (util3 & PDV_TRIGINT)
    {
      EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("trigint interrupt status %x\n", status));
      edt_set(edt_p, PDV_CMD, PDV_FV_CLR);

      /* EDT event func wakeup for pdv acquire interrupt -- RWH 03/00 */
      edt_broadcast_event(edt_p, EDT_PDV_EVENT_TRIGINT);
    }
    else
    {
      EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("frame valid interrupt status %x\n", status));
      edt_set(edt_p, PDV_CMD, PDV_FV_CLR);

      /* EDT event func wakeup for pdv acquire interrupt -- RWH 03/00 */
      edt_broadcast_event(edt_p, EDT_PDV_EVENT_FVAL);
    }

    if (edt_p->m_DoneOnFval)
    {
      EdtTrace(edt_p, 0xbbbbbbbb);
      edt_p->m_HasFvalDone = 1;
    }
    had_devint = 1;
  }

  if (status & LHS_INTERRUPT)
  {
    EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("strobe interrupt"));
    edt_clear(edt_p, PDV_SERIAL_DATA_CNTL, LHS_INTEN);
    edt_or(edt_p, PDV_SERIAL_DATA_CNTL, LHS_INTEN);

    /* EDT event func wakeup for pdv strobe interrupt -- Mark 02/99 */
    edt_broadcast_event(edt_p, EDT_PDV_STROBE_EVENT);
    had_devint = 1;
  }
  if (status & PDV_RECEIVE_RDY)
  {
    edt_clear(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_RX_INT);
    edt_wakeup_serial(edt_p);
    had_devint = 1;
  }

  EDTPRT(edt_p,
         DBG_SERIAL,
         TRL_INFO,
         ("Devint Before testing TX %x against status %x\n", PDV_TRANSMIT_RDY, status));

  if (status & PDV_TRANSMIT_RDY)
  {
    if (ser_p->m_Wr_count > 0)
    {
      u_short ch;

      ch = ser_p->m_Wr_buf[ser_p->m_Wr_consumer];
      edt_set(edt_p, PDV_SERIAL_DATA, ch & 0xff);
      /* this may avoid an extra clock causing write to data */
      (void)edt_get(edt_p, PDV_SERIAL_DATA_STAT);
      if (++ser_p->m_Wr_consumer == PDV_SSIZE)
        ser_p->m_Wr_consumer = 0;
      ser_p->m_Wr_count--;
      if (ser_p->m_Wr_count == 0)
      {
        /* Shut down the transmit cycle */
        edt_clear(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_TX_INT);
      }
      had_devint = 1;
      EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("Setting out char to %c\n", ch));
    }
    else
    {
      /* Shut down the transmit cycle */
      EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("Clearing PDV_EN_TX_INT\n"));

      edt_clear(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_TX_INT);
    }
  }
  EDTPRT(edt_p, DBG_DPC, TRL_INFO, ("DeviceInt returns %d\n", had_devint));

  return had_devint;
}

int pdv_DeviceInt(Edt_Dev* edt_p)
{
  return pdv_DeviceIntPdv(edt_p);
}

DRV_EXPORT int pdvdrv_serial_init(Edt_Dev* edt_p)
{
  edt_serial* ser_p;

  EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("serial init called for channel %d\n", edt_p->m_DmaChannel));

  ser_p = edt_p->m_serial;
  ser_p->m_Wr_producer = 0;
  ser_p->m_Wr_consumer = 0;
  ser_p->m_Wr_count = 0;

  return 0;
}

DRV_EXPORT void pdv_dep_init(Edt_Dev* edt_p)
{
  (void)pdvdrv_serial_init(edt_p);
  EDTPRT(edt_p, DBG_SERIAL, TRL_INFO, ("serial init done\n"));
  edt_p->m_DepSet = 0;
}

DRV_EXPORT void pdvdrv_set_fval_done(Edt_Dev* edt_p, int turnon)
{
  EdtTrace(edt_p, 0xcccccccc);
  EdtTrace(edt_p, turnon);

  edt_p->m_DoneOnFval = (unsigned char)turnon;

  if (turnon)
  {
    edt_set(edt_p, PDV_CMD, PDV_FV_CLR); /* first clear */
    edt_or(edt_p, PDV_UTIL3, PDV_TRIGINT);
    /* enable this at GRAB */
    /* edt_or(edt_p, PDV_UTIL2, PDV_INT_ENFV) ;*/
    edt_or(edt_p, PDV_SERIAL_DATA_CNTL, PDV_EN_DEV_INT);
    edt_enable_remote_interrupt(edt_p);
  }
  else
  {
    edt_set(edt_p, PDV_CMD, PDV_FV_CLR); /* first clear */
    edt_clear(edt_p, PDV_UTIL3, PDV_TRIGINT);
    /* edt_clear(edt_p, PDV_UTIL2, PDV_INT_ENFV) ;*/
  }
}

#define C_IDLE       0
#define C_START_CONT 1
#define C_CONT       2
#define C_WAIT       3
#ifdef PDV
static char* snames[] = {
    "IDLE",
    "STARTCONT",
    "CONT",
    "WAIT",
};

static u_int laststate = 0xfff;
#endif

void pdvdrv_step_continuous(Edt_Dev* edt_p)
{
  /* we are called when dma is done or when todo changes */
  if (edt_p->m_Freerun || edt_p->m_Done != edt_p->m_LastDone || edt_p->m_Todo != edt_p->m_LastTodo)
  {
    edt_p->m_LastDone = edt_p->m_Done;
    edt_p->m_LastTodo = edt_p->m_Todo;

    switch (edt_p->m_ContState)
    {
    case C_WAIT:
      /* now have one more frame */
      if (edt_p->m_Done >= edt_p->m_Started /*
													|| (edt_p->m_Todo > edt_p->m_Done+1) */)
      {
        /* fall through */
        edt_p->m_ContState = C_IDLE;
        EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("wait->idle %x\n", edt_p->m_Done));
      }
      else
        break;
    case C_IDLE:
      if (edt_p->m_Todo > edt_p->m_Done + 1 || edt_p->m_Freerun)
      {
        edt_set(edt_p, PDV_DATA_PATH, edt_p->m_SaveDpath | PDV_CONTINUOUS);
        edt_set(edt_p, edt_p->m_StartDmaDesc, edt_p->m_StartDmaVal);
        edt_p->m_ContState = C_START_CONT;
        EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("idle->startcont %x %x\n", edt_p->m_Done, edt_p->m_Todo));
      }
      else
      {
        if (edt_p->m_Todo > edt_p->m_Done)
        {
          edt_set(edt_p, edt_p->m_StartDmaDesc, edt_p->m_StartDmaVal);
          edt_p->m_Started = edt_p->m_Done + 1;
          edt_p->m_ContState = C_WAIT;
          EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("idle->wait %x %x\n", edt_p->m_Done, edt_p->m_Todo));
        }
      }
      break;
    case C_START_CONT:
      if (!(edt_p->m_Todo > edt_p->m_Done + 1 || edt_p->m_Freerun))
      {
        edt_set(edt_p, PDV_CMD, PDV_CLR_CONT);
        /* to wait one frame till no longer cont */
        /* edt_p->m_Started = edt_p->m_Done + 1 ; */
        edt_p->m_Started = edt_p->m_Done;
        edt_p->m_ContState = C_WAIT;
        EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("startcont->wait %x %x\n", edt_p->m_Done, edt_p->m_Todo));
      }
      else
      {
        EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("startcont->cont %x %x\n", edt_p->m_Done, edt_p->m_Todo));
        edt_p->m_ContState = C_CONT;
      }
      break;
    case C_CONT:
      if (!(edt_p->m_Todo > edt_p->m_Done + 1 || edt_p->m_Freerun))
      {
        edt_set(edt_p, PDV_CMD, PDV_CLR_CONT);
        /* to wait one frame till no longer cont */
        /* edt_p->m_Started = edt_p->m_Done + 1 ; */
        edt_p->m_Started = edt_p->m_Done;
        edt_p->m_ContState = C_WAIT;
        EDTPRT(edt_p, DBG_DMA, TRL_INFO, ("cont->wait %x %x\n", edt_p->m_Done, edt_p->m_Todo));
      }
      break;
    }
    if (edt_p->m_ContState != laststate)
    {
      EDTPRT(edt_p,
             DBG_DMA,
             TRL_INFO,
             ("%s: %d done %d todo %d started %d loaded\n",
              snames[edt_p->m_ContState],
              edt_p->m_Done,
              edt_p->m_Todo,
              edt_p->m_Started,
              edt_p->m_Loaded));
      laststate = edt_p->m_ContState;
      EdtTrace(edt_p, TR_CONTSTATE);
      EdtTrace(edt_p, edt_p->m_ContState);
    }
  }
}

void edtdrv_pdv_set_continuous(Edt_Dev* edt_p, int state)
{
  edt_start_critical_src(edt_p, 30);

  if (state)
  {
    edt_p->m_LastTodo = edt_p->m_Todo;
    edt_p->m_LastDone = edt_p->m_Done;
    edt_p->m_Continuous = state;
    edt_p->m_ContState = 0;
  }
  else
  {
    EdtTrace(edt_p, 0x5aa5);
    edt_p->m_Continuous = state;
    edt_p->m_Todo = edt_p->m_Done;
    edt_p->m_Started = edt_p->m_Done;
  }

  edt_end_critical_src(edt_p, 30);
}
