/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * Header file for declarations & defines related to loading the Xilinx
 * bitfile.
 */
#ifndef _EDT_BF_H_
#define _EDT_BF_H_

/* MAXPATH, etc. */
#include "edt_os_utils.h"

#ifdef __cplusplus
extern "C" {
#endif

/* shorthand debug level */
#define DEBUG0 PDVLIB_MSG_FATAL
#define DEBUG1 EDTLIB_MSG_INFO_1
#define DEBUG2 EDTLIB_MSG_INFO_2

#ifdef DOXYGEN_SHOW_UNDOC
/**
 * @weakgroup edt_undoc
 * @{
 */
#endif

/** Number of bytes to read in for bitfile preamble. */
#define BFH_EXTRASIZE 128

/** Structure for bitfile header fields */
typedef struct {
  char filename[MAXPATH];
  uint8_t fi[8];
  /** Bitfile name extracted from the vendor-specific header. Xilinx ISE bitfiles
   * historically had a ".ncd" extension. */
  char ncdname[MAXPATH];
  uint8_t id[32];
  /** Date string when bitfile was generated. */
  uint8_t date[16];
  /** Time string when bitfile was generated. */
  uint8_t time[16];
  uint32_t dsize; /* get_long_size puts 4-byte dsize here */
  uint32_t filesize;
  uint32_t key;
  /** XilinxMagic, AlteraMagic, or Altera2Magic */
  int magic;
  uint32_t data_start;
  char promstr[256];
  uint8_t extra[BFH_EXTRASIZE];
} EdtBitfileHeader;

/** Structure used to access UI bitfile. */
typedef struct EdtBitfile {
  int is_file;
  char* filename;

  uint32_t full_buffer_size; /* current size of valid data */
  uint32_t buffer_allocated; /* actual allocated size - can be bigger than buffer_size */
  uint8_t* full_buffer;      /* data */
  uint32_t cur_index;
  uint8_t* data;      /* start of actual data */
  uint32_t data_size; /* size - header */
  HANDLE f;

  EdtBitfileHeader hdr;

} EdtBitfile;

/**
 * Structure describing a list of bitfiles
 * and their associated headers.
 */
typedef struct _bitfile_list {
  int nbfiles;
  EdtBitfileHeader* bitfiles;
} EdtBitfileList;

#define MAX_CHIPS_PER_ID 16

/** Structure describing the FPGAs on a board. */
typedef struct {
  uint32_t id;
  const char* fpga_0[MAX_CHIPS_PER_ID];
  const char* fpga_1[MAX_CHIPS_PER_ID];
} EdtBoardFpgas;

extern EDTAPI EdtBoardFpgas board_chips[];
extern EDTAPI EdtBoardFpgas mezz_chips[];

/**
 * @brief Initialize a bitfile list to zero.
 *
 * @param bf
 *   Pointer to the bitfile list to initialize.
 */
EDTAPI void bf_init(EdtBitfileList* bf);

/**
 * @brief Frees the memory for a list of bitfiles.
 *
 * @param bf
 *   Pointer to the bitfile list to destroy.
 */
EDTAPI void bf_destroy(EdtBitfileList* bf);

/**
 * @brief Adds a bitfile header to a list of bitfiles.
 *
 * @param bf
 *   Pointer to the bitfile list to add to.
 * @param bfh
 *   Pointer to the bitfile header to add.
 */
EDTAPI void bf_add_entry(EdtBitfileList* bf, EdtBitfileHeader* bfh);

/**
 * @brief Checks if a bitfile exists and adds it to the
 * bitfile list if it does.
 *
 * @param bf
 *   Pointer to the bitfile list to add to.
 * @param fname
 *   Name of the bitfile to check for.
 */
EDTAPI void bf_check_and_add(EdtBitfileList* bf, const char* fname);

/**
 * @brief Allocates @p data to the largest bitfile size found in the @p bf list.
 *
 * Deallocate with @ref edt_bitfile_destroy()
 *
 * @param bf
 *   Pointer to the bitfile list to read.
 * @param[out] data
 *   Bitfile struct with EdtBitfile.full_buffer set to an allocated array of
 *   size EdtBitfile.buffer_allocated bytes.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int bf_allocate_max_buffer(EdtBitfileList* bf, EdtBitfile* data);

/**
 * @brief Gets the base directory based on the device type
 * (PDV vs. PCD).
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param in
 *   The name of the directory. Optional, if set, it will be returned.
 * @param out
 *   The resulting directory. Optional, if set, will overwritten with the home
 *   directory value and returned.
 *
 * @return The resulting directory.
 */
EDT_LIB_LOCAL const char* edt_bitload_basedir(EdtDev edt_p, const char* in, char* out);

/**
 * @brief Gets all possible bitfiles with name fname and stores them in list
 * @p bf. If @p fpga_hint is present, it will be first in the list.
 *
 * @param basedir
 *   The base directory to look for the bitfile in.
 * @param devdir
 *   The device directory to look for the bitfile in.
 * @param fname
 *   The name of the bitfile.
 * @param fpga_list
 *   The list of available FPGAs.
 * @param id
 *   Either a board id or a mezzanine id, which is used to look up FPGA type(s)
 *   in the @p fpga_list.
 * @param channel
 *   Index into the mezz_chips array.
 * @param bf
 *   The list of bitfiles to add to.
 * @param fpga_hint
 *   The name of the fpga to check for.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_get_bitfile_list(const char* basedir,
                                const char* devdir,
                                const char* fname,
                                EdtBoardFpgas* fpga_list,
                                int id,
                                int channel,
                                EdtBitfileList* bf,
                                char* fpga_hint);

/**
 * @brief Loads a mezzanine bitfile.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param basedir
 *   The base directory to look for the bitfile in.
 * @param bitname
 *   The name of the bitfile.
 * @param bitload_flags
 *   Flags to use when loading the bitfile.
 * @param skip_load
 *   Currently not used.
 * @param channel
 *   Index into the mezz_chips array.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_load_mezzfile(EdtDev edt_p,
                             const char* basedir,
                             const char* bitname,
                             int bitload_flags,
                             int skip_load,
                             int channel);

/**
 * @brief Extract the FPGA header (Xilinx, OR faked one for Altera, etc.)
 *
 * @param bp
 *   Struct that will be populated with the FPGA file information.
 * @param bfh
 *   Struct that will be populated with the FPGA file header information.
 *
 * @return -1 on error, otherwise the address of the beginning of the FPGA data
 * (after the header).
 */
EDTAPI int edt_get_bitfile_header(EdtBitfile* bp, EdtBitfileHeader* bfh);

/**
 * @brief Program the mezzanine bitfile. Wraps @ref edt_program_mezzanine_direct().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param buf
 *   Pointer to the bitfile data.
 * @param size
 *   Size of the bitfile data.
 * @param channel
 *   Which channel the data is directed to.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_program_mezzanine(EdtDev edt_p, const uint8_t* buf, uint32_t size, int channel);

/**
 * @brief Program the interface FPGA. Typically only called by edt_bitload.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param buf
 *   Buffer containing the data to be loaded.
 * @param size
 *   Number of bytes to load from buffer.
 * @param do_sleep
 *   Enable (1) or disable (0) delays to slow flashing down in case of host
 *   speed issues.
 *
 * @return 0 on success, nonzero on failure
 */
EDT_LIB_LOCAL int edt_program_flash_direct(EdtDev edt_p,
                                           const uint8_t* buf,
                                           int size,
                                           int do_sleep);

/**
 * @brief Program the mezzanine bitfile.
 *
 * @param edt_p
 *   Pointer to the device structure.
 * @param buf
 *   Pointer to the bitfile data.
 * @param size
 *   Size of the bitfile data.
 * @param channel
 *   Which channel the data is directed to.
 *
 * @return 0 on success, -1 on failure.
 */
EDT_LIB_LOCAL int edt_program_mezzanine_direct(EdtDev edt_p,
                                               const uint8_t* buf,
                                               int size,
                                               int channel);

/**
 * @addtogroup dma_init
 * @{
 */

/**
 * Searches for and loads a FPGA bit file into an EDT board. Searches under
 * `<basedir>/bitfiles/xxx`, or if a PDV board, in the appropriate sub-directory.
 * Quits after the first successful load.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param basedir
 *   Base directory to start looking for the file.
 * @param fname
 *   Name of the bitfile to load.
 * @param flags
 *   Misc flag bits -- should be combination of BITLOAD_FLAGS_*.
 *   (This variable was formerly `rcam` which is obsolete.)
 * @param skip
 *   If nonzero, don't actually load the file, just find the files (debugging).
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_bitload(EdtDev edt_p, const char* basedir, const char* fname, int flags, int skip);

/** @} */ /* end dma_init */

/* flag bits to use in flags param to edt_bitload */
// #define BITLOAD_FLAGS_NOFS 0x1 /**< embedded for vxworks */
#define BITLOAD_FLAGS_OVR      0x2 /**< Override GS/SS bitfile size constraint */
#define BITLOAD_FLAGS_CH1      0x8 /**< Program OCM channel 1 if BITLOAD_FLAGS_OCM is set */
#define BITLOAD_FLAGS_READBACK 0x20
#define BITLOAD_FLAGS_SLEEP    0x40

/** Changes programming algorithm to program mezzanine boards. */
#define BITLOAD_FLAGS_MEZZANINE        0x4
#define _DEPRECATED_BITLOAD_FLAGS_OCM  0x4  /**< @deprecated Use BITLOAD_FLAGS_MEZZANINE. */
#define _DEPRECATED_BITLOAD_FLAGS_SRXL 0x10 /**< @deprecated Use BITLOAD_FLAGS_MEZZANINE. */

/**
 * @brief Opens a bitfile and populates the bfd structure.
 *
 * @param bfd
 *   The structure to be populated.
 * @param name
 *   The path of the file to open.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_bitfile_load_file(EdtBitfile* bfd, const char* name);

/**
 * @brief Initialize an EdtBitfile data structure.
 *
 * @param bitfile
 *   The structure to be initialized.
 */
EDTAPI void edt_bitfile_init(EdtBitfile* bitfile);

/**
 * @brief Closes the associated file handle, zeros fields, and frees allocated
 * memory for the given EdtBitfile structure.
 *
 * @param bfd
 *   The bitfile to be destroyed.
 */
EDTAPI void edt_bitfile_destroy(EdtBitfile* bfd);

/**
 * Normalizes a bitfile name to ensure it is valid.
 *
 * @param name
 *   Name of bitfile before normalization.
 * @param bitname
 *   The normalized bitfile name.
 */
EDTAPI void ensure_bitfile_name(const char* name, char* bitname);

/**
 * @brief Get the FPGA header from an array, typically as read out from an FPGA.
 *
 * @param ba
 *   Array containing data bytes presumably read from the FPGA.
 * @param header
 *   String that will be populated with the header info in a printable format.
 * @param size
 *   Size of the formatted @p header string.
 *
 * @returns 0 on success, -1 on failure.
 */
EDTAPI int edt_get_x_array_header(uint8_t* ba, char* header, int* size);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _EDT_BF_H_ */
