/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2000 Engineering Design Team, Inc.
 * @brief Generalized error handling for edt and pdv libraries (header file)
 */
#ifndef EDT_ERROR_H
#define EDT_ERROR_H

#include "edt_os_utils.h"

#include <stdarg.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef enum EdtReturn {
  RET_SUCCESS = 0,
  ERROR_GENERIC = -1,
  ERROR_BADCHANNEL = -2,
  ERROR_CONNECTION = -3,
  ERROR_BADPARAM = -4,
  ERROR_BADPATH = -5,
  ERROR_BADDEVICE = -6,
  ERROR_PERMISSION = -7,
  ERROR_NOCODE = -8,
  ERROR_UNDEFINED = -9,
  ERROR_NULL_POINTER = -10,
  ERROR_NO_FILE = -11,
  ERROR_RESPONSE = -12
} EdtReturn;

/**
 * @defgroup msg EDT Message Handler Library
 *
 * Provides generalized error- and message-handling for the edt and pdv
 * libraries.
 *
 * These routines provide a way for application programs to intercept
 * and handle edtlib and pdvlib error, warning, and debug messages, but
 * you can also use them for application messages.
 *
 * By default, output goes to the console (stdout), but you can
 * substitute user-defined functions -- for example, a function that
 * pops up a window to display text. You can set different message
 * levels for different output, and multiple message handles can exist
 * within an application, with different message handlers associated
 * with them.
 *
 * Predefined message flags are described in the "Defines" section of
 * this document. Those starting with \c EDTAPP_MSG_ are for general
 * application use, those starting with \c EDTLIB_MSG_ are for libedt
 * messages, and those beginning with \c PDVLIB_MSG_ are for libpdv
 * messages.  Application programmers can define other flags in the
 * 0x1000 to 0x1000000 range.
 *
 * Message levels are defined by flag bits, and each bit can be set or
 * cleared individually. So, for example, to have a message-handler
 * called only for fatal and warning application messages, specify
 * <tt> EDTAPP_MSG_FATAL | EDTAPP_MSG_WARNING </tt>.
 *
 * As you can see, the edt and pci dv libraries have their own message
 * flags. These can be turned on and off from within an application,
 * and also by setting the environment variables \c EDTDEBUG and \c PDVDEBUG,
 * respectively, to values greater than zero.
 *
 * Application programs ordinarily specify combinations of either the
 * \c EDTAPP_MSG_ or \c EDT_MSG_ flags for their messages.
 *
 * \b Files
 *
 * \arg \e edt_error.h header file (automatically included if \e edtinc.h is
 *     included)
 * \arg \e edt_error.c: message subroutines
 *
 * The #EdtMsgHandler structure is defined in \e edt_error.h.  For
 * compatibility with possible future changes, do not access structure
 * elements directly; instead always use the error subroutines.
 *
 * @{
 */

/**
 * @brief An EdtMsgFunction is a function which outputs a message if that
 * message's level is high enough.
 *
 * @param target
 *    This stores extra info useful to the specific function
 *    defined. In the default message handler setup by edt_msg_init,
 *    the function used expects target to be a FILE pointer.
 * @param level
 *    The message level associated with with the message.
 * @param message
 *    The message which can be output by the function.
 */
typedef int (*EdtMsgFunction)(void* target, int level, const char* message);

/**
 * Structure used by the \ref msg "Message Handler Library" to control
 * the output of messages.
 */
typedef struct edt_msg_handler_s {
  EdtMsgFunction func;
  int level;
  /** The file the default handler function sends output to (\c stderr). */
  FILE* file;
  /** Flag set by #edt_msg_set_name to indicate that libedt is responsible
   * for closing the file */
  unsigned char own_file;
  void* target;

} EdtMsgHandler;

/* default initialization - uses stdout and stderr */

/**
 * @brief Initializes a message handler with default values. The message file is
 * initialized to @c stderr. The output subroutine pointer is set to @c
 * fprintf (to write output to the console). The message level is set to
 * EDT_MSG_WARNING | EDT_MSG_FATAL.
 *
 * @param msg_p
 *   Pointer to message handler structure to initialize.
 *
 * @see edt_msg_output
 */
EDTAPI void edt_msg_init(EdtMsgHandler* msg_p);

/**
 * @brief Initializes a message handler to use the specified file and level.
 * This takes a pointer to a FILE which has been opened by the caller.
 *
 * @param msg_p
 *    Pointer to message handler structure to initialize.
 * @param file
 *    FILE pointer returned by e.g. fopen().
 * @param level
 *    The level that future messages must match against if
 *    they are to be handled by the \a msg_p handler.
 */
EDTAPI void edt_msg_init_file(EdtMsgHandler* msg_p, FILE* file, int level);

/**
 * @brief Closes and frees up memory associated with a message handler. Use only
 * on message handlers that have been explicitly initialized by
 * edt_msg_init. Do not try to close the default message handler.
 * If the message handler has been configured to use a file which the
 * user opened, through functions such as #edt_msg_init_file or
 * #edt_msg_set_file, then the user is responsible for closing that
 * file after calling this function.
 *
 * @param msg_p
 *    Pointer to message handler to close, which was initialized by #edt_msg_init
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI void edt_msg_close(EdtMsgHandler* msg_p);

/**
 * @brief Sets the "message level" flag bits that determine whether to call the
 * message handler for a given message. The flags set by this function
 * are ANDed with the flags set in each edt_msg call, to determine
 * whether the call goes to the message function and actually results in
 * any output.
 *
 * @param msg_p
 *    Pointer to message handler.
 * @param newlevel
 *    The new level to set in the message handler.
 */
EDTAPI void edt_msg_set_level(EdtMsgHandler* msg_p, int newlevel);

/**
 * @brief Gets the message level that messages must match in order to be
 * handled by the message handler \a msg_p. The level is a combination
 * of flags OR'ed together as described in the overview.
 *
 * @param msg_p
 *    Pointer to message handler.
 */
EDTAPI int edt_msg_get_level(EdtMsgHandler* msg_p);

/**
 * @brief Sets the function to call when a message event occurs. The default
 * message function is \c fprintf() (which outputs to \c stderr);
 * this routine allows programmers to substitute any type of
 * message handler (pop-up callback, file write, etc).
 *
 * For an example of how this could be used, see #edt_msg.
 *
 * @param msg_p
 *    Pointer to message handler.
 * @param f
 *    The function to call when a message event occurs.
 *
 * @see edt_msg_set_level, edt_msg
 */
EDTAPI void edt_msg_set_function(EdtMsgHandler* msg_p, EdtMsgFunction f);

/**
 * @brief Sets the output file pointer for the message handler.
 *
 * The user still owns the file, so they are responsible for closing it
 * after this message handler is done with it, such as after this function
 * is called again, or after #edt_msg_close is called.
 *
 * @param msg_p
 *    Pointer to message handler.
 * @param f
 *    FILE pointer to an opened file, to which the messages should be output.
 */
EDTAPI void edt_msg_set_file(EdtMsgHandler* msg_p, FILE* f);

/**
 * @brief Sets the target in the message handler.
 *
 * @param msg_p
 *    Pointer to message handler.
 * @param t
 *    Stores extra info useful to the specific function
 *    defined. In the default message handler setup by edt_msg_init,
 *    the function used expects target to be a FILE pointer.
 *    The target would usually be an object that messages are sent to, such
 *    as a window, but exactly what it will be depends on what the message
 *    handler's function expects.
 *
 * @see edt_msg_set_function, EdtMsgFunction
 */
EDTAPI void edt_msg_set_target(EdtMsgHandler* msg_p, void* t);

/**
 * @brief Sets the output file to the named file.
 *
 * @param msg_p
 *    Pointer to message handler.
 * @param name
 *    The name of a file to open. Future messages will be written to that file.
 */
EDTAPI void edt_msg_set_name(EdtMsgHandler* msg_p, const char* name);

/**
 * @brief Submits a message to the default message handler, which will
 * conditionally (based on the flag bits) send the message
 * to the default message handler function. This function uses
 * the default message handler, and is equivalent to calling
 * edt_msg_output(edt_msg_default_handle(), ...). To submit a message
 * for handling by other than the default message handle, use
 * #edt_msg_output.
 *
 * @param level
 *    An integer variable that contains flag bits indicating
 *    what 'level' message it is. Flag bits are described in the overview.
 * @param format
 *    A string and arguments describing the format. Uses
 *    vsprintf to print formatted text to a string, and sends the result to
 *    the handler subroutine. Refer to the @c printf manual page for
 *    formatting flags and options.
 *
 * @return 0 on success, -1 on failure.
 */
EDT_STR_FORMAT_ATTR(__printf__, 2, 3)
EDTAPI int edt_msg(int level, const char* format, ...);

EDT_STR_FORMAT_ATTR(__printf__, 3, 4)
EDTAPI int edt_msg_output(EdtMsgHandler* msg_p, int level, const char* format, ...);

/**
 * @brief Conditionally outputs a system perror using the default message handler.
 *
 * This function is equivalent to calling
 * edt_msg_output_perror(edt_msg_default_handle(), level, msg);
 *
 * @param level
 *    Message level for the current message, as described in the overview.
 * @param msg
 *    Message to concatenate to the system error message.
 *
 * @return 0 on success, -1 on failure.
 * @see edt_perror, edt_msg_output_perror
 */
EDTAPI int edt_msg_perror(int level, const char* msg);

/**
 * @brief Conditionally (based on the flag bits) outputs \p message, followed by
 * the last system error message, to \p msg_p.
 *
 * To output to the default message handler, use #edt_msg_perror.
 *
 * @param msg_p
 *    Pointer to message handler, initialized by #edt_msg_init.
 * @param level
 *    Message level for the current message, as described in the overview.
 * @param msg
 *    Message to concatenate to the system error message.
 *
 * @see edt_perror
 */
EDTAPI int edt_msg_output_perror(EdtMsgHandler* msg_p, int level, const char* message);

/**
 * @brief Writes to the specified EdtMsgHandler a caller-specified message
 * (in the printf-style format) followed by the last system error message.
 *
 * @param msg_p
 *    Pointer to message handler, initialized by #edt_msg_init
 * @param level
 *    The EDT Message level. This function will only output the
 *    message if level is greater than or equal to that set by #edt_msg_init
 *    or #edt_msg_set_level.
 * @param format
 *    A printf() style format string. Like printf(), it should be
 *    followed by arguments to match the format.
 */
EDT_STR_FORMAT_ATTR(__printf__, 3, 4)
EDTAPI int edt_msg_output_printf_perror(EdtMsgHandler* msg_p, int level, const char* format, ...);

/**
 * @brief Gets the default message handler.
 *
 * This is useful if you want to modify the default handler's behaviour,
 * with functions such as #edt_msg_set_level, #edt_msg_set_function,
 * #edt_msg_set_file, #edt_msg_set_name, or #edt_msg_set_target.
 */
EDTAPI EdtMsgHandler* edt_msg_default_handle(void);

/**
 * @brief Gets the message level that messages must match in order to be
 * handled by the default message handler. The level is a combination
 * of flags OR'ed together as described in the overview.
 *
 * The equivalent function for a user defined message handler is
 * #edt_msg_get_level.
 */
EDTAPI int edt_msg_default_level(void);

/**
 * @brief Returns pointer to most recent msg string.
 */
EDTAPI char* edt_msg_last_error(void);

/**
 * @brief Simple verbose level setting output message simplified - levels 0 - 3
 *
 * @param verbose
 *    Level setting. One of:
 *        level 0 - fatal only
 *        level 1 - warnings
 *        level 2 - info level 1
 *        level 3 - info level 2
 */
EDTAPI void edt_set_verbosity(int verbose);

/**
 * @brief Return simple verbose level setting.
 */
EDTAPI int edt_get_verbosity(void);

/**
 * Predefined message flags. EDTAPP_MSG are for general purpose application
 * use EDTLIB_MSG are for libedt messages. PDVLIB are for libpdv messages.
 * Application programmers can define other flags in the 0x1000 to
 * 0x1000000 range.
 */

/** Fatal-error messages in applications. */
#define EDTAPP_MSG_FATAL 0x1

/** Warning messages in applications */
#define EDTAPP_MSG_WARNING 0x2

/** First level info messages in applications. */
#define EDTAPP_MSG_INFO_1 0x4

/** Second level info messages in applications. */
#define EDTAPP_MSG_INFO_2 0x8

/** Fatal-error messages in libedt. */
#define EDTLIB_MSG_FATAL 0x10
/** Warning messages in libedt. */
#define EDTLIB_MSG_WARNING 0x20

/** Informative messages in libedt. */
#define EDTLIB_MSG_INFO_1 0x40

/** Debugging messages in libedt. */
#define EDTLIB_MSG_INFO_2 0x80

/** Fatal-error messages in libpdv. */
#define PDVLIB_MSG_FATAL 0x100

/** Warning messages in libpdv. */
#define PDVLIB_MSG_WARNING 0x200

/** Informative messages in libpdv. */
#define PDVLIB_MSG_INFO_1 0x400

/** Debugging messages in libpdv. */
#define PDVLIB_MSG_INFO_2 0x800

#define EDT_MSG_ALWAYS 0x80000000

#define EDT_MSG_FATAL EDTAPP_MSG_FATAL | EDTLIB_MSG_FATAL | PDVLIB_MSG_FATAL

#define EDT_MSG_WARNING EDTAPP_MSG_WARNING | EDTLIB_MSG_WARNING | PDVLIB_MSG_WARNING

#define EDT_MSG_INFO_1 EDTAPP_MSG_INFO_1 | EDTLIB_MSG_INFO_1 | PDVLIB_MSG_INFO_1

#define EDT_MSG_INFO_2 EDTAPP_MSG_INFO_2 | EDTLIB_MSG_INFO_2 | PDVLIB_MSG_INFO_2

/** @} */ /* end group msg */

#ifdef __cplusplus
}
#endif

#endif /* EDT_ERROR_H */
