/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2013 Engineering Design Team, Inc.
 * @brief FPGA header read / write utilities.
 */

#ifndef EDT_FLASH_H
#define EDT_FLASH_H

#include "libedt.h"
#include "edt_portability.h"

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Prom stuff, mostly for pciload.
 */
#define PCI_ID_SIZE    128
#define OSN_SIZE       32
#define ESN_SIZE       64
#define PCIE_INFO_SIZE 1024
#define PCIE_PID_SIZE  128
#define MACADDR_SIZE   12
#define MAX_MACADDRS   16
/* Should be ((MACADDR_SIZE+1) * MAX_MACADDRS) + 4 for tag + 4 for the count
 * + comma, even number. */
#define MACLIST_SIZE                       216
#define EXTRAADDR_SIZE                     4
#define EXTRATAG_SIZE                      4
#define OPTSN_SIZE                         32
#define PROM_EXTRA_SIZE                    512
#define E_SECTOR_SIZE                      0x04000
#define AMD_SECTOR_SIZE                    0x10000
#define AMD_SECTOR_SIZE2                   0x20000
#define SPI_SECTOR_SIZE                    0x100000
#define MIC_N25Q064A13ESE40G_PROMINFO_ADDR 0x7f0000
#define EDTMACOUI                          0x00251C

/*
 * Legal device proms for pciload and other prom detect code.
 */
#define PROM_UNKN            0
#define AMD_4013E            1
#define AMD_4013XLA          2
#define AMD_4028XLA          3
#define AMD_XC2S150          4
#define AMD_XC2S200_4M       5
#define AMD_XC2S200_8M       6
#define AMD_XC2S100_8M       7
#define AMD_XC2S300E         8
#define SPI_XC3S1200E        9
#define AMD_XC5VLX30T        10 /* WAS SPI_XC5VLX30T */
#define AMD_XC5VLX50T        11 /* WAS SPI_XC5VLX50T */
#define AMD_EP2SGX30D        12
#define AMD_XC5VLX70T        13
#define AMD_XC5VLX30T_A      14
#define AMD_XC6SLX45         15
#define AMD_EP2SGX30D_A      16
#define AMD_EP2AGX45D        17
#define AMD_5SGXMA3K2F40C3   18
#define AMD_5SGXMA5K2F40C3   19
#define AMD_5SGXMA7K2F40C3   20
#define AMD_5SGXMA8K2H40C3N  21
#define AMD_5SGXMA7K1F40C2   22
#define AMD_5SGXEA2K1F40C2ES 23
#define AMD_5SGXMA9K2H40C2   24
#define MIC_N25Q064A13ESE40G 25
#define AMD_5AGZME1          26
#define MIC_N25Q256A13EF840  27

#define AMD_S29GL512S_XC7K160T 34 /* PCIe CML-ECL */
#define AMD_S29GL512S_XC7K325T 35 /* PCIe CML-ECL */
#define AMD_S29GL512S_XC7K410T 36 /* PCIe CML-ECL */
/** @deprecated Use RFX_RADIO_MIC_MT25QL256ABA1EW9 instead. */
#define HH_RB_MIC_MT25QL256ABA1EW9     RFX_RADIO_MIC_MT25QL256ABA1EW9
#define RFX_RADIO_MIC_MT25QL256ABA1EW9 38
#define AMD_XCKU11P                    39 /* Added for PE8G3_WBX_ID */

/* Device (chip) programming types for pciload etc. */
#define FTYPE_X    0
#define FTYPE_BT   1
#define FTYPE_LTX  2
#define FTYPE_SPI  3
#define FTYPE_BT2  4
#define FTYPE_F16  5
#define FTYPE_MIC  6
#define FTYPE_F16A 7

/* Magic numbers for devices (chips). */
typedef enum { UnknownMagic, XilinxMagic = 1, AlteraMagic = 2, Altera2Magic = 3 } FPGAMagic;

/* FPGA bitfile size constants */
#define MIN_BIT_SIZE_ANY    0x4000
#define MIN_BIT_SIZE_X      0x8000
#define MIN_BIT_SIZE_XLA    0x14000
#define MIN_BIT_SIZE_BT     0x20000
#define MIN_BIT_SIZE_SPI    0x60000
#define MIN_BIT_SIZE_AMD512 0x300000

/* Flash ROM control register (`CTRL`) bits */
#define EDT_ROM_JUMPER 0x01
#define EDT_5_VOLT     0x02

/* Write Control command, allows for writing the CTRL 4-bit register in the
 * flash rom FPGA interface. Written to the EDT_FLASHROM_DATA register.
 * e.g. edt_reg_write(edt_p, EDT_FLASHROM_DATA, (EDT_ROM_CMD_WRITE_CTRL | <data to write>)
 */
#define EDT_ROM_CMD_WRITE_CTRL (0x02 << 16)

/*
 * Status bits.
 *
 * Bits 2-7 are Xilinx ID for 4028 and above.
 * Shift down 2 and we get a number for the boot controller.
 */
#define STAT_PROTECTED    0x01
#define STAT_5V           0x02
#define STAT_IDMASK       0x7c
#define STAT_IDSHFT       2
#define IS_DEFAULT_SECTOR -6167

/*
 * Command bits for the 4028xla boot controller.
 */
#define BT_READ      0x0100
#define BT_WRITE     0x0200
#define BT_INC_ADD   0x0400
#define BT_A0        0x0800
#define BT_A1        0x1000
#define BT_RSVD      0x2000
#define BT_REINIT    0x4000
#define BT_EN_READ   0x8000
#define BT_LD_LO     BT_WRITE
#define BT_LD_MID    BT_WRITE | BT_A0
#define BT_LD_HI     BT_WRITE | BT_A1
#define BT_LD_ROM    BT_WRITE | BT_A0 | BT_A1
#define BT_RD_ROM    BT_READ | BT_A0 | BT_A1 | BT_EN_READ
#define BT_RD_FSTAT  BT_READ | BT_EN_READ
#define BT_RD_PALVER BT_READ | BT_A0 | BT_EN_READ
#define BT_MASK      0xff00

/**
 * Embedded information.
 *
 * @see @ref edt_parse_devinfo()
 */
typedef struct {
  int clock;
  char sn[11];  /**< Board serial number. */
  char pn[11];  /**< Board part number. */
  char opt[15]; /**< Board option string. */
  int rev;      /**< Board revision number. */
  char ifx[11];
  char optsn[11];
  char maclist[MACLIST_SIZE];
} Edt_embinfo;

typedef struct {
  char fpga[32];                   /* MUST match a pkg directory name */
  char promdesc[32];               /* description may include size */
  uint16_t stat;                   /* status bits after shift */
  uint16_t statx;                  /* extended status bits for 16-bit devices */
  char busdesc[8];                 /* PCI, PCIe, ... */
  int ftype;                       /* FTYPE_X, FTYPE_BT, FTYPE_LTX, FTYPE_SPI */
  int magic;                       /* magic code (not the magic # itself but a code */
  uint32_t sectorsize;             /* sector size in bytes */
  uint32_t sectsperseg;            /* sectors per segment */
  uint32_t nsegments;              /* number of segments (banks) */
  int defaultseg;                  /* default unprotected segment */
  uint32_t (*id_addr)(void*, int); /* id address function pointer */
  int load_seg0;                   /* 2d default unprotected segment (3v) */
  int load_seg1;                   /* 2d default unprotected segment (5v) */
} Edt_prominfo;

typedef struct {
  char type[4];
  uint32_t size;
} EdtPromParmBlock;

/** @deprecated (>v6.0.0) edt_get_parms_block() now returns NULL for all errors.
 * This value is not used. */
#define BAD_PARMS_BLOCK ((EdtPromParmBlock*)0xffffffff)

#define EdtPromParmData(p) ((uint8_t*)p + sizeof(EdtPromParmBlock))

typedef struct {
  char id[PCI_ID_SIZE];
  char esn[ESN_SIZE];
  char osn[OSN_SIZE];

  /* EXTRA area is variable size up to PROM_EXTRA_SIZE */
  /* size is int at beginning (top or bottom) of extra area */
  /* preceded by 4 char key "XTR:" */

  int extra_size; /* total size of parm blocks */
  int nblocks;    /* how many blocks */

  uint8_t extra_buf[PROM_EXTRA_SIZE];

  Edt_embinfo ei;
  char optsn[ESN_SIZE];
  char maclist[MACLIST_SIZE];
} EdtPromData;

typedef struct _prom_addr {
  uint32_t id_addr;
  uint32_t esn_addr;
  uint32_t osn_addr;
  uint32_t extra_data_addr;
  uint32_t extra_tag_addr;
  uint32_t extra_size_addr;
  uint32_t extra_size;
  uint32_t optsn_addr;
  uint32_t maclist_addr;
} EdtPromIdAddresses;

/**
 * @addtogroup prominfo
 * @{
 */

EDTAPI int edt_parse_esn(const char* str, Edt_embinfo* ei);

/**
 * @brief Parse the board's embedded information string.
 *
 * @param str
 *  An embedded information string, from edt_get_esn, etc.
 * @param[out] ei
 *  @ref Edt_embinfo structure into which the the parsed information will be put.

 * @return 0 on success, -1 on error (input string not valid or too long).
 *
 * @see edt_readinfo, edt_get_esn, edt_fmt_pn
 */
EDTAPI int edt_parse_devinfo(const char* str, Edt_embinfo* ei);

/**
 * @brief Retrieve the board's manufacturer and OEM embedded information strings
 * from the PCI Xilinx information header.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param esn
 *   A buffer of size ESN_SIZE or larger to receive the EDT information string.
 * @param osn
 *   A buffer of size OSN_SIZE or larger to receive the OEM information string.
 *
 * @see edt_get_esn, edt_get_osn, edt_parse_devinfo, edt_fmt_pn
 */
EDTAPI void edt_get_sns(EdtDev edt_p, char* esn, char* osn);

/**
 * @brief Retrieve the board's manufacturer and OEM embedded information strings
 * from the PCI Xilinx information header.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param esn
 *   A buffer of size ESN_SIZE or larger to receive the EDT information string.
 * @param osn
 *   A buffer of size OSN_SIZE or larger to receive the OEM information string.
 * @param sector
 *   Which sector of the PROM to retreive data from.
 *
 * @see edt_get_esn, edt_get_osn, edt_get_sns, edt_parse_devinfo, edt_fmt_pn
 */
EDTAPI void edt_get_sns_sector(EdtDev edt_p, char* esn, char* osn, int sector);

/**
 * @brief Retrieve the board's OEM embedded information string from the PCI
 * Xilinx information header.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param[out] osn
 *   A buffer of size OSN_SIZE or larger to receive the information string.
 *
 * @see edt_get_sns, edt_get_esn, edt_fmt_pn
 */
EDTAPI void edt_get_osn(EdtDev edt_p, char* osn);

/**
 * @brief Retrieve the board's EDT embedded information string from the PCI
 * Xilinx information header.
 *
 * @param edt_p
 *   The EDT open device handle.
 * @param[out] esn
 *   A buffer of size ESN_SIZE or larger to receive the information string.
 *
 * @see edt_get_sns, edt_get_osn, edt_parse_devinfo, edt_fmt_pn
 */
EDTAPI void edt_get_esn(EdtDev edt_p, char* esn);

/**
 * @brief Takes a part number portion of the embedded information string, as
 * retrieved from the PCI FPGA information header via edt_get_sns; reformats
 * and returns it in a "human readable" form (dashes added).
 *
 * This is mainly for use by pciload and other EDT utility programs, but also
 * can be used by user apps if desired.
 *
 * @param pn
 *  The part number without dashes.
 * @param[out] str
 *  A buffer to hold the output string, which must be at least 13 characters long.
 *
 * @see edt_get_esn, edt_parse_esn
 */
EDTAPI char* edt_fmt_pn(const char* pn, char* pn_str);

EDTAPI void edt_print_dev_flashstatus(EdtDev edt_p, uint16_t stat, int sector);

EDTAPI void edt_init_promdata(EdtPromData* pdata);
EDTAPI void edt_init_parmblock(EdtPromParmBlock* block, const char* type, int datasize);
EDTAPI EdtPromParmBlock* edt_add_parmblock(EdtPromData* pdata, const char* type, int datasize);
EDTAPI EdtPromParmBlock* edt_get_parms_block(EdtPromData* pdata, const char* id);

EDTAPI uint32_t edt_flash_get_promaddrs(EdtDev edt_p,
                                        int promcode,
                                        int segment,
                                        EdtPromIdAddresses* paddr);

/**
 * @brief Extract the name of the on-board firmware in the device's FPGA PROM,
 * minus the extension.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param[out] name
 *  A buffer of size PCI_ID_SIZE or larger to receive the FPGA filename.
 *
 * @see edt_flash_get_fname_auto
 */
EDTAPI int edt_flash_get_fname(EdtDev edt_p, char* name);

/**
 * @brief Extract the name of the on-board firmware in the device's FPGA PROM,
 * minus the version and extension.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param[out] name
 *  A buffer of size PCI_ID_SIZE or larger to receive the FPGA filename.
 *
 * @see edt_flash_get_fname
 */
EDTAPI int edt_flash_get_fname_auto(EdtDev edt_p, char* name);

EDTAPI void edt_flash_program_prominfo(EdtDev edt_p, int promcode, int sector, EdtPromData* pdata);
EDTAPI void edt_flash_byte_program(EdtDev edt_p, uint32_t addr, uint8_t data, int ftype);
EDTAPI void edt_flash_verify(EdtDev edt_p, uint32_t addr, uint8_t* data, int nbytes, int ftype);
EDTAPI void edt_flash_reset(EdtDev edt_p, int ftype);
EDTAPI void edt_flash_print16(EdtDev edt_p, uint32_t addr, int ftype);

/**
 * @brief Find the flash prom and flash status (jumper positions) for the device.
 * Finds out which flash PROM is used and queries the device PROM status --
 * see EPinfo stat field for status bits values
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param[out] stat
 *  The device status register value (if stat is NULL, it will be ignored).
 *
 * @return The prom device code.
 * @see edt_get_flashsize, edt_get_prominfo
 */
EDTAPI int edt_flash_prom_detect(EdtDev edt_p, uint16_t* stat);

EDTAPI int edt_flash_is_protected(EdtDev edt_p);

EDTAPI void edt_dump_flash(EdtDev edt_p, uint32_t addr, uint32_t size);

/**
 * @brief Get the flash PROM bank size.
 *
 * Note that this may be the full size of the PROM or the size of one program
 * bank, if the device is split up into multiple banks (program sectors).
 *
 * @param promcode
 *  The prom code, from #edt_flash_prom_detect.
 *
 * @return The size in bytes of the flash PROM programming bank.
 */
EDTAPI uint32_t edt_get_flashsize(int promcode);

/**
 * @return The highest numbered known promcode.
 */
EDTAPI int edt_get_max_promcode(void);

/**
 * @brief Returns the pciload device information structure for the specific
 * flash PROM code given.
 *
 * @param promcode
 *  The flash PROM code, from #edt_flash_prom_detect.
 *
 * @return A pointer to the #Edt_prominfo struct.
 */
EDTAPI Edt_prominfo* edt_get_prominfo(int promcode);

/**
 * @brief Returns the FPGA manufacturer name string.
 *
 * @param edt_p
 *  The EDT open device handle.
 *
 * @return the name of the FPGA type; one of "Xilinx", "Altera", or "" if unknown.
 */
EDTAPI const char* edt_get_fpga_mfg(EdtDev edt_p);

EDTAPI uint8_t edt_flash_read8(EdtDev edt_p, uint32_t addr, int ftype);

/**
 * @brief Returns a printable name for FTYPE values.  Mainly for debug output.
 *
 * @param ftype
 *   An FTYPE value.
 *
 * @return A string containing the flash type.
 */
EDTAPI const char* edt_flash_type_string(int ftype);

/**
 * @brief Read the board's FPGA configuration file id and serial numbers.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param promcode
 *  The device prom code, from #edt_flash_prom_detect.
 * @param sect
 *  The sector from which to read the prom (-1 for the default).
 * @param[out] idstr
 *  A buffer of size PCIE_PID_SIZE FPGA to receive configuration file ID string.
 * @param[out] devinfo
 *  A buffer of size ESN_SIZE to receive the device information string including
 *  serial_number part number, and other information.
 * @param[out] osn
 *  A buffer of size OSN_SIZE to receive the OEM's part number string, if present.
 *
 * @see  edt_parse_devinfo, edt_flash_prom_detect
 */
EDTAPI void edt_readinfo(EdtDev edt_p,
                         int promcode,
                         int sect,
                         char* idstr,
                         char* devinfo,
                         char* oemsn);

/**
 * @brief Get the onboard PROM device info as written to unused PROM space
 * by EDT before shipping, or thereafter via -I or -i options to
 * pciload.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param promcode
 *  The device prom code, from #edt_flash_prom_detect.
 * @param segment
 *  Which segment to read the info from; for the default, call #edt_get_prominfo,
 *  and use the defaultseg element from the Edt_prominfo struct it returns.
 * @param[out] pdata
 *   A pointer to an EdtPromData struct, into which the info will be read.
 */
EDTAPI void edt_read_prom_data(EdtDev edt_p, int promcode, int segment, EdtPromData* pdata);
EDTAPI uint32_t edt_get_id_addr(int promcode, int segment);

EDTAPI int edt_program_flash(EdtDev edt_p, const uint8_t* buf, int size, int do_sleep);

EDTAPI int edt_program_flash_start(EdtDev edt_p);
EDTAPI void edt_program_flash_chunk(EdtDev edt_p, const uint8_t* buf, int xfer, int do_sleep);
EDTAPI int edt_program_flash_end(EdtDev edt_p);
EDTAPI int edt_get_hw_rev(EdtDev edt_p);

EDTAPI int edt_sector_erase(EdtDev edt_p, uint32_t sector, uint32_t sec_size, int type);
EDTAPI uint32_t edt_flash_writesize(int ftype);

EDTAPI void edt_flash_block_program(EdtDev edt_p,
                                    uint32_t addr,
                                    const uint8_t* data,
                                    unsigned int nbytes,
                                    int ftype);

/**
 * @brief Read a string out of the FPGA, of a given size, from a given address.
 * Used for Xilinx ID id string and serial number strings.
 *
 * @param edt_p
 *  The EDT open device handle.
 * @param addr
 *  The address to start reading the string.
 * @param[out] buf
 *  The buffer to read the string into.
 * @param size
 *  The number of bytes to read.
 *
 * @return The buffer passed in.
 */
EDTAPI uint8_t* edt_flash_block_read(EdtDev edt_p,
                                     uint32_t addr,
                                     uint8_t* buf,
                                     unsigned int size,
                                     int ftype);

EDT_LIB_LOCAL int edt_flash_get_ftype(EdtDev edt_p);

/** @} */ /* end prominfo group */

#ifdef __cplusplus
}
#endif

#endif /* EDT_FLASH_H */
