/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 */
#ifndef _EDT_OPTSTRING_H_
#define _EDT_OPTSTRING_H_

#ifdef __cplusplus
extern "C" {
#endif

#define EDT_BITFILE_VERSION     0x0201007c
#define EDT_BITFILE_STRING      0x0101007e
#define EDT_OLD_BITFILE_STRING  0x0101007d
#define EDT_MEZZ_BITFILE_STRING 0x010100A0

/**
 * @brief Get board description.
 *
 * @note This function is primarily used for EDT's mezzanine cards, such
 * as the SS or GS boards. To get board and bitfile information from a
 * non-mezzanine card, such as an A5, S5, or VisionLink F1/F4, it may be simpler
 * to use @ref edt_read_prom_data().
 *
 * This attempts to learn as much as possible about the board referred to
 * by edt_p. If the bitfile that's loaded has an option string defined,
 * fill it in and parse its fields if possible. If edt_p is not channel 0,
 * channel 0 is opened, examined, and its fields copied to edt_p. The
 * bfd structure within the EdtDev structure is filled in as much as possible.
 *
 * @param edt_p
 *    The open EDT device handle.
 * @param force_mezzanine
 *    Attempt to load a bitfile to characterize the mezzanine board if no bitfile is loaded.
 *
 * @returns 0 on success, -1 on failure
 */
EDTAPI int edt_get_board_description(EdtDev edt_p, int force_mezzanine);

/**
 * @brief Prints all known fields from an EdtDev device.
 *
 * @param edt_p
 *    The open EDT device handle.
 */
EDTAPI void edt_print_board_description(EdtDev edt_p);

/**
 * @brief Get identifier for board associated with mezz_id.
 *
 * @param mezz_id
 *    The mezzanine id returned by edt_get_board_id.
 *
 * @returns A static string identifying the board associated with mezz_id.
 */
EDTAPI const char* edt_mezz_name(int mezz_id);

EDTAPI int edtdev_channels_from_type(EdtDev edt_p);

/**
 * @brief Check the current bitfile to see if it's a test file.
 *
 * @param edt_p
 *    The open EDT device handle.
 *
 * @returns 1 if test file, 0 otherwise
 */
EDTAPI int edt_is_test_file_loaded(EdtDev edt_p);

/**
 * @brief Read a new-type option string from the FPGA and copy
 * it into the EDT device handle's internal bitfile descriptor.
 *
 * @ref EdtBitfileDescriptor.optionstr and
 * @ref EdtBitfileDescriptor.revision_register members.
 *
 * @param edt_p
 *    The open EDT device handle.
 *
 * @returns 0 for success, -1 for failure (string not present)
 *
 * @see @ref edt_get_esn(), @ref edt_get_osn(), @ref edt_get_sns(), @ref edt_parse_devinfo()
 */
EDTAPI int edt_read_option_string(EdtDev edt_p);

/**
 * @brief Parse option string and populate fields in bitfile descriptor.
 *
 * @param s Option string.
 * @param bfd Bitfile descriptor to populate.
 *
 * @returns 0 on success, -1 on failure
 */
EDTAPI int edt_parse_option_string(char* s, EdtBitfileDescriptor* bfd);

/**
 * @brief Get read-only access to the @ref EdtBitfileDescriptor structure in the
 * device handle.
 *
 * @pre Call @ref edt_get_board_description() and @pre edt_read_option_string()
 * first to read information from the FPGA.
 *
 * @param edt_p
 *    The open EDT device handle.
 *
 * @return Bitfile descriptor (bfd) pointer or NULL on failure.
 */
EDTAPI const EdtBitfileDescriptor* edt_bitfile_get_descriptor(EdtDev edt_p);

#if defined(PCD)
/**
 * @brief Get read-only access to the @ref EdtMezzDescriptor structure in the
 * device handle.
 *
 * @pre Call @ref edt_get_full_board_id() or @ref edt_get_board_id() first, to
 * read information from the FPGA and parse it into the descriptor structure.
 *
 * @param edt_p
 *    The open EDT device handle.
 *
 * @return Mezzanine bitfile descriptor pointer or NULL on failure.
 */
EDTAPI const EdtMezzDescriptor* edt_bitfile_get_mezz_descriptor(EdtDev edt_p);
#endif

/**
 * @brief Get the last @ref EdtBitfileDescriptor.optionstr read from device.
 *
 * @pre Call @ref edt_read_option_string() first, to read the option string from
 * the FPGA.
 *
 * @param edt_p
 *    The open EDT device handle.
 *
 * @return Bitfile option string.
 *
 * @see  @ref edt_parse_devinfo()
 */
EDTAPI const char* edt_bitfile_option_string(EdtDev edt_p);

/**
 * @brief Get the last bitfile @ref EdtBitfileDescriptor.revision_register
 * value read from device.
 *
 * @pre Call @ref edt_read_option_string() first, to read the bitfile revision
 * from the FPGA.
 *
 * @param edt_p
 *    The open EDT device handle.
 *
 * @return Bitfile revision string.
 *
 * @see @ref edt_get_esn(), @ref edt_parse_devinfo()
 */
EDTAPI uint16_t edt_bitfile_revision(EdtDev edt_p);

#ifdef __cplusplus
}
#endif

#endif
