/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * Utility functions with OS-specific implementations.
 */

#ifndef EDT_OS_UTILS_H
#define EDT_OS_UTILS_H

#include "edt_types.h"
#include "edt_portability.h"

#ifdef _WIN32
#ifndef WIN32
#define WIN32
#endif
#endif

/* Include OS-Specific stuff in these headers */
#if defined(WIN32)
#include "edt_os_nt.h"
#elif defined(__linux__)
#include "edt_os_lnx.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* System time functions */

/**
 * @brief Return the time in seconds since the last call to edt_dtime.
 *
 * Each time edt_dtime is called, it returns the time delta since the last time it was called.
 * Therefore the first time edt_dtime is called in a process the delta may be meaningless. Time is
 * in seconds and fractions of a second.
 *
 * This call uses system time, which is system dependent and may be too coarse for some
 * applications. For more accurate time stamping consider the IRIG option and an external IRIG
 * timestamp source, available with some EDT boards.
 *
 * @return delta time, as a double floating point
 */
EDTAPI double edt_dtime(void);

/**
 * @brief Gets the seconds + fractional seconds since the Epoch, OS independent.
 * Uses system time; accuracy is OS dependent.
 *
 * @return The timestamp in seconds; a double value with tenths/hundredths/etc on the right of the
 * decimal point.
 */
EDTAPI double edt_timestamp(void);

/**
 * @brief Causes the process to sleep for the specified number of milliseconds.
 *
 * @param msecs
 *    The number of milliseconds for the process to sleep.
 *
 * @return 0 on success, -1 on failure. If an error occurs, call
 * #edt_perror to get the system error message.
 */
EDTAPI void edt_msleep(int msecs);

/**
 * @brief Causes process to sleep for specified number of microseconds.
 *
 * @param usecs
 *    Number of microseconds for process to sleep.
 *
 * @return 0 on success; -1 on error
 */
EDTAPI void edt_usleep(int usecs);

EDTAPI void edt_usec_busywait(uint32_t usec);

/**
 * @brief Convenience routine to free the memory allocated with #edt_alloc.
 *
 * @param ptr
 *    Address of memory buffer to free.
 */
EDTAPI void edt_free(void* ptr);

/**
 * @brief Convenience routine to allocate memory in a system-independent way.
 * The buffer returned is page aligned. Uses VirtualAlloc on Windows NT
 * systems, posix_memalign on UNIX-based systems.
 *
 * @param size
 *    Number of bytes of memory to allocate.
 *
 * @return The address of the allocated memory, or NULL on error. If
 * NULL, use #edt_perror to print the error.
 */
EDT_DEALLOCATE_WITH(edt_free)
EDTAPI void* edt_alloc(size_t size);

/**
 * @brief Cross platform function to close a directory handle.
 *
 * @param h
 *    Open directory handle.
 */
EDTAPI void edt_closedir(DIRHANDLE h);

/**
 * @brief Cross platform function to open a directory for reading.
 *
 * @param dirname
 *    The name of the directory to open.
 *
 * @return Directory handle upon success or NULL on failure.
 */
EDT_DEALLOCATE_WITH(edt_closedir)
EDTAPI DIRHANDLE edt_opendir(const char* dirname);

/**
 * @brief Cross platform function to read the next directory entry.
 *
 * @param h
 *    A handle for the directory, returned by #edt_opendir.
 * @param name
 *    A pointer to a string into which the name of the next
 * file in the directory will be written. The string must be allocated
 * by the caller to at least 256 bytes.
 *
 * @return h if entries remain or NULL if this is the last entry.
 */
EDTAPI int edt_readdir(DIRHANDLE h, char* name);

EDTAPI HANDLE edt_open_datafile(const char* path,
                                const char* name,
                                uint8_t writing,
                                uint8_t direct,
                                uint8_t truncate);
EDTAPI void edt_close_datafile(HANDLE f);
EDTAPI int edt_write_datafile(HANDLE f, void* p, int bufsize);
EDTAPI int edt_read_datafile(HANDLE f, void* p, int bufsize);

/**
 * @brief Change the slashes in str from forward to back or vice versa depending
 * on whether it's Windows or Unix.
 *
 * @param str
 *    String to process.
 */
EDTAPI void edt_correct_slashes(char* str);
EDTAPI void edt_fwd_to_back(char* str);
EDTAPI void edt_back_to_fwd(char* str);

EDTAPI uint64_t edt_get_file_position(HANDLE f);
EDTAPI uint64_t edt_get_file_size(HANDLE f);
EDTAPI uint64_t edt_file_seek(HANDLE f, uint64_t pos);

/**
 * @brief Determines file access, independent of operating system.
 *
 * This a convenience routine that maps to access() on Unix/Linux
 * systems and _access() on Windows systems.
 *
 * @param fname
 *    Path name of file to check access permissions.
 * @param perm
 *    Permission flag(s) to test for. See documentation
 * for access() (Unix/Linux) or _access() (Windows) for valid values.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_access(const char* fname, int perm);

EDTAPI void edt_init_critical(edt_critical_section_t* cs);
EDTAPI void edt_start_critical(edt_critical_section_t* cs);
EDTAPI void edt_end_critical(edt_critical_section_t* cs);

#ifdef __cplusplus
}
#endif

#endif /* EDT_OS_UTILS_H */
