/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2023 Engineering Design Team, Inc.
 * @brief Defines and macros for compiler portability or platform compatibility.
 */

#ifndef EDT_PORTABILITY_H
#define EDT_PORTABILITY_H

/**
 * By default, linking to a shared library is assumed. To link to a static
 * library, users should define `EDT_API_LIBSHARED=0` to disable shared library
 * exports in header files.
 */
#ifndef EDT_API_LIBSHARED
#define EDT_API_LIBSHARED 1
#endif

/*==== Library Symbol Export Macros =========================================*/
#if defined(__GNUC__) && (__GNUC__ >= 4)
#define EDT_HELPER_DLL_IMPORT __attribute__((visibility("default")))
#define EDT_HELPER_DLL_EXPORT __attribute__((visibility("default")))
#define EDT_HELPER_DLL_LOCAL  __attribute__((visibility("hidden")))
#elif defined(_MSC_VER) || defined(__CYGWIN__)
#define EDT_HELPER_DLL_IMPORT __declspec(dllimport)
#define EDT_HELPER_DLL_EXPORT __declspec(dllexport)
#define EDT_HELPER_DLL_LOCAL
#else
#define EDT_HELPER_DLL_IMPORT
#define EDT_HELPER_DLL_EXPORT
#define EDT_HELPER_DLL_LOCAL
#endif

#if defined(EDT_API_LIBSHARED) && (EDT_API_LIBSHARED == 1)
/* Defined when building shared library (not when using it). */
#ifdef EDT_API_LIBSHARED_EXPORTS
#define EDT_LIB_API EDT_HELPER_DLL_EXPORT
#define EDTAPI      EDT_HELPER_DLL_EXPORT
#else
#define EDT_LIB_API EDT_HELPER_DLL_IMPORT
#define EDTAPI      EDT_HELPER_DLL_IMPORT
#endif /* EDT_API_LIBSHARED_EXPORTS */
#define EDT_LIB_LOCAL EDT_HELPER_DLL_LOCAL
#else
#define EDTAPI
#define EDT_LIB_API
#define EDT_LIB_LOCAL
#endif /* EDT_API_LIBSHARED */

/*==== Macros with Compiler-specific definitions ============================*/

/**
 * @def EDT_DEALLOCATE_WITH
 * The memory allocated by the decorated function should be free'd by calling
 * the function `x`.
 *
 * - GCC function attribute "malloc(<deallocator>)"
 */

/**
 * @def EDT_STR_FORMAT_ATTR
 *
 * A wrapper for adding the 'format' function attribute to allow the compiler
 * to type-check the format string and variadic arguments.
 *
 * - GCC function attribute "format"
 *
 * @param archetype
 *   Style of formatting arguments, such as 'printf' or 'scanf'.
 * @param fmt_string_index
 *   Function parameter index of the format string.
 * @param first_to_check
 *   Function parameter index of the first variadic argument. Should be zero (0)
 *   for functions where arguments are not available to be checked.
 */

/**
 * @def EDT_CHECK_RETURN
 * The caller should inspect the return value of the decorated function.
 *
 * Dev note: Both the function declaration and definition should be decorated,
 * for MSVC analysis.
 *
 * - GCC function attribute "warn_unused_result"
 * - Microsoft Visual Studio property "_Check_return_" and warning C6031
 */

/**
 * @def EDT_DEPRECATED_FUNC
 * The decorated function is deprecated and may be removed in a future release.
 *
 * - GCC function attribute "deprecated"
 * - Microsoft Visual Studio "__declspec(deprecated)" and warning C4996
 */

/**
 * @def EDT_DEPRECATED_TYPE
 * The decorated typedef is deprecated and may be removed in a future release.
 *
 * - GCC type attribute "deprecated"
 * - MSVC "__declspec(deprecated)"
 */

#define EDT_DEPRECATED_FUNC_MSG_(ver_str, msg) "EDT deprecated as of package v" ver_str ". " msg

/*---- Microsoft Visual Studio ----------------------------------------------*/
#if defined(_MSC_VER)

#define EDT_DEALLOCATE_WITH(x)

#define EDT_STR_FORMAT_ATTR(archetype, fmt_string_index, first_to_check)

#if (_MSC_VER >= 1700)
#define EDT_CHECK_RETURN _Check_return_
#else
#define EDT_CHECK_RETURN
#endif

#define EDT_DEPRECATED_FUNC(as_of_version, explanation) \
  __declspec(deprecated(EDT_DEPRECATED_FUNC_MSG_(as_of_version, explanation)))

#define EDT_DEPRECATED_TYPE(as_of_version, explanation) \
  __declspec(deprecated(EDT_DEPRECATED_FUNC_MSG_(as_of_version, explanation)))

/*---- GCC ------------------------------------------------------------------*/
#elif defined(__GNUC__) || defined(__DOXYGEN__)

#if (__GNUC__ >= 11)
#define EDT_DEALLOCATE_WITH(x) __attribute__((__malloc__(x)))
#else
#define EDT_DEALLOCATE_WITH(x)
#endif

#define EDT_STR_FORMAT_ATTR(archetype, fmt_string_index, first_to_check) \
  __attribute__((__format__(archetype, fmt_string_index, first_to_check)))

#define EDT_CHECK_RETURN __attribute__((__warn_unused_result__))

#define EDT_DEPRECATED_FUNC(as_of_version, explanation) \
  __attribute__((__deprecated__(EDT_DEPRECATED_FUNC_MSG_(as_of_version, explanation))))

#define EDT_DEPRECATED_TYPE(as_of_version, explanation) \
  __attribute__((__deprecated__(EDT_DEPRECATED_FUNC_MSG_(as_of_version, explanation))))

/*---- Unknown --------------------------------------------------------------*/
#else
#error Unknown compiler.
#endif

#endif /* EDT_PORTABILITY_H */
