/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * A utility / example program to send multiple TIFF format images through the EDT PCI DV
 * CameraLink Simulator (CLS) or an equivalent to a Camera Link receiver.
 *
 * Has hooks for esoteric things: variable sized images, margin fills, and formatted image lists.
 * For a simpler application with the same purpose see: simple_clsend.c
 *
 * Prerequisites:
 *   Setup the CLS via the `clsiminit` application.
 */

#include "edtinc.h"
#include "clsim_lib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>

#ifdef __linux__
#include <strings.h> /* strcasecmp() */
#endif

/* Fill value for the left (A) and right (B) margins as described in the CLS User Guide.
 *  "-1" (0xFF) to use the pixel value found in the first pixel of the first line in the image. */
#define DEFAULT_FILLA 0xff
#define DEFAULT_FILLB 0xff
/* Vertical gap to have between the current image and the next image. */
#define DEFAULT_VGAP 400
/* Pixel specification of where in the camera's image to start the image from the file.
 *  "-1" indicates to center the image file within the camera's image. */
#define DEFAULT_HSTART (-1)
/* File containing a list of images to use. */
#define DEFAULT_LISTFILE "imagelist.txt"
/* Number of ring buffers to use; 4 provides max efficiency. */
#define NUMBUFS 4

static EdtMsgHandler logger;
static EdtMsgHandler* const lp = &logger;

/* Message Logger: minimum verbosity. */
#define ML_MIN_VRB_Msk 1
/* Message Logger: only DMA messages. */
#define ML_CHECK_DMA_Msk 2
/* Message Logger: only image related messages. */
#define ML_CHECK_IMG_Msk 4
/* Message Logger: only frame valid messages. */
#define ML_CHECK_FV_Msk 8
/* Message Logger: only event messages. */
#define ML_CHECK_EVENT_Msk 16
/* Message Logger: only image read messages. */
#define ML_CHECK_IMG_READ_Msk 32
/* Message Logger: maximum verbosity. */
#define ML_MAX_VRB_Msk 0x80

/* Contains information about an image. */
typedef struct {
  char filename[MAX_PATH];
  int width;
  int height;
  short depth;
  int size;
  short bands;  // image dimensions
  int hStart;   // horizontal start position of image
  int fillA;    // left margin fill value
  int fillB;    // right margin fill value
  int vgap;     // desired amount of vertical blanking in number of rasters
  double rate;
  int image_size;
  unsigned char* data;
} image_info_t;

/* Data used by setup_clsim_event to configure for images as frame valid goes high. */
typedef struct {
  PdvDev pdv_p;
  image_info_t* image_details;
  int sending;
  int next_setup;
  int next_load;
} event_data_t;

/* Forward Declarations */

static void usage(const char* progname, const char* errstr);
static int get_list(image_info_t** details, const char* filename, const image_info_t* defaults);
static int check_list_same_size(image_info_t* list, int nimages);
static int fill_buffer(image_info_t* img_details, uint8_t* buffer, int max_size);
static int load_next_imagefile(int current_image,
                               image_info_t* image_details,
                               int total_images,
                               int buffer_size);
static int load_next_image(PdvDev pdv_p,
                           int current_image,
                           image_info_t* image_details,
                           int total_images,
                           uint8_t** buffers,
                           int nbuffers,
                           int buffer_size);
static void setup_clsim(PdvDev pdv_p, image_info_t image);
static void setup_clsim_event(void* p);

/* Variables */

static int Rgb_4tap = 0;

/**
 * Open a number of TIFF images and send them though the simulator.
 */
int main(int argc, const char* argv[])
{
  // Set initial (default) values.
  PdvDev pdv_p = NULL;
  int unit = 0;
  int channel = 0;

  const char* image_list_file = DEFAULT_LISTFILE;  // file with list of images to send
  image_info_t* image_details = NULL;              // list of images & parameters
  int image_list_send_count = -1;  // number of images to transfer (images in list, or value of -t)
  int image_list_count;            // number of actual images in list
  uint32_t image_list_configured_count;  // number of images found in the image_list

  uint32_t loops = 1;  // times to loop through and send image list
  int loop_indefinitely = 0;

  uint32_t num_buffers = NUMBUFS;
  uint8_t** rb_buffers;  // array of buffers used for ring buffer
  int buffer_size;       // size of each buffer: width * height
  uint32_t current_image;
  uint32_t current_buffer;

  event_data_t cbinfo_clsim_event;

  int cam_depth = 0;
  int cam_height = 0;
  int cam_width = 0;

  int verbosity = ML_MIN_VRB_Msk;

  int variable_size = FALSE;

  // -1 specifies not-set, and will be changed to DEFAULT_* later.
  image_info_t defaults = {.filename = "FAKE_FILE",
                           .width = 0,
                           .height = 0,
                           .depth = -1,
                           .size = -1,
                           .bands = -1,
                           .hStart = -1,
                           .fillA = -1,
                           .fillB = -1,
                           .vgap = 0,
                           .rate = 20.0,
                           .image_size = 0,
                           .data = NULL};

  double total_time;
  int started = 0;
  int usemem = 0;
  const char* progname = argv[0];

  // Initialize the logger.
  edt_msg_init(&logger);

  // Parse user arguments.
  --argc;
  ++argv;
  while (argc && argv[0][0] == '-')
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: option 'u' requires an argument\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        unit = atoi(argv[0]);
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: option 'c' requires an argument\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        channel = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: option 'c' requires an argument\n");
        exit(1);
      }
      break;

    case 'N':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 'N' requires argument\n");
        exit(1);
      }
      num_buffers = atoi(argv[0]);
      break;

    case 'i':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 'i' requires argument\n");
        exit(1);
      }
      image_list_file = argv[0];
      break;

    case 'l':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 'l' requires argument\n");
        exit(1);
      }
      loops = atoi(argv[0]);
      break;

    case 't':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 't' requires argument\n");
        exit(1);
      }
      image_list_send_count = atoi(argv[0]);
      break;

      /*
         case 'd': // image depth
         ++argv;
         --argc;
         if (argc < 1) {
         usage(progname, "ERROR: option 'd' requires argument\n");
         exit(1);
         }
         depth = atoi(argv[0]);
         break;
       */

    case 'r':
      if (strcmp(argv[0], "-rgb4") == 0)
        Rgb_4tap = 1;
      break;

    case 'm':
      usemem = 1;
      break;

    case 'h':
      usage(progname, NULL);
      exit(0);
      break;

    case 'v':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 'v' requires argument\n");
        exit(1);
      }
      verbosity |= strtol(argv[0], 0, 0);
      break;

    case 'A':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 'A' requires argument\n");
        exit(1);
      }
      defaults.fillA = strtol(argv[0], 0, 0);
      break;

    case 'B':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "ERROR: option 'B' requires argument\n");
        exit(1);
      }
      defaults.fillB = strtol(argv[0], 0, 0);
      break;

    case '-':
      if (strcmp(argv[0], "--help") == 0)
      {
        usage(progname, NULL);
        return 0;
      }
      else
      {
        usage(progname, "ERROR: unknown option\n");
        exit(1);
      }
      break;

    default:
      usage(progname, NULL);
      exit(1);
    }
    argv++;
    argc--;
  }

  // Turn off TIFF warnings.
  TIFFSetWarningHandler(NULL);

  // Set the logger level.
  edt_msg_set_level(&logger, verbosity);

  // Open specified device for use.
  pdv_p = pdv_open_channel("pdv", unit, channel);
  if (pdv_p == NULL)
  {
    edt_msg_perror(EDTAPP_MSG_WARNING, "ERROR: Could not open EDT device");
    exit(1);
  }

  edt_set_wtimeout(pdv_p, 0);
  edt_set_rtimeout(pdv_p, 0);

  // Validate that the open device is a CLS.
  if (!pdv_is_simulator(pdv_p))
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: unit %d is not a Camera Link Simulator!\n", unit);
    exit(1);
  }

  // Flush the fifo.
  edt_flush_fifo(pdv_p);

  // Setup fake camera frame dimensions/settings.
  cam_width = pdv_get_width(pdv_p);
  cam_height = pdv_get_height(pdv_p);
  cam_depth = pdv_get_depth(pdv_p);
  if ((cam_depth == 24) && Rgb_4tap)
  {
    // Emulate 24 bit RGB with 32 bit; last 8 bits are throwaway.
    cam_depth = 32;
  }
  defaults.vgap = pdv_cls_get_vgap(pdv_p);
  edt_msg_output(lp, ML_MAX_VRB_Msk, "Vgap = %d\n", defaults.vgap);

  // Configure image data to come from host DMA.
  pdv_cls_set_datacnt(pdv_p, 0);

  // Validate number of images to send.
  image_list_configured_count = get_list(&image_details, image_list_file, &defaults);
  if (image_list_send_count < 0 || image_list_send_count > (int)image_list_configured_count)
  {
    image_list_send_count = image_list_configured_count;
  }
  if (image_list_send_count < 1)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: no images specified for sending.\n");
    exit(1);
  }
  image_list_count = image_list_send_count;
  variable_size = !check_list_same_size(image_details, image_list_configured_count);
  edt_msg_output(lp,
                 ML_MIN_VRB_Msk,
                 "%s images configured: %d\n",
                 (variable_size) ? "variable size" : "constant size",
                 image_list_configured_count);

  // Create ring buffers.
  buffer_size = cam_width * pdv_bytes_per_line(cam_height, cam_depth);
  if (buffer_size == 0)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Simulator not initialized\n");
    exit(1);
  }
  edt_configure_ring_buffers(pdv_p, buffer_size, (int)num_buffers, EDT_WRITE);
  rb_buffers = (uint8_t**)edt_buffer_addresses(pdv_p);

  // Setup looping; calculate expected images or set indefinite.
  if (loops != 0)
    image_list_send_count *= loops;
  else if (loops == 0)
    loop_indefinitely = 1;

  edt_msg_output(lp,
                 ML_MAX_VRB_Msk,
                 "width: %d height: %d buffer_size: 0x%x bytes\n",
                 cam_width,
                 cam_height,
                 buffer_size);

  // If specified, read all images into memory first.
  if (usemem)
  {
    int i;

    for (i = 0; i < image_list_count; i++)
    {
      load_next_imagefile(i, image_details, image_list_configured_count, buffer_size);
    }
  }

  rb_buffers = (uint8_t**)edt_buffer_addresses(pdv_p);

  // Fill all except the last ring buffer; get the details about the images stored in those buffers.
  for (current_image = 0;
       current_image < num_buffers - 1 && (int)current_image < image_list_send_count;
       ++current_image)
  {
    load_next_image(pdv_p,
                    current_image,
                    image_details,
                    image_list_configured_count,
                    rb_buffers,
                    (int)num_buffers,
                    buffer_size);
  }

  cbinfo_clsim_event.pdv_p = pdv_p;
  cbinfo_clsim_event.image_details = image_details;

  // Configure CLS for the first image.
  setup_clsim(pdv_p, image_details[0]);

  if (variable_size)
    edt_set_event_func(pdv_p,
                       EDT_PDV_EVENT_FVAL,
                       (EdtEventFunc)setup_clsim_event,
                       &cbinfo_clsim_event,
                       1);

  // Image transfer.
  edt_start_buffers(pdv_p, current_image);
  started = current_image;
  for (current_buffer = 0; ((int)current_buffer < (int)image_list_send_count) || loop_indefinitely;
       ++current_buffer, ++current_image)
  {
    int t1 = 0, t2 = 0;
    int load_detail;
    if (verbosity != ML_MIN_VRB_Msk)
      edt_msg_output(lp, 0xfe, "----------------------------------------------\n");

    total_time = edt_timestamp();

    cbinfo_clsim_event.sending = current_buffer;
    if ((int)current_buffer + 1 >= (int)image_list_send_count && !loop_indefinitely)
      cbinfo_clsim_event.next_setup = -1;
    else
      cbinfo_clsim_event.next_setup = current_buffer + 1;
    cbinfo_clsim_event.next_load = current_image;

    // Start sending out the current buffer.
    edt_ref_tmstamp(pdv_p, 0xabcd0001);
    edt_msg_output(lp, ML_CHECK_DMA_Msk, "Sending %d\n", current_buffer);
    if (started < (int)image_list_send_count || loop_indefinitely)
    {
      edt_start_buffers(pdv_p, 1);
      started++;
    }

    // Load the next image to be sent.
    if ((int)current_image < (int)image_list_send_count || loop_indefinitely)
    {
      load_next_image(pdv_p,
                      current_image,
                      image_details,
                      image_list_configured_count,
                      rb_buffers,
                      num_buffers,
                      buffer_size);
    }
    load_detail = current_image % image_list_configured_count;

    // Check for timeouts.
    t1 = edt_timeouts(pdv_p);
    edt_wait_for_buffers(pdv_p, 1);
    t2 = edt_timeouts(pdv_p);
    if (t1 != t2)
      edt_msg_output(lp, ML_MIN_VRB_Msk, "Apparent timeout\n");

    total_time = edt_timestamp() - total_time;

    edt_msg_output(lp,
                   ML_MIN_VRB_Msk,
                   "\rSent %6d %-20s Loaded %6d %-20s Rate %10f",
                   current_buffer,
                   image_details[current_buffer % image_list_configured_count].filename,
                   load_detail,
                   image_details[load_detail].filename,
                   image_details[load_detail].rate / 1000000.0);
  }
  printf("\n");

  free(image_details);
  pdv_close(pdv_p);
  return 0;
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param errstr A hint as to what is invalid about the call.
 */
static void usage(const char* progname, const char* errstr)
{
  if (errstr && *errstr)
    puts(errstr);
  printf("Usage: %s [options]\n", progname);
  printf("Options:\n");
  printf("-u <unit>            unit (board) number (default 0)\n");
  printf("-c <channel>         channel number (0 [default] or 1)\n");
  printf("-i <listfile>        input file with list of images and parameters.\n");
  printf("                      (Default file is '%s'). Use '-i -' to specify stdin.\n",
         DEFAULT_LISTFILE);
  printf("-l <loops>            loop through image set loops times\n");
  printf("-rgb4                emulate RGB with 4-taps for RGB (R+G+B+dummy, needed for PCI\n");
  printf("                      boards, and PCIe boards with pre-10/2012 FW)\n");
  printf("-m                   load all images into memory up front (CAUTION: must have\n");
  printf("                     adequate memory for the #images in the list!)\n");
  printf("-t <#images>         number of images to send (should be the same or\n");
  printf("                      less than the number found in image list)\n");
  /* TODO: User s/b able to override depth, so images not matching won't be sent (or included
   * in internal image list)
   * puts("-d <depth>           bits per pixel (default: depth of first image)"); */
  printf("-A <fillA, hex>      left fill value if image width < cam width (default 0x%x)\n",
         DEFAULT_FILLA);
  printf("-B <fillB, hex>      right fill value if image width < cam width (default 0x%x)\n",
         DEFAULT_FILLB);
  printf("-h                   print this help message\n");
  printf("-N <#buffers>        number of ring buffers to use (default %d)\n", NUMBUFS);
  printf("-v <mask>            set verbosity to debug output mask ... \n\
                      Verbosity can be mask of: \n\
                      ML_MIN_VRB_Msk         1 (print only the minimum) \n\
                      ML_CHECK_DMA_Msk       2 \n\
                      ML_CHECK_IMG_Msk       4 \n\
                      ML_CHECK_FV_Msk        8 \n\
                      ML_CHECK_EVENT_Msk     16 \n\
                      ML_CHECK_IMG_READ_Msk  32 \n\
                      ML_MAX_VRB_Msk         0x80 (print everything)\n");
}

/**
 * Check the specified string for comment or blank character.
 *
 * @param str The string to check.
 *
 * @return 1 (True) if all blanks or first nonblank is '#', 0 (False) otherwise.
 */
static int comment_or_blank(char* str)
{
  unsigned int i;

  for (i = 0; i < strlen(str); i++)
  {
    if (str[i] == '#')
      return 1;

    if ((str[i] != ' ') && (str[i] != '\t') && (str[i] != '\r') && (str[i] != '\n'))
      return 0;
  }
  return 1;
}

/**
 * Parse an image listing.
 *  ex: `filename.tif filla:value fillb:value hstart:value vgap:value #comment`
 *
 * @param line Character array to parse.
 *   Line parameters: hstart, vgap, fillA, fillB
 *   Line parameters are case and order insensitive.
 *   Line parameter pairs are specified as "parameter:value".
 * @param[out] img_info
 *   A structure containing the parsed information; for unspecified fields,
 *   defaults are used.
 * @param defaults
 *   Structure populated with default values to be overwritten after parse.
 */
static void process(const char* const line,
                    image_info_t* img_info,
                    const image_info_t* const defaults)
{
  const char* tok = line;
  size_t tok_len = 0;

  memcpy(img_info, defaults, sizeof(*img_info));

  tok_len = strcspn(tok, " ");
  if (tok_len >= 2)
  {
    size_t str_length;

    if (tok_len + 1 >= sizeof(img_info->filename) - 1)
    {
      fprintf(stderr,
              "ERROR: Filename too long (actual: %zu, expected <= %zu)\n",
              tok_len + 1,
              sizeof(img_info->filename) - 1);
      exit(1);
    }

    strncpy(img_info->filename, tok, tok_len + 1);
    img_info->filename[sizeof(img_info->filename) - 1] = '\0';

    str_length = strlen(img_info->filename);
    if (img_info->filename[str_length - 1] == '\n')
    {
      img_info->filename[str_length - 1] = '\0';
    }

    // Jump over the space or newline.
    tok += tok_len + 1;
    while ((tok_len = strcspn(tok, " ")) >= 2)
    {
      if (strncasecmp(tok, "fillA:", 6) == 0)
      {
        img_info->fillA = (int)strtod(tok + 6, NULL);
      }
      else if (strncasecmp(tok, "fillB:", 6) == 0)
      {
        img_info->fillB = (int)strtod(tok + 6, NULL);
      }
      else if (strncasecmp(tok, "hStart:", 7) == 0)
      {
        img_info->hStart = (int)strtod(tok + 7, NULL);
      }
      else if (strncasecmp(tok, "vgap:", 5) == 0)
      {
        img_info->vgap = (int)strtod(tok + 5, NULL);
      }
      else if (tok[0] == '#')
      {
        // Done processing.
        break;
      }

      tok += tok_len + 1;
    }
  }
  else
  {
    // Error out because the file specified can not be found.
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: line without filename: \"%s\"\n", line);
    exit(1);
  }

  return;
}

/**
 * Validate the TIFF image; ensure it meets current requirements.
 *  Image must be <= 16 bits/pixel and 1 sample/pixel.
 *  Image must also define width and height.
 *
 * @param image The image to validate.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
static int is_valid_tiff(TIFF* image)
{
  int value = 0;

  if (TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &value) == 0)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: undefined number of bits per sample\n");
    return 0;
  }
  else if (value > 16)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: unsupported number of bits per sample: %d\n", value);
    return 0;
  }

  if (TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &value) == 0)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: undefined number of samples per pixel\n");
    return 0;
  }

  else if ((value != 1) && (value != 3))
  {
    // TODO: support 3 sample images (rgb) as long as depth isn't > 8.
    edt_msg_output(lp,
                   ML_MIN_VRB_Msk,
                   "ERROR: unsupported number of samples per pixel: %d\n",
                   value);
    return 0;
  }

  if (TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &value) == 0)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Image does not define its width\n");
    return 0;
  }

  if (TIFFGetField(image, TIFFTAG_IMAGELENGTH, &value) == 0)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Image does not define its height\n");
    return 0;
  }

  return 1;
}

/**
 * Retrieve information for tiff image.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *
 * @return 0 (True) on successful retrieval.
 */
static int edt_get_tiff_info(image_info_t* img_details)
{
  TIFF* image;

  edt_msg_output(lp,
                 ML_MAX_VRB_Msk,
                 "%s(): processing image: [%s]\n",
                 __func__,
                 img_details->filename);

  // Open the image.
  if ((image = TIFFOpen(img_details->filename, "r")) == NULL)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Could not open image %s\n", img_details->filename);
    exit(42);
  }

  // Check that the image is of supported type (TIFF) and dimensions.
  if (!is_valid_tiff(image))
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Invalid image; exiting.\n");
    exit(42);
  }

  TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &img_details->width);
  TIFFGetField(image, TIFFTAG_IMAGELENGTH, &img_details->height);

  TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &img_details->depth);
  TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &img_details->bands);

  if (img_details->width == 0 || img_details->height == 0 || img_details->depth == 0 ||
      img_details->bands == 0)
  {
    printf("Error: invalid size parameter(s) -- has clsiminit been run?\n");
    exit(1);
  }

  TIFFClose(image);

  return 0;
}

/**
 * Collect list of images and their parameters from file with specified name.
 *
 * @param details An array of structures containing information about each image in the imagelist.
 *   This function allocates the array; caller is responsible for freeing it.
 * @param filename The name of the image list file.
 * @param defaults Structure populated with default values to be overwritten after collection.
 *
 * @return The number of elements in the allocated array.
 */
static int get_list(image_info_t** details,
                    const char* filename,
                    const image_info_t* const defaults)
{
  char line[2 * MAX_PATH] = {'\0'};
  int starting_size = 100; /* starting starting_size */
  int actual_size = 0;     /* size actually used */
  FILE* list_file;
  image_info_t* img_details;

  img_details = (image_info_t*)calloc(starting_size, sizeof(image_info_t));
  if (img_details == NULL)
  {
    edt_msg_output(lp, EDTAPP_MSG_FATAL, "ERROR: calloc() failed in get_list()\n");
    exit(1);
  }

  if (strcmp(filename, "-") == 0)
  {
    list_file = stdin;
  }
  else
  {
    list_file = fopen(filename, "r");
    if (list_file == NULL)
    {
      edt_msg_output_printf_perror(lp, EDTAPP_MSG_FATAL, "Error opening file '%s'", filename);
      return 0;
    }
  }

  while (fgets(line, sizeof(line), list_file) != NULL)
  {
    if (actual_size == starting_size)
    {
      starting_size *= 2;
      img_details = (image_info_t*)realloc(img_details, starting_size * sizeof(image_info_t));
      if (img_details == NULL)
      {
        edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: realloc() failed in get_list()\n");
        exit(1);
      }
    }
    if (!comment_or_blank(line))
    {
      process(line, &img_details[actual_size], defaults);

      /* Note that if fillA or fillB is negative, the left
       * or right pixel value from the first line of the image
       * will be used (which is done in the function that opens the image) */
      if (img_details[actual_size].hStart < 0)
        img_details[actual_size].hStart = DEFAULT_HSTART;
      if (img_details[actual_size].vgap < 0)
        img_details[actual_size].vgap = DEFAULT_VGAP;
      img_details[actual_size].data = NULL;
      img_details[actual_size].size = 0;

      edt_get_tiff_info(&img_details[actual_size]);

      ++actual_size;
    }
  }

  if (list_file != stdin)
  {
    fclose(list_file);
  }

  *details = img_details;
  return actual_size;
}

/**
 * Verify that an image information list is
 * (a) correctly sized to the expected image count and
 * (b) all of the images within the list have the same dimensions.
 *
 * @param list The list of images and their parameters.
 * @param nimages The expected image count.
 *
 * @return 1 (True) if the sizes match, 0 (False) otherwise.
 */
static int check_list_same_size(image_info_t* list, int nimages)
{
  int i = 0;
  int w = 0;
  int h = 0;
  int d = 0;

  if (nimages == 0)
    return 0;

  w = list->width;
  h = list->height;
  d = list->depth;

  for (i = 0; i < nimages; i++)
  {
    if (list[i].width != w || list[i].height != h || list[i].depth != d)
      return 0;
  }

  return 1;
}

/**
 * Stores the image found in img_details->filename to the specified buffer.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *   If fill* are set to -1, replace with values from first image.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer.
 */
static int fill_buffer(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  TIFF* image;

  uint32_t width = 0, height = 0, depth = 0, bands = 0;
  tsize_t imageOffset = 0;
  tsize_t result = 0;
  int stripMax = 0, stripCount = 0;
  tmsize_t stripSize = 0;

  unsigned long image_size; /* in bytes */

  edt_msg_output(lp,
                 ML_MAX_VRB_Msk,
                 "fill_buffer(): processing image: [%s]\n",
                 img_details->filename);

  // Open the image.
  if ((image = TIFFOpen(img_details->filename, "r")) == NULL)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Could not open image %s\n", img_details->filename);
    exit(42);
  }

  // Check that the image is of supported type (TIFF) and dimensions.
  if (!is_valid_tiff(image))
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "ERROR: Invalid image; exiting.\n");
    exit(42);
  }

  TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetField(image, TIFFTAG_IMAGELENGTH, &height);
  TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &depth);
  TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &bands);

  if (width == 0 || height == 0 || depth == 0 || bands == 0)
  {
    printf("Error: invalid size parameter(s) -- has clsiminit been run?\n");
    exit(1);
  }

  // Display image information.
  edt_msg_output(lp,
                 ML_MAX_VRB_Msk,
                 "fill_buffer(): Image: %s Width: %d Height: %d Depth: %d Bands %d\n",
                 img_details->filename,
                 width,
                 height,
                 depth,
                 bands);

  // CLS doesn't do RGB; emulate by sending 4 channels with 1 throwaway.
  if ((bands == 3) && Rgb_4tap)
    image_size = width * height * pdv_bits_to_bytes(depth) * (bands + 1);
  else
    image_size = width * height * pdv_bits_to_bytes(depth) * bands;

  edt_msg_output(lp,
                 ML_MAX_VRB_Msk,
                 "fill_buffer(): Width*Height*depth_bytes*channels=%lu (0x%lx)\n",
                 image_size,
                 image_size);

  // Read in the possibly multiple strips.
  stripMax = TIFFNumberOfStrips(image);
  stripSize = TIFFStripSize(image);

  imageOffset = 0;

  if (image_size > (uint32_t)max_size)
  {
    edt_msg_output(lp,
                   ML_MIN_VRB_Msk,
                   "ERROR: Image size (0x%lx bytes) is larger "
                   "than buffer size (0x%x bytes)\n",
                   image_size,
                   max_size);
    exit(42);
  }

  // CLS doesn't do RGB, emulate by sending 4 bytes with 1 throwaway (inserts 0 every 4th byte).
  if ((bands == 3) && Rgb_4tap)
  {
    tsize_t i;
    uint8_t* bp = buffer;
    uint8_t* stripbuf = (uint8_t*)malloc(stripSize);

    for (stripCount = 0; stripCount < stripMax; stripCount++)
    {
      if ((result = TIFFReadEncodedStrip(image, stripCount, stripbuf, (tsize_t)-1)) == (tsize_t)-1)
      {
        edt_msg_output(lp,
                       ML_MIN_VRB_Msk,
                       "ERROR: Read error on input strip number %d\n",
                       stripCount);
        exit(42);
      }

      for (i = 0; i < result; i += 3)
      {
        *bp++ = stripbuf[i];
        *bp++ = stripbuf[i + 1];
        *bp++ = stripbuf[i + 2];
        *bp++ = 0;
      }
    }
    free(stripbuf);
  }
  else
  {
    for (stripCount = 0; stripCount < stripMax; stripCount++)
    {
      if ((result = TIFFReadEncodedStrip(image, stripCount, buffer + imageOffset, (tsize_t)-1)) ==
          -1)
      {
        edt_msg_output(lp,
                       ML_MIN_VRB_Msk,
                       "ERROR: Read error on input strip number %d\n",
                       stripCount);
        exit(42);
      }

      imageOffset += result;
    }
  }

  img_details->width = width;
  img_details->height = height;

  if (img_details->fillA == -1)
    img_details->fillA = buffer[0];

  if (img_details->fillB == -1)
    img_details->fillB = buffer[width - 1];

  TIFFClose(image);

  return image_size;
}

/**
 * Allocate for and load current_image into local memory buffer (not ring buffer).
 *
 * @param current_image The index of the image to load into local buffer.
 * @param img_details List of structures containing image details.
 * @param total_images The number of image_info_t's in  image_details.
 * @param buffer_size The size of the local buffer.
 *
 * @return The size of the image loaded to the local buffer.
 */
static int load_next_imagefile(int current_image,
                               image_info_t* image_details,
                               int total_images,
                               int buffer_size)
{
  edt_msg_output(lp,
                 ML_CHECK_IMG_Msk,
                 "Loading image #%d of %d (%s)\n",
                 current_image + 1,
                 total_images,
                 image_details[current_image].filename);

  if ((image_details[current_image].data = (unsigned char*)edt_alloc(buffer_size)) == NULL)
  {
    printf("Error: could not allocate buffer (possible memory overflow)\n");
    exit(1);
  }

  image_details[current_image].size =
      fill_buffer(&image_details[current_image], image_details[current_image].data, buffer_size);

  return image_details[current_image].size;
}

/**
 * Load the current_image into the appropriate ring-buffer.
 *
 * If current_image is larger than total_images, it wraps around.
 *
 * @param pdv_p Valid EDT device pointer to the CLS.
 * @param current_image The index of the image to load into local buffer.
 * @param img_details List of structures containing image details.
 * @param total_images The number of image_info_t's in  image_details.
 * @param buffers The array of ring buffers in use.
 * @param nbuffers The number of available buffers.
 * @param buffer_size The size of the local buffer.
 *
 * @return The size of the image loaded to the ring buffer.
 */
static int load_next_image(PdvDev pdv_p,
                           int current_image,
                           image_info_t* image_details,
                           int total_images,
                           uint8_t** buffers,
                           int nbuffers,
                           int buffer_size)
{
  int image_size;
  double t2;
  int current_buffer = current_image % nbuffers;
  current_image = current_image % total_images;

  edt_msg_output(lp,
                 ML_CHECK_IMG_Msk,
                 "Filling buffer %d with image %d of %d (%s)\n",
                 current_buffer,
                 current_image + 1,
                 total_images,
                 image_details[current_image].filename);

  edt_dtime();

  if (image_details[current_image].size)
  {
    memcpy(buffers[current_buffer],
           image_details[current_image].data,
           image_details[current_image].size);
    image_size = image_details[current_image].size;
  }
  else
  {
    image_size = fill_buffer(&image_details[current_image], buffers[current_buffer], buffer_size);
  }

  t2 = edt_dtime();

  image_details[current_image].rate = ((double)buffer_size) / t2;

  // Resize buffer to correct size according to image.
  edt_msg_output(lp,
                 ML_CHECK_IMG_Msk,
                 "load_next_image(): Setting buffer %d to size %d\n",
                 current_buffer,
                 image_size);
  edt_set_buffer_size(pdv_p, current_buffer, image_size, EDT_WRITE);

  return image_size;
}

/**
 * Set up configuration registers on simulator.
 *
 * The following will be set for each image:
 *  FillA: left margin fill
 *  FillB: right margin fill
 *  vgap: number rasterlines of vertical blanking (VcntMax - Vactv)
 *  hStart: horizontal start position of image within raster (-1 is centered between margins)
 *  height of image
 *
 * @param pdv_p Valid EDT device pointer to the CLS.
 * @param image Structure containing information about the image to be transferred.
 */
static void setup_clsim(PdvDev pdv_p, image_info_t image)
{
  int rvStart, rvEnd;
  int max_width;

  int clocks;

  int taps = (edt_reg_read(pdv_p, PDV_CL_DATA_PATH) >> 4) + 1;
  int num_bytes = ((edt_reg_read(pdv_p, PDV_CL_DATA_PATH) & 15) + 7) / 8;

  int hlvstart = edt_reg_read(pdv_p, PDV_CLSIM_HLVSTART);
  int hlvend = edt_reg_read(pdv_p, PDV_CLSIM_HLVEND);

  int depth_bytes = pdv_bits_to_bytes(pdv_get_depth(pdv_p));

  // CLS doesn't do RGB, emulate by sending 4 bytes with 1 throwaway.
  if (taps == 3)
    taps = 4;

  /*
   * 1 tap = 1 byte / clock.
   * width is in pixels
   * depth_bytes is in bytes / pixel
   *
   * So, (width * depth_bytes) / taps
   *
   *   = (width * depth_bytes) * 1/taps
   *
   *               bytes     clocks
   *   =  pixels * -----  *  ------  = clocks
   *               pixel     byte
   */

  if (depth_bytes == 3)
    depth_bytes = 4;

  clocks = (image.width * depth_bytes) / (taps * num_bytes);

  max_width = hlvend - hlvstart;

  edt_msg_output(lp, ML_CHECK_IMG_Msk, "setup_clsim(): file: %s ", image.filename);
  edt_msg_output(lp, ML_CHECK_IMG_Msk, "height: %d width: %d ", image.height, image.width);
  edt_msg_output(lp, ML_CHECK_IMG_Msk, "vgap: %d ", image.vgap);
  edt_msg_output(lp, ML_CHECK_IMG_Msk, "filla: %d fillb: %d ", image.fillA, image.fillB);
  edt_msg_output(lp, ML_CHECK_IMG_Msk, "hStart: %d\n", image.hStart);

  edt_msg_output(lp, ML_CHECK_IMG_Msk, "setup_clsim(): max_width: %d\n", max_width);

  if (image.hStart == -1)
  {
    // Adjust margins so that image is centered between fill values.
    rvStart = (max_width / 2) - (clocks / 2) + hlvstart;
    rvEnd = rvStart + clocks;
  }
  else
  {
    // Use supplied settings.
    rvStart = image.hStart;
    rvEnd = rvStart + clocks;
  }

  edt_msg_output(lp, ML_CHECK_IMG_Msk, "setup_clsim(): rvStart: %d rvEnd: %d\n", rvStart, rvEnd);

  pdv_cls_set_fill(pdv_p, (uint8_t)image.fillA, (uint8_t)image.fillB);
  pdv_cls_set_height(pdv_p, image.height, image.vgap);
  pdv_cls_set_rven(pdv_p, 1);
  pdv_cls_set_readvalid(pdv_p, (uint16_t)rvStart, (uint16_t)rvEnd);
}

/**
 * Configure the CLS.
 *
 * @param p Pointer containing event data.
 */
static void setup_clsim_event(void* p)
{
  event_data_t* data = (event_data_t*)p;
  static int count = 0;

  static double last_time = 0.0;
  double t = edt_timestamp();
  double dt = t - last_time;
  last_time = t;

  if (!data)
  {
    edt_msg_output(lp, ML_MIN_VRB_Msk, "%s(): Event %4d, NULL data pointer!\n", __func__, count);
    return;
  }

  edt_msg_output(lp,
                 ML_CHECK_EVENT_Msk,
                 "Event %4d : setting up %4d width = %5d height = %5d dt = %10.9f \n",
                 count,
                 data->next_setup,
                 data->image_details[data->next_setup].width,
                 data->image_details[data->next_setup].height,
                 dt);

  count++;
  if (data->next_setup >= 0)
    setup_clsim(data->pdv_p, data->image_details[data->next_setup]);
  else
    edt_msg_output(lp, ML_CHECK_EVENT_Msk, "Not setting up at end of list\n");

  return;
}
