/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2017 Engineering Design Team, Inc.
 *
 * A utility / example program to send commands to serial camera over AIA serial lines.
 *
 * Requirements:
 *   Camera in use must have serial capabilities.
 */

#include "edtinc.h"

/* Size of the buffer to use for serial communication. */
#define SERBUFSIZE 512

/* Forward Declarations */

static void usage(const char* progname, const char* err);
static int PdvSerialWriteRead(int argc,
                              const char* argv[],
                              int unit,
                              int channel,
                              int hexin,
                              int do_wait,
                              int do_printflush,
                              int inter,
                              int timeout,
                              int verbose,
                              int baud,
                              int readonly);
static void strip_newline(char* const s);
static void print_ascii_string(char* buf);
static void printify(char* buf, char* dest, size_t n, int show_nonprint);

/* Variables */

static int basler_framing = 0;
static int duncan_framing = 0;

/**
 * Send serial data to a connected device.
 */
int main(int argc, const char* argv[])
{
  int status;  // Return status from command.
  int unit = 0;
  int hexin = 0;
  int inter = 0;
  int verbose = 0;
  int do_wait = 0;
  int do_printflush = 0;
  int timeout = 0;
  int channel = 0;
  int readonly = 0;
  int baud = 0;
  const char* progname = argv[0];

  --argc;
  ++argv;
  while ((argc > 0) && ((argv[0][0] == '-') || (argv[0][0] == '/')))
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: option 'u' requires a numeric argument\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        unit = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: option 'u' requires a numeric argument\n");
        exit(1);
      }
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error - channel # expected\n");
        exit(-1);
      }
      channel = atoi(argv[0]);
      break;

    case 'x':
      hexin = 1;
      break;

    case 'w':
      do_wait = 1;
      break;

    case 'p':
      do_printflush = 1;
      break;

    case 't':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: option 't' requires argument\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        timeout = atoi(argv[0]);
      break;

    case 'b':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: option 'b' requires argument\n");
        exit(1);
      }
      baud = atoi(argv[0]);
      break;

    case 'v':
      verbose = 1;
      break;

    case 'r':
      readonly = 1;
      break;

    case 'B':
      basler_framing = 1;
      break;

    case 'D':
      duncan_framing = 1;
      break;

    case 'h':
    case '?':
      usage(progname, "");
      exit(0);
      break;

    case '-':
      if (strcmp(argv[0], "--help") == 0)
      {
        usage(progname, "");
        exit(0);
      }
      else
      {
        fprintf(stderr, "unknown option: %s\n", argv[0]);
        usage(progname, "");
        exit(1);
      }
      break;

    default:
      usage(progname, "Error: unknown option\n");
      exit(1);
    }
    argc--;
    argv++;
  }

  if ((argc < 1) && !readonly && !do_printflush)
    inter = 1;

  status = PdvSerialWriteRead(argc,
                              argv,
                              unit,
                              channel,
                              hexin,
                              do_wait,
                              do_printflush,
                              inter,
                              timeout,
                              verbose,
                              baud,
                              readonly);

  return status;
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param err A hint as to what is invalid about the call.
 */
static void usage(const char* progname, const char* err)
{
  printf("%s", err);
  printf("Usage:  %s [options] \"CMD\"\n", progname);
  printf(" Where CMD is the command(s) to send; ASCII (default) or hex.\n");
  printf("-? | /? | -h | --help  print this help message\n");
  printf("-u <unit>      unit (board) number (default 0)\n");
  printf("-c <channel>   channel number (0 [default] or 1)\n");
  printf("-b <baudrate>  baudrate to use\n");
  printf("-v             be verbose\n");
  printf("-w             wait before read\n");
  printf("-r             read only\n");
  printf("-p             print what's in the buffer on start\n");
  printf("-t <0-9>       serial timeout override value (0 through 9)\n");
  printf("-B             Basler framing mode; CMD should be hex bytes separated by spaces\n");
  printf("                do not specify framing info; STX and checksum added automatically\n");
  printf("-D             Duncan framing mode; CMD should be hex bytes separated by spaces\n");
  printf("                do not specify framing info; STX and checksum added automatically\n");
  printf("-x             hex mode; CMD should be hex bytes, separated by spaces\n");
}

/**
 * Read or write to the connected devices serial port.
 *
 * @param argc The argument count from main().
 * @param argv The argument list from main().
 * @param unit Unit number of device to interact with.
 * @param channel Channel number of device to interact with.
 * @param hexin Is input hexadecimal; 1 for True, 0 for False.
 * @param do_wait Wait before reading; 1 for True, 0 for False.
 * @param do_printflush Print initial contents of buffer; 1 for True, 0 for False.
 * @param inter Use in interactive mode; 1 for True, 0 for False.
 * @param timeout Milliseconds to wait before serial timeout occurs.
 * @param verbose Be verbose; 1 for True, 0 for False.
 * @param baud The baudrate to use for serial interaction.
 * @param readonly Execute as read only; 1 for True, 0 for False.
 *
 * @return 0 or 1 on success, -1 on error.
 */
static int PdvSerialWriteRead(int argc,
                              const char* argv[],
                              int unit,
                              int channel,
                              int hexin,
                              int do_wait,
                              int do_printflush,
                              int inter,
                              int timeout,
                              int verbose,
                              int baud,
                              int readonly)
{
  int i = 0;
  int ret = 0;
  int nbytes = 0;
  int length = 0;
  uint8_t hbuf[SERBUFSIZE] = {'\0'};
  char tmpbuf[SERBUFSIZE + 4] = {'\0'};
  char getbuf[SERBUFSIZE + 1] = {'\0'};
  char* ibuf_p = NULL;
  uint8_t lastbyte = 0, waitc = 0;
  char buf[SERBUFSIZE + 1] = {'\0'};
  char bs[32][3];
  EdtDev ed = NULL;

  (void)argc;

  // Open a handle to the device.
  ed = pdv_open_channel(EDT_INTERFACE, unit, channel);
  if (ed == NULL)
  {
    edt_msg_perror(EDTAPP_MSG_WARNING, EDT_INTERFACE);
    return -1;
  }

  pdv_serial_read_enable(ed);

  // Serial_timeout can be specified within the config file or via the '-t' override flag. The CLI
  // flag will take precident. If 'readonly' is set but 'timeout' is unspecified, then the value is
  // set to 60000. Otherwise use the camera config file's value, which defaults to 1000.
  if (readonly && (timeout == 0))
    timeout = 60000;
  else if (timeout < 1)
    timeout = pdv_serial_get_timeout(ed);

  if (verbose)
    printf("serial timeout %d\n", timeout);

  if (inter)
  {
    ibuf_p = getbuf;
    printf("\nEnter command (Ctrl-C to quit)\n\n");
  }
  else
  {
    // Set to something non-null, to allow entrance into the while loop below.
    snprintf(getbuf, sizeof(getbuf), "%s\n", argv[0]);
    ibuf_p = getbuf;
  }

  if (baud)
  {
    pdv_serial_set_baud(ed, baud);
  }

  if ((pdv_serial_read(ed, buf, SERBUFSIZE) > 0) && do_printflush)
  {
    printf("%s\n", buf);
    exit(0);
  }

  do
  {
    if (inter)
    {
      printf("> ");
      if (strlen(fgets(ibuf_p, SERBUFSIZE, stdin)) > SERBUFSIZE - 1)
      {
        printf("Max command string length (%d) exceeded\n", SERBUFSIZE - 1);
        continue;
      }
    }

    if (!readonly)
    {
      if (hexin)
      {
        unsigned int val;
        i = 0;

        strip_newline(ibuf_p);

        while (*ibuf_p)
        {
          while ((*ibuf_p == ' ') || (*ibuf_p == '\t'))
            ++ibuf_p;

          if (*ibuf_p == '\0')
            break;

          if (sscanf(ibuf_p, "%x", &val) != 1)
          {
            printf("error reading input byte %d\n", i);
            i = 0;
            break;
          }

          if (val > 0xff)
          {
            printf("Hex string format error -- expect hex bytes separated by spaces, e.g. '00 a0 "
                   "ff ...' \n");
            i = 0;
            break;
          }
          else
            hbuf[i++] = (unsigned char)val;

          while ((*ibuf_p != ' ') && (*ibuf_p != '\t') && (*ibuf_p != '\0'))
            ++ibuf_p;
        }

        // Use pdv_serial_binary_command, because the alternative `pdv_serial_write` prepends 'c' if
        // using a FOI (fiber optic interface).
        if (i)
          pdv_serial_binary_command(ed, hbuf, i);
      }
      else if (duncan_framing)
      {
        nbytes = sscanf(ibuf_p,
                        "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
                        bs[0],
                        bs[1],
                        bs[2],
                        bs[3],
                        bs[4],
                        bs[5],
                        bs[6],
                        bs[7],
                        bs[8],
                        bs[9],
                        bs[10],
                        bs[11],
                        bs[12],
                        bs[13],
                        bs[14],
                        bs[15]);

        for (i = 0; i < nbytes; i++)
        {
          if (strlen(bs[i]) > 2)
          {
            printf("duncan command format error\n");
            break;
          }
          hbuf[i] = (strtoul(bs[i], NULL, 16) & 0xff);
        }

        pdv_serial_write_duncan(ed, hbuf, nbytes);
        // edt_msleep(10000);
      }
      else if (basler_framing)
      {
        nbytes = sscanf(ibuf_p,
                        "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
                        bs[0],
                        bs[1],
                        bs[2],
                        bs[3],
                        bs[4],
                        bs[5],
                        bs[6],
                        bs[7],
                        bs[8],
                        bs[9],
                        bs[10],
                        bs[11],
                        bs[12],
                        bs[13],
                        bs[14],
                        bs[15]);

        for (i = 0; i < nbytes; i++)
        {
          if (strlen(bs[i]) > 2)
          {
            printf("basler command format error\n");
            break;
          }
          hbuf[i] = (strtoul(bs[i], NULL, 16) & 0xff);
        }

        pdv_serial_write_basler(ed, hbuf, nbytes);
        // edt_msleep(10000);
      }
      else
      {
        snprintf(tmpbuf, sizeof(tmpbuf) - 1, "%s\r", ibuf_p);
        if (verbose)
          printf("writing <%s>\n", ibuf_p);
        pdv_serial_command(ed, tmpbuf);
      }
    }

    if (duncan_framing)
    {
      ret = pdv_serial_read_duncan(ed, (uint8_t*)buf);
      printf("resp <");
      for (i = 0; i < ret; i++)
        printf("%02x%s", buf[i], (i == ret - 1) ? ">\n" : " ");
    }
    else
    {
      pdv_serial_wait(ed, timeout, 64);

      if (do_wait)
      {
        printf("return to read response: ");
        getchar();
      }

      // Read and print the serial response; looping until either 'waitchar' is received, or no
      // more characters arrive. Format output as appropriate for ASCII, HEX, or Pulnix STX/ETX
      // format.
      do
      {
        ret = pdv_serial_read(ed, buf, SERBUFSIZE);
        if (verbose)
          printf("read returned %d\n", ret);

        if (*buf)
          lastbyte = buf[strlen(buf) - 1];

        if (ret != 0)
        {
          buf[ret + 1] = 0;
          if (hexin || basler_framing || duncan_framing)
          {
            int j;
            printf("resp <");
            for (j = 0; j < ret; j++)
              printf("%s%02x", j ? " " : "", buf[j]);
            printf(">\n");
          }
          else
          {
            // Simple ASCII.
            print_ascii_string(buf);
          }
          length += ret;
        }

        if (pdv_serial_get_waitchar(ed, &waitc) && (lastbyte == waitc))
          ret = 0;  // Jump out if 'waitchar' is enabled or received.
        else
          ret = pdv_serial_wait(ed, 500, 64);
      } while (ret > 0);
    }

    printf("\n");

  } while (inter);

  pdv_close(ed);

  return (1);
}

/**
 * Remove newline characters (`\n` or `\r`) from a provided string.
 * If found the character is replaced with NULL.
 *
 * @param s The string to strip.
 */
static void strip_newline(char* const s)
{
  int len = (int)strlen(s);
  if (len && ((s[len - 1] == '\n') || (s[len - 1] == '\r')))
    s[len - 1] = '\0';
}

/**
 * Pretty print an ASCII string; replacing `\r` or [Ctrl]-[M] with `\n` for consistency.
 *
 * @param buf The buffer to print.
 */
static void print_ascii_string(char* buf)
{
  char tmpbuf[SERBUFSIZE];

  printify(buf, tmpbuf, strlen(buf), 1);

  fputs(tmpbuf, stdout);
  fflush(stdout);
}

/**
 * Display non-printable characters; ACK, NAK, etc.
 *
 * @param buf The input buffer to display.
 * @param dest A buffer to hold the result (including typically non-printable chars).
 * @param n The size of the provided buffer.
 * @param show_noprint Flag to display (1) or ignore (0) typically non-printable characters.
 */
static void printify(char* buf, char* dest, size_t n, int show_nonprint)
{
  int has_lf = FALSE;
  size_t i = 0;
  size_t j = 0;

  for (i = 0; i < n; i++)
  {
    if (buf[i] == 10)
      has_lf = TRUE;
  }

  for (i = 0, j = 0; i < n; i++)
  {
    if (buf[i] == 13)
    {
      if (!has_lf)
        dest[j++] = '\n';
    }
    else if (buf[i] == 10 || (buf[i] >= ' ' && buf[i] < 127))
      dest[j++] = buf[i];
    else if (show_nonprint)
    {
      switch (buf[i])
      {
      case 0x06:
        strcpy((char*)dest + j, "<ACK>");
        break;
      case 0x15:
        strcpy((char*)dest + j, "<NAK>");
        break;
      case 0x02:
        strcpy((char*)dest + j, "<STX>");
        break;
      case 0x03:
        strcpy((char*)dest + j, "<ETX>");
        break;
      default:
        sprintf((char*)dest + j, "<%02x>", buf[i]);
      }

      j = strlen((char*)dest);
    }

    dest[j] = 0;
  }
}
