/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * A utility / example program to send multiple TIFF format images through the EDT Camera
 * Link Simulator (CLS) to a Camera Link frame grabber, such as the EDT VisionLink F4.
 *
 * For a more complex application (including variable width images, left/right fill, etc)
 * with the same purpose see: send_tiffs.c
 *
 * See the EDT CLS Simulator users guide for detailed information on this application.
 *
 * Prerequisites:
 *   Setup the CLS via the `clsiminit` application.
 */

#include "edtinc.h"
#include "clsim_lib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>

#ifdef __linux__
#include <strings.h> /* strcasecmp() */
#endif

/* File containing a list of images to use. */
#define DEFAULT_LISTFILE "imagelist.txt"

/* Contains available image formats. */
typedef enum { fmt_raw, fmt_ras, fmt_gif, fmt_jpg, fmt_tif, fmt_bmp, fmt_unk } ImgFmt;

/* Contains information about each image stored in imagelist. */
/** Information about an image. */
typedef struct {
  /** Filesystem path where the image was read from */
  char* pathname;
  /** Width (pixels) */
  int width;
  /** Height (pixels) */
  int height;
  /** Bit depth (bits) */
  short depth;
  /** Size of the image data buffer (bytes) */
  int size;
  /** Number of (tiff) image bands or strips */
  short bands;
  int image_size;
  /** Detected image format */
  ImgFmt format;
  /** Pointer to the image data buffer. */
  unsigned char* data;
} image_info_t;

/* Function Declarations */

void usage(const char* progname, const char* errstr);
int get_image_info_file(const char* fname, image_info_t** imagelist);
ImgFmt check_extension(const char* fname);
int get_images_info_dir(const char* dirpath, image_info_t** imagelist);
int get_image_info(ImgFmt fmt, image_info_t* image_p);
int get_tif_info(image_info_t* img_details);
int is_valid_tiff(TIFF* image);
int get_gif_info(image_info_t* img_details);
int get_ras_info(image_info_t* img_details);
int get_raw_info(image_info_t* img_details);
int get_bmp_info(image_info_t* img_details);
int get_jpg_info(image_info_t* img_details);
int get_images_info_list(const char* listpath, image_info_t** imagelist);
char* strip_nl(char* s);
int comment_or_blank(const char* str);
void error_exit(const char* errstr);
void free_image_info_list(image_info_t* imagelist, const size_t cnt);
int check_image_sizes(image_info_t* list, unsigned int nimages);
int load_next_image(PdvDev pdv_p,
                    unsigned int current_image,
                    image_info_t* image_details,
                    unsigned int total_images,
                    uint8_t** buffers,
                    unsigned int nbuffers,
                    unsigned int buffer_size);
long unsigned int fill_buffer_tif(image_info_t* img_details, uint8_t* buffer, int max_size);
int fill_buffer_gif(image_info_t* img_details, uint8_t* buffer, int max_size);
int fill_buffer_bmp(image_info_t* img_details, uint8_t* buffer, int max_size);
int fill_buffer_jpg(image_info_t* img_details, uint8_t* buffer, int max_size);
int fill_buffer_ras(image_info_t* img_details, uint8_t* buffer, int max_size);
int fill_buffer_raw(image_info_t* img_details, uint8_t* buffer, int max_size);
unsigned int edt_min(unsigned int l, unsigned int r);

/* Variables */

static int Rgb_4tap = 0;

/**
 * Open a number of TIFF images and send them though the simulator.
 */
int main(int argc, const char* argv[])
{
  // Set initial (default values).
  int unit = 0, channel = 0;     // default simulator device
  unsigned int loops = 1;        // number of times to loop through and send images
  int loop_indefinitely = 0;     // loop forever; set true if loops argument is 0
  unsigned int num_buffers = 4;  // number of ring buffers to use
  unsigned int num_images = 0;   // number of images to transfer (count of image directory contents)
  unsigned int total_images = 0;  // number of images to transfer * number of loops
  const char* imagedir = NULL;    // pathname of directory from which to read and transfer images
  const char* imagelist = NULL;   // pathname of list from which to read and transfer images
  const char* imagefile = NULL;   // pathname of single image file if desired
  unsigned int buffer_size = 0;   // size of each buffer: width * height
  unsigned int current_image = 0;
  unsigned int current_buffer = 0;
  int exitval = 0;
  int cam_width = 0, cam_height = 0, cam_depth = 0;
  PdvDev pdv_p;                        // pointer to simulator device structure
  uint8_t** buffers = NULL;            // array of buffers used for ring buffer
  image_info_t* image_details = NULL;  // list of images and their parameters

  // // -1 specifies not-set, and will be changed to DEFAULT_* later
  // image_info_t defaults = {
  //     "FAKE_FILE", /* pathname */
  //     0,           /* width */
  //     0,           /* height */
  //     -1,          /* depth */
  //     -1,          /* size */
  //     -1,          /* bands */
  //     0,           /* image_size */
  //     fmt_unk,     /* format */
  //     NULL         /* data */
  // };

  unsigned int started = 0;
  int usemem = 0;
  const char* progname = argv[0];

  // Parse user arguments.
  --argc;
  ++argv;
  while (argc && argv[0][0] == '-')
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'u' requires an argument.\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        unit = atoi(argv[0]);
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'c' requires an argument.\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        channel = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'c' requires an argument.\n");
        exit(1);
      }
      break;

    case 'N':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'N' requires an argument.\n");
        exit(1);
      }
      {
        unsigned long tmp_long = strtoul(argv[0], NULL, 0);
        if ((tmp_long < 1) || (tmp_long > UINT_MAX))
        {
          printf("ERROR: 'N' value range error.\n");
          exit(1);
        }
        num_buffers = (unsigned int)tmp_long;
      }
      break;

    case 'd':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'd' requires an argument.\n");
        exit(1);
      }
      imagedir = argv[0];
      break;

    case 'i':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'i' requires an argument.\n");
        exit(1);
      }
      imagelist = argv[0];
      break;

    case 'f':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'f' requires an argument.\n");
        exit(1);
      }
      imagefile = argv[0];
      break;

    case 'l':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'l' requires an argument.\n");
        exit(1);
      }
      {
        unsigned long tmp_long = strtoul(argv[0], NULL, 0);
        if (tmp_long > UINT_MAX)
        {
          printf("ERROR: 'l' value range error.\n");
          exit(1);
        }
        loops = (unsigned int)tmp_long;
      }
      break;

    case 'm':
      usemem = 1;
      break;

    case 'r':
      if (strcmp(argv[0], "-rgb4") == 0)
        Rgb_4tap = 1;
      break;

    case 'h':
      usage(progname, NULL);
      exit(0);
      break;

    case '-':
      if (strcmp(argv[0], "--help") == 0)
      {
        usage(progname, NULL);
        return 0;
      }
      else
      {
        printf("ERROR: Unknown option \"%s\".", argv[0]);
        usage(progname, NULL);
        exit(1);
      }
      break;

    default:
      usage(progname, NULL);
      exit(1);
    }
    argv++;
    argc--;
  }

  // If neither "d" nor "i" are specified the images directory might be the last argument (for
  // historical reasons), otherwise (if there is no final argument) use the default filename.
  if (!(imagedir || imagelist || imagefile))
  {
    if (argc)
      imagedir = argv[0];
    else
      imagelist = DEFAULT_LISTFILE;
  }

  if (imagefile)
    exitval = get_image_info_file(imagefile, &image_details);
  else if (imagedir)
    exitval = get_images_info_dir(imagedir, &image_details);
  else if (imagelist)
    exitval = get_images_info_list(imagelist, &image_details);

  if (exitval < 1)
    error_exit("Error: no valid images found, check list or directory pathnames.");

  num_images = (unsigned int)exitval;

  if ((pdv_p = pdv_open_channel("pdv", unit, channel)) == NULL)
  {
    char errmsg[256] = {'\0'};
    snprintf(errmsg,
             sizeof(errmsg),
             "Error: Could not open EDT unit %d channel %d.",
             unit,
             channel);
    edt_msg_perror(EDTAPP_MSG_WARNING, errmsg);

    free_image_info_list(image_details, num_images);
    exit(1);
  }

  // Validate that the open device is a CLS.
  if (!pdv_is_simulator(pdv_p))
  {
    free_image_info_list(image_details, num_images);
    pdv_close(pdv_p);
    error_exit("Error: specified pdv unit is not a Camera Link Simulator!");
  }

  cam_width = pdv_get_width(pdv_p);
  cam_height = pdv_get_height(pdv_p);
  cam_depth = pdv_get_depth(pdv_p);

  edt_set_wtimeout(pdv_p, 0);
  edt_set_rtimeout(pdv_p, 0);

  pdv_flush_fifo(pdv_p);

  // Emulate 24 bit RGB with 32 bit; last 8 bits are throwaway.
  if ((cam_depth == 24) && Rgb_4tap)
    cam_depth = 32;

  // Determine ring buffer size.
  if ((buffer_size = (unsigned)(cam_height * pdv_bytes_per_line(cam_width, cam_depth))) == 0)
  {
    printf("Error: Simulator not initialized. Use clsiminit or equivalent to initialize the "
           "simulator before running %s.\n",
           progname);
    free_image_info_list(image_details, num_images);
    pdv_close(pdv_p);
    exit(1);
  }

  if (!check_image_sizes(image_details, num_images))
  {
    free_image_info_list(image_details, num_images);
    pdv_close(pdv_p);
    error_exit("Error: image / configuration size mismatch!");
  }

  // Configure image data to come from host DMA.
  pdv_cls_set_datacnt(pdv_p, 0);

  // Setup looping; calculate expected images or set indefinite.
  if (loops != 0)
  {
    // Run through imagelist 'loops' times.
    total_images = num_images * loops;
  }
  else
  {
    // Loop forever.
    loop_indefinitely = 1;
    total_images = num_images;
  }

  printf("width: %d height: %d buffer_size: 0x%x bytes\n", cam_width, cam_height, buffer_size);

  // Create ring buffers.
  edt_configure_ring_buffers(pdv_p, buffer_size, num_buffers, EDT_WRITE);

  buffers = (uint8_t**)edt_buffer_addresses(pdv_p);

  // If specified, read all images into memory first.
  if (usemem)
  {
    unsigned int i;

    for (i = 0; i < num_images; i++)
    {
      if (!load_next_image(pdv_p, i, image_details, num_images, NULL, num_buffers, buffer_size))
      {
        free_image_info_list(image_details, num_images);
        pdv_close(pdv_p);
        error_exit("Error: Unsupported image type.");
      }
    }
  }

  // Fill all except the last ring buffer; get details about the images stored in those buffers.
  for (current_image = 0; current_image < num_buffers - 1 && current_image < total_images;
       ++current_image)
  {
    load_next_image(pdv_p,
                    current_image,
                    image_details,
                    num_images,
                    buffers,
                    num_buffers,
                    buffer_size);
  }

  if (current_image != edt_min(num_buffers - 1, total_images))
  {
    printf("Error: current_image != first buffer!\n");
    exitval = 1;
  }

  edt_start_buffers(pdv_p, current_image);

  started = current_image;

  for (current_buffer = 0; current_buffer < total_images || loop_indefinitely;
       ++current_buffer, ++current_image)
  {
    unsigned int load_detail;
    int t1 = 0, t2 = 0;  // timeout counters

    // Image transfer.
    if (started < total_images || loop_indefinitely)
    {
      // Start current_buffer transferring.
      edt_start_buffers(pdv_p, 1);
      started++;
    }
    if (current_image < total_images || loop_indefinitely)
    {
      // Load the next image to be sent.
      load_next_image(pdv_p,
                      current_image,
                      image_details,
                      num_images,
                      buffers,
                      num_buffers,
                      buffer_size);
    }
    load_detail = current_image % num_images;

    // Check for timeouts.
    t1 = edt_timeouts(pdv_p);
    edt_wait_for_buffers(pdv_p, 1);
    t2 = edt_timeouts(pdv_p);
    if (t1 != t2)
    {
      printf("Apparent timeout.\n");
    }

    printf("\rSent %6d %-20s Loaded %6d %-20s",
           current_buffer,
           image_details[current_buffer % num_images].pathname,
           load_detail,
           image_details[load_detail].pathname);
  }

  free_image_info_list(image_details, num_images);
  pdv_close(pdv_p);
  exit(exitval);
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param errstr A discription of what is invalid about the call.
 */
void usage(const char* progname, const char* errstr)
{
  if (errstr && *errstr)
    puts(errstr);
  printf("Usage: %s [options] dirname\n", progname);
  printf("Last argument (dirname) is required, and is the name of a directory where all images to "
         "send are stored\n");
  printf("Options:\n");
  printf("-u <unit>         unit (board) number (default 0)\n");
  printf("-c <channel>      channel number (0 [default] or 1)\n");
  printf("-l <loops>        loop through image set loops times\n");
  printf("-m                load all images into memory up front (CAUTION: must have\n");
  printf("                   adequate memory for the #images in the list!)\n");
  printf("-rgb4             emulate RGB with 4-taps for RGB (R+G+B+dummy, needed for PCI\n");
  printf("                   boards, and PCIe boards with pre-10/2012 FW)\n");
  printf("-d <directory>    input directory containing images\n");
  printf("                   if specified, cycles through all valid images in the directory\n");
  printf("-i <listfile>     input file with list of images\n");
  printf("                   if specified, rotates through valid images in the file\n");
  printf("                   format for imagelist file is one filename (path)  per line\n");
  printf("-t <#images>      number of images to send (should be the same or\n");
  printf("                   less than the number found in image source)\n");
  printf("-h                print this help message\n");

  return;
}

/**
 * Retrieve the specified image's information.
 *
 * @param fname Name of the image file whose information is to be gathered.
 * @param imagelist An array of structures containing information about each image in the imagelist.
 *
 * @return 1 if specified file is found and meets format specs, otherwise -1.
 */
int get_image_info_file(const char* fname, image_info_t** imagelist)
{
  image_info_t* image_p;
  ImgFmt fmt;

  // Allocate room for the image file.
  if ((image_p = *imagelist = (image_info_t*)edt_alloc(1 * sizeof(image_info_t))) == NULL)
    return -1;

  // Only collect information for files with valid extensions.
  if ((fmt = check_extension(fname)) != fmt_unk)
  {
    char fullpath[256] = {'\0'};
    strcpy(fullpath, fname);
    edt_correct_slashes(fullpath);
    image_p->pathname = (char*)malloc(strlen(fullpath) + 1);
    strcpy(image_p->pathname, fullpath);
    image_p->format = fmt;

    if (get_image_info(fmt, image_p) != 0)
      return -1;
    ++image_p;
  }

  return 1;
}

/**
 * Validate image type of specified file, based on extension.
 *
 * @param fname Name of the image file to validate.
 *
 * @return Enum representing the image type. @ref fmt_unk if extension matches nothing.
 */
ImgFmt check_extension(const char* fname)
{
  const char* ext = NULL;
  size_t len = strlen(fname);

  if (len < 5) /* must be ?.ext */
    return fmt_unk;
  ext = &fname[len - 4];

  if (strcasecmp(ext, ".tif") == 0)
    return fmt_tif;
  if (strcasecmp(ext, ".gif") == 0)
    return fmt_gif;
  if (strcasecmp(ext, ".raw") == 0)
    return fmt_raw;
  if (strcasecmp(ext, ".ras") == 0)
    return fmt_ras;
  if (strcasecmp(ext, ".bmp") == 0)
    return fmt_bmp;
  if (strcasecmp(ext, ".jpg") == 0)
    return fmt_jpg;
  return fmt_unk;
}

/**
 * Collect information for all valid images within a specified directory.
 *
 * Uses the OS independent edt_* directory open / read / close functions.
 *
 * @param dirpath Path to a directory containing images to validate and collect information upon.
 * @param imagelist An array of structures containing information about each image in the imagelist.
 *
 * return number of images found, or -1 if error
 */
int get_images_info_dir(const char* dirpath, image_info_t** imagelist)
{
  image_info_t* image_p = NULL;
  int nimages = 0;
  DIRHANDLE dirp;
  char d_name[256] = {'\0'};

  // Open the specified directory.
  if ((dirp = edt_opendir(dirpath)) == NULL)
    return -1;

  // Read the directory entries one at a time to tally the valid image filename extensions.
  while (edt_readdir(dirp, d_name))
  {
    if (check_extension(d_name) != fmt_unk)
      ++nimages;
  }

  // EDT does not have the capability to re-read an open directory, must close and then re-open.
  edt_closedir(dirp);

  // Allocate room for the previously tallied valid files.
  if ((image_p = *imagelist = (image_info_t*)edt_alloc(nimages * sizeof(image_info_t))) == NULL)
    return -1;

  // Re-open specified directory.
  if ((dirp = edt_opendir(dirpath)) == NULL)
    return -1;

  // Read the directory entries one at a time to collect each ones information.
  while (edt_readdir(dirp, d_name))
  {
    char fullpath[256 + 1] = {'\0'};
    ImgFmt fmt;

    // Only collect information for files with valid extensions.
    if ((fmt = check_extension(d_name)) != fmt_unk)
    {
      size_t fullpath_len = 0;

      snprintf(fullpath, sizeof(fullpath), "%s/%s", dirpath, d_name);
      edt_correct_slashes(fullpath);

      fullpath_len = strlen(fullpath) + 1;
      image_p->pathname = (char*)malloc(fullpath_len);
      strncpy(image_p->pathname, fullpath, fullpath_len);
      image_p->format = fmt;

      if (get_image_info(fmt, image_p) != 0)
        return -1;
      ++image_p;
    }
  }

  return nimages;
}

/**
 * Retrieve information for a specified image.
 *
 * @param fmt An enum representing the image type.
 * @param image_p Storage for the image details.
 *   Must have pathname specified.
 *
 * @return 0 on successful retrieval.
 */
int get_image_info(ImgFmt fmt, image_info_t* image_p)
{
  // Only tiff format is implemented; fill in the switch statement with other formats.
  switch (fmt)
  {
  case fmt_tif:
    get_tif_info(image_p);
    break;

  case fmt_gif:
    get_gif_info(image_p);
    break;

  case fmt_ras:
    get_ras_info(image_p);
    break;

  case fmt_raw:
    get_raw_info(image_p);
    break;

  case fmt_bmp:
    get_bmp_info(image_p);
    break;

  case fmt_jpg:
    get_jpg_info(image_p);
    break;

    // Add your format here.

  case fmt_unk:
  default:
    return -1;
  }
  return 0;
}

/**
 * Retrieve information for tiff image.
 *
 * @param img_details Storage for the image details.
 *   Must have pathname specified.
 *
 * @return 0 on successful retrieval.
 */
int get_tif_info(image_info_t* img_details)
{
  TIFF* image;

  // printf("%s(): processing image: [%s]\n", __func__, img_details->pathname);

  // Turn off tiff warnings.
  TIFFSetWarningHandler(NULL);

  // Open the TIFF image.
  if ((image = TIFFOpen(img_details->pathname, "r")) == NULL)
  {
    printf("Error: Could not open image: %s\n", img_details->pathname);
    exit(42);
  }

  // Check that the image is of supported type (TIFF) and dimensions.
  if (!is_valid_tiff(image))
  {
    error_exit("Error: Invalid image found.");
  }

  TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &img_details->width);
  TIFFGetField(image, TIFFTAG_IMAGELENGTH, &img_details->height);

  TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &img_details->depth);
  TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &img_details->bands);

  TIFFClose(image);

  return 0;
}

/**
 * Validate the TIFF image; ensure it meets current requirements.
 *  Image must be <= 16 bits/pixel and 1 sample/pixel.
 *  Image must also define width and height.
 *
 * @param image The image to validate.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
int is_valid_tiff(TIFF* image)
{
  int value = 0;

  if (TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &value) == 0)
  {
    printf("Error: Undefined number of bits per sample.\n");
    return 0;
  }
  else if (value > 16)
  {
    printf("Error: Unsupported number of bits per sample: %d\n", value);
    return 0;
  }

  if (TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &value) == 0)
  {
    printf("Error: Undefined number of samples per pixel.\n");
    return 0;
  }

  else if ((value != 1) && (value != 3))
  {
    // TODO: support 3 sample images (rgb) as long as depth isn't greater than 8.
    printf("Error: Unsupported number of samples per pixel: %d\n", value);
    return 0;
  }

  if (TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &value) == 0)
  {
    printf("Error: Image does not define its width.\n");
    return 0;
  }

  if (TIFFGetField(image, TIFFTAG_IMAGELENGTH, &value) == 0)
  {
    printf("Error: Image does not define its height.\n");
    return 0;
  }

  return 1;
}

/* STUBS - fill in the following to add different image formats; see get_tif_info for example. */

/**
 * Retrieve information for gif image - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
int get_gif_info(image_info_t* img_details)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  return 0;
}

/**
 * Retrieve information for ras image - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
int get_ras_info(image_info_t* img_details)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  return 0;
}

/**
 * Retrieve information for raw image - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
int get_raw_info(image_info_t* img_details)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  return 0;
}

/**
 * Retrieve information for bmp image - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
int get_bmp_info(image_info_t* img_details)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  return 0;
}

/**
 * Retrieve information for jpg image - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 *
 * @return 1 (True) if image is valid, 0 (False) otherwise.
 */
int get_jpg_info(image_info_t* img_details)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  return 0;
}

/*
 * Collect image information for images listed within a file.
 *
 * @param listpath The file containing a list of images to gather information on.
 * @param imagelist An array of structures containing information about each image.
 *
 * @return Count of the images found, or -1 on failure.
 */
int get_images_info_list(const char* listpath, image_info_t** imagelist)
{
  image_info_t* image_p;
  int nimages = 0;
  char f_name[256];
  FILE* fp;

  // Open the list file.
  if ((fp = fopen(listpath, "r")) == NULL)
  {
    edt_msg_perror(EDTAPP_MSG_WARNING, listpath);
    exit(1);
  }

  // Read the file and tally valid files; ignoring comments and correcting for invalid characters.
  while (fgets(f_name, 255, fp))
  {
    strip_nl(f_name);
    edt_correct_slashes(f_name);

    if ((!comment_or_blank(f_name)) && (check_extension(f_name) != fmt_unk) &&
        (edt_access(f_name, 0) == 0))
      ++nimages;
  }
  if (!nimages)
    return 0;

  // Go back to the beginning of the list file.
  fseek(fp, 0, SEEK_SET);

  // Allocate room for the previously tallied valid files.
  if ((image_p = *imagelist = (image_info_t*)edt_alloc(nimages * sizeof(image_info_t))) == NULL)
  {
    printf("Error allocating space for %d image entries.\n", nimages);
    exit(1);
  }

  // Read the list file again; ignoring comments and correcting for invalid characters.
  while (fgets(f_name, 255, fp))
  {
    ImgFmt fmt;

    strip_nl(f_name);
    edt_correct_slashes(f_name);

    // Only collect information for files with valid extensions.
    if ((!comment_or_blank(f_name)) && ((fmt = check_extension(f_name)) != fmt_unk) &&
        (edt_access(f_name, 0) == 0))
    {
      edt_correct_slashes(f_name);
      image_p->pathname = (char*)malloc(strlen(f_name) + 1);
      strcpy(image_p->pathname, f_name);
      image_p->format = fmt;

      if (get_image_info(fmt, image_p) != 0)
        return -1;
      ++image_p;
    }
  }

  return nimages;
}

/**
 * Strip the newline off a string by replacing any terminating (`\r` or `\n`)
 * with a NULL character.
 *
 * @param s String to strip any terminating characters from.
 *
 * @return Same string pointer, with `\r` and `\n` chars replaced.
 */
char* strip_nl(char* s)
{
  int len = (int)strlen(s);
  if (len && ((s[len - 1] == '\n') || (s[len - 1] == '\r')))
    s[len - 1] = '\0';
  return s;
}

/**
 * Check the specified string for comment or blank character.
 *
 * @param str The string to check.
 *
 * @return 1 (True) if all blanks or first nonblank is '#', 0 (False) otherwise.
 */
int comment_or_blank(const char* str)
{
  int i;

  for (i = 0; i < (int)strlen(str); i++)
  {
    if (str[i] == '#')
      return 1;

    if ((str[i] != ' ') && (str[i] != '\t') && (str[i] != '\r') && (str[i] != '\n'))
      return 0;
  }
  return 1;
}

/**
 * Helper function to specify error string before exiting with failure.
 *
 * @param errstr String containing a specialized error message.
 */
void error_exit(const char* errstr)
{
  puts(errstr);
  exit(1);
}

/**
 * Free a list of images; all images within the list and then list itself.
 *
 * @param list The list of images and their parameters.
 * @param cnt The number of images in the list.
 */
void free_image_info_list(image_info_t* imagelist, const size_t cnt)
{
  size_t i;
  for (i = 0; i < cnt; i++)
  {
    image_info_t* image_p = &(imagelist[i]);

    // Free pathname.
    free(image_p->pathname);
    image_p->pathname = NULL;
  }

  // Free list.
  edt_free(imagelist);
}

/**
 * Verify that an image information list is
 * (a) correctly sized to the expected image count and
 * (b) all of the images within the list have the same dimensions.
 *
 * @param list The list of images and their parameters.
 * @param nimages The expected image count.
 *
 * @return 1 (True) if the sizes match, 0 (False) otherwise.
 */
int check_image_sizes(image_info_t* list, unsigned int nimages)
{
  unsigned int i;

  for (i = 0; i < nimages; i++)
  {
    if (list[i].width != list[0].width || list[i].height != list[0].height ||
        list[i].depth != list[0].depth)
      return 0;
  }

  return 1;
}

/**
 * Load the current_image into memory OR the appropriate ring-buffer.
 *
 * If the index of current_image is larger than total_images it will wrap around.
 *   (current_image = current_image % total_images)
 *
 * @param pdv_p Valid EDT device pointer to the CLS.
 * @param current_image The index of the image to load into local buffer.
 * @param img_details List of structures containing image details.
 * @param total_images The number of elements in @p image_details list.
 * @param buffers The array of ring buffers in use.
 * @param nbuffers The number of elements in @p buffers array.
 * @param buffer_size The size of image data buffers to allocate, if @p buffers is NULL.
 *
 * @return If successful the size of the image loaded, 0 on failure.
 */
int load_next_image(PdvDev pdv_p,
                    unsigned int current_image,
                    image_info_t* image_details,
                    unsigned int total_images,
                    uint8_t** buffers,
                    unsigned int nbuffers,
                    unsigned int buffer_size)
{
  int image_size;
  int current_buffer = current_image % nbuffers;
  uint8_t* buf_p = NULL;

  current_image = current_image % total_images;

  // printf("Filling buffer %d with image %d of %d (%s)\n", current_buffer, current_image + 1,
  // total_images, image_details[current_image].pathname);

  if (buffers)
  {
    // Read and load buffers directly.
    buf_p = buffers[current_buffer];
  }
  else
  {
    // Load buffers into memory for later processing.
    if ((buf_p = image_details[current_image].data = (unsigned char*)edt_alloc(buffer_size)) ==
        NULL)
    {
      printf("Error allocating space for: %s\n", image_details[current_image].pathname);
      exit(1);
    }
  }

  if (image_details[current_image].size)
  {
    memcpy(buffers[current_buffer],
           image_details[current_image].data,
           image_details[current_image].size);
    image_size = image_details[current_image].size;
    edt_set_buffer_size(pdv_p, current_buffer, image_size, EDT_WRITE);
  }

  switch (image_details->format)
  {
  case fmt_tif:
    image_size = fill_buffer_tif(&image_details[current_image], buf_p, buffer_size);
    break;
  case fmt_gif:
    image_size = fill_buffer_gif(&image_details[current_image], buf_p, buffer_size);
    break;
  case fmt_bmp:
    image_size = fill_buffer_bmp(&image_details[current_image], buf_p, buffer_size);
    break;
  case fmt_jpg:
    image_size = fill_buffer_jpg(&image_details[current_image], buf_p, buffer_size);
    break;
  case fmt_raw:
    image_size = fill_buffer_raw(&image_details[current_image], buf_p, buffer_size);
    break;
  case fmt_ras:
    image_size = fill_buffer_ras(&image_details[current_image], buf_p, buffer_size);
    break;

  case fmt_unk:
  default:
    return 0;
  }

  return image_size;
}

/**
 * Stores the tiff image found in img_details->pathname to the specified buffer.
 *
 * @param img_details Storage for the image details.
 *   Must have pathname specified.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer.
 */
long unsigned int fill_buffer_tif(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  TIFF* image;

  uint32_t width = 0, height = 0, depth = 0, bands = 0;
  tsize_t imageOffset = 0;
  tsize_t result = 0;
  int stripMax = 0, stripCount = 0;
  tmsize_t stripSize = 0;

  unsigned long image_size; /* in bytes */

  // printf("fill_buffer_tif(): processing image: [%s]\n", img_details->pathname);

  // Open the TIFF image.
  if ((image = TIFFOpen(img_details->pathname, "r")) == NULL)
  {
    printf("Error: Could not open image: %s\n", img_details->pathname);
    exit(42);
  }

  // Check that the image is of supported type (TIFF) and dimensions.
  if (!is_valid_tiff(image))
  {
    printf("Error: Invalid image; exiting.\n");
    exit(42);
  }

  TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetField(image, TIFFTAG_IMAGELENGTH, &height);
  TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &depth);
  TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &bands);

  // Display image information.
  // printf("fill_buffer_tif(): Image: %s Width: %d Height: %d Depth: %d Bands %d\n",
  // img_details->pathname, width, height, depth, bands);

  // CLS doesn't do RGB; emulate by sending 4 channels with 1 throwaway.
  if ((bands == 3) && Rgb_4tap)
    image_size = width * height * pdv_bits_to_bytes(depth) * (bands + 1);
  else
    image_size = width * height * pdv_bits_to_bytes(depth) * bands;

  // printf("fill_buffer_tif(): Width*Height*depth_bytes*channels=%d (0x%x)\n", image_size,
  // image_size);

  // Read in the possibly multiple strips.
  stripMax = TIFFNumberOfStrips(image);
  stripSize = TIFFStripSize(image);

  imageOffset = 0;

  if (image_size > (uint32_t)max_size)
  {
    printf("Error: Image size (0x%lx bytes) is larger "
           "than buffer size (0x%x bytes).\n",
           image_size,
           max_size);
    exit(42);
  }

  // CLS doesn't do RGB, emulate by sending 4 bytes with 1 throwaway (inserts 0 every 4th byte).
  if ((bands == 3) && Rgb_4tap)
  {
    tsize_t i;
    uint8_t* bp = buffer;
    uint8_t* stripbuf = (uint8_t*)malloc(stripSize);

    for (stripCount = 0; stripCount < stripMax; stripCount++)
    {
      if ((result = TIFFReadEncodedStrip(image, stripCount, stripbuf, (tsize_t)-1)) == (tsize_t)-1)
      {
        printf("Error: Read error on input strip number %d.\n", stripCount);
        free(stripbuf);
        exit(42);
      }

      for (i = 0; i < result; i += 3)
      {
        *bp++ = stripbuf[i];
        *bp++ = stripbuf[i + 1];
        *bp++ = stripbuf[i + 2];
        *bp++ = 0;
      }
    }
    free(stripbuf);
  }
  else
  {
    for (stripCount = 0; stripCount < stripMax; stripCount++)
    {
      if ((result = TIFFReadEncodedStrip(image, stripCount, buffer + imageOffset, (tsize_t)-1)) ==
          -1)
      {
        printf("Error: Read error on input strip number %d.\n", stripCount);
        exit(42);
      }

      imageOffset += result;
    }
  }

  img_details->width = width;
  img_details->height = height;

  TIFFClose(image);

  return image_size;
}

/* STUBS - fill in the following add different image formats; see get_tif_info for example. */

/**
 * Stores the gif image found in img_details->filename to the specified buffer - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer, -1 on error.
 */
int fill_buffer_gif(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  (void)buffer;
  (void)max_size;
  return -1;
}

/**
 * Stores the bmp image found in img_details->filename to the specified buffer - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer, -1 on error.
 */
int fill_buffer_bmp(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  (void)buffer;
  (void)max_size;
  return -1;
}

/**
 * Stores the jpg image found in img_details->filename to the specified buffer - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer, -1 on error.
 */
int fill_buffer_jpg(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  (void)buffer;
  (void)max_size;
  return -1;
}

/**
 * Stores the ras image found in img_details->filename to the specified buffer - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer, -1 on error.
 */
int fill_buffer_ras(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  (void)buffer;
  (void)max_size;
  return -1;
}

/**
 * Stores the raw image found in img_details->filename to the specified buffer - STUB.
 *
 * @param img_details Storage for the image details.
 *   Must have filename specified.
 * @param buffer A pointer to a buffer in which to store the image.
 * @param max_size The size of the buffer in bytes.
 *
 * @return The size in bytes of image loaded into buffer, -1 on error.
 */
int fill_buffer_raw(image_info_t* img_details, uint8_t* buffer, int max_size)
{
  printf("WARNING: Unimplemented stub -- %s()\n", __func__);

  (void)img_details;
  (void)buffer;
  (void)max_size;
  return -1;
}

/**
 * Determine the minimum value of the two specified.
 *
 * @param l One of two integer values to determine the lesser of.
 * @param r The other of two integer values to determine the lesser of.
 *
 * @return The lesser of the two values.
 */
unsigned int edt_min(unsigned int l, unsigned int r)
{
  if (l < r)
    return l;
  return r;
}
