/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2012 Engineering Design Team, Inc.
 *
 * An example program to show usage of EDT PDV library to acquire and optionally save images
 * from EDT VisionLink products, making use of the IRIG2 footer to check: the end of frame magic
 * number, the frame synchronization, and (if applicable) the IRIG timestamp.
 *
 * For a non-IRIG acquisition example see: simple_take.c
 * For simple frame synchronization verification see the "method_framesync" directive in the
 *  Camera Configuration Guide, or pdv_enable_framesync() in the API.
 *
 * Requirements:
 *   An image source must be provided to the EDT board.
 *
 * Driver 5.3.9.3+: IRIG2 frame validation functionality is now embedded within the library,
 *  activated via config file directive "method_header_type" (see: Camera Configuration Guide).
 */

#include "edtinc.h"
#include "pdv_irig.h"
#include "libdvu.h"

/* Function Declarations */

static void usage(const char* progname);
static void save_image(uint8_t* image_p,
                       int width,
                       int height,
                       int depth,
                       char* basename,
                       int count);

/**
 * Acquire and save images from an EDT frame grabber.
 */
int main(int argc, const char* argv[])
{
  int i = 0;
  int unit = 0;
  int overruns = 0;
  int last_timeouts = 0;
  int recovering_timeout = FALSE;
  const char* progname = NULL;
  const char* cameratype = NULL;
  char bmpfname[128] = {'\0'};
  int numbufs = 4;
  int started = 0;
  uint8_t* image_p = NULL;
  PdvDev pdv_p = NULL;
  int loops = 1;
  int width = 0, height = 0, depth = 0;
  char edt_devname[128] = {'\0'};
  int channel = 0;

  progname = argv[0];

  // Parse user arguments.
  --argc;
  ++argv;
  while (argc && ((argv[0][0] == '-') || (argv[0][0] == '/')))
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        unit = atoi(argv[0]);
      else
        strncpy(edt_devname, argv[0], sizeof(edt_devname) - 1);
      break;

    case 'c':
      ++argv;
      --argc;
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        channel = atoi(argv[0]);
      break;

    case 'N':
      ++argv;
      --argc;
      numbufs = atoi(argv[0]);
      break;

    case 'i':
    case 'b':
      ++argv;
      --argc;
      strcpy(bmpfname, argv[0]);
      break;

    case 'l':
      ++argv;
      --argc;
      loops = atoi(argv[0]);
      break;

    case '-':
      if (strcmp(argv[0], "--help") == 0)
      {
        usage(progname);
        exit(0);
      }
      else
      {
        fprintf(stderr, "Unknown option: %s\n", argv[0]);
        usage(progname);
        exit(1);
      }
      break;

    case '?':
    case 'h':
      usage(progname);
      exit(0);

    default:
      fprintf(stderr, "Unknown flag -'%c'.\n", argv[0][1]);
      usage(progname);
      exit(1);
    }
    argc--;
    argv++;
  }

  // Open the requested interface.
  //  Note: EDT_INTERFACE is defined in edtdef.h (included via edtinc.h)
  if (edt_devname[0])
  {
    unit = edt_parse_unit_channel(edt_devname, edt_devname, EDT_INTERFACE, &channel);
  }
  else
  {
    strcpy(edt_devname, EDT_INTERFACE);
  }

  if ((pdv_p = pdv_open_channel(edt_devname, unit, channel)) == NULL)
  {
    char errstr[128 + 24] = {'\0'};
    snprintf(errstr, sizeof(errstr), "pdv_open_channel(%s, %d, %d)", edt_devname, unit, channel);
    edt_msg_perror(EDTAPP_MSG_WARNING, errstr);
    return (1);
  }

  // Save image size and camera type for later use.
  width = pdv_get_width(pdv_p);
  height = pdv_get_height(pdv_p);
  depth = pdv_get_depth(pdv_p);
  cameratype = pdv_get_camera_type(pdv_p);

  // Enable IRIG2 footer.
  //  Note: will provide magic number and framecount regardless of IRIG timesource presence.
  //  Note: can also enable in config file via "method_header_type: IRIG".
  pdv_set_header_type(pdv_p, PDV_HDR_TYPE_IRIG2, 2, 0);

  // Allocate 4 buffers for PDV ring buffer.
  pdv_multibuf(pdv_p, numbufs);

  printf("Reading %d image%s from '%s'\nwidth %d height %d depth %d\n",
         loops,
         loops == 1 ? "" : "s",
         cameratype,
         width,
         height,
         depth);

  //  Prestart the first image or images outside the loop to get the pipeline going.
  if (pdv_get_force_single(pdv_p))
  {
    // Start a single image; some cameras (e.g. ones that need a gap between images or that take a
    // serial command to start every image) don't tolerate queueing of multiple images.
    pdv_start_images(pdv_p, 1);
    started = 1;
  }
  else
  {
    // Start multiple images.
    pdv_start_images(pdv_p, numbufs);
    started = numbufs;
  }

  for (i = 0; i < loops; i++)
  {
    int timeouts = 0;
    int overrun;

    // Get the image and immediately start the next one (if not the last time through the loop).
    // Processing (saving to a file in this case) can then occur in parallel with the next
    // acquisition.
    printf("Image %d        \r", i + 1);
    fflush(stdout);
    image_p = pdv_wait_images(pdv_p, 1);

    if (EDT_DEVID_HAS_IRIGB(edt_device_id(pdv_p)) &&
        (pdv_get_header_type(pdv_p) == PDV_HDR_TYPE_IRIG2))
    {
      Irig2Record* irp = (Irig2Record*)(image_p + pdv_get_header_offset(pdv_p));

      printf("magic %s framecnt %d\n", irp->magic == IRIG2_MAGIC ? "OK " : "BAD", irp->framecnt);
    }
    else
      printf("Not pcie8 or header type not enabled in config.\n");

    overrun = edt_reg_read(pdv_p, PDV_STAT) & PDV_OVERRUN;
    if (overrun)
      ++overruns;

    if (i < loops - started)
    {
      pdv_start_images(pdv_p, 1);
    }
    timeouts = pdv_timeouts(pdv_p);

    // Check for timeouts or data overruns.
    //  Timeouts occur when data is lost, camera isn't hooked up, etc, and application should
    //  always check for them. Data overruns usually occur as a result of a timeout but should
    //  be checked for separately since ROI can sometimes mask timeouts.
    if (timeouts > last_timeouts)
    {
      // pdv_timeout_cleanup helps recover gracefully after a timeout, particularly if multiple
      // buffers were prestarted.
      pdv_timeout_restart(pdv_p, TRUE);
      last_timeouts = timeouts;
      recovering_timeout = TRUE;
      printf("\nTimeout....\n");
    }
    else if (recovering_timeout)
    {
      pdv_timeout_restart(pdv_p, TRUE);
      recovering_timeout = FALSE;
      printf("\nRestarted....\n");
    }
    if (*bmpfname)
      save_image(image_p, width, height, depth, bmpfname, i);
  }
  puts("");

  printf("%d images %d timeouts %d overruns\n", loops, last_timeouts, overruns);

  // Timeouts indicate there is a problem.
  if (last_timeouts)
    printf("Check camera and connections.\n");

  pdv_close(pdv_p);

  exit(0);
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 */
static void usage(const char* progname)
{
  printf("Usage: %s [options]\n", progname);
  printf("Options:\n");
  printf("-u <unit>       unit (board) number (default 0)\n");
  printf("-c <channel>    channel number (0 [default] or 1)\n");
#ifdef _NT_
  printf("-b <fname>      specifies output (MS bitmap file) name\n");
#else
  printf("-i <fname>      specifies output (Sun Raster file) name \n");
#endif
  printf("-l <loops>      number of loops (images to take)\n");
  printf("-N <#buffers>   number of ring buffers to use (default 4)\n");
  printf("-h              print this help message\n");
}

/**
 * Save an image; constructs appropriate name.
 *
 * @param image_p Buffer containing image data.
 * @param s_width Width of the image to save.
 * @param s_height Height of the image to save.
 * @param s_depth Depth of the image to save.
 * @param tmpname Output file name.
 * @param count Image index number.
 */
static void save_image(uint8_t* image_p,
                       int s_width,
                       int s_height,
                       int s_depth,
                       char* tmpname,
                       int count)
{
  int s_db = pdv_bits_to_bytes(s_depth);
  char fname[256];

#ifdef _NT_
  uint8_t* bbuf = NULL;
  if ((strcmp(&tmpname[strlen(tmpname) - 4], ".bmp") == 0) ||
      (strcmp(&tmpname[strlen(tmpname) - 4], ".BMP") == 0))
    tmpname[strlen(tmpname) - 4] = '\0';
  sprintf(fname, "%s%02d.bmp", tmpname, count);
#else
  if ((strcmp(&tmpname[strlen(tmpname) - 4], ".ras") == 0) ||
      (strcmp(&tmpname[strlen(tmpname) - 4], ".RAS") == 0))
    tmpname[strlen(tmpname) - 4] = '\0';
  sprintf(fname, "%s%02d.ras", tmpname, count);
#endif

  // Write bmp file (Windows) or Sun Raster (Unix/Linux); switch on number of bytes per pixel.
  switch (s_db)
  {
  case 1:
#ifdef _NT_

    edt_dvu_write_bmp(fname, image_p, s_width, s_height);
#else
    printf("Writing %dx%dx%d raster file to: %s\n", s_width, s_height, s_depth, fname);

    edt_dvu_write_rasfile(fname, image_p, s_width, s_height);
#endif
    break;

  case 2:
    printf("Converting %dx%dx%d image to 8 bits, writing to: %s\n",
           s_width,
           s_height,
           s_depth,
           fname);

#ifdef _NT_
    if (!bbuf)
      bbuf = (uint8_t*)pdv_alloc(s_width * s_height);

    if (bbuf == NULL)
    {
      edt_msg_perror(EDTAPP_MSG_WARNING, "Data buf malloc.");
      exit(1);
    }
    edt_dvu_word_to_byte((uint16_t*)image_p, bbuf, s_width * s_height, s_depth);
    edt_dvu_write_bmp(fname, bbuf, s_width, s_height);
#else
    edt_dvu_write_rasfile16(fname, image_p, s_width, s_height, s_depth);
#endif
    break;

  case 3:
    printf("Writing %dx%dx%d bmp file to: %s\n", s_width, s_height, s_depth, fname);

#ifdef _NT_
    edt_dvu_write_bmp24(fname, image_p, s_width, s_height);
#else
    edt_dvu_write_rasfile24(fname, image_p, s_width, s_height);
#endif

    break;

  default:
    printf("Error: Invalid image depth for file write!\n");
    break;
  }
}
