/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2002 Engineering Design Team, Inc.
 *
 * An example program to show usage of EDT PDV library to acquire and save images from EDT
 * VisionLink products. This program takes a finite number of images from the camera at the maximum
 * frame rate and copies them to memory until done acquiring all images. Once all images have been
 * acquired, it saves them each in a separate file.
 *
 * For other operations, including error detection, diagnostics, and tuning the acquisition in
 * various ways see: take.c or simple_take.c
 *
 * Requirements:
 *   An image source must be provided to the EDT board.
 */
#include "edtinc.h"

/* Output file; described and specified per OS. */
#ifdef _NT_
// Windows; save images as bitmap format.
#define EXT     ".bmp"
#define EXTDESC "Windows Bitmap"
#else
// Linux; save images as Sun Raster format.
#define EXT     ".ras"
#define EXTDESC "Sun Raster"
#endif

/* Function Declarations */

static void usage(const char* progname, const char* err);
static void resolve_fname(const char* fname,
                          char* newname,
                          const size_t newname_size,
                          const char* ext);
static int write_image_file(char* tmpfname,
                            uint8_t* image_p,
                            int s_width,
                            int s_height,
                            int s_depth);

/* Variables */

static int ASnum = 0;

/**
 * Acquire and save images from an EDT frame grabber.
 */
int main(int argc, const char* argv[])
{
  int i = 0;
  int unit = 0;
  int timeouts = 0;
  const char* progname = argv[0];
  const char* cameratype = NULL;
  char bmpfname[128] = {'\0'};
  char rawfname[128] = {'\0'};
  char tmpfname[280] = {'\0'};
  uint8_t** bufs = NULL;
  uint8_t* image_p = NULL;
  PdvDev pdv_p = NULL;
  char errstr[128 + 30] = {'\0'};
  int loops = 1;
  int numbufs = 1;
  int width = 0, height = 0, depth = 0, imagesize = 0;
  double dtime = 0;
  char edt_devname[128] = {'\0'};
  int channel = 0;

  --argc;
  ++argv;
  while (argc && ((argv[0][0] == '-') || (argv[0][0] == '/')))
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'u' requires a numeric argument.\n");
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        unit = atoi(argv[0]);
      else
        strncpy(edt_devname, argv[0], sizeof(edt_devname) - 1);
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        channel = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
      }
      break;

    case 'i':
    case 'b':
      ++argv;
      --argc;
      strcpy(bmpfname, argv[0]);
      break;

    case 'f':
      ++argv;
      --argc;
      strcpy(rawfname, argv[0]);
      break;

    case 'l':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'l' requires a numeric argument.\n");
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        numbufs = loops = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'l' requires a numeric argument.\n");
      }
      break;

    case '?':
    case 'h':
      usage(progname, "");
      break;

    default:
      fprintf(stderr, "Error: Unknown flag -'%c'.\n", argv[0][1]);
      usage(progname, "");
    }
    argc--;
    argv++;
  }

  // Open the requested interface.
  //  Note: EDT_INTERFACE is defined in edtdef.h (included via edtinc.h)
  if (edt_devname[0])
  {
    unit = edt_parse_unit_channel(edt_devname, edt_devname, EDT_INTERFACE, &channel);
  }
  else
  {
    strcpy(edt_devname, EDT_INTERFACE);
  }

  if ((pdv_p = pdv_open_channel(edt_devname, unit, channel)) == NULL)
  {
    snprintf(errstr,
             sizeof(errstr) - 1,
             "pdv_open_channel(%s, %d, %d)",
             edt_devname,
             unit,
             channel);
    edt_msg_perror(EDTAPP_MSG_WARNING, errstr);
    return (1);
  }

  // Save image size and camera name for later use.
  width = pdv_get_width(pdv_p);
  height = pdv_get_height(pdv_p);
  depth = pdv_get_depth(pdv_p);
  imagesize = pdv_get_image_size(pdv_p);
  cameratype = pdv_get_camera_type(pdv_p);

  if (width <= 0)
  {
    printf("%s%d: Invalid image size. Make sure device has been initialized.\n", edt_devname, unit);
    exit(1);
  }

  // Allocate 4 buffers for optimal PDV ring buffer pipeline.
  pdv_multibuf(pdv_p, 4);

  // Allocate a list of image buffer pointers.
  bufs = (uint8_t**)malloc(loops * sizeof(uint8_t*));
  for (i = 0; i < loops; i++)
  {
    // Allocate each image buffer pointer; edt_alloc page aligns.
    if ((bufs[i] = (uint8_t*)edt_alloc(imagesize)) == NULL)
    {
      printf("Buffer allocation FAILED (probably too many images specified).\n");
      exit(1);
    }
  }

  printf("Reading %d image%s from: '%s'", loops, loops == 1 ? "" : "s", cameratype);
  fflush(stdout);

  // Initialize time for checking later.
  (void)edt_dtime();

  // Start DMAing all image buffers in the ring.
  pdv_start_images(pdv_p, numbufs);
  for (i = 0; i < loops; i++)
  {
    image_p = pdv_wait_images(pdv_p, 1);
    memcpy(bufs[i], image_p, imagesize);
  }
  dtime = edt_dtime();
  printf(" Done.\n");

  printf("%f frames/sec\n", (double)(loops) / dtime);
  timeouts = pdv_timeouts(pdv_p);

  puts("");
  for (i = 0; i < loops; i++)
  {
    if (*bmpfname)
    {
      resolve_fname(bmpfname, tmpfname, sizeof(tmpfname), EXT);
      write_image_file(tmpfname, bufs[i], width, height, depth);
    }
    else if (*rawfname)
    {
      resolve_fname(rawfname, tmpfname, sizeof(tmpfname), ".raw");
      printf("Writing %dx%dx%d raw file to: %s\n", width, height, depth, tmpfname);

      (void)edt_dvu_write_raw(imagesize, bufs[i], tmpfname);
    }
  }

  printf("%d images %d timeouts\n", loops, timeouts);

  // Timeouts indicate there is a problem.
  if (timeouts)
  {
    printf("Warning: got %d timeouts (incomplete images).\n", timeouts);
    printf("Check camera and connections.\n");
  }
  pdv_close(pdv_p);

  exit(0);
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param err A description of what is invalid about the call.
 */
static void usage(const char* progname, const char* err)
{
  if (err && *err)
    puts(err);
  printf("Usage: %s [options]\n", progname);
  printf("Options:\n");
  printf("-u <unit>       unit (board) number (default 0)\n");
  printf("-c <channel>    channel number (0 [default] or 1)\n");
#ifdef _NT_
  printf("-b <fname>      specifies output (MS bitmap file) name\n");
#else
  printf("-i <fname>      specifies output (Sun Raster file) name \n");
#endif
  printf("-f <fname>      specifies output (raw file) name\n");
  printf("                 (bytes per pixel: 8-bit devices = 1, 10-bit to 16-bit devices = 2)\n");
  printf("-l <loops>      number of loops (images to take, default 1)\n");
  printf("-h              print this help message\n");
}

/**
 * Generate appropriate name for saving an image.
 *
 * @param fname Specified base name for file.
 * @param newname Storage for the newly created file name.
 * @param newname_size Maximum length of the new file name.
 * @param ext The extension to use.
 */
static void resolve_fname(const char* fname,
                          char* newname,
                          const size_t newname_size,
                          const char* ext)
{
  size_t l = strlen(fname);
  char tmpfname[256];
  char tmpext[5];

  strcpy(tmpfname, fname);
  if ((l > 4) && (tmpfname[l - 4] == '.'))
  {
    strcpy(tmpext, &tmpfname[l - 4]);
    tmpfname[l - 4] = '\0';
  }
  else
    strcpy(tmpext, ext);

  snprintf(newname, newname_size, "%s%04d%s", tmpfname, ASnum++, tmpext);
}

/**
 * Save an image to provided name.
 *
 * @param tmpname Output file name.
 * @param image_p Buffer containing image data.
 * @param s_width Width of the image to save.
 * @param s_height Height of the image to save.
 * @param s_depth Depth of the image to save.
 *
 * @return 0 on success, -1 on failure.
 */
static int write_image_file(char* tmpfname,
                            uint8_t* image_p,
                            int s_width,
                            int s_height,
                            int s_depth)
{
#ifdef _NT_
  uint8_t* bbuf = NULL;
#endif
  int s_db = pdv_bits_to_bytes(s_depth);

  // Write bmp file (Windows) or Sun Raster (Unix/Linux). Switch on number of bytes per pixel.
  switch (s_db)
  {
  case 1:
    printf("Writing %dx%dx%d %s file to: %s\n", s_width, s_height, s_depth, EXTDESC, tmpfname);

#ifdef _NT_
    edt_dvu_write_bmp(tmpfname, image_p, s_width, s_height);
#else
    edt_dvu_write_rasfile(tmpfname, image_p, s_width, s_height);
#endif
    break;

  case 2:
    printf("Converting %dx%dx%d image to 8 bits, writing to: %s\n",
           s_width,
           s_height,
           s_depth,
           tmpfname);

#ifdef _NT_
    if (!bbuf)
      bbuf = (uint8_t*)pdv_alloc(s_width * s_height);

    if (bbuf == NULL)
    {
      edt_msg_perror(EDTAPP_MSG_WARNING, "Data buf malloc.");
      exit(1);
    }
    edt_dvu_word_to_byte((u_short*)image_p, bbuf, s_width * s_height, s_depth);
    edt_dvu_write_bmp(tmpfname, bbuf, s_width, s_height);
#else
    edt_dvu_write_rasfile16(tmpfname, image_p, s_width, s_height, s_depth);
#endif
    break;

  case 3:
    printf("Writing %dx%dx%d bmp file to: %s\n", s_width, s_height, s_depth, tmpfname);

#ifdef _NT_
    edt_dvu_write_bmp24(tmpfname, image_p, s_width, s_height);
#else
    edt_dvu_write_rasfile24(tmpfname, image_p, s_width, s_height);
#endif

    break;

  default:
    printf("Error: Invalid image depth for file write!\n");
    return -1;
    break;
  }

  return 0;
}
