/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2007 Engineering Design Team, Inc.
 *
 * A utility / example program to show usage of EDT PDV library to acquire and save images from
 * EDT frame grabber products such as the VisionLink F4. This application specifically showcases
 * performance optimization strategies which take advantage of ring buffers for image aquisition and
 * subsiquent processing.
 *
 * For more complex operations, including error detection, diagnostics,
 * changing camera exposure times, and tuning the acquisition in various
 * ways, refer to the "take" utility.
 *
 * Requirements:
 *   An image source must be provided to the EDT board.
 */

#include "edtinc.h"

/* Function Declarations */

static void usage(const char* progname, const char* errmsg);
static void save_image(uint8_t* image_p,
                       int width,
                       int height,
                       int depth,
                       char* basename,
                       int count);

/**
 * Acquire and save images from an EDT frame grabber board.
 */
int main(int argc, const char* argv[])
{
  int i;
  int unit = 0;
  int overruns = 0;
  int timeouts = 0, last_timeouts = 0;
  int recovering_timeout = FALSE;
  const char* progname = NULL;
  const char* cameratype = NULL;
  char bmpfname[128] = {'\0'};
  int numbufs = 4;
  int started = 0;
  uint8_t* image_p = NULL;
  PdvDev pdv_p = NULL;
  char errstr[128 + 30] = {'\0'};
  int loops = 1;
  int width = 0, height = 0, depth = 0;
  char edt_devname[128] = {'\0'};
  int channel = 0;

  progname = argv[0];

  // Process command line arguments.
  --argc;
  ++argv;
  while (argc && ((argv[0][0] == '-') || (argv[0][0] == '/')))
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 1)
        usage(progname, "Error: Option 'u' requires an argument.\n");
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
        unit = atoi(argv[0]);
      else
        strncpy(edt_devname, argv[0], sizeof(edt_devname) - 1);
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        channel = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
      }
      break;

    case 'N':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'N' requires a numeric argument.\n");
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        numbufs = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'N' requires a numeric argument.\n");
      }
      break;

    case 'b':
      ++argv;
      --argc;
      strcpy(bmpfname, argv[0]);
      break;

    case 'l':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'l' requires a numeric argument.\n");
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        loops = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'l' requires a numeric argument.\n");
      }
      break;

    case '-':
      if (strcmp(argv[0], "--help") == 0)
      {
        usage(progname, "");
      }
      else
      {
        fprintf(stderr, "Error: Unknown option: %s.\n", argv[0]);
        usage(progname, "");
      }
      break;

    case '?':
    case 'h':
      usage(progname, "");
      break;

    default:
      fprintf(stderr, "Error: Unknown flag -'%c'.\n", argv[0][1]);
      usage(progname, "");
    }
    argc--;
    argv++;
  }

  // Open the requested interface.
  //  Note: EDT_INTERFACE is defined in edtdef.h (included via edtinc.h)
  if (edt_devname[0])
  {
    unit = edt_parse_unit_channel(edt_devname, edt_devname, EDT_INTERFACE, &channel);
  }
  else
  {
    strcpy(edt_devname, EDT_INTERFACE);
  }

  if ((pdv_p = pdv_open_channel(edt_devname, unit, channel)) == NULL)
  {
    snprintf(errstr, sizeof(errstr), "pdv_open_channel(%s, %d, %d)", edt_devname, unit, channel);
    edt_msg_perror(EDTAPP_MSG_WARNING, errstr);
    return (1);
  }

  pdv_flush_fifo(pdv_p);

  // Save image size and camera name for later use.
  width = pdv_get_width(pdv_p);
  height = pdv_get_height(pdv_p);
  depth = pdv_get_depth(pdv_p);
  cameratype = pdv_get_camera_type(pdv_p);

  // Setup the PDV ring buffer.
  pdv_multibuf(pdv_p, numbufs);

  printf("Reading %d image%s from '%s'\nwidth %d height %d depth %d.\n",
         loops,
         loops == 1 ? "" : "s",
         cameratype,
         width,
         height,
         depth);

  // Prestart the first image or images outside the loop to get the pipeline going.
  if (pdv_get_force_single(pdv_p))
  {
    // Start a single image. Some cameras (e.g. ones that need a gap between images or that take a
    // serial command to start every image) don't tolerate queueing of multiple images.
    pdv_start_images(pdv_p, 1);
    started = 1;
  }
  else
  {
    // Start multiple images.
    pdv_start_images(pdv_p, numbufs);
    started = numbufs;
  }

  for (i = 0; i < loops; i++)
  {
    int overrun;

    // Get the image and immediately start the next one (if not the last time through the loop).
    // Processing (saving to a file in this case) can then occur in parallel with the next
    // acquisition.
    printf("image %d\r", i + 1);
    fflush(stdout);
    image_p = pdv_wait_images(pdv_p, 1);

    // Check for overruns. Data overruns can occur as a result of a timeout
    // or due to the application not keeping up with incoming data. Overruns
    // should be checked separately from timeouts since ROI can sometimes
    // mask timeouts.
    overrun = edt_reg_read(pdv_p, PDV_STAT) & PDV_OVERRUN;
    if (overrun)
    {
      ++overruns;
      printf("WARNING: %d overruns\n", overruns);
    }

    if (i < loops - started)
    {
      pdv_start_images(pdv_p, 1);
    }
    timeouts = pdv_timeouts(pdv_p);

    // Check for timeouts. Timeouts occur when data is lost, camera isn't hooked up, etc,
    // and applications should always check for them.
    if (timeouts > last_timeouts)
    {
      // pdv_timeout_restart helps recover gracefully after a timeout, particularly if multiple
      // buffers were prestarted.
      pdv_timeout_restart(pdv_p, TRUE);
      last_timeouts = timeouts;
      recovering_timeout = TRUE;
      printf("\nTimeout....\n");
    }
    else if (recovering_timeout)
    {
      pdv_timeout_restart(pdv_p, TRUE);
      recovering_timeout = FALSE;
      printf("\nRestarted....\n");
    }
    if (*bmpfname)
      save_image(image_p, width, height, depth, bmpfname, (loops > 1 ? i : -1));
  }
  puts("");

  printf("%d images %d timeouts %d overruns\n", loops, last_timeouts, overruns);

  // Timeouts indicate there is a problem.
  if (last_timeouts)
    printf("Check camera and connections.\n");
  pdv_close(pdv_p);

  if (overruns || timeouts)
    exit(2);
  exit(0);
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param errmsg A description of what is invalid about the call.
 */
static void usage(const char* progname, const char* errmsg)
{
  if (errmsg && *errmsg)
    puts(errmsg);
  printf("Usage: %s [options]\n", progname);
  printf("Options:\n");
  printf("-u <unit>      unit (board) number (default 0)\n");
  printf("-c <channel>   channel number (default 0)\n");
  printf("-N <numbufs>   number of ring buffers to use (default 4)\n");
  printf("-b <fname>     specifies output (MS bitmap file) name\n");
  printf("-l <loops>     number of loops (images to take)\n");
  printf("-h             print this help message\n");
}

/**
 * Save an image; constructs appropriate name.
 *
 * @param image_p Buffer containing image data.
 * @param s_width Width of the image to save.
 * @param s_height Height of the image to save.
 * @param s_depth Depth of the image to save.
 * @param tmpname Output file name.
 * @param count Image index number.
 */
static void save_image(uint8_t* image_p,
                       int s_width,
                       int s_height,
                       int s_depth,
                       char* tmpname,
                       int count)
{
  int s_db = pdv_bits_to_bytes(s_depth);
  char fname[256];

  uint8_t* bbuf = NULL;
  if ((strcmp(&tmpname[strlen(tmpname) - 4], ".bmp") == 0) ||
      (strcmp(&tmpname[strlen(tmpname) - 4], ".BMP") == 0))
    tmpname[strlen(tmpname) - 4] = '\0';

  if (count >= 0)
    sprintf(fname, "%s_%03d.bmp", tmpname, count);
  else
    sprintf(fname, "%s.bmp", tmpname);

  switch (s_db)
  {
  case 1:
    edt_dvu_write_bmp(fname, image_p, s_width, s_height);
    printf("Writing %dx%dx%d bitmap file to: %s", s_width, s_height, s_depth, fname);
    break;

  case 2:
    printf("Converting %dx%dx%d image to 8 bits, writing to: %s",
           s_width,
           s_height,
           s_depth,
           fname);

    if (!bbuf)
      bbuf = (uint8_t*)pdv_alloc(s_width * s_height);

    if (bbuf == NULL)
    {
      edt_msg_perror(EDTAPP_MSG_WARNING, "Data buf malloc");
      exit(1);
    }
    edt_dvu_word_to_byte((uint16_t*)image_p, bbuf, (s_width * s_height), s_depth);
    edt_dvu_write_bmp(fname, bbuf, s_width, s_height);
    break;

  case 3:
    printf("Writing %dx%dx%d bmp file to: %s\n", s_width, s_height, s_depth, fname);

    edt_dvu_write_bmp24(fname, image_p, s_width, s_height);
    break;

  default:
    printf("ERROR: Invalid image depth for file write!\n");
    break;
  }
}
