/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * Library to read a 2 wire protocol device such as a serial EEPROM or a
 * diagnostic port on a serial transceiver.
 *
 * The port must be implemented as a register address like the PCISS/GS OCM
 *
 * - bit 1 is the two wire clock, it is always an output
 * - bit 2 is the two wire serial data out, it is enable onto the serial data when
 * - bit 3 enables bit 2 onto the serial data bus when low
 * - bit 4 reads back the stat of the serial data wire, it should follow bit 2 when bit 3 is on
 *
 * The OCM has three 2-wire device addresses:
 * 1. channel 0 SFP transceiver at 0x23
 * 2. channel 1 SFP transceiver at 0x33
 * 3. serial EEPROM calibration - [!] currently unimplemented
 *
 * Some two wire devices have 1 byte address others 2 etc. The
 * @ref EdtTwoWire.address_length field designates number of bytes.
 *
 * The page write size varies from device to device. Setting to 1 will work on
 * anything but writes will be slow.
 */

#ifndef _LIB_TWO_WIRE_H
#define _LIB_TWO_WIRE_H

#ifdef __cplusplus
extern "C" {
#endif

#include "edtinc.h"

/**
 * @brief Two-Wire interface options, for use in @ref EdtTwoWire config struct.
 */
typedef enum {
  /**
   * Enable support for I2C targets which may hold the SCL line until data is
   * ready ("clock stretching"). This will cause the software to tristate the
   * FPGA I2C SCL line during a transaction and wait until the line actually goes
   * high.
   */
  TWO_WIRE_OPT_CLK_STRETCH = (1 << 0),
} EdtTwoWireOptions;

/** Two-wire interface configuration. */
typedef struct edt_two_wire {
  uint32_t read_address;
  uint32_t write_address;
  uint32_t sclk_mask;
  uint32_t tri_state_mask;
  uint32_t sda_out_mask;
  uint32_t sda_in_mask;
  uint32_t scl_tri_state_mask;
  uint32_t scl_in_mask;
  /** Option flags for this interface. */
  EdtTwoWireOptions options;
  /** Number of bytes of address/command.*/
  uint32_t address_length;
  uint32_t page_write_size;
} EdtTwoWire;

/** ID of SFP-type transceivers. Section 4.2 in @cite SFF8024 */
#define TWO_WIRE_SFP 0x03
/** ID of XFP-type transceivers. Section 4.2 in @cite SFF8024 */
#define TWO_WIRE_XFP 0x06

/** I2C Chip ID of SFF transceiver devices. */
#define TWO_WIRE_SERIAL_ID_DEVICE 0x50

/**
 * @brief Reset two wire interface in FPGA.
 *
 * Send up to 9 SCLKs while reading SDA until SDA is high. Then enable SDA and
 * bring it low to make an I2C start condition.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param tw_ptr
 *   Two-wire configuration.
 *
 * @returns Zero if successful.
 */
EDTAPI int edt_two_wire_reset(EdtDev edt_p, const EdtTwoWire* tw_ptr);

/**
 * @brief Block read transaction.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param tw_ptr
 *   Two-wire configuration.
 * @param chip_id
 *   I2C chip ID of the target device.
 * @param address
 *   Address/command to send to target. Number of bits in address configured by
 *   @ref EdtTwoWire.address_length.
 * @param len_data
 *   Number of bytes to read from target device.
 * @param[out] data
 *   Data read from target device. Must be at least @ref len_data bytes.
 *
 * @returns Number of bytes actually read from target.
 */
EDTAPI uint16_t edt_two_wire_read(EdtDev edt_p,
                                  const EdtTwoWire* tw_ptr,
                                  uint8_t chip_id,
                                  uint32_t address,
                                  uint16_t len_data,
                                  uint8_t* data);

/**
 * @brief Write a single byte.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param tw_ptr
 *   Two-wire configuration.
 * @param chip_id
 *   I2C chip ID of the target device.
 * @param address
 *   Address/command to send to target. Number of bits in address configured by
 *   @ref EdtTwoWire.address_length.
 * @param data
 *   Byte to send to target.
 *
 * @returns Zero if successful.
 */
EDTAPI int edt_two_wire_write_byte(EdtDev edt_p,
                                   const EdtTwoWire* tw_ptr,
                                   uint8_t chip_id,
                                   uint32_t address,
                                   uint8_t data);

/**
 * @brief Write to a two-wire I2C or SPI device.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device_id
 *   7-bit I2C chip id, or address extension for SPI.
 * @param address
 *   Target address/command.
 * @param value
 *   Data to write to the target.
 */
EDTAPI void edt_serial_dev_reg_write(EdtDev edt_p,
                                     uint32_t base_desc,
                                     uint8_t device_id,
                                     uint8_t address,
                                     uint8_t value);

/**
 * @brief Read from a two-wire I2C or SPI device.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device_id
 *   7-bit I2C chip id, or address extension for SPI.
 * @param address
 *   Target address/command.
 *
 * @return Data read from target.
 */
EDTAPI uint8_t edt_serial_dev_reg_read(EdtDev edt_p,
                                       uint32_t base_desc,
                                       uint8_t device_id,
                                       uint8_t address);

/**
 * @brief Block read from a two-wire I2C or SPI device.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device_id
 *   7-bit I2C chip id, or address extension for SPI.
 * @param address
 *   Target address/command.
 * @param length
 *   Number of bytes to read from target device.
 * @param[out] block
 *   Array of data read from target. Must be at least @ref length bytes.
 *
 * @return Number of bytes read from target.
 */
EDTAPI uint16_t edt_serial_dev_reg_read_block(EdtDev edt_p,
                                              uint32_t base_desc,
                                              uint8_t device_id,
                                              uint8_t address,
                                              uint16_t length,
                                              uint8_t* block);

/**
 * @brief Block write to a two-wire I2C or SPI device.
 *
 * Address in @ref address will be incremented and sent before each data byte.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device_id
 *   7-bit I2C chip id, or address extension for SPI.
 * @param address
 *   Target address/command.
 * @param length
 *   Number of bytes to write to target device.
 * @param block
 *   Data to send. Must be at least @ref length bytes.
 *
 * @return Number of bytes sent to target.
 */
EDTAPI uint16_t edt_serial_dev_reg_write_block(EdtDev edt_p,
                                               uint32_t base_desc,
                                               uint8_t device_id,
                                               uint8_t address,
                                               uint16_t length,
                                               uint8_t* block);

/**
 * @brief Read/modify/write bits in a data byte on target device.
 *
 * Useful to set specific bits in a bitfield of a target device register.
 * For example, to set a value of 3 to a 2-bit field with an offset of 4 bits,
 * pass 3 to @ref value, 4 to @ref offset, and 0x3 to @ref mask.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device_id
 *   7-bit I2C chip id, or address extension for SPI.
 * @param address
 *   Target address/command.
 * @param value
 *   Data value to write in the bitfield.
 * @param offset
 *   Bit offset in the target register. @ref value will be left-shifted by this
 *   amount.
 * @param mask
 *   Bit mask to clear in byte read from target. Typically indicates the width
 *   of the bitfield. The mask will be shifted by @ref offset.
 *
 * @return Copy of the value written to the target.
 */
EDTAPI uint8_t edt_serial_dev_set_bits(EdtDev edt_p,
                                       uint32_t base_desc,
                                       uint8_t device_id,
                                       uint8_t address,
                                       uint8_t value,
                                       uint8_t offset,
                                       uint8_t mask);

/**
 * Basic descriptor for registers.
 * This code will also work for standard edt registers as well.
 */
typedef struct {
  /** First register address of the bit field. */
  uint32_t reg;
  /** Least-significant bit offset of the field in the register word.
   * Note that if the field spans multiple registers, this should be the offset
   * in the last register.
   */
  uint8_t bitstart;
  /** Number of bits in the field. This can span multiple bytes. */
  uint8_t nbits;
  /** Name of the bitfield. Shown by @ref edt_two_wire_reg_dump() .*/
  const char* name;
  uint8_t offset;
} EdtRegisterDescriptor;

/**
 * @brief Read arbitrary value up to 32-bits wide from an 8-bit interface.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param map
 *   Descriptor of register to read from.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device
 *   7-bit I2C chip id, or address extension for SPI.
 *
 * @returns Data value read from target device.
 */
EDTAPI uint32_t edt_get_two_wire_value(EdtDev edt_p,
                                       const EdtRegisterDescriptor* map,
                                       uint32_t base_desc,
                                       uint8_t device);

/**
 * @brief Write arbitrary value up to 32-bits wide.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param map
 *   Descriptor of register to read from.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device
 *   7-bit I2C chip id, or address extension for SPI.
 * @param value
 *   Data value to write to target.
 *
 * @returns Zero if successful.
 */
EDTAPI int edt_set_two_wire_value(EdtDev edt_p,
                                  const EdtRegisterDescriptor* map,
                                  uint32_t base_desc,
                                  uint32_t device,
                                  uint32_t value);

/**
 * @brief Read all registers included in a given register map and print to STDOUT.
 *
 * The entire 8-bit data value read from the target registers is shown. The
 * field offsets and shifts specified in the register map are ignored.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param map
 *   Array of register descriptors to read from. Last element in the map MUST
 *   be all zeros to indicate the end of the array.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device
 *   7-bit I2C chip id, or address extension for SPI.
 *
 * @returns Zero if successful.
 */
EDTAPI int edt_two_wire_reg_dump_raw(EdtDev edt_p,
                                     const EdtRegisterDescriptor* map,
                                     uint32_t base_desc,
                                     uint32_t device);
/**
 * @brief Read register fields from a register map and print to STDOUT.
 *
 * Only shows the values of the register fields defined in the provided
 * register map.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param map
 *   Array of register descriptors to read from. Last element in the map MUST
 *   be all zeros to indicate the end of the array.
 * @param base_desc
 *   The base address of the FPGA two-wire interface.
 * @param device
 *   7-bit I2C chip id, or address extension for SPI.
 *
 * @returns Zero if successful.
 */
EDTAPI int edt_two_wire_reg_dump(EdtDev edt_p,
                                 const EdtRegisterDescriptor* map,
                                 uint32_t base_desc,
                                 uint32_t device);

#ifdef __cplusplus
}
#endif

#endif /* _LIB_TWO_WIRE_H */
