/* SPDX-License-Identifier: LicenseRef-NI-PL */

/**
 * @file
 *
 * @brief The EDT-specific library to interact with Camera Link serial ports
 *     on EDT VisionLink devices.
 *
 * The "clallserial.dll" library, defined by the Camera Link (tm) standard,
 * will find and load this EDT-specific library. Users should be interacting
 * with "clallserial.dll", not this library.
 *
 * ---------------------------------------------------------------------------
 *  (clallser.h, created 12/03/2003 @ 14:27:00)
 *
 *  Copyright (c) 2004. National Instruments Corporation
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted, provided that each of the following
 *  conditions are met. By using the software in any manner, you agree
 *  to each of the following:
 *
 *    - All redistributions of the software must be accompanied with the
 *      above copyright notice (provided however that for redistributions
 *      in binary form, the copyright notice may be omitted), the above
 *      preamble, this list of conditions, and the disclaimer set forth below.
 *
 *    - Except for the copyright notice required above, neither the name
 *      or trademarks of National Instruments Corporation (NI) nor the names
 *      of its contributors may be used in any manner (including, but not
 *      limited to, using the same to endorse or promote products derived from
 *      this software) without the specific prior written permission of NI.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" WITHOUT WARRANTY OF ANY KIND. NO WARRANTIES, EITHER EXPRESS
 *  OR IMPLIED, ARE MADE WITH RESPECT TO THE SOFTWARE, INCLUDING, BUT
 *  NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 *  FOR A PARTICULAR PURPOSE, TITLE OR NON-INFRINGEMENT, OR ANY OTHER
 *  WARRANTIES THAT MAY ARISE FROM USAGE OF TRADE OR COURSE OF DEALING.
 *  THE COPYRIGHT HOLDERS AND CONTRIBUTORS DO NOT WARRANT, GUARANTEE, OR
 *  MAKE ANY REPRESENTATIONS REGARDING THE USE OF OR THE RESULTS OF THE
 *  USE OF THE SOFTWARE IN TERMS OF CORRECTNESS, ACCURACY, RELIABILITY,
 *  OR OTHERWISE AND DO NOT WARRANT THAT THE OPERATION OF THE SOFTWARE
 *  WILL BE UNINTERRUPTED OR ERROR FREE.  THE ENTIRE RISK AS TO THE
 *  PERFORMANCE OF THE SOFTWARE IS WITH YOU. IN NO EVENT SHALL THE
 *  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 *  THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * MODIFIED 2023 by Engineering Design Team, Inc.
 */
#ifndef CLSEREDT_H
#define CLSEREDT_H

//==============================================================================
//  Macros and Constants
//==============================================================================

#if (defined(__cplusplus) || defined(__cplusplus__))
extern "C" {
#endif

#if defined _WIN32 || defined __CYGWIN__
#define CLALLSERIAL_IMPORT __declspec(dllimport)
#define CLALLSERIAL_EXPORT __declspec(dllexport)
#elif defined(__GNUC__) && (__GNUC__ >= 4)
#define CLALLSERIAL_IMPORT __attribute__((visibility("default")))
#define CLALLSERIAL_EXPORT __attribute__((visibility("default")))
#else
#define CLALLSERIAL_IMPORT
#define CLALLSERIAL_EXPORT
#endif

#if defined(CLSEREDT_EXPORTS)
#define CLSEREDT_API CLALLSERIAL_EXPORT
#elif defined(CLSEREDT_STATIC)
#define CLSEREDT_API
#else
#define CLSEREDT_API CLALLSERIAL_IMPORT
#endif

#ifndef CLALLSERIALCC
#if defined _WIN32 || defined __CYGWIN__
#define CLALLSERIALCC __stdcall
#else
#define CLALLSERIALCC
#endif /* if Windows */
#endif /* if CLALLSERIALCC not defined*/

#ifndef _CL_hSerRef_DEFINED_
#define _CL_hSerRef_DEFINED_
/** Generic pointer type */
typedef void* hSerRef;
#endif

#ifndef _CL_INT32_DEFINED_
#define _CL_INT32_DEFINED_
/** (signed) 32-bit integer type */
typedef int CLINT32;
#endif

#ifndef _CL_UINT32_DEFINED_
#define _CL_UINT32_DEFINED_
/** (unsigned) 32-bit integer type */
typedef unsigned int CLUINT32;
#endif

#ifndef _CL_INT8_DEFINED_
#define _CL_INT8_DEFINED_
/** 8-bit integer type */
typedef char CLINT8;
#endif

#ifndef CL_DLL_VERSION_NO_VERSION
//------------------------------------------------------------------------------
//  CL serial DLL versions
//------------------------------------------------------------------------------
#define CL_DLL_VERSION_NO_VERSION 1  // Not a CL dll
#define CL_DLL_VERSION_1_0        2  // Oct 2000 compliant
#define CL_DLL_VERSION_1_1        3  // Oct 2001 compliant
#endif

//------------------------------------------------------------------------------
//  Baud Rates
//------------------------------------------------------------------------------
#ifndef CL_BAUDRATE_9600
#define CL_BAUDRATE_9600   (1U)
#define CL_BAUDRATE_19200  (1U << 1U)
#define CL_BAUDRATE_38400  (1U << 2U)
#define CL_BAUDRATE_57600  (1U << 3U)
#define CL_BAUDRATE_115200 (1U << 4U)
#define CL_BAUDRATE_230400 (1U << 5U)
#define CL_BAUDRATE_460800 (1U << 6U)
#define CL_BAUDRATE_921600 (1U << 7U)
#endif

//------------------------------------------------------------------------------
//  Error Codes
//------------------------------------------------------------------------------
#ifndef CL_ERR_NO_ERR
/** No error. */
#define CL_ERR_NO_ERR 0
/** Size of the buffer passed in is not large enough. */
#define CL_ERR_BUFFER_TOO_SMALL (-10001)
/** Requested manufacturer DLL does not exist. */
#define CL_ERR_MANU_DOES_NOT_EXIST (-10002)
/** Requested port is currently in use. */
#define CL_ERR_PORT_IN_USE (-10003)
/** Operation timed out. */
#define CL_ERR_TIMEOUT (-10004)
/** Index is not valid. */
#define CL_ERR_INVALID_INDEX (-10005)
/** Serial reference (pointer) is not valid. */
#define CL_ERR_INVALID_REFERENCE (-10006)
/** A description for the provided error code was not found. */
#define CL_ERR_ERROR_NOT_FOUND (-10007)
/** Baud rate is unsupported. */
#define CL_ERR_BAUD_RATE_NOT_SUPPORTED (-10008)
/** A memory allocation failed. */
#define CL_ERR_OUT_OF_MEMORY (-10009)
/** Could not find the cameralink registry key in (Windows) system registry. */
#define CL_ERR_REGISTRY_KEY_NOT_FOUND (-10010)
/** Pointer is invalid. */
#define CL_ERR_INVALID_PTR (-10011)
/** Could not load a DLL. Potential reasons include missing API functions or a lack of memory. */
#define CL_ERR_UNABLE_TO_LOAD_DLL (-10098)
/** Function does not exists in the manufacturer's library. */
#define CL_ERR_FUNCTION_NOT_FOUND (-10099)

// EDT-specific errors

/** EDT board has not been initalized yet. */
#define CL_ERR_EDT_NOT_INITIALIZED (-12301)
#endif

//==============================================================================
// Imported functions, clseredt.dll version
//==============================================================================

/**
 * @brief Discard any bytes that are available in the serial input buffer.
 *
 * @param[in] serialRef
 *     The value from @ref clSerialInit that describes the port to be flushed.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clFlushPort(hSerRef serialRef);

/**
 * @brief Converts an error code to error text.
 *
 * @param[in] mfgName
 *     The manufacturer name in a NULL-terminated buffer, as returned from
 *     @ref clGetPortInfo.
 * @param[in] errorCode
 *     The error code to get a text description of.
 * @param[out] errorText
 *     A caller-allocated buffer which returns the NULL-terminated error text.
 * @param[in,out] errorTextSizeconst
 *     On success, contains the number of bytes written into the buffer,
 *     including the NULL-termination character. This value should be the size
 *     in bytes of the error text buffer passed in. On CL_ERR_BUFFER_TOO_SMALL,
 *     contains the size of the buffer needed to write the error text.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clGetErrorText(const CLINT8* mfgName,
                                    CLINT32 errorCode,
                                    CLINT8* errorText,
                                    CLUINT32* errorTextSizeconst);

/**
 * @brief An alias of @ref clGetNumSerialPorts()
 *
 * Due to inconsistency in the spec between VB and C++ versions, this function
 * was added to the clser EDT library. It simply calls @ref clGetNumSerialPorts().
 *
 * @extends clGetNumSerialPorts
 */
CLSEREDT_API CLINT32 clGetNumPorts(CLUINT32* numPorts);

/**
 * @brief Get the number of Camera Link serial ports supported by this (EDT) library.
 *
 * @note (EDT library) Always assumes base mode and returns 2 ports per board.
 *     If medium/full mode cameras are in use, only the even-numbered serial
 *     port indexes should be used.
 *
 * @param[out] numPorts
 *     Number of serial supported serial ports.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clGetNumSerialPorts(CLUINT32* numPorts);

/**
 * @brief Get the number of bytes available in the serial RX buffer.
 *
 * @param[in] serialRef
 *     Serial port reference pointer. @see clSerialInit()
 * @param[out] numBytes
 *     Number of bytes avaliable to be read.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clGetNumBytesAvail(hSerRef serialRef, CLUINT32* numBytes);

/**
 * @brief Get information about the serial port.
 *
 * @param[in] serialIndex
 *     Index of the serial port to query. Should be in the range of 0 to N-1,
 *     where N is the value returned by @ref clGetNumSerialPorts().
 * @param[out] manufacturerName
 *     Pointer of a caller-allocated buffer which this function will write
 *     the NULL-terminated manufacturer name.
 * @param[in,out] nameBytes
 *     Pass in the size of the @ref nameBytes buffer. Returns the number of
 *     bytes written to the buffer, including the NULL terminating character.
 *     If #CL_ERR_BUFFER_TOO_SMALL is returned, this will be set to the buffer
 *     size that would be needed to write all the data.
 * @param[out] portID
 *     EDT-specific identifier of the serial port.
 * @param[in,out] IDBytes
 *     Pass in the size of the @ref portID buffer. Returns the number of bytes
 *     written io the buffer, including the NULL terminating character.
 *     If #CL_ERR_BUFFER_TOO_SMALL is returned, this will be set to the buffer
 *     size that would be needed to write all the data.
 * @param[out] version
 *     Version of the Camera Link(tm) specification which this library is
 *     compatible with.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clGetPortInfo(CLUINT32 serialIndex,
                                   CLINT8* manufacturerName,
                                   CLUINT32* nameBytes,
                                   CLINT8* portID,
                                   CLUINT32* IDBytes,
                                   CLUINT32* version);

/**
 * @brief Get a list of baud rates that are valid for a given serial interface.
 *
 * @param[in] serialRef
 *     Serial port reference pointer. @see clSerialInit()
 * @param[out] baudRates
 *     Supported baud rates, represented as a bitfield. Refer to the
 *     `CL_BAUDRATE_*` defines.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clGetSupportedBaudRates(hSerRef serialRef, CLUINT32* baudRates);

/**
 * @brief Close the serial interface and free associated resources. Upon return
 *     @ref serialRef is no longer usable.
 *
 * @param[in] serialRef
 *     Serial port reference pointer. @see clSerialInit()
 */
CLSEREDT_API void clSerialClose(hSerRef serialRef);

/**
 * @brief Initialize the device indicated by the @ref serialIndex and return
 *     a pointer handle to the serial port.
 *
 * @param[in] serialIndex
 *     Index of the serial port to open. Should be in the range of 0 to N-1,
 *     where N is the value returned by @ref clGetNumSerialPorts().
 *     The order of serial devices is vendor-specific.
 * @param[out] serialRefPtr
 *     If successful, this will return a handle to the serial device.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clSerialInit(CLUINT32 serialIndex, hSerRef* serialRefPtr);

/**
 * @brief Read from the serial device.
 *
 * @param[in] serialRef
 *     Serial port reference pointer. @see clSerialInit()
 * @param[out] buffer
 *     A caller-allocated buffer. The data read from the serial port will be
 *     written to this buffer if the operation is successful.
 * @param[in,out] bufferSize
 *      Pass in the size, in bytes, of @ref buffer. If the read operation is
 *      successful, the actual number of bytes written into @ref buffer is returned.
 * @param[in] serialTimeout
 *     Indicates the read operation timeout, in milliseconds.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clSerialRead(hSerRef serialRef,
                                  CLINT8* buffer,
                                  CLUINT32* bufferSize,
                                  CLUINT32 serialTimeout);
/**
 * @brief Write data to the serial device.
 *
 * @param[in] serialRef
 *     Serial port reference pointer. @see clSerialInit()
 * @param[in] buffer
 *     Data to write to the serial port.
 * @param[in,out] bufferSize
 *     Pass in the size, in bytes, of @ref buffer. If the write operation is
 *     successful, this contains the actual number of bytes written.
 * @param[in] serialTimeout
 *     Indicates the write operation timeout, in milliseconds.
 *
 * @return #CL_ERR_NO_ERR if successful.
 */
CLSEREDT_API CLINT32 clSerialWrite(hSerRef serialRef,
                                   CLINT8* buffer,
                                   CLUINT32* bufferSize,
                                   CLUINT32 serialTimeout);
/**
 * @brief Set the baud rate of the serial port.
 *
 * Query the baud rates supported by this port by calling
 * @ref clGetSupportedBaudRates().
 *
 * For example:
 *
 * @code{.c}
 *     hSerRef ser = NULL;
 *     clSerialInit(0, &ser);
 *
 *     uint32_t supported_baud_rates = 0;
 *     clGetSupportedBaudRates(ser, &supported_baud_rates)
 *
 *     if (CL_BAUDRATE_115200 & supported_baud_rates)
 *         clSetBaudRate(ser, CL_BAUDRATE_115200);
 *     else if (CL_BAUDRATE_9600 & supported_baud_rates)
 *         clSetBaudRate(ser, CL_BAUDRATE_9600);
 * @endcode
 *
 * @param[in] serialRef
 *     Serial port reference pointer. @see clSerialInit()
 * @param[in] baudRate
 *     The baud rate to set. Should be one of the `CL_BAUDRATE_*` defines.
 * @return CLSEREDT_API
 */
CLSEREDT_API CLINT32 clSetBaudRate(hSerRef serialRef, CLUINT32 baudRate);

#if (defined(__cplusplus) || defined(__cplusplus__))
}
#endif

#endif /* CLSEREDT_H */
