/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 1997 Engineering Design Team, Inc.
 * @brief EDT digital imaging utilities library header file
 */

#ifndef LIBDVU_H
#define LIBDVU_H

#include "edt_portability.h"

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Writes a 24-bit per pixel RGB data buffer as a raw data file (no
 * formatting).
 *
 * @param imagesize
 *    Number of bytes in the image.
 * @param imagebuf
 *    Pointer to image data buffer.
 * @param fname
 *    Output file name.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_dvu_write_raw(int imagesize, uint8_t* imagebuf, char* fname);

/**
 * @brief Writes an 8-bit per pixel data buffer as a grayscale Windows bitmap file.
 *
 * @param fname
 *    Filename.
 * @param buffer
 *    Data buffer, one byte per pixel.
 * @param width
 *    Number of pixels (bytes) per line.
 * @param height
 *    Number of lines in the image.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see #edt_dvu_write_bmp24
 */
EDTAPI int edt_dvu_write_bmp(char* fname, uint8_t* buffer, int w, int h);

/**
 * @brief Writes a 24-bit per pixel RGB data buffer as a Windows bitmap file.
 *
 * @param fname
 *    Filename.
 * @param buffer
 *    Data buffer, one byte per pixel.
 * @param width
 *    Number of pixels (bytes) per line.
 * @param height
 *    Number of lines in the image.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see #edt_dvu_write_bmp
 */
EDTAPI int edt_dvu_write_bmp24(char* fname, uint8_t* buffer, int w, int h);

/**
 * @brief Utility routine that outputs a 1-band, 8-bit image to a Sun raster format
 * file (regardless of platform), with stride. This function can be used to
 * output a partial image. For example, to output the top left 100x100 pixels of
 * the image, specify \a x_size of 100 and \a y_size * 100, and an \a istride of
 * the actual width of the image.
 *
 * @note Use #edt_dvu_write_rasfile to output a full image in the most efficient way.
 *
 * @param fname
 *    The name of the output file.
 * @param addr
 *    The address of the image data (8 bits per pixel).
 * @param x_size
 *    Width in pixels of image.
 * @param y_size
 *    Height in pixels of image.
 * @param istride
 *    Number of pixels (bytes) to skip from one the beginning of one line
 * to the beginning of the next (this should just be the image width, unless you
 * want something weird like a diagonally skewed image).
 *
 * @return
 *	 0 on success, -1 on failure
 */
EDTAPI int edt_dvu_write_image(char* fname, uint8_t* addr, int x_size, int y_size, int istride);

/**
 * @brief Writes a 24-bit per pixel RGB data buffer as a Sun Raster format file, with
 * stride.
 *
 * @note To output a full image in the most efficient way, use #edt_dvu_write_rasfile24.
 *
 * @param fname
 *    The name of output file.
 * @param addr
 *    Data buffer, three bytes per pixel (RGB).
 * @param x_size
 *    Number of pixels per line.
 * @param y_size
 *    Number of lines in the image.
 * @param istride
 *    Number of pixels to skip between sucessive lines.
 *
 * @return
 *   0 on success, -1 on failure
 */
EDTAPI int edt_dvu_write_image24(char* fname, uint8_t* addr, int x_size, int y_size, int istride);

/**
 * @brief Utility routine that outputs a 1-band, 8-bit image to a Sun raster format
 * file (regardless of platform).
 *
 * @param fname
 *    The name of the output file.
 * @param addr
 *    The address of the image data (8 bits per pixel).
 * @param x_size
 *    Width in pixels of image.
 * @param y_size
 *    Height in pixels of image.
 *
 * @return
 *	 0 on success, -1 on failure
 */
EDTAPI int edt_dvu_write_rasfile(char* fname, uint8_t* addr, int x_size, int y_size);

/**
 * @brief Converts 1 band, 10-16 bit image to a sun raster format file and
 * writes to a file.
 *
 * @param fname
 *    The name of the output file.
 * @param addr
 *    The address of the image data (8 bits per pixel).
 * @param x_size
 *    Width in pixels of image.
 * @param y_size
 *    Height in pixels of image.
 * @param depth_bits
 *    Number of bits per pixel
 *
 * @return
 *	 0 on success, -1 on failure
 */
EDTAPI int edt_dvu_write_rasfile16(char* fname, uint8_t* addr, int x_size, int y_size, int depth);

/**
 * @brief Writes a 24-bit per pixel RGB data buffer as a Sun Raster format file.
 *
 * @param fname
 *    File name.
 * @param addr
 *    Data buffer, three bytes per pixel (RGB).
 * @param x_size
 *    Number of pixels per line.
 * @param y_size
 *    Number of lines in the image.
 *
 * @return
 *	 0 on success, -1 on failure
 */
EDTAPI int edt_dvu_write_rasfile24(char* fname, uint8_t* addr, int x_size, int y_size);

EDTAPI int edt_dvu_word_to_byte(uint16_t* wbuf, uint8_t* bbuf, int count, int shift);

EDTAPI void edt_dvu_long_to_charbuf(uint32_t val, uint8_t* buf);

#ifdef __cplusplus
}
#endif

#endif /* LIBDVU_H */
