/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 * @brief EDT PCI device library.
 *
 * Provides a 'C' language interface to the EDT PCI DMA cards to simplify the
 * ring buffer method of reading data. All routines access a specific device,
 * whose handle is created and returned by the @ref edt_open() routine.
 */

#ifndef INCLUDE_edtlib_h
#define INCLUDE_edtlib_h

#include "edt_portability.h"
#include "edt_os_utils.h"

#include "driver/edt_pci_devices.h"
#include "driver/edt_shared_types.h"

#include <stdint.h>
#include <stdbool.h>

#if defined(__linux__)
#include <pthread.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef DOXYGEN_SHOW_UNDOC
/**
 * @internal
 * @defgroup edt_undoc Undocumented / Uncategorized
 */

/**
 * @weakgroup dma
 * @{
 */
#endif

/*
 * Try to standardize on size of various strings
 */
#define EDT_STRBUF_SIZE  128
#define EDT_PATHBUF_SIZE 256
#define EDT_BIGBUF_SIZE  512

#define MAX_DMA_BUFFERS 2048

/**
 * @brief Structure to set up phase locked loop parameters.
 *
 * Some of these values go into the AV9110-02 PLL chip
 * and others set up pre-scalars in the FPGA.
 */
typedef struct {
  int m; /**< AV9110 reference frequency divide range 3-127. */
  int n; /**< AV9110 VCO feedback frequency divide range 3-127. */
  int v; /**< AV9110 VCO feedback frequency pre-scalar range 1 or 8. */
  int r; /**< AV9110 VCO output divider 1, 2, 4, 8. */
  int h; /**< Xilinx high speed divider (VCO output) 1, 3, 5, 7. */
  int l; /**< Xilinx divide by n 1-64. */
  int x; /**< Xilinx AV9110 pre-scale of 30 MHz oscillator 1-256. */
} edt_pll;

/** Flag for @ref edt_configure_ring_buffers() indicates data will be
 * transferred from the EDT card to the Host. */
#define EDT_READ 0
/** Flag for @ref edt_configure_ring_buffers() indicates data will be
 * transferred from the Host to the EDT card. */
#define EDT_WRITE 1

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

/** Event callback function type. */
typedef void (*EdtEventFunc)(void*);

/** Event handler configuration. */
typedef struct edt_event_handler {
  struct edt_event_handler* next; /**< In case we want linked lists. */
  EdtEventFunc callback;          /**< Function to call when event occurs. */
  struct edt_device* owner;       /**< Reflexive pointer. */
  uint32_t event_type;
  void* data;         /**< Closure pointer for callback. */
  uint8_t active;     /**< Flag for a graceful death. */
  uint8_t continuous; /**< Flag for continuing events. */
#if defined(_WIN32) || defined(__DOXYGEN__)
  thread_t wait_thread;  /**< Thread handle. */
  event_t wait_event;    /**< Kernel event handle. */
  event_t closing_event; /**< Abort event handle. */
  uint32_t thrdid;       /**< Thread id number. */
#endif
#if defined(__linux__) || defined(__DOXYGEN__)
  pthread_t thrdid; /**< Thread id number. */
#endif

} EdtEventHandler;

/** EDT PCI board information. @see @ref edt_detect_boards(). */
typedef struct {
  char type[8];           /**< Family e.g. "pcd". */
  int id;                 /**< Board unit number. */
  enum EDT_PCI_PID bd_id; /**< Board PCI product id. */
  int promcode;           /**< FPGA flash chip type e.g. AMD_XC5VLX30T */
} EdtBoardInfo;

enum EdtBitfileOptionBoardType {
  EDT_UNKNOWN_TYPE = 0,
  EDT_SS_TYPE = 1,
  EDT_GS_TYPE = 2,
  EDT_CD_TYPE = 3,
  EDT_LX_TYPE = 4,
  EDT_LC_TYPE = 5
};

typedef struct edt_option_string_fields {
  enum EdtBitfileOptionBoardType board_type;
  int DMA_channels;
#if defined(PCD)
  char mezzanine_type[68];
#endif
  char filename[68];
  int version_number;
  int rev_number;
  char date[50];
  int custom_DMA_channels;
  int available_DMA_channels;
} EdtOptionStringFields;

typedef struct edt_bitfile_descriptor {
  edt_bitpath bitfile_name;
#if defined(PCD)
  edt_bitpath mezz_name0;
  edt_bitpath mezz_name1;
#endif
  char optionstr[68];
#if defined(PCD)
  char mezz_optionstr0[32];
  char mezz_optionstr1[32];
#endif

  EdtOptionStringFields ostr;

  /** Value of the EDT_BITFILE_VERSION register. */
  int revision_register;
  int string_type;

} EdtBitfileDescriptor;

/**
 * @brief Values that represent a port on a card. This is to enforce typed
 * access in C++.
 */
typedef enum edt_io_port {
  EDT_NO_PORT = -1,
  EDT_IO_PORT_0 = 0,
  EDT_IO_PORT_1 = 1,
  EDT_IO_PORT_2 = 2,
  EDT_IO_PORT_3 = 3,
  EDT_IO_PORT_4 = 4,
  EDT_IO_PORT_5 = 5,
  EDT_IO_PORT_6 = 6,
  EDT_IO_PORT_7 = 7,
  EDT_IO_PORT_8 = 8
} EdtIOPort;

/**
 * @brief EDT device handle (an opaque pointer).
 *
 * @see @ref edt_open(), @ref edt_open_device(), @ref edt_close(),
 * @ref docgroup_libedt_openclose
 */
typedef struct edt_device* EdtDev;
#define EDT_DEV_TYPE_DEFINED 1

/**
 * @addtogroup docgroup_libedt_openclose
 * @{
 */

/**
 * @brief Shut down all pending I/O operations, closes the device or channel
 * and frees all resources associated with the device handle.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure. If an error occurs, call
 * @ref edt_perror() to get the system error message.
 *
 * @see @ref edt_open_device()
 */
EDTAPI int edt_close(EdtDev edt_p);

/**
 * @brief Open the specified EDT board and sets up the device handle. Equivalent
 * to `edt_open_device(device_name, unit, channel: 0, verbose: 1)`.
 *
 * @param device_name
 *   Pass @ref EDT_INTERFACE. Open a dummy device for testing by passing "dmy".
 * @param unit
 *   Unit number of the device. Unit numbers start at 0.
 *
 * @return The opened EDT device handle if successful, or NULL if unsuccessful.
 * Use @ref edt_perror() to print an error message.
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(edt_close)
EDTAPI EdtDev edt_open(const char* device_name, int unit);

/**
 * @brief Same as @ref edt_open(), but with fewer log messages. Equivalent
 * to `edt_open_device(device_name, unit, channel: 0, verbose: 0)`.
 *
 * @param device_name
 *   Pass @ref EDT_INTERFACE. Open a dummy device for testing by passing "dmy".
 * @param unit
 *   Unit number of the device. Unit numbers start at 0.
 *
 * @return The opened EDT device handle if successful, or NULL if unsuccessful.
 * Use @ref edt_perror() to print an error message.
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(edt_close)
EDTAPI EdtDev edt_open_quiet(const char* device_name, int unit);

/**
 * @brief Open a specific DMA channel on the specified EDT board. Equivalent
 * to `edt_open_device(device_name, unit, channel, verbose: 1)`.
 *
 * @param device_name
 *   Pass @ref EDT_INTERFACE. Open a dummy device for testing by passing "dmy".
 * @param unit
 *   Unit number of the device. Unit numbers start at 0.
 * @param channel
 *   DMA channel number, starting from 0.
 *
 * @return The opened EDT device handle if successful, or NULL if unsuccessful.
 * Use @ref edt_perror() to print an error message.
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(edt_close)
EDTAPI EdtDev edt_open_channel(const char* device_name, int unit, int channel);

/**
 * @brief Open a specific EDT board and DMA channel.
 *
 * @param device_name
 *   Pass @ref EDT_INTERFACE. Open a dummy device for testing by passing "dmy".
 * @param unit
 *   Unit number of the device. Unit numbers start at 0.
 * @param channel
 *   DMA channel number, starting from 0.
 * @param verbose
 *   Enable (true / non-zero) messages on successful open, or error messages on
 *   failure. Pass false or 0 to disable.
 *
 * @return The opened EDT device handle if successful, or NULL if unsuccessful.
 * Use @ref edt_perror() to print an error message.
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(edt_close)
EDTAPI EdtDev edt_open_device(const char* device_name, int unit, int channel, int verbose);

/**
 * @brief Set the @p port number, as distinct from the DMA channel.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param port
 *   Port number enum.
 *
 * @see @ref edt_get_port(), @ref edt_device_channel()
 */
EDTAPI void edt_set_port(EdtDev edt_p, EdtIOPort port);

/**
 * @brief Get the port number set by @ref edt_set_port().
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Port number enum.
 */
EDTAPI EdtIOPort edt_get_port(EdtDev edt_p);

/** @} */ /* end docgroup_libedt_openclose */

/** @addtogroup dma_inout
 * @{
 */

/**
 * @brief Perform a DMA read from the EDT device.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param buf
 *   Buffer to read data in to.
 * @param size
 *   Number of bytes to read. @p buf must be equal to or greater than this size.
 *
 * @returns Number of bytes read if successful. -1 on failure.
 */
EDTAPI int edt_read(EdtDev edt_p, void* buf, unsigned int size);

/**
 * @brief Perform a DMA write to the EDT device.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param buf
 *   Data to write to the device.
 * @param size
 *   Number of bytes to write. @p buf must be equal to or greater than this size.
 *
 * @returns Number of bytes written if successful. -1 on failure.
 */
EDTAPI int edt_write(EdtDev edt_p, void* buf, unsigned int size);

EDTAPI void edt_set_buffer_granularity(EdtDev edt_p, unsigned int granularity);

/**
 * @brief Prepare and configure ring buffers for DMA to/from the EDT device.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bufsize
 *   Size of each buffer, in bytes.
 * @param numbufs
 *   Number of buffers. Must be 1 or greater. Four is recommended for most
 *   applications.
 * @param write_flag
 *   Whether this DMA channel is configured for input (read from device) or
 *   output (write to device). Pass either @ref EDT_READ or @ref EDT_WRITE.
 *
 * @return 0 on success, -1 on failure. Get system error message with @ref edt_perror()
 */
EDTAPI int edt_configure_ring_buffers(EdtDev edt_p,
                                      unsigned int bufsize,
                                      unsigned int numbufs,
                                      unsigned int write_flag);

/**
 * @brief Prepare and configure ring buffers for DMA, allocating buffers as a
 * single large block, setting the buffer addresses from within that block.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bufsize
 *   Size of individual buffers, in bytes
 * @param numbufs
 *   Number of buffers to create in the ring.
 * @param write_flag
 *   Either @ref EDT_WRITE or @ref EDT_READ.
 * @param header_size
 *   If non-zero, additional memory will be allocated for each buffer for
 *   Header data. The location of this header region is determined by the
 *   argument @p header_before.
 * @param header_before
 *   If non-zero, the header region is placed before the DMA buffer. Otherwise
 *   it comes after the DMA buffer.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_configure_block_buffers(EdtDev edt_p,
                                       int bufsize,
                                       int numbufs,
                                       int write_flag,
                                       int header_size,
                                       int header_before);

/**
 * @brief Register a callback to populate the ring buffer with data when
 * @ref edt_wait_for_buffers() is called.
 *
 * @param edt_p
 *   The EDT device handle, opened as a dummy device.
 * @param callback
 *   Function to invoke when @ref edt_wait_for_buffers() is called.
 */
EDTAPI void edt_set_dmy_wait_for_buffers_callback(EdtDev edt_p,
                                                  void (*callback)(EdtDev edt_p, void* buf));

/**
 * @brief Register a callback to read a simulated register when
 * @ref edt_reg_read() is called.
 *
 * @param edt_p
 *   The EDT device handle, opened as a dummy device.
 * @param callback
 *   Function to invoke when @ref edt_reg_read() is called.
 */
EDTAPI void edt_set_dmy_reg_read_callback(EdtDev edt_p,
                                          uint32_t (*callback)(EdtDev edt_p, uint32_t reg_desc));

/**
 * @brief Register a callback to write a simulated register when
 * @ref edt_reg_write() is called.
 *
 * @param edt_p
 *   The EDT device handle, opened as a dummy device.
 * @param callback
 *   Function to invoke when @ref edt_reg_write() is called.
 */
EDTAPI void edt_set_dmy_reg_write_callback(EdtDev edt_p,
                                           void (*callBack)(EdtDev edt_p,
                                                            uint32_t reg_desc,
                                                            uint32_t reg_value));

/**
 * @brief Memory-map the entire DMA channel's ring of buffers.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @returns On success, a pointer to the mapped area. On failure, NULL.
 */
EDT_CHECK_RETURN
EDTAPI void* edt_map_dmamem(EdtDev edt_p);

/**
 * @brief Disable EDT device ring buffers. Pending DMA is cancelled and
 * all data buffers are released.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure. Get system error message with @ref edt_perror().
 */
EDTAPI int edt_disable_ring_buffers(EdtDev edt_p);

/**
 * @brief Get the number of buffers allocated even if by other process (for
 * monitoring from a separate call to @ref edt_open()).
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Number of allocated buffers, or -1 on failure.
 */
EDTAPI int edt_get_numbufs(EdtDev edt_p);

/**
 * @brief Stop any active DMA then set the ring buffer to restart on buffer
 * @p bufnum.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bufnum
 *   The index of the ring buffer at which to start the next DMA. A number
 *   larger than the number of buffers will set current done count to @p bufnum
 *   modulo the number of buffers.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_reset_ring_buffers(EdtDev edt_p, uint32_t bufnum);

/**
 * @brief Stop any active DMA then set the ring buffer to restart on the
 * @b current buffer.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_abort_dma(EdtDev edt_p);

/**
 * @brief Stop any active DMA then set the ring buffer to restart on the
 * @b next buffer.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_abort_current_dma(EdtDev edt_p);

/**
 * @brief Stop DMA transfer after the current buffer has completed. Ring
 * buffer mode remains active, and transfers may be resumed by calling
 * @ref edt_start_buffers().
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_stop_buffers(EdtDev edt_p);

/**
 * @brief Start DMA transfers for @p count number of buffers.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param count
 *   Number of buffers in a ring to release to the driver for transfer. A value
 *   of 0 starts "free running" mode where transfers run continuously until
 *   @ref edt_stop_buffers() is called. If @p count is greater than the number
 *   of buffers in the ring, DMA continues looping through the buffers until the
 *   total count has been satisfied.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_start_buffers(EdtDev edt_p, unsigned int count);

/**
 * @brief Change the size or direction of one buffer in a ring.
 *
 * @warning Improper use can have nasty side effects. Prefer to use separate
 * DMA channels with different size or direction configuration, instead of
 * modifying a channel's individual buffer.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param index
 *   Index of ring buffer to change.
 * @param size
 *   New buffer size to use.
 * @param write_flag
 *   Direction; either @ref EDT_WRITE or @ref EDT_READ.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_set_buffer_size(EdtDev edt_p,
                               unsigned int index,
                               uint32_t size,
                               uint32_t write_flag);

/**
 * @brief Get the allocated size of the specified buffer.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param buffer_idx
 *   Index of the buffer.
 *
 * @return The buffer size in bytes, or 0 if the specified index is
 * invalid.
 */
EDTAPI unsigned int edt_allocated_size(EdtDev edt_p, int buffer_idx);

/**
 * @brief Get the total buffer size for block of buffers, in which the memory
 * allocation size is rounded up so all buffers start on a page boundary.
 * This is used to allocate a single contiguous block of DMA buffers.
 *
 * @note The total size of a ring buffer is this value multiplied by the number of
 * buffers in the ring.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bufsize
 *   Size of a buffer.
 * @param header_size
 *   Optional size of an additional header region.
 *
 * @return Full buffer size, in bytes.
 *
 * @see @ref edt_configure_block_buffers()
 */
EDTAPI int edt_get_total_bufsize(EdtDev edt_p, int bufsize, int header_size);

/**
 * @brief Blocks until the specified number of buffers have completed.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param count
 *   Number of buffers to block for. Completed buffers are numbered relatively;
 *   start each call with 1.
 *
 * @return Address of last completed buffer on success; NULL on failure.
 */
EDTAPI void* edt_wait_for_buffers(EdtDev edt_p, int count);

/**
 * @brief Causes application-defined events to show up in the same timeline
 * as driver events when the event history is listed by running `setdebug -g`.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param val
 *   An arbitrary value meaningful to the application.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_ref_tmstamp(EdtDev edt_p, uint32_t val);

/**
 * @brief Get the seconds and nanoseconds timestamp of when DMA was completed
 * on the buffer specified by \a bufnum.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] ts
 *   Timestamp representing the system time at which the buffer completed.
 * @param bufnum
 *   Either buffer index or, if greater than number of buffers in the ring,
 *   number of buffers completed.
 *
 * @return On success, fills in timestamp @p ts and returns 0. On failure,
 * returns -1.
 */
EDTAPI int edt_get_timestamp(EdtDev edt_p, edt_timespec_t* ts, unsigned int bufnum);

/**
 * @brief Wait for the next buffer that finishes DMA. Depending on how often
 * this routine is called, buffers that have already completed DMA might
 * be skipped.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Returns a pointer to the buffer, or NULL on failure.
 */
EDTAPI void* edt_wait_for_next_buffer(EdtDev edt_p);

/**
 * @brief Like @ref edt_last_buffer() but also returns the time at which the DMA
 * was complete on this buffer.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] ts
 *   Timestamp representing the system time at which the buffer completed.
 *
 * @return A pointer to the last buffer, or NULL on failure.
 */
EDTAPI void* edt_last_buffer_timed(EdtDev edt_p, edt_timespec_t* ts);

/**
 * @brief Wait for the last buffer that has been transferred. Useful if an
 * application cannot keep up with buffer transfer. If this routine is called
 * a second time before another buffer has been transferred, it will block
 * waiting for the next transfer to complete.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return A pointer to the last buffer, or NULL on failure.
 */
EDTAPI void* edt_last_buffer(EdtDev edt_p);

/**
 * @brief Block until the specified number of buffers have completed.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param count
 *   Number of buffers for which to block. Completed buffers are numbered
 *   cumulatively starting with 0 when the EDT device is opened.
 * @param[out] ts
 *   Timestamp representing the system time at which the last buffer completed.
 *
 * @return A pointer to the last completed buffer, or NULL on failure.
 */
EDTAPI void* edt_wait_buffers_timed(EdtDev edt_p, int count, edt_timespec_t* ts);

/**
 * @brief Set which buffer should be started next. Usually done to recover after
 * a timeout, interrupt, or error.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bufnum
 *   The index of the buffer to start next.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_set_buffer(EdtDev edt_p, uint32_t bufnum);

/**
 * @brief Get a pointer to the next buffer scheduled for output DMA, in order
 * to fill the buffer with data.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return A pointer to the buffer, or NULL on failure.
 */
EDTAPI void* edt_next_writebuf(EdtDev edt_p);

/**
 * @brief Get the number of DMA buffers that have been waited for.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Number of buffers that userspace has waited for.
 */
EDTAPI uint32_t edt_dma_buffers_done_waiting(EdtDev edt_p);

/**
 * @brief Get an array of ring buffer pointers.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return An array of pointers to the ring buffers for this DMA channel.
 * The array is indexed from zero to `n-1` where `n` is the number of ring
 * buffers set in @ref edt_configure_ring_buffers().
 */
EDTAPI void** edt_buffer_addresses(EdtDev edt_p);

/**
 * @brief Returns the cumulative count of completed buffer transfers in ring
 * buffer mode.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The number of completed buffer transfers. -1 is returned if ring
 * buffer mode is not configured or some other error occurred.
 */
EDTAPI uint32_t edt_done_count(EdtDev edt_p);

/**
 * @brief Checks, without blocking, if the specified number of buffers have
 * completed.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param count
 *   Number of buffers. Must be 1 or greater.
 *
 * @return Returns the address of the ring buffer corresponding to @p count
 * if it has completed DMA, or NULL if count buffers are not yet
 * complete.
 */
EDTAPI void* edt_check_for_buffers(EdtDev edt_p, uint32_t count);

/**
 * @brief Get the value of the `PCD_DIRA` and `PCD_DIRB` registers.
 * The value from `PCD_DIRB` is shifted up 8 bits.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Combined value read from direction registers.
 */
EDTAPI uint16_t edt_get_direction(EdtDev edt_p);

/**
 * @brief Set the register and value to use at the start of DMA, as set by
 * @ref edt_startdma_action().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   Register description of which register to use as in @ref edtreg.h.
 * @param val
 *   Value to write to register described by @p desc.
 */
EDTAPI void edt_startdma_reg(EdtDev edt_p, uint32_t desc, uint32_t val);

/**
 * @brief Sets the register and value to use at the end of dma, as set by
 * @ref edt_enddma_action().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   Register description of which register to use as in @ref edtreg.h.
 * @param val
 *   Value to write to register described by @p desc.
 */
EDTAPI void edt_enddma_reg(EdtDev edt_p, uint32_t desc, uint32_t val);

/**
 * @brief Specifies when to perform the action at the start of a DMA transfer
 * as specified by @ref edt_startdma_reg().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param val
 *   One of @c EDT_ACT_NEVER, @c EDT_ACT_ONCE, or @c EDT_ACT_ALWAYS.
 */
EDTAPI void edt_startdma_action(EdtDev edt_p, uint32_t val);

/**
 * @brief Specifies when to perform the action at the end of a DMA transfer as
 * specified by @ref edt_enddma_reg().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param val
 *   One of @c EDT_ACT_NEVER, @c EDT_ACT_ONCE, or @c EDT_ACT_ALWAYS.
 */
EDTAPI void edt_enddma_action(EdtDev edt_p, uint32_t val);

/* Deal with a hardware timeout. */

/**
 * @brief Cause the driver to perform the same actions as it would on a
 * timeout. Partially filled FIFOs will be flushed and DMA will be aborted.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success; -1 on failure
 */
EDTAPI int edt_do_timeout(EdtDev edt_p);

/**
 * @brief Get the number of read and write timeouts that have occurred
 * since the last call of @ref edt_open().
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The number of read and write timeouts that have occurred
 * since the last call of @ref edt_open().
 */
EDTAPI int edt_timeouts(EdtDev edt_p);

/**
 * @brief Get the number of bytes transferred at last timeout.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The number of bytes transferred at last timeout.
 */
EDTAPI uint32_t edt_get_timeout_count(EdtDev edt_p);

/**
 * @brief Set the driver behavior on a timeout.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param action
 *   The action to perform when a timeout occurs.
 *   Values documented in @ref EdtDmaTimeoutAction.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_set_timeout_action(EdtDev edt_p, EdtDmaTimeoutAction action);

/**
 * @brief Get the number of good bits in the last long word of a read buffer
 * after the last timeout. This routine is called after a timeout if the timeout
 * action is set to @ref EDT_TIMEOUT_BIT_STROBE.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Number 0-31 represents the number of good bits in the last
 * 32-bit word of the read buffer associated with the last timeout. 0 on failure.
 */
EDTAPI int edt_get_timeout_goodbits(EdtDev edt_p);

/**
 * @brief Get the current number of good bits in the last long word of a
 * read buffer (0 through 31).
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Number 0-31 representing the number of good bits in the
 * last 32-bit word of the current read buffer. 0 on failure.
 */
EDTAPI int edt_get_goodbits(EdtDev edt_p);

/**
 * @brief Define a function to call when an event occurs.
 *
 * Use this routine to send an application-specific function when required;
 * for example, when DMA completes, allowing the application to continue
 * executing until the event of interest occurs.
 *
 * If you wish to receive notification of one event only, and then disable
 * further event notification, send a final argument of 0 (see the continue
 * parameter described below). This disables event notification at the time of
 * the callback to your function.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param event_type
 *   The event that causes the function to be called.
 * @param func
 *   The function to call when the event occurs.
 * @param data
 *   Data to pass to the @p func function; usually @c edt_p.
 * @param continuous
 *   Flag to enable or disable continued event notification. A value of 0 causes
 *   @ref edt_remove_event_func() to be called as the event is triggered.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_set_event_func(EdtDev edt_p,
                              enum edt_kernel_event event_type,
                              EdtEventFunc func,
                              void* data,
                              int continuous);

/**
 * @brief Remove an event function previously set with @ref edt_set_event_func().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param event_type
 *   The event that causes the function to be called. Valid events are listed
 *   in @ref edt_set_event_func().
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_remove_event_func(EdtDev edt_p, enum edt_kernel_event event_type);

#ifdef _NT_
EDTAPI int edt_get_kernel_event(EdtDev edt_p, enum edt_kernel_event event_num);
#endif

EDTAPI int edt_enable_event(EdtDev edt_p, enum edt_kernel_event event_type);
EDTAPI int edt_reset_event_counter(EdtDev edt_p, enum edt_kernel_event event_type);
EDTAPI int edt_wait_event(EdtDev edt_p, enum edt_kernel_event event_type, int timeoutval);

/**
 * @brief Get the number of buffers that the driver has been told to acquire.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Number of buffers started via @ref edt_start_buffers().
 */
EDTAPI uint32_t edt_get_todo(EdtDev edt_p);

/**
 * @brief Determine if DMA has written to more buffers in the ring than user
 * space code has "waited" (@ref edt_wait_for_buffers()) for. Indicates user
 * space code cannot keep up with the DMA transfer rate.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 1 (true) when overrun has occurred, 0 (false) otherwise.
 */
EDTAPI int edt_ring_buffer_overrun(EdtDev edt_p);

/**
 * @brief Set the burst enable flag, determining whether the DMA master
 * transfers as many words as possible at once, or transfers them one at
 * a time as soon as the data is acquired.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param enable
 *   A value of 1 turns the flag on (the default); 0 turns it off.
 *
 * @return 0 if successful.
 */
EDTAPI int edt_set_burst_enable(EdtDev edt_p, int enable);

/**
 * @brief Get the value of the burst enable flag.
 *
 * @see @ref edt_set_burst_enable()
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @retval 1 if burst transfers are enabled
 * @retval 0 if disabled
 * @retval -1 on error
 */
EDTAPI int edt_get_burst_enable(EdtDev edt_p);

/**
 * @brief Set the timeout for data receive calls.
 *
 * Affects @ref edt_wait_for_buffers() and @ref edt_read().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param timeout_ms
 *   Milliseconds to wait before timing out. A value of 0 disables the timeout,
 *   causing read calls to wait forever (blocking).
 *
 * @return 0 on success; -1 on error
 */
EDTAPI int edt_set_rtimeout(EdtDev edt_p, int timeout_ms);

/**
 * @brief Set the timeout for data transmit calls.
 *
 * Affects @ref edt_wait_for_buffers() and @ref edt_write().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param timeout_ms
 *   Milliseconds to wait before timing out. A value of 0 disables the timeout,
 *   causing write calls to wait forever (blocking).
 *
 * @return 0 on success; -1 on error
 */
EDTAPI int edt_set_wtimeout(EdtDev edt_p, int timeout_ms);

/**
 * @brief Gets the current read timeout value: the number of milliseconds to
 * wait for DMA reads to complete before returning.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The number of milliseconds in the current read timeout period,
 * or -1 on error.
 */
EDTAPI int edt_get_rtimeout(EdtDev edt_p);

/**
 * @brief Gets the current write timeout value: the number of milliseconds to
 * wait for DMA reads to complete before returning.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The number of milliseconds in the current write timeout period,
 * or -1 on error.
 */
EDTAPI int edt_get_wtimeout(EdtDev edt_p);

/**
 * @brief Get DMA overflow counter from the driver. Reset by @ref edt_reset_counts().
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @retval >=0 Overflow count read from the driver.
 * @retval -1 Error occurred.
 */
EDTAPI uint32_t edt_overflow(EdtDev edt_p);

/** @} */ /* end dma_inout */

/** @addtogroup docgroup_libedt_registers
 * @{
 */

/**
 * @brief Read a device register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   The name of the register to read. Use the names provided in the register
 *   descriptions in documentation for the card you are using. Or, register
 *   descriptor values can be constructed with macros such as @ref EDT_BAR1_REG4
 *   or @ref EDT_IND_REG_1.
 *
 * @return The (4-byte) value of register.
 */
EDTAPI uint32_t edt_reg_read(EdtDev edt_p, uint32_t desc);

/**
 * @brief Write a device register.
 *
 * @warning Use this routine with care; it writes directly to the hardware.
 * An incorrect value can crash your system, possibly causing loss of data.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   The name of the register to write. Use the names provided in the register
 *   descriptions in documentation for the card you are using. Or, register
 *   descriptor values can be constructed with macros such as @ref EDT_BAR1_REG4
 *   or @ref EDT_IND_REG_1.
 * @param value
 *   The value to write to the register.
 */
EDTAPI void edt_reg_write(EdtDev edt_p, uint32_t desc, uint32_t val);

/**
 * @brief Perform a register read-modify-write operation in the driver, @b OR-ing
 * the @p mask with the register value.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   The name of the register to modify.
 * @param mask
 *   The value to OR with the register.
 *
 * @return The value written back to the register.
 */
EDTAPI uint32_t edt_reg_or(EdtDev edt_p, uint32_t desc, uint32_t mask);

/**
 * @brief Perform a register read-modify-write operation in the driver, @b AND-ing
 * the @p mask with the register value.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   The name of the register to modify.
 * @param mask
 *   The value to AND with the register.
 *
 * @return The value written back to the register.
 */
EDTAPI uint32_t edt_reg_and(EdtDev edt_p, uint32_t desc, uint32_t mask);

/**
 * @brief Toggle the register bits specified in @p mask, off (0) then on (1).
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   The name of the register to modify.
 * @param mask
 *   Bits set in this mask indicate which bits in the register to toggle.
 */
EDTAPI void edt_reg_clearset(EdtDev edt_p, uint32_t desc, uint32_t mask);

/**
 * @brief Toggle the register bits specified in @p mask, on (1) then off (0).
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param desc
 *   The name of the register to modify.
 * @param mask
 *   Bits set in this mask indicate which bits in the register to toggle.
 */
EDTAPI void edt_reg_setclear(EdtDev edt_p, uint32_t desc, uint32_t mask);

/**
 * @brief Write 8-bit value to a user interface ("UI") FPGA register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Integer offset into user interface FPGA space, or @ref edt_reg_write() style
 *   register descriptor.
 * @param data
 *   8-bit value to write to the register.
 */
EDTAPI void edt_intfc_write8(EdtDev edt_p, uint32_t offset, uint8_t data);

/**
 * @brief Read 8-bit value from a user interface ("UI") FPGA register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Integer offset into user interface FPGA space, or @ref edt_reg_read() style
 *   register descriptor.
 *
 * @return 8-bit value read from the register.
 */
EDTAPI uint8_t edt_intfc_read8(EdtDev edt_p, uint32_t offset);

/**
 * @brief Write 16-bit value to a user interface ("UI") FPGA register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Integer offset into user interface FPGA space, or @ref edt_reg_write() style
 *   register descriptor.
 * @param data
 *   16-bit value to write to the register.
 */
EDTAPI void edt_intfc_write16(EdtDev edt_p, uint32_t offset, uint16_t data);

/**
 * @brief Read 16-bit value from a user interface ("UI") FPGA register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Integer offset into user interface FPGA space, or @ref edt_reg_read() style
 *   register descriptor.
 *
 * @return 16-bit value read from the register.
 */
EDTAPI uint16_t edt_intfc_read16(EdtDev edt_p, uint32_t offset);

/**
 * @brief Write 32-bit value to a user interface ("UI") FPGA register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Integer offset into user interface FPGA space, or @ref edt_reg_read() style
 *   register descriptor.
 * @param data
 *   32-bit value to write to the register.
 */
EDTAPI void edt_intfc_write32(EdtDev edt_p, uint32_t offset, uint32_t data);

/**
 * @brief Read 32-bit value from a user interface ("UI") FPGA register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Integer offset into user interface FPGA space, or @ref edt_reg_read() style
 *   register descriptor.
 *
 * @return 32-bit value read from the register.
 */
EDTAPI uint32_t edt_intfc_read32(EdtDev edt_p, uint32_t offset);

/**
 * @brief A convenience routine to write registers in the PCI BAR1 region.
 * Equivalent to `edt_reg_write(edt_p, EDT_BAR1_REG4(offset)), data)`.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Byte offset into PCI BAR1 register memory, addressing a 32-bit value.
 *   The least-significant two bits are ignored (masked off).
 * @param data
 *   32-bit value to write to the register.
 */
EDTAPI void edt_bar1_write(EdtDev edt_p, uint32_t offset, uint32_t data);

/**
 * @brief A convenience routine to read registers in the PCI BAR1 region.
 * Equivalent to `edt_reg_read(edt_p, EDT_BAR1_REG4(offset)))`.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Byte offset into PCI BAR1 register memory, addressing a 32-bit value.
 *   The least-significant two bits are ignored (masked off).
 *
 * @return 32-bit value read from the register.
 */
EDTAPI uint32_t edt_bar1_read(EdtDev edt_p, uint32_t offset);

/** @} */ /* end docgroup_libedt_registers */

/** @addtogroup dma_fifo
 * @{
 */

/**
 * @brief Flush the board's input and output FIFOs to allow new data
 * transfers to start from a known state.
 *
 * @param edt_p
 *   The open EDT device handle.
 */
EDTAPI void edt_flush_fifo(EdtDev edt_p);

/**
 * @brief Specify whether and when to flush the FIFOs before DMA transfer. By
 * default, the FIFOs are not flushed.  However, certain applications
 * may require flushing before a given DMA transfer, or before each
 * transfer.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param flag
 *   Tells whether and when to flush the FIFOs. Valid values are:
 *   @arg @c EDT_ACT_NEVER don't flush before DMA transfer (default)
 *   @arg @c EDT_ACT_ONCE   flush before the start of the next DMA transfer
 *   @arg @c EDT_ACT_ALWAYS flush before the start of every DMA transfer
 *
 * @return 0 on success; -1 on error.
 */
EDTAPI int edt_set_firstflush(EdtDev edt_p, int flag);

EDTAPI void edt_flush_channel(EdtDev edt_p, int channel);

/**
 * @brief Set DMA channel enable bits in @c SSD16_CHEN or @c SSD16_CHEN32 register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param mask
 *   If a bit in this mask is set to 1, the corresponding DMA channel will be
 *   enabled. Other channel enable bits will be unaffected.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_enable_channels(EdtDev edt_p, uint32_t mask);

/**
 * @brief Set DMA channel enable bit in @c SSD16_CHEN or @c SSD16_CHEN32 register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param channel
 *   Index of the channel to enable.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_enable_channel(EdtDev edt_p, unsigned int channel);

/**
 * @brief Clear DMA channel enable bits in @c SSD16_CHEN or @c SSD16_CHEN32 register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param mask
 *   If a bit in this mask is set to 1, the corresponding DMA channel will be
 *   disabled. Other channel enable bits will be unaffected.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_disable_channels(EdtDev edt_p, uint32_t mask);

/**
 * @brief Clear DMA channel enable bit in @c SSD16_CHEN or @c SSD16_CHEN32 register.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param channel
 *   Index of the channel to disable.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_disable_channel(EdtDev edt_p, unsigned int channel);

/** @} */ /* end dma_fifo */

EDTAPI EdtDmaWaitStatus edt_get_wait_status(EdtDev edt_p);
EDTAPI int edt_set_timeout_ok(EdtDev edt_p, unsigned int val);
EDTAPI int edt_get_timeout_ok(EdtDev edt_p);

EDTAPI int edt_set_autodir(EdtDev edt_p, int val);

EDTAPI int edt_get_tracebuf(EdtDev edt_p, uint32_t* addr);
EDTAPI int edt_set_debug(EdtDev edt_p, int count);
EDTAPI int edt_get_debug(EdtDev edt_p);

/** @addtogroup docgroup_libedt_utility
 * @{
 */

/**
 * @brief Reset timeout and overflow counters in the driver.
 *
 * @param edt_p
 *   The open EDT device handle.
 */
EDTAPI void edt_reset_counts(EdtDev edt_p);

/**
 * @brief Get the 'unit' number of the specified device.
 *
 * @param edt_p
 *   EDT device handle to inspect.
 *
 * @return Unit number (>=0) on success, -1 on failure.
 */
EDTAPI int edt_device_unit(EdtDev edt_p);

/**
 * @brief Get the 'channel' number of the specified device.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Channel number (>=0) on success, -1 on failure.
 */
EDTAPI int edt_device_channel(EdtDev edt_p);

/**
 * @brief Get the PCI device ID (PID) of the EDT device.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return PCI PID number.
 */
EDTAPI enum EDT_PCI_PID edt_device_id(EdtDev edt_p);

/**
 * @brief Get a descriptive string of a given board ID.
 *
 * @param id
 *   PCI device id. See @ref edt_device_id().
 *
 * @return A descriptive string, such as "PCIe8 A5 10G", "Visionlink F4",
 * or "unknown".
 */
EDTAPI const char* edt_idstr(enum EDT_PCI_PID id);

#if defined(PCD)
/**
 * @brief Get the mezzanine ID value from the EDT device handle.
 *
 * Unlike @ref edt_get_mezz_id(), this does not query the driver. This is useful
 * to get any error-related mezz ID's that @ref edt_get_full_board_id() may have
 * set when it was querying hardware.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Value of @ref EdtMezzDescriptor.id in the EDT device handle.
 */
EDTAPI unsigned int edt_device_mezz_id(EdtDev edt_p);
#endif

/**
 * @brief Parse an EDT device name string. Same as @ref edt_parse_unit_channel()
 * but without the channel number.
 *
 * @param[in] str
 *   Device name string from command line. Should be either a unit number ("0")
 *   or device/unit concatenation ("pcd0," "pcd1," etc.).
 * @param[out] dev
 *   Array to hold the device name string; filled in by ths routine.
 * @param[in] default_dev
 *   Device name to use if none is given in the @p str argument. If NULL,
 *   EDT_INTERFACE will be used.
 *
 * @return Unit number, or -1 on error. The first device is unit 0.
 */
EDTAPI int edt_parse_unit(const char* str, char* dev, const char* default_dev);

/**
 * @brief Parse an EDT device name string. Fills in the name of the device,
 * with the default_device if specified, or a default determined by the
 * package, and returns a unit number.  Designed to facilitate a
 * flexible device/unit command line argument scheme for application
 * programs.  Most EDT example/utility programs use this subroutine to
 * allow users to specify either a unit number alone or a device/unit
 * number concatenation.
 *
 * @param[in] str
 *   Device name string from command line. Should be either a unit number ("0")
 *   or device/unit concatenation ("pcd0," "pcd1," etc.).
 * @param[out] dev
 *   Array to hold the device name string; filled in by ths routine.
 * @param[in] default_dev
 *   Device name to use if none is given in the @p str argument. If NULL,
 *   EDT_INTERFACE will be used.
 * @param[out] channel
 *   The channel number parsed from @p str. If this is NULL or parsed channel
 *   number is -1, this is ignored and unchanged.
 *
 * @return Unit number, or -1 on error. The first device is unit 0.
 */
EDTAPI int edt_parse_unit_channel(const char* str,
                                  char* dev,
                                  const char* default_dev,
                                  int* channel);

/**
 * @brief Read the default part number->fpga cross-reference file edt_parts.xpn in the current
 * directory, and provides the FPGA if a match is found.
 *
 * @param[in] part_number
 *   EDT part number to return information about.
 * @param[out] fpga
 *   A character array into which the FPGA type will be stored. For example
 *   'xc2s100e' will be returned for the part_number '01901933'). An array of
 *   128 bytes will be more than enough for the foreseeable future.
 *
 * @return 1 if found 8 or 10 digit match, 0 if not.
 */
EDTAPI int edt_find_xpn(const char* part_number, char* fpga);

/**
 * @brief Read a part number->fpga cross-reference file and provides the FPGA
 * and base serial number if a match is found.
 *
 * @param[in] path
 *   Path of the xref FPGA file (typ. "edt_parts.xpn").
 * @param[in] pn
 *   EDT part number to search for. Either 8 to 10 characters, meaning either
 *   with or without the last 2 characters, which is the revision number.
 *   Example: `"01901933"`.
 * @param[out] fpga
 *   If a match is found, the FPGA part number will be returned here. Should be
 *   at least 64 bytes. If NULL, this parameter will be ignored.
 * @param[out] sn
 *   The base serial number, if found. Should be at least 64 bytes. If NULL,
 *   this parameter will be ignored.
 * @param[out] mac_type
 *   The MAC address board type, if found. Should be at least 8 bytes.
 *   If NULL, this parameter will be ignored.
 * @param[out] mac_offset
 *   The MAC address offset found. Should be at least 8 bytes. If NULL this
 *   parameter will be ignored.
 * @param[out] mac_count:
 *   The MAC addresses for this board, if found. Should be at least 8 bytes.
 *   If NULL this parameter will be ignored.
 * @param[out] rsvd1
 *   Reserved. Always pass NULL.
 * @param[out] rsvd2
 *   Reserved. Always pass NULL.
 *
 * @return Number of parameters successfully assigned, or 0 if none or error.
 */
EDTAPI int edt_get_xref_info(const char* path,
                             const char* pn,
                             char* fpga,
                             char* sn,
                             char* mac_type,
                             char* mac_offset,
                             char* mac_count,
                             char* desc,
                             char* rsvd1,
                             char* rsvd2);

/**
 * @brief Get an operating system dependent error number.
 *
 * `errno` on UNIX, `GetLastError()` on Windows.
 *
 * @return OS-dependent error number.
 */
EDTAPI unsigned int edt_errno(void);

/**
 * @brief Get base installation directory of PCD or PDV package.
 *
 * Home directory is based on EDT environment variable for the driver type, if set.
 * Driver type depends on the version of EDT package / SDK in use. Drivertype / environment
 * variable associations follow:
 *
 * "pcd":    $EDT_PCD_DIR
 * "pdv":    $EDT_PDV_DIR
 *
 * By default, the installation script installs packages in `/opt/EDT<type>` (Linux)
 * or `C:\EDT\<type>` (Windows), so for example with a PCD package and default
 * install directory, this subroutine would return "/opt/EDTpcd" or "C:\EDT\pcd".
 *
 * @param edt_p pointer to edt device structure returned by #edt_open or #edt_open_channel
 *
 * @return value of the appropriate $XXXHOME environment variable, or CWD (".") if not set.
 */
EDTAPI const char* edt_home_dir(EdtDev edt_p);

EDTAPI const char* edt_timestring(edt_timespec_t* ts);

/**
 * @brief Performs a UNIX-like `system()` call which passes the argument
 * string to a shell or command interpreter, then returns the exit status
 * so that errors can be detected. On Windows `spawnv()` will be used.
 *
 * @param cmdstr
 *   The command string to execute.
 *
 * @return The value returned by the underlying system call. 0 should indicate
 * the command completed successfully.
 */
EDT_CHECK_RETURN
EDTAPI int edt_system(const char* cmdstr);

/* Subroutines to set driver last bitfile loaded; used in bitload */

/**
 * @brief Set @p bitpath as the currently loaded "user interface" FPGA firmware
 * ("bitfile") in the driver.
 *
 * For use by @ref edt_bitload() and the @c bitload application.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bitpath
 *   Firmware (bitfile) file name. Should be no more than 128 bytes.
 *
 * @return 0 on success, -1 on failure.
 * @see @ref edt_get_bitpath(), @ref edt_set_mezz_bitpath()
 */
EDTAPI int edt_set_bitpath(EdtDev edt_p, const char* bitpath);

/**
 * @brief Get file path from the driver of the currently loaded interface bitfile.
 *
 * @ref edt_bitload() sets this when an interface bitfile is successfully loaded.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] bitpath
 *   The file path read from the driver will be returned here. Should be at least
 *   128 bytes (@ref edt_bitpath).
 * @param size
 *   Size of @p bitpath.
 *
 * @return 0 on success, -1 on failure.
 * @see @ref edt_set_bitpath()
 */
EDTAPI int edt_get_bitpath(EdtDev edt_p, char* bitpath, int size);

/**
 * @brief Same as @ref edt_get_bitpath() but returns the file name without the
 * directory.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] bitfile
 *   The file name read from the driver will be returned here. Should be at least
 *   128 bytes.
 * @param size
 *   Size of @p bitfile.
 *
 * @return 0 on success, -1 on failure.
 * @see @ref edt_set_bitpath()
 */
EDTAPI int edt_get_bitname(EdtDev edt_p, char* bitfile, int size);

#if defined(PCD)
/**
 * @brief Obtains path of the currently loaded mezzanine bitfile from the
 * driver. @ref edt_bitload() sets this string when a mezzanine bitfile is
 * successfully loaded.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] bitpath
 *   The file path read from the driver will be returned here. Should be at least
 *   128 bytes.
 * @param size
 *   Size of @p bitpath.
 * @param channel
 *   Mezzanine channel index.
 *
 * @return 0 on success, -1 on failure.
 * @see @ref edt_get_bitpath()
 */
EDTAPI int edt_get_mezz_chan_bitpath(EdtDev edt_p, char* bitpath, int size, int channel);

/**
 * @brief Set pathname to the currently loaded mezzanine bitfile in the driver.
 *
 * For use by @ref edt_bitload().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bitpath
 *   The file path to save in the driver. Must be at least 128 bytes.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_set_mezz_bitpath(EdtDev edt_p, const char* bitpath);

/**
 * @brief Obtains pathname to the currently loaded mezzanine bitfile from the
 * driver. @ref edt_bitload() sets this when a mezzanine bitfile is successfully
 * loaded.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] bitpath
 *   The file path read from the driver will be returned here. Should be at least
 *   128 bytes.
 * @param size
 *   Size of @p bitpath.
 *
 * @return 0 on success, -1 on failure.
 * @see @ref edt_get_bitpath()
 */
EDTAPI int edt_get_mezz_bitpath(EdtDev edt_p, char* bitpath, int size);

/**
 * @brief Get the mezzanine board ID. Equivalent to calling
 * @ref edt_get_full_board_id() with NULL as all the pointer arguments.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Mezzanine ID.
 */
EDTAPI unsigned int edt_get_board_id(EdtDev edt_p);

/**
 * @brief Set the mezzanine ID in the driver.
 *
 * Expects mezzanine ID to be properly set in the @p edt_p handle's internal
 * @ref EdtMezzDescriptor structure. Use @ref edt_device_mezz_id() to get the
 * internal mezz id value that would be saved to the driver by this function.
 *
 * @note For use by EDT library functions like @ref edt_get_board_description().
 * Should not be used by customer applications.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int edt_set_mezz_id(EdtDev edt_p);

/**
 * @brief Get the mezzanine ID from the driver.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Mezzanine ID (>=0) on success, -1 on failure.
 */
EDTAPI int edt_get_mezz_id(EdtDev edt_p);

#endif /* PCD */

/**
 * @param Get the version of the EDT driver. The version string is the same
 * format as that returned by @ref edt_get_library_version().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] version_str
 *   Driver's version string will be returned here. Should be at least 64 bytes
 *   (@ref edt_version_string).
 * @param size
 *   Size, in bytes, of @p version_str buffer.
 *
 * @return 0 on success.
 * @see @ref edt_get_library_version()
 */
EDTAPI int edt_get_driver_version(EdtDev edt_p, char* version_str, size_t size);

/**
 * @brief Get the full build ID of the EDT driver. The build ID string is the
 * same format as that returned by @ref edt_get_library_buildid().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] build_str
 *   Driver's build info string will be returned here. Should be at least 128
 *   bytes.
 * @param size
 *   Size, in bytes, of @p build_str buffer.
 *
 * @return 0 on success.
 * @see @ref edt_get_library_buildid()
 */
EDTAPI int edt_get_driver_buildid(EdtDev edt_p, char* build_str, size_t size);

/**
 * @brief Get the version string of the EDT library. Example: `"1.2.3-rc4"`
 *
 * @param[out] version_str
 *   Version string from @ref libedt_get_version_str() will be copied into this
 *   buffer.
 * @param size
 *   Size, in bytes, of @p version_str buffer.
 *
 * @return 0 on success.
 * @see @ref edt_get_driver_version()
 */
EDTAPI int edt_get_library_version(char* version_str, size_t size);

/**
 * @brief Get the EDT library version as a packed number. Example v5.6.1 would
 * be represented as `50601`.
 *
 * @return Packed library version number.
 */
EDTAPI unsigned int edt_get_version_number(void);

/**
 * @brief Get the build ID of the EDT library. Example: `"20210214124523.asdf1234 64 bit"`
 *
 * @param[out] build_str
 *   Build string from data in @ref libedt_get_version() will be saved into this
 *   buffer.
 * @param size
 *   Size, in bytes, of @p build_str buffer.
 * @see @ref edt_get_driver_buildid()
 */
EDTAPI int edt_get_library_buildid(char* build_str, size_t size);

/**
 * @brief Ensure the EDT library and the kernel driver are the same version.
 *
 * While uncommon, it is possible for package updates to fail, causing an older
 * EDT driver to be left in the system. User applications may wish to run this
 * sanity check early, to ensure the library and driver match before they
 * encounter errors due to an incompatibility.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return True (!= 0) if the versions match. False (0) if they differ.
 */
EDTAPI int edt_check_version(EdtDev edt_p);

/**
 * @brief Gets information about active DMA.
 *
 * Determine whether this or another open process has enabled DMA or
 * image acquisition on any channel of a specific board (unit).
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] dmainfo
 *   DMA information result.
 *
 * @return The value of `used_dma`, `alloc_dma` and `active_dma` masks OR'd
 * together.
 */
EDTAPI uint32_t edt_get_dma_info(EdtDev edt_p, edt_dma_info_t* dmainfo);

/** @} */ /* end docgroup_libedt_utility */

/**
 * @brief Read a 32-bit word directly from the board's PCI configuration space.
 *
 * Note that this does not allow for reading past the first 64 bytes.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param offset
 *   Offset in the PCI configuration space to read from. Must be 32-bit aligned
 *   (lower 2 bits clear). Max value is 0x3C (60).
 *
 * @return Value read from config space, or UINT32_MAX if offset is invalid.
 */
EDTAPI uint32_t edt_pci_config_read(EdtDev edt_p, uint8_t offset);

/**
 * @brief Attempt to reboot the PCI FPGA.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_pci_reboot(EdtDev edt_p);

EDTAPI int edt_set_sglist(EdtDev edt_p, uint32_t bufnum, uint32_t* log_list, uint32_t log_entrys);

/**
 * @brief Get the number of bytes read so far into the current buffer.
 *
 * Can be used to monitor how much data has been read into the buffer
 * during acquisition.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param[out] cur_buffer
 *   Current buffer's index is returned here. Pass NULL if buffer number is not
 *   needed.
 *
 * @return The number of bytes transferred.
 */
EDTAPI uint32_t edt_get_bufbytecount(EdtDev edt_p, uint32_t* cur_buffer);

/**
 * @brief Check if currently running on a little-endian machine.
 * @return True (1) if machine is little-endian.
 */
EDTAPI int edt_little_endian(void);

EDTAPI int edt_resume(EdtDev edt_p);

/**
 * @brief Unmap a PCI BAR region, previously mapped with @ref edt_mapmem().
 *
 * On Linux, all pages containing a part of the indicated range are unmapped,
 * and subsequent references to these pages will generate segmentation fault
 * errors.
 *
 * On Windows, if the indicated range does not match the range originally
 * requested (e.g. @p size does not match) then nothing is un-mapped and an
 * error is returned.
 *
 * @param addr
 *   Virtual address to unmap. Must be a multiple of the page size (but @p size
 *   need not be).
 * @param size
 *   Size of the region to unmap.
 *
 * @return 0 on success.
 */
EDTAPI int edt_unmapmem(EdtDev edt_p, void* addr, uint32_t size);

/**
 * @brief Map regions of PCI BAR space into the virtual address space of the
 * calling process.
 *
 * Calls `mmap()` on UNIX, or a driver ioctl on Windows.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param addr
 *   Selects the region and offset of a BAR to mmap.
 *   @arg `(0 >= addr <0x10000)` PCI BAR0 region.
 *   @arg `(0x10000 >= addr < PDMA_OFFSET)` PCI BAR1 region.
 *   @arg `(addr >= PDMA_OFFSET)` Undefined, do not use.
 * @param size
 *   Size (in bytes) of the region to memory map. Use @ref edt_get_mappable_size()
 *   to get the size of a given PCI BAR.
 *
 * @return Pointer to the mapped area on success. NULL on error.
 *
 * @see @ref edt_map_dmamem(), @ref edt_get_mappable_size()
 */
EDT_DEALLOCATE_WITH(edt_unmapmem)
EDTAPI void* edt_mapmem(EdtDev edt_p, uint32_t addr, uint32_t size);

/**
 * @brief Get the size of a given PCI base address register (BAR) region.
 *
 * Similar to Linux kernel's `pci_resource_len()` function.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param bar
 *   Index of the BAR to query. Must be either 0 or 1.
 *
 * @return The byte length of the PCI @p bar region.
 */
EDTAPI uint32_t edt_get_mappable_size(EdtDev edt_p, int bar);

/**
 * @brief Change the maximum number of ring buffers that can be allocated.
 *
 * The absolute maximum is 1024.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param newmax
 *   New maximum, up to 1024.
 *
 * @return 0 on success.
 * @see @ref edt_configure_ring_buffers(), @ref edt_get_max_buffers()
 */
EDTAPI int edt_set_max_buffers(EdtDev edt_p, int newmax);

/**
 * @brief Get the maximum number of ring buffers that can be allocated.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The maximum number of buffers that can be allocated, or -1 on failure.
 * @see @ref edt_configure_ring_buffers(), @ref edt_set_max_buffers()
 */
EDTAPI int edt_get_max_buffers(EdtDev edt_p);

/**
 * @brief Set flags in kernel driver to change the way kernel buffers are
 * created and used.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param onoff
 *   Bitfield of @ref EDT_USER_BUFS, @ref EDT_COPY_KBUFS, @ref EDT_MMAP_KBUFS.
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int edt_set_kernel_buffers(EdtDev edt_p, uint32_t onoff);

/**
 * @brief Get the current kbufs state.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Current kbuf state.
 */
EDTAPI int edt_get_kernel_buffers(EdtDev edt_p);

EDTAPI int edt_set_mmap_buffers(EdtDev edt_p, int onoff);
EDTAPI int edt_get_mmap_buffers(EdtDev edt_p);

EDTAPI void edt_set_dump_reg_access(bool on);
EDTAPI bool edt_get_dump_reg_access(void);
EDTAPI void edt_set_dump_reg_address(uint32_t reglow, unsigned int n, bool on);
EDTAPI void edt_set_dump_ir_access(bool on);

/**
 * @brief Free memory containing board list created & returned by any of the
 * edt_detect_boards functions.
 *
 * @see @ref edt_detect_boards(), @ref edt_detect_boards_id(), @ref edt_detect_boards_ids()
 */
EDTAPI void edt_free_detected_boards_list(void);

/**
 * @brief Detect all EDT boards in the system. Designed for use by pciload &
 * pciload-like applications.
 *
 * @note This is a wrapper of @ref edt_detect_boards_ids().
 *
 * @param unit
 *   A specific unit to find, or -1 for any unit that matches.
 * @param[out] nunits
 *   Returns the number of units found.
 * @param verbose
 *   Enable (1) verbose output.
 *
 * @return Array of zero or more @ref EdtBoardInfo structs, each with a detected
 * board that meets the given parameters. NULL on error. Free with
 * @ref edt_free_detected_boards_list().
 */
EDTAPI const EdtBoardInfo* edt_detect_boards(int unit, int* nunits, int verbose);

/**
 * @brief Detect any EDT boards in the system that match the given ids.
 * Designed for use by pciload & pciload-like applications.
 *
 * @param unit
 *   A specific unit to find, or -1 for any unit that matches.
 * @param[in] idlist
 *   List of device ids to search for. The last ID in the list @b must be
 *   @ref EDT_PCI_UNKNOWN_ID to mark the end of the list. Pass NULL to search
 *   for all/any EDT devices.
 * @param[out] nunits
 *   Returns the number of units found.
 * @param verbose
 *   Enable (1) verbose output.
 *
 * @return Array of zero or more @ref EdtBoardInfo structs, each with a detected
 * board that meets the given parameters. NULL on error. Free with
 * @ref edt_free_detected_boards_list().
 */
EDTAPI const EdtBoardInfo* edt_detect_boards_ids(int unit,
                                                 const enum EDT_PCI_PID* idlist,
                                                 int* nunits,
                                                 int verbose);

/* routines to enable/disable register accesses tracing in the driver */

/**
 * @brief Enable or disable tracing a single reg access in the driver.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param reg_def
 *   If @p reg_def ends in 0xffff, trace all register accesses of that
 *   type.
 * @param state
 *   1 enable tracing, 0 disable tracing.
 */
EDTAPI void edt_set_trace_regs(EdtDev edt_p, uint32_t reg_def, unsigned int state);

/**
 * @brief Enable global tracing of register accesses. Individual registers or
 * register types must still be enabled using @ref edt_set_trace_regs().
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param state
 *   Enable (1) or disable (0) global register trace flag in the driver.
 */
EDTAPI void edt_trace_regs_enable(EdtDev edt_p, uint32_t state);

/**
 * @brief Get the number of (DMA) channels avaliable for this board type.
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return The number of channels for this board's device ID.
 * @see @ref edt_device_id()
 */
EDTAPI int edtdev_channels_from_type(EdtDev edt_p);

EDTAPI void edt_set_intr_mask(EdtDev edt_p, uint32_t state);
EDTAPI uint32_t edt_get_intr_mask(EdtDev edt_p);
EDTAPI void edt_set_remote_intr(EdtDev edt_p, unsigned int onoff);
EDTAPI unsigned int edt_get_remote_intr(EdtDev edt_p);

#ifdef PCD

EDTAPI void pcd_set_abortdma_onintr(EdtDev edt_p, int flag);

#endif /* PCD */

/** @} */ /* end weakgroup dma */

/**
 * @internal
 * @brief Operating system-independent routine for making an IOCTL system call
 * to an EDT device driver.
 *
 * @warning Customers are discouraged from calling this routine directly and
 * encouraged to use the other functions documented in this library which will
 * make appropriate IOCTL calls internally.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param code
 *   IOCTL code defined in driver/edt_ioctl.h header.
 * @param arg
 *   IOCTL argument is depenant on the code itself.
 *
 * @return 0 on success, -1 on error.
 */
EDTAPI int edt_ioctl(EdtDev edt_p, int code, void* arg);

/**
 * @brief Get DMA burst size (between EDT DMA FIFOs and the PCI core).
 *
 * @param edt_p
 *   The open EDT device handle.
 *
 * @return Current dma burst size.
 */
EDTAPI uint32_t edt_get_dma_burst_size(EdtDev edt_p);

/**
 * @brief Set DMA burst size (between EDT DMA FIFOs and the PCI core).
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param new_size
 *   The new DMA burst size to set.
 */
EDTAPI void edt_set_dma_burst_size(EdtDev edt_p, uint32_t new_size);

#ifdef __cplusplus
}
#endif

#endif /* INCLUDE_edtlib_h */
