/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2007 Engineering Design Team, Inc.
 *
 * Interface for retrieving IRIG timecode information from EDT boards.
 */

#ifndef INCLUDE_libedt_timing_h
#define INCLUDE_libedt_timing_h

#include <edt_portability.h>

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/** Timing register offset for SRXL-2 */
#define TIMECODE_SRXL2_BASE_REGISTER 0x5c
/** Timing register offset for Visionlink F1 */
#define TIMECODE_VL_F1_BASE_REGISTER 0xb1
/** Timing register offset for Visionlink F4 */
#define TIMECODE_VL_F4_BASE_REGISTER 0xb1
/** Timing register offset for S5 */
#define TIMECODE_S5_BASE_REGISTER 0x6d
/** Timing register offset for A5/TacWrap */
#define TIMECODE_A5_BASE_REGISTER 0x6d
/** (compatibility) Timing register offset for VL F1/F4 */
#define PDV_IRIG_SPI_BASE 0xb1

#define TIMECODE_BASE_REGISTER TIMECODE_VL_F4_BASE_REGISTER

/** Raw timecode format */
typedef struct _ts_raw_t {
  uint32_t seconds : 6;
  uint32_t minutes : 6;
  uint32_t hours : 5;
  uint32_t days : 9;
  uint32_t years : 6;
} ts_raw_t;

// timecode_configuration

/**
 * @brief Opens the EDT SPI master.
 *
 * @param edt_interface
 *    String containing the type of board; "pcd", "pdv"
 * @param unit
 *    Board unit number.
 * @param spi_reg_base
 *    Sets unit spi register base address.
 *
 * @return A pointer to the EdtDev structure if successful or NULL if the open
 * fails or the device is not attached to the SPI master on the timing board.
 *
 * This EDT device handle should be used with edt_spi_ prefixed subroutines.
 *
 * See the included example program timing_test.c.
 *
 * @ingroup timecode_configuration
 */
EDTAPI EdtDev edt_spi_open(const char* edt_interface, int unit, uint32_t spi_reg_base);

/**
 * @brief Closes the SPI master EDT device handle.
 *
 * @param edt_p
 *    The EDT open device handle.
 *
 * @return The value from edt_close().
 *
 * @ingroup timecode_configuration
 */
EDTAPI int edt_spi_close(EdtDev edt_p);

/**
 * @brief Set the base address of the SPI interface used with EDT (IRIG) timing
 * functions.
 *
 * @param edt_p
 *   EDT device handle to update.
 * @param spi_reg_base
 *    Sets unit spi register base address.
 *
 * @return 0 if successful
 */
EDTAPI int edt_spi_set_base_address(EdtDev edt_p, uint32_t spi_reg_base);

/**
 * @brief Enable or disable timecode production from the MSP430.
 *
 * @param edt_p The EDT open device handle.
 * @param enable An integer containing 1 to enable timestamp production, 0 to disable it.
 *
 * @return 0 for success, -1 for failure.
 *
 * @ingroup timecode_configuration
 */
EDTAPI int edt_set_timecode_enable(EdtDev edt_p, uint8_t enable);

/**
 * @brief Select an internal or external clock source for the MSP430 and state the clock rate.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param clock_sel
 *    An integer containing 0 for internal clock and 1 for external clock select.
 * @param clock_hz
 *    An integer stating the number of Hertz for the selected clock.
 *
 * @ingroup timecode_configuration
 */
EDTAPI void edt_set_msp430_clock(EdtDev edt_p, int clock_sel, int clock_hz);

/**
 * @brief Enable or disable raw timecode format from the MSP430.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param enable
 *    An integer containing 1 to enable raw timestamp format, 0 to disable it.
 *
 * @ingroup timecode_configuration
 */
EDTAPI void edt_set_timecode_raw(EdtDev edt_p, int enable);

/**
 * @brief Set a seconds offset for timecode production.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param seconds
 *    An integer containing the number of seconds to add to the timecode.
 *
 * @ingroup timecode_configuration
 */
EDTAPI void edt_set_timecode_seconds_offset(EdtDev edt_p, uint32_t seconds);

/**
 * @brief Enable the msp430 IRIG-B firmware programmable year mode and set the year.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param year
 *    The year >= 2000.
 *
 * @ingroup timecode_configuration
 */
EDTAPI void edt_enable_timecode_programmable_year(EdtDev edt_p, uint16_t year);

/**
 * @brief Disable the msp430 IRIG-B firmware programmable year mode.
 *
 * @param edt_p
 *    The EDT open device handle.
 *
 * @ingroup timecode_configuration
 */
EDTAPI void edt_disable_timecode_programmable_year(EdtDev edt_p);

// timecode_display

EDTAPI uint8_t edt_spi_get_stat(EdtDev edt_p);

/**
 * @brief Prints the timecode in ascii once per second for N loops.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param loops
 *    The number of seconds to print.
 *
 * @ingroup timecode_display
 */
EDTAPI void edt_spi_display_time(EdtDev edt_p, int loops);

// timecode_update

/**
 * @brief Get the msp430 IRIG-B firmware version
 *
 * @param edt_p
 *    The EDT open device handle.
 *
 * @return An integer containing the firmware version number
 *
 * @ingroup timecode_update
 */
EDTAPI int edt_get_timecode_version(EdtDev edt_p);

/**
 * @brief Puts the EDT SPI msp430 into flash-loader mode.
 *
 * @param edt_p
 *    The EDT open device handle.
 *
 * @return Zero on success, -1 on fault.
 *
 * @ingroup timecode_update
 */
EDTAPI int edt_spi_invoke_flash_loader(EdtDev edt_p);

// timecode_utility

/**
 * @brief Polls the EDT SPI master for a single byte.
 *
 * @param edt_p
 *    The EDT open device handle.
 *
 * @return The byte obtained from the SPI interface.
 *
 * @ingroup timecode_utility
 */
EDTAPI uint8_t edt_spi_get_byte(EdtDev edt_p);

/**
 * @brief Sends one byte to the EDT SPI master.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param ch
 *    The byte to send to the SPI master.
 *
 * @return The byte sent to the SPI interface.
 *
 * @ingroup timecode_utility
 */
EDTAPI uint8_t edt_spi_put_byte(EdtDev edt_p, uint8_t ch);

/**
 * @brief Flushes the EDT SPI input and output fifos, discarding all data.
 *
 * @param edt_p
 *    The EDT open device handle.
 *
 * @ingroup timecode_utility
 */
EDTAPI void edt_spi_flush_fifo(EdtDev edt_p);

/**
 * @brief Send a single line of ascii string to the msp430 over SPI.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param str
 *    The null-terminated string to send to the EDT SPI interface.
 *
 * @return A pointer to the null-terminated argument string.
 *
 * @ingroup timecode_utility
 */
EDTAPI char* edt_spi_putstr(EdtDev edt_p, char* const str);

EDTAPI uint8_t edt_crc16_lowbyte(uint8_t* buf, int len);
EDTAPI uint8_t edt_crc16_highbyte(uint8_t* buf, int len);

/**
 * @brief After a packet-start byte is received, call this routine to input the packet and check the
 * CRC.
 *
 * @param edt_p
 *    The EDT open device handle.
 * @param byte
 *    Pointer to an unsigned char array of 64-bytes which will contain the packet on success.
 */
EDTAPI void edt_spi_send_packet(EdtDev edt_p, uint8_t* cmdbuf);

#ifdef __cplusplus
}
#endif

#endif /* INCLUDE_libedt_timing_h */
