/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @author EDT <tech@edt.com>
 * @copyright Copyright (c) 2023 Engineering Design Corporation, Inc.
 *
 * @details
 * This file provides both compile-time macros and run-time functions to get
 * the version of the EDT library.
 *
 * Compile-time macros may be used to write pre-processor logic to support
 * building against different versions of the EDT library.
 *
 * The version getter functions may be useful to check the library version in
 * use at runtime.
 */
#ifndef LIBEDT_VERSION_H
#define LIBEDT_VERSION_H

#ifdef __cplusplus
extern "C" {
#endif

#include "edt_portability.h"

#include <stdint.h>

/** @brief @b Major version number of the EDT library */
#define LIBEDT_VERSION_MAJOR 6
/** @brief @b Minor version number of the EDT library */
#define LIBEDT_VERSION_MINOR 2
/** @brief @b Patch version number of the EDT library */
#define LIBEDT_VERSION_PATCH 0

/**
 * @brief EDT library version structure.
 */
typedef struct libedt_version {
  uint32_t major;              /**< @brief Major version number */
  uint32_t minor;              /**< @brief Minor version number */
  uint32_t patch;              /**< @brief Patch version number */
  const char* prerelease;      /**< @brief Pre-release string, such as "rc0" */
  uint32_t vcs_id;             /**< @brief Internal version control system id */
  uint64_t build_timestamp;    /**< @brief Build time, as UNIX timestamp */
  const char * build_time_str; /**< @brief Build time, as compact string */
} libedt_version_t;

/**
 * @brief Get detailed EDT library version information.
 *
 * @returns Pointer to read-only version structure.
 */
EDTAPI const libedt_version_t* libedt_get_version(void);

/**
 * @brief Get EDT library version as a string.
 *
 * Example: "1.2.3-rc4"
 *
 * @note For informational purposes only. The format of this string is subject
 *     to change at any time.
 *
 * @returns String pointer.
 */
EDTAPI const char* libedt_get_version_str(void);

/**
 * @brief Get EDT library version as a detailed string.
 *
 * Example: "1.2.3-rc4+20210214124523.asdf1234"
 *
 * @note For informational purposes only. The format of this string is subject
 *     to change at any time.
 *
 * @returns String pointer.
 */
EDTAPI const char* libedt_get_version_str_full(void);

/**
 * @brief Get simple version string of compiler used to build the EDT library.
 *
 * @note For informational purposes only. The accuracy or format of this string
 *     is subject to change at any time.
 *
 * @returns String pointer.
 */
EDTAPI const char* libedt_compiled_by(void);

#ifdef __cplusplus
}
#endif

#endif /* LIBEDT_VERSION_H */
