/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 * @brief EDT PDV digital imaging library.
 *
 * Provides a 'C' language interface to the EDT PCI bus Digital Imaging
 * Camera/Device Capture family of products. All routines access a specific
 * device, whose handle is created and returned by the @ref pdv_open() routine.
 *
 */

#ifndef LIBPDV_H
#define LIBPDV_H

#ifdef __cplusplus
extern "C" {
#endif

#include "edt_portability.h"
#include "edt_os_utils.h"
#include "driver/edt_shared_types.h"

#include <stdint.h>
#include <stdbool.h>

/** @brief PDV device handle, an opaque pointer.
 *
 * For convenience, we use PdvDev in PDV applications, and EdtDev
 * in applications for all other EDT applications.  PdvDev and EdtDev
 * are interchangable, some functions from edtlib can be called using
 * a PdvDev handle.
 */
typedef struct edt_device* PdvDev;

#if !defined(EDT_DEV_TYPE_DEFINED)
typedef PdvDev EdtDev;
#define EDT_DEV_TYPE_DEFINED 1
#endif

/** Definition for header data for buffers. */
typedef enum edt_pdv_header_type {
  PDV_HDR_TYPE_NONE = 0, /**< No header. */
  PDV_HDR_TYPE_IRIG1 = 1,
  PDV_HDR_TYPE_FRAMECOUNT = 2,
  PDV_HDR_TYPE_IRIG2 = 3,     /**< IRIG2 format timing information. */
  PDV_HDR_TYPE_BUFHEADER = 4, /* from general data (wrap) */
} EdtPdvHeaderType;

/** Position of header relative to the bounds of a DMA buffer.
 *
 * For disk I/O speed, header memory may be allocated with ring-buffer memory
 * even if the header is not in the DMA stream. This covers all possible
 * relationships of header memory to image data.
 */
typedef enum edt_pdv_header_position {
  PDV_HDR_POS_NONE,
  /** Header data should be allocated contiguously before DMA buffer. */
  PDV_HDR_POS_BEFORE,
  /** Header data should be at the beginning of DMA buffer. */
  PDV_HDR_POS_BEGIN,
  /** Header data should be in the middle of DMA buffer. */
  PDV_HDR_POS_MIDDLE,
  /** Header data should be at the end of DMA buffer. */
  PDV_HDR_POS_END,
  /** Header data should be allocated contiguously after DMA buffer. */
  PDV_HDR_POS_AFTER,
  /** Header data is somewhere else in memory. */
  PDV_HDR_POS_SEPARATE
} EdtPdvHeaderPosition;

static inline int pdv_bits_to_bytes(int bits)
{
  return (bits + 7) / 8;
}

/** Serial command flags. Generally not used in application code. */
typedef enum edt_pdv_serial_cmd_flag {
  /** Tells the driver not to wait for a response on FOI before returning. */
  PDV_SER_CMD_FLAG_NORESP = 1
} EdtPdvSerCmdFlag;

/**
 * @addtogroup updown
 * @{
 */

/**
 * @brief Closes the specified device and frees the image memory and
 * other resources associated with the device handle.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 0 if successful, -1 if unsuccessful.
 */
EDTAPI int pdv_close(PdvDev pdv_p);

/**
 * @brief Opens channel 0 of an EDT Framegrabber for application access.
 *
 * To open a specific channel on multi-channel device, see @ref pdv_open_channel().
 *
 * @param dev_name
 *  The name of the device.  For EDT imaging boards use @ref EDT_INTERFACE.
 *
 * @param unit
 *  The unit number of the device (board). The first device is 0.
 *
 * @return The opened PDV device handle if successful, or NULL if unsuccessful.
 *
 * @see @ref pdv_open_channel()
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(pdv_close)
EDTAPI PdvDev pdv_open(const char* dev_name, int unit);

/**
 * @brief Open the pdv device, with option to suppress non-existent device console output.
 *
 * @param dev_name
 *  The name of the device.  For EDT imaging boards use @ref EDT_INTERFACE.
 * @param unit
 *  The unit number of the device (board). The first device is 0.
 * @param channel
 *  The channel of the specified unit to open. The first channel is 0.
 * @param verbose
 *  Enable verbose output (1) or not (0).
 *
 * @return The opened PDV device handle if successful, or NULL if unsuccessful.
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(pdv_close)
EDTAPI PdvDev pdv_open_device(const char* dev_name, int unit, int channel, int verbose);

/**
 * @brief Opens an EDT Framegrabber channel for application access.
 *
 * @param dev_name
 *  The name of the device.  For EDT imaging boards use @ref EDT_INTERFACE.
 * @param unit
 *  The unit number of the device (board). The first device is 0.
 * @param channel
 *  The channel of the specified unit to open. The first channel is 0.
 *
 * @return The opened PDV device handle if successful, or NULL if unsuccessful.
 *
 * @see @ref pdv_open()
 */
EDT_CHECK_RETURN
EDT_DEALLOCATE_WITH(pdv_close)
EDTAPI PdvDev pdv_open_channel(const char* dev_name, int unit, int channel);

/** @} */ /* end updown */

/**
 * @addtogroup debug
 * @{
 */

/**
 * @brief Sets the debug level. For values, see the @ref msg.
 *
 * @param debug
 *  The debug level.
 */
EDTAPI void pdv_set_debug(int debug);

/**
 * @brief Gets the debug level, as set by pdv_set_debug or outside environment variables.
 * For values, see the \ref msg.
 *
 * @return The debug level.
 */
EDTAPI int pdv_get_debug(void);
/** @} */ /* end debug */

/**
 * @addtogroup acquisition
 * @{
 */

/**
 * @brief Reads image data from the EDT framegrabber board.
 * pdv_read is not supported on all platforms and is included mainly for historical reasons.
 * Consider using @ref pdv_image() or @ref pdv_start_images() instead.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param buf
 *   The data buffer to receive the image.
 * @param size
 *   The size, in bytes, of the data buffer; ordinarily, width * height * bytes per pixel
 *
 * @return The number of bytes read.
 *
 * @see @ref pdv_image(), @ref pdv_start_images()
 */
EDTAPI int pdv_read(PdvDev pdv_p, unsigned char* buf, unsigned long size);

/**
 * @brief Start image acquisition if not already started, then wait for and return
 * the address of the next available image.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The address of the image buffer that has been acquired.
 *
 * @see @ref pdv_start_images(), @ref pdv_wait_images()
 */
EDTAPI uint8_t* pdv_image(PdvDev pdv_p);

/**
 * @brief Sets the number of multiple buffers to use in ring buffer continuous mode,
 * and allocates them. This routine allocates the buffers itself, in kernel or
 * low memory as required by the EDT device driver for optimal DMA.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param numbufs
 *   The number of buffers to allocate.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_buffer_addresses()
 */
EDTAPI int pdv_multibuf(PdvDev pdv_p, int numbufs);

/**
 * @brief Starts multiple image acquisition.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param count
 *   Number of images to start. A value of 0 starts freerun. To stop freerun,
 *   call @ref pdv_start_images again with a @p count of 1 or more.
 *
 * @see @ref pdv_wait_images(), @ref pdv_multibuf()
 */
EDTAPI void pdv_start_images(PdvDev pdv_p, int count);

/**
 * @brief Returns the addresses of the buffers allocated by the last call to
 * @ref pdv_multibuf(). See @ref pdv_wait_images() for a description and example
 * of use.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return An array of pointers to image buffers. The size of the array is equal
 * to the number of buffers allocated.
 *
 * @see @ref pdv_multibuf()
 */
EDTAPI uint8_t** pdv_buffer_addresses(PdvDev pdv_p);

/**
 * @brief Waits for the images started by @ref pdv_start_images.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param count
 *   Number of images to wait for before returning. If @c count is greater than
 *   the number of buffers set by @ref pdv_multibuf, only the last @c count
 *   images will be available when this function returns.
 *
 * @return The address of the last image.
 *
 * @see @ref pdv_wait_images_timed_raw()
 */
EDTAPI uint8_t* pdv_wait_images(PdvDev pdv_p, int count);

/**
 * @brief Waits for one or more images started by @ref pdv_start_images.
 * Optionally controls post-processing and returns a timestamp.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param count
 *   Number of images to wait for before returning. If @c count is greater than
 *   the number of buffers set by @ref pdv_multibuf, only the last @c count
 *   images will be available when this function returns.
 * @param[out] ts
 *   If non-null, returns the time when the last image DMA transfer finished.
 *   Pass NULL to skip getting the last image's timestamp.
 * @param skip_deinterleave
 *   If true, any de-interleaving post pocessing will be skipped.
 *
 * @return The address of the last image.
 *
 * @see @ref pdv_start_images(), @ref pdv_wait_images(), @ref pdv_decode_timestamp()
 */
EDTAPI uint8_t* pdv_wait_images_timed_raw(PdvDev pdv_p,
                                          int count,
                                          edt_timespec_t* ts,
                                          bool skip_deinterleave);

/**
 * @brief Waits for the last image that has been acquired.
 * A convenience wrapper of @ref pdv_wait_last_image_timed_raw().
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] num_skipped
 *   Pointer to an integer which will be filled in with number of images
 *   skipped, if any. Can be set to NULL to ignore.
 *
 * @return Address of the image.
 *
 * @see @ref pdv_start_images(), @ref pdv_wait_images(), @ref pdv_wait_last_image_timed_raw()
 */
EDTAPI uint8_t* pdv_wait_last_image(PdvDev pdv_p, int* num_skipped);

/**
 * @brief Waits for the last image that has been acquired.
 * Optionally controls post-processing and returns a timestamp.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] num_skipped
 *   Pointer to an integer which will be filled in with number of images
 *   skipped, if any. Can be set to NULL to ignore.
 * @param[out] ts
 *   If non-null, returns the time when the last image DMA transfer finished.
 *   Pass NULL to ignore. See @ref pdv_wait_images_timed_raw for a detailed
 *   explanation.
 * @param skip_deinterleave
 *   If true, any de-interleave post pocessing will be skipped.
 *
 * @return The address of the last image.
 *
 * @see @ref pdv_start_images(), @ref pdv_wait_images(), @ref pdv_wait_last_image()
 */
EDTAPI uint8_t* pdv_wait_last_image_timed_raw(PdvDev pdv_p,
                                              int* num_skipped,
                                              edt_timespec_t* ts,
                                              bool skip_deinterleave);

/**
 * @brief Returns a pointer to the last image that was acquired (non-blocking).
 * It will return a pointer to the same buffer if called a second time
 * with no new images acquired.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The address of the last image acquired.
 *
 * @see @ref pdv_wait_last_image(), @ref pdv_wait_last_image_timed_raw(),
 *      @ref pdv_wait_images(), @ref edt_done_count()
 */
EDTAPI uint8_t* pdv_get_last_image(PdvDev pdv_p);

/**
 * @brief Returns a pointer to the last image that was acquired (non-blocking).
 * Identical to the @ref pdv_get_last_image(), except that it skips any
 * image deinterleave method defined by the \b method_interlace
 * config file directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The address of the last image acquired.
 *
 * @see @ref pdv_get_last_image()
 */
EDTAPI uint8_t* pdv_get_last_image_raw(PdvDev pdv_p);

/**
 * @brief Waits for the _next_ image, skipping any previously started images.
 * A convenience wrapper of @ref pdv_wait_next_image_timed_raw().
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] num_skipped
 *   Pointer to an integer which will be filled in with number of images
 *   skipped, if any. Can be set to NULL to ignore.
 *
 * @return The address of the image.
 *
 * @see @ref pdv_start_images(), @ref pdv_wait_images(),
 *      @ref pdv_wait_next_image_timed_raw()
 */
EDTAPI uint8_t* pdv_wait_next_image(PdvDev pdv_p, int* num_skipped);

/**
 * @brief Waits for the _next_ image, skipping any previously started images.
 * Optionally controls post-processing and returns a timestamp.
 *
 * The format of the returned data and timestamp is described in
 * @ref pdv_wait_images_timed_raw.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] num_skipped
 *   Pointer to an integer which will be filled in with number of images
 *   skipped, if any. Can be set to NULL to ignore.
 * @param[out] ts
 *   If non-null, returns the time when the last image DMA transfer finished.
 *   Pass NULL to ignore. See @ref pdv_wait_images_timed_raw for a detailed
 *   explanation.
 * @param skip_deinterleave
 *   If true, any de-interleave post processing will be skipped.
 *
 * @return The address of the image.
 *
 * @see @ref pdv_start_images(), @ref pdv_wait_images(), @ref pdv_wait_next_image()
 */
EDTAPI uint8_t* pdv_wait_next_image_timed_raw(PdvDev pdv_p,
                                              int* num_skipped,
                                              edt_timespec_t* ts,
                                              bool skip_deinterleave);

/**
 * @brief Decode the timestamp from @ref pdv_wait_images_timed_raw and related
 * functions.
 *
 * Converts the timestamp returned by the _timed subroutines to floating-point,
 * representing the value in seconds.
 *
 * @param ts
 *   The timestamp returned from @ref pdv_wait_images_timed_raw or similar
 *   function.
 *
 * @return The timestamp as a double-precision floating-point value.
 *
 * @see @ref pdv_wait_images_timed_raw(), @ref pdv_wait_last_image_timed_raw(),
 *      @ref pdv_wait_next_image_timed_raw()
 */
EDTAPI double pdv_decode_timestamp(const edt_timespec_t* ts);

/**
 * @brief Performs setup for continuous transfers.
 *
 * Shouldn't need to be called by user apps since @ref pdv_start_images(), etc.
 * call it already. But it is in some EDT example applications from before
 * this was the case.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @see @ref pdv_stop_continuous()
 */
EDTAPI void pdv_setup_continuous(PdvDev pdv_p);

/**
 * @brief Performs un-setup for continuous transfers.
 *
 * Shouldn't need to be called by user apps since other subroutines
 * (e.g. @ref pdv_timeout_restart()) now call it as needed. But it is still in
 * some EDT example applications from before this was the case.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @see @ref pdv_setup_continuous()
 */
EDTAPI void pdv_stop_continuous(PdvDev pdv_p);

/**
 * @brief Gets the status of the continuous flag.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if continuous, 0 if not.
 *
 * @see @ref pdv_setup_continuous(), @ref pdv_stop_continuous()
 */
EDTAPI int pdv_in_continuous(PdvDev pdv_p);

/**
 * @brief Flushes the board's input FIFOs, to allow new data transfers to start
 * from a known state.
 *
 * This subroutine effectively resets the board, so calling it after every
 * image will degrade performance and is not recommended. Additionally, resetting
 * after a timeout, involves more than just flushing the FIFOs -- therefore we
 * recommend using @ref pdv_timeout_restart() to reset (which calls this, among
 * other things).
 *
 * @param pdv_p
 *   The open PDV device handle.
 */
EDTAPI void pdv_flush_fifo(PdvDev pdv_p);

/**
 * @brief Sets the length of time to wait for data on acquisition before timing out.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   @arg `> 0` -- The number of milliseconds to wait for timeout.
 *   @arg `0` -- Block forever waiting for data.
 *   @arg `-1` -- Revert to automatic timeouts.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see \b user_timeout directive in the <a
 * href="https://edt.com/downloads/camera-configuration-guide.pdf">Camera Configuration Guide</A>
 */
EDTAPI int pdv_set_timeout(PdvDev pdv_p, int value);

/**
 * @brief Gets the length of time to wait for data on acquisition before timing out.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The timeout value, in milliseconds.
 */
EDTAPI int pdv_get_timeout(PdvDev pdv_p);

/**
 * @brief Returns the number of times the device timed out (frame didn't transfer
 * completely or at all) since the device was opened.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of timeouts since the device was opened.
 *
 * @see @ref pdv_set_timeout(), @ref pdv_get_timeout(),
 *      @ref pdv_timeout_restart(), @ref pdv_enable_framesync()
 */
EDTAPI int pdv_timeouts(PdvDev pdv_p);

/**
 * @brief Cleans up after a timeout, particularly when you've prestarted multiple
 * buffers or if you've forced a timeout with @ref edt_do_timeout().
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param restart
 *   Whether to immediately restart acquiring.
 *
 * @return The number of buffers left undone.
 *
 * @see @ref pdv_timeouts()
 */
EDTAPI int pdv_timeout_restart(PdvDev pdv_p, int restart);

/**
 * @brief Determines whether data overran on the last aquire.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of bytes of data remaining from last aquire.
 *         0 indicates no overrun.
 */
EDTAPI int pdv_overrun(PdvDev pdv_p);

/**
 * @brief Search an image buffer for a valid frame.
 * If the beginning of the image was not at the start of the image buffer, stop
 * and re-start continuous image aquisition.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param image
 *   Image buffer to inspect.
 * @param imagesize
 *   Total size (in bytes) of the image in @p imagebuf.
 * @param verbose
 *   Pass @c true (1) to enable more log messages.
 */
EDTAPI void pdv_check_frame(PdvDev pdv_p, uint16_t* image, uint32_t imagesize, int verbose);

/**
 * @brief Returns the post-processed data corresponding to a given raw buffer.
 * It does not cause post-processing to happen, it only returns the data if it already
 * exists.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param buf
 *   A pointer to the raw data buffer.
 * @param bufnum
 *   The index of the data buffer in the buffer array.
 *
 * @return A pointer to the post-processed buffer.
 */
EDTAPI uint8_t* pdv_get_interleave_data(PdvDev pdv_p, uint8_t* buf, int bufnum);

/**
 * @brief Returns the value of the \b force_single flag. This flag is 0 by default,
 * and is set by the \b force_single directive in the config file.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @return Value of the force_single flag.
 *
 * @see @ref pdv_multibuf(),
 *      \b force_single
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      camera configuration</A> directive.
 */
EDTAPI int pdv_get_force_single(PdvDev pdv_p);

/**
 * @brief Enables frame valid done functionality on the board.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param enable
 *   Whether to enable of disable frame valid done.
 *
 * @see @ref pdv_get_lines_xferred(), \b fval_done directive in the
 * <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 * Camera Configuration Guide</A>
 */
EDTAPI void pdv_set_fval_done(PdvDev pdv_p, int enable);

/**
 * @brief Returns the frame valid done enable state.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return  Whether the frame valid done feature is enabled.
 *
 * @see @ref pdv_set_fval_done()
 */
EDTAPI int pdv_get_fval_done(PdvDev pdv_p);

/**
 * @brief Gets the number of lines transferred during the last acquire.
 * Typically only used in line scan applications where the actual number
 * of lines transferred into a given buffer is unknown at the time of
 * the acquire.
 *
 * Note that if acquires are continuously being
 * queued (as in \c pdv_start_images(pdv_p, n) where \e n is
 * greater than 1), the number of lines tranferred may not reflect
 * the last finished acquire.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of lines transferred on the last acquire.
 *
 * @see @ref pdv_get_width_xferred()
 */
EDTAPI int pdv_get_lines_xferred(PdvDev pdv_p);

/**
 * @brief Gets the number of pixels transferred during the last line transferred.
 * Typically only used in line scan applications where the actual number
 * of pixels transferred per line may not be known.
 *
 * Note that if lines are continuously being transferred (the normal case),
 * the number of pixels tranferred may not reflect the last finished line.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of pixels transferred on the last line.
 *
 * @see @ref pdv_get_lines_xferred()
 */
EDTAPI int pdv_get_width_xferred(PdvDev pdv_p);

/**
 * @brief Gets the number of frame valid transitions that have been seen by
 * the board since the last time the board/channel was initialized or
 * the last time @ref pdv_cl_reset_fval_counter() was called.
 *
 * The number returned is NOT the number of frames read in; the counter
 * on the board counts all frame ticks seen whether images are being read
 * in or not. As such this subroutine can be useful in determining whether
 * a camera is connected (among other things), assuming that the camera
 * is a freerun camera or has a continuous external trigger.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of frame valid transitions seen.
 *
 * @see @ref pdv_cl_reset_fval_counter()
 */
EDTAPI int pdv_cl_get_fval_counter(PdvDev pdv_p);

/**
 * @brief Resets the frame valid counter to zero.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @see @ref pdv_cl_get_fval_counter()
 */
EDTAPI void pdv_cl_reset_fval_counter(PdvDev pdv_p);
/** @} */ /* end acquisition */

/** @addtogroup settings
 * @{
 */

/**
 * Gets the type of the camera, as set by \e initcam from the camera
 * configuration file's camera description directives. This is a
 * concatenation of \b camera_class, \b camera_model, and \b camera_info,
 * directives.
 *
 * @note the camera class, model and info are for information only, and are not
 * used by the driver or library. They are intended for use by applications to
 * allow the user to browse and select a specific camera configuration.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return A string representing the camera type.
 *
 * @see @ref pdv_get_camera_class(), @ref pdv_get_camera_model(),
 *      @ref pdv_get_camera_info(),
 *       \b camera_class, \b camera_model, \b camera_info directives in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>
 */
EDTAPI const char* pdv_get_camera_type(PdvDev pdv_p);

/**
 * @brief Gets the class of the camera (usually the manufacturer name),
 * as set by \e initcam from the camera_config file \b camera_class directive.
 *
 * @note the camera class, model and info are for information only, and are not
 * used by the driver or library. They are intended for use by applications to
 * allow the user to browse and select  a specific camera configuration.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return A string representing the camera class.
 *
 * @see @ref pdv_get_camera_type(), \b camera_class directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>
 */
EDTAPI const char* pdv_get_camera_class(PdvDev pdv_p);

/**
 * @brief Gets the model of the camera, as set by initcam from the camera_config
 * file \b camera_model directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return A string representing the camera model.
 *
 * @see \b camera_model directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>
 */
EDTAPI const char* pdv_get_camera_model(PdvDev pdv_p);

/**
 * @brief Gets the string set by the \b camera_info configuration file directive.
 *
 * See @ref pdv_get_camera_type() for more information on camera strings.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return A string representing the camera info.
 *
 * @see \b camera_info directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>
 */
EDTAPI const char* pdv_get_camera_info(PdvDev pdv_p);

/**
 * @brief Sets width and reallocates buffers accordingly.
 *
 * Since we rarely ever set width and not height, you should normally
 * just use @ref pdv_set_image_size() to set both.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The new width in pixels.
 *
 * @return 0 on success, -1 on error.
 */
EDTAPI int pdv_set_width(PdvDev pdv_p, int value);

/**
 * @brief Gets the width of the image (number of pixels per line), based on the
 * camera in use.
 *
 * If the width has been changed by setting a region of interest, the
 * modified values are returned; use @ref pdv_get_cam_width() to get the
 * unchanged width.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The width in pixels of images returned from an aquire.
 *
 * @see @ref pdv_get_cam_width(), \b width directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_width(PdvDev pdv_p);

/**
 * @brief Gets the number of bytes per line.
 *
 * The result will be different than @ref pdv_get_width() if the
 * pixel depth is not 8 bits.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The pitch in bytes of images returned from an aquire.
 */
EDTAPI int pdv_get_pitch(PdvDev pdv_p);

/**
 * @brief Sets height and reallocates buffers accordingly.
 *
 * Since we rarely ever set height and not width, you should normally
 * just use @ref pdv_set_image_size() to set both.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The new height in pixels.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_set_height(PdvDev pdv_p, int value);

/**
 * @brief Gets the height of the image (number of lines), based on the camera
 * in use.
 *
 * If the height has been changed by setting a region of interest, the
 * modified values are returned; use @ref pdv_get_cam_height() to get the
 * unchanged height.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The height in pixels of images returned from an acquire.
 *
 * @see @ref pdv_get_cam_height(), \b height directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_height(PdvDev pdv_p);

/**
 * @brief Gets the height of a single image when the system is configured for multiple
 *        images per buffer.
 *
 * Unless the frame_height directive is specified in the config file this will
 * return zero.
 *
 * This setting is useful in some cases where special handling
 * of image data by an application is used such as multiple frames per image.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The height of a single image frame in pixels.
 *
 * @see @ref pdv_get_height(), @ref pdv_get_cam_height()
 */
EDTAPI int pdv_get_frame_height(PdvDev pdv_p);

/**
 * @brief Sets the bit depth, extended depth, and camera link data path.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param depth
 *   The new depth value.
 * @param extdepth
 *   The new extended depth value.
 * @param dpath
 *   The new camera link data path value. Passing zero (0) will cause an appropriate
 *   value to be calculated automatically.
 *
 * @return 0 on success
 *
 * @see @ref pdv_cl_set_base_channels(), @ref pdv_get_depth(), @ref pdv_get_extdepth(),
 *      @c depth, @c extdepth, @c CL_DATA_PATH_NORM directives in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide </a>
 */
EDTAPI int pdv_set_depth_extdepth_dpath(PdvDev pdv_p, int depth, int extdepth, unsigned int dpath);

/**
 * @brief Gets the depth of the image (number of bits per pixel), as set in the
 *        configuration file for the camera in use.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of bits per pixel in the image.
 *
 * @see @ref pdv_set_depth_extdepth_dpath(), @ref pdv_get_extdepth(),
 *      \b depth directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_depth(PdvDev pdv_p);

/**
 * @brief Gets the extended depth of the camera.
 *
 * The extended depth is the number of valid bits per pixel that the
 * camera outputs, as set by \e initcam from the configuration file
 * \b edtdepth directive.
 * Note that if \b depth is set differently than \b extdepth, the actual
 * number of bits per pixel passed through by the EDT framegrabber board will be
 * different. For example, if \b extdepth is 10 but \b depth is 8, the
 * board will only transfer one byte per pixel, even though the camera is
 * outputting two bytes per pixel.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The extended depth.
 *
 * @see @ref pdv_get_depth(), \b extdepth directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_extdepth(PdvDev pdv_p);

/**
 * @brief Set the number of channels (taps) and horizontal and vertical alignment of the taps.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param htaps
 *   The number of horizontal taps.
 * @param vtaps
 *   The number of vertical taps.
 *
 * @see @ref pdv_set_depth_extdepth_dpath(),
 *      \b hskip, \b vskip and \b CL_DATA_PATH_NORM directives in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI void pdv_cl_set_base_channels(PdvDev pdv_p, int htaps, int vtaps);

/**
 * @brief Set the device's exposure method and CC line state.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param method
 *   The exposure method.
 * @param mcl
 *   The mode control (CC line) state.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_get_shutter_method()
 */
EDTAPI int pdv_set_shutter_method(PdvDev pdv_p, int method, unsigned int mcl);

/**
 * @brief Returns the shutter (expose) timing method and mode control (CC) state.
 *
 * See \ref pdv_set_shutter_method for an explanation of the return value
 * (shutter method) and @p mcl parameter;
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] mcl
 *   Returns mode control (CC line) state. Pass NULL to skip reading the current state.
 *
 * @return The shutter (expose) timing method.
 *
 * @see @ref pdv_set_shutter_method()
 */
EDTAPI int pdv_get_shutter_method(PdvDev pdv_p, unsigned int* mcl);

/**
 * @brief Returns the interleave method, as set from the \b method_interlace directive in the
 * configuration file (from @ref pdv_initcam()). This method is used to determine how
 * the image data will be rearranged (if at all) before being returned from @ref pdv_wait_images()
 * or @ref pdv_read().
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The interleave method.
 *
 * @see \b method_interlace directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_interlace_method(PdvDev pdv_p);

/**
 * @brief Sets the exposure time, using the method defined by the directives in the camera
 *        configuration file, if set.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The exposure time. For AIA_MCL or AIA_MCL_100US, the valid range is 0-25500. For other
 *   methods, valid range and increments are camera-dependent.
 *
 * @return 0 if successful, -1 if unsuccessful.
 * @see @ref pdv_set_shutter_method(), @ref pdv_get_shutter_method(),
 *      \b MODE_CNTL_NORM, \b serial_exposure and \b method_camera_shutter_timing
 *      directives in the <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_set_exposure(PdvDev pdv_p, int value);

/**
 * @brief Gets the exposure time on the digital imaging device.
 *
 * Applies only when using board-controlled shutter timing for which
 * shutter timing methods have been programmed into the library.  The
 * valid range is camera-dependent.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return Exposure time, in milliseconds.
 *
 * @see @ref pdv_set_exposure(), \b method_camera_shutter_timing directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_exposure(PdvDev pdv_p);

/**
 * @brief Sets the gain on the input device.
 *
 * Applies only to cameras for which extended control capabilities have
 * been added to the library, or that have a serial command protocol that
 * has been configured using the \b serial_gain configuration directive.
 * Unless you know that one of the above has been implemented for your camera,
 * it is usually safest to just send the specific serial commands via
 * @ref pdv_serial_command() or @ref pdv_serial_write().
 *
 * The valid range is -128 to 128.  The actual range is camera-dependent.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The gain setting.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_get_gain(), \b seriial_gain directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_set_gain(PdvDev pdv_p, int value);

/**
 * @brief Gets the gain on the device.
 *
 * Applies only to cameras for which extended control capabilities have
 * been written into the library, such as the Kodak Megaplus i series.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The gain value.
 *
 * @see @ref pdv_set_gain(), \b seriial_gain directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_gain(PdvDev pdv_p);

/**
 * @brief Sets the black level (offset) on the input device.
 *
 * Applies only to cameras for which extended control capabilities have
 * been added to the library (see the source code), or that have a serial
 * command protocol that has been configured using the \b serial_offset
 * configuration directive. Unless you know that one of the above has
 * been implemented for your camera, it is usually safest to just send
 * the specific serial commands via @ref pdv_serial_command() or
 * @ref pdv_serial_write().
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The black level value. The valid range is camera-dependent.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_get_blacklevel(), \b serial_offset directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_set_blacklevel(PdvDev pdv_p, int value);

/**
 * @brief Gets the black level (offset) on the imaging device.
 *
 * Applies only to cameras for which extended control capabilities have
 * been written into the library, such as the Kodak Megaplus i series.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The black level value.
 */
EDTAPI int pdv_get_blacklevel(PdvDev pdv_p);

/**
 * @brief Set binning on the camera to the specified values, and recalculate the
 * values that will be returned by @ref pdv_get_width(), @ref pdv_get_height(),
 * and @ref pdv_get_image_size().
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param xval
 *   The x binning value. Usually 1, 2, 4 or 8. Default is 1.
 * @param yval
 *   The y binning value. Usually 1, 2, 4 or 8. Default is 1.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see \b serial_binning directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_set_binning(PdvDev pdv_p, int xval, int yval);

/**
 * @brief Set the time to wait between acquiring images.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param delay_ms
 *   The delay in milliseconds.
 */
EDTAPI void pdv_set_start_delay(PdvDev pdv_p, int delay_ms);

/**
 * @brief Get the time to wait between acquiring images.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The delay in milliseconds.
 */
EDTAPI int pdv_get_start_delay(PdvDev pdv_p);

/**
 * @brief Enables on-board region of interest.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param flag
 *   Nonzero to enable region of interest; 0 to disable it.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_set_roi() for an example.
 */
EDTAPI int pdv_enable_roi(PdvDev pdv_p, int flag);

/**
 * @brief Read if region of interest (ROI) is enabled.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return Nonzero if the region of interest is enabled.
 */
EDTAPI int pdv_get_roi_enabled(PdvDev pdv_p);

/**
 * @brief Sets a rectangular region of interest, supporting cropping.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param hskip
 *    The X coordinate of the upper left corner of the region of interest.
 * @param hactv
 *    The width (number of pixels per line) of the region of interest.
 * @param vskip
 *    The Y coordinate of the upper left corner of the region of interest.
 * @param vactv
 *    The height (number of lines per frame) of the region of interest.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_enable_roi(), \b vskip, \b vactv, \b hskip, \b hactv directives in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_set_roi(PdvDev pdv_p, int hskip, int hactv, int vskip, int vactv);

/**
 * @brief Set ROI to camera width/height; adjust ROI width to be a multiple of 4,
 *        and enable ROI.
 *
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_auto_set_roi(PdvDev pdv_p);

/**
 * @brief Sets placeholder for original full camera frame width, unaffected
 *  by ROI changes and usually only called by pdv_initcam.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The width of the camera's sensor in pixels.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_set_cam_width(PdvDev pdv_p, int value);

/**
 * @brief Sets placeholder for original full camera frame height, unaffected
 * by ROI changes and usually only called by pdv_initcam.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param value
 *   The height of the camera's sensor in pixels.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_set_cam_height(PdvDev pdv_p, int value);

/**
 * @brief Gets the minimum allowable exposure value for this camera, as set by
 * \e initcam from the camera_config file \b shutter_speed_min directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The minimum exposure value.
 *
 * @see \b shutter_speed_min directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_min_shutter(PdvDev pdv_p);

/**
 * @brief Gets the maximum allowable exposure value for this camera, as set by
 * \e initcam from the camera_config file \b shutter_speed_max directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The maximum exposure value.
 *
 * @see \b shutter_speed_max directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_max_shutter(PdvDev pdv_p);

/**
 * @brief Gets the minimum allowable gain value for this camera, as set by \e initcam
 * from the camera configuration file \b gain_min directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The minimum gain value.
 *
 * @see \b gain_min directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_min_gain(PdvDev pdv_p);

/**
 * @brief Gets the maximum allowable gain value for this camera, as set by \e initcam
 * from the camera configuration file \b gain_max directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The maximum gain value.
 *
 * @see \b gain_max directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_max_gain(PdvDev pdv_p);

/**
 * @brief Gets the minimum allowable offset (black level) value for this camera, as set
 * by \e initcam from the camera configuration file \b offset_min directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The minimum offset value.
 *
 * @see \b offset_min directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_min_offset(PdvDev pdv_p);

/**
 * @brief Gets the maximum allowable offset (black level) value for this camera, as set
 * by \e initcam from the camera configuration file \b offset_max directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The maximum offset value.
 *
 * @see \b offset_max directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_max_offset(PdvDev pdv_p);

/**
 * @brief Tell the EDT framegrabber hardware to invert each pixel before transferring
 * it to the host computer's memory.
 *
 * This is implemented in firmware and has no impact on performance.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param val
 *   Nonzero to enable invert mode; 0 to disable it.
 */
EDTAPI void pdv_set_invert(PdvDev pdv_p, int val);

/**
 * @brief Get the state of the hardware invert register enable bit.
 * See @ref pdv_set_invert for details on this feature.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return Nonzero if invert mode is enabled.
 */
EDTAPI int pdv_get_invert(PdvDev pdv_p);

/**
 * @brief Enable hardware overwrite of first two bytes of the frame with a counter.
 *
 * The counter increments by one for every frame received by the framegrabber.
 * Disabling this also resets the counter to zero, unless framesync mode is
 * also enabled (see @ref pdv_enable_framesync()).
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param val
 *   Nonzero to enable the counter; 0 to disable it.
 */
EDTAPI void pdv_set_firstpixel_counter(PdvDev pdv_p, int val);

/**
 * @brief Query state of the hardware first pixel counter register enable bit.
 * See @ref pdv_set_firstpixel_counter() for details on this feature.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return Nonzero if the counter is enabled.
 */
EDTAPI int pdv_get_firstpixel_counter(PdvDev pdv_p);

/**
 * @brief Sets the header (or footer) type.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param header_type
 *   Header type, from the @ref edt_pdv_header_type enum.
 * @param irig_offset
 *   Timecode offset in seconds. Typically set to 2 seconds.
 *   Ignored if header_type is @ref PDV_HDR_TYPE_NONE.
 * @param irig_raw
 *   0 = Enable UNIX timecode format. 1 = Enable raw timecode format.
 *   Ignored if header_type is @ref PDV_HDR_TYPE_NONE.
 *
 * @return 0 in success, -1 on failure.
 *
 * @see @ref pdv_set_header_size(), \b method_header_type directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>,  and the Timestamp appendix in the
 *      User's Guide.
 */
EDTAPI int pdv_set_header_type(PdvDev pdv_p,
                               EdtPdvHeaderType header_type,
                               int irig_offset,
                               int irig_raw);

/**
 * @brief Get the header (or footer) type.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The header type as an edt_pdv_header_type enum.
 *
 * @see @ref pdv_set_header_type(), \b method_header_type directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI EdtPdvHeaderType pdv_get_header_type(PdvDev pdv_p);

/**
 * @brief Sets the header (or footer) size, in bytes, for the device.
 *
 * This can also be done by using the \b header_size directive in the
 * camera configuration file.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param header_size
 *   The new header size in bytes.
 *
 * @see @ref pdv_get_header_size(), \b header_size directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI void pdv_set_header_size(PdvDev pdv_p, int header_size);

/**
 * @brief Returns the currently defined header or footer size.
 *
 * This is usually set in the configuration file with the directive
 * \b header_size.  It can also be set by calling @ref pdv_set_header_size().
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return Current header size.
 *
 * @see @ref pdv_set_header_size(), \b header_size directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_header_size(PdvDev pdv_p);

/**
 * @brief Sets the header (or footer) position.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param header_position
 *   The starting point for the header.
 *
 * @see @ref pdv_get_header_offset(), @ref pdv_set_header_offset(),
 *      @ref EdtPdvHeaderPosition.
 */
EDTAPI void pdv_set_header_position(PdvDev pdv_p, EdtPdvHeaderPosition header_position);

/**
 * @brief Returns the header or footer position value.
 *
 * These values can be set in the configuration file with the
 * \b method_header_position directive.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The header position.
 *
 * @see @ref pdv_get_header_offset(), \b method_header_position directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI EdtPdvHeaderPosition pdv_get_header_position(PdvDev pdv_p);

/**
 * @brief Sets the byte offset of the header data in the allocated buffer.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param header_offset
 *   The header offset.
 */
EDTAPI void pdv_set_header_offset(PdvDev pdv_p, int header_offset);

/**
 * @brief Returns the byte offset of the header in the buffer.
 *
 * The byte offset is determined by the header position value.  If header_position
 * is PDV_HEADER_BEFORE, the offset is 0; if header_position is PDV_HEADER_AFTER
 * (i.e. not really a header but a footer), the offset is the image size.  If
 * header_position is PDV_HEADER_MIDDLE, the header offset can be set using the
 * \b header_offset directive in the camera_configuration file, or by calling
 * @ref pdv_set_header_offset().
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return A byte offset from the beginning of the buffer.
 *
 * @see @ref pdv_get_header_position(), @ref pdv_set_header_offset()
 */
EDTAPI int pdv_get_header_offset(PdvDev pdv_p);

/**
 * @brief Sets the boolean value for whether the image header is included in the DMA
 * from the camera.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param header_dma
 *   True to include the header, false if not.
 *
 * @see @ref pdv_get_header_dma()
 */
EDTAPI void pdv_set_header_dma(PdvDev pdv_p, bool header_dma);

/**
 * @brief Returns the current setting for flag which determines whether the header
 * (or footer) size is to be added to the DMA size.  This is true if the camera/device
 * returns header information at the beginning or end of its transfer.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return True if header is included in DMA size.
 */
EDTAPI bool pdv_get_header_dma(PdvDev pdv_p);

/**
 * @brief Sets the width and height of the image.  Tells the driver what
 *  width and height (in pixels) to expect from the camera.
 *
 * This call is ordinarily unnecessary in an application program, because the
 * width and height are set automatically when \e initcam runs. Exceptions
 * can occur, however; for example, if the camera's output size can be
 * changed while running, or if the application performs setup that
 * supersedes initcam.  This routine is provided for these special
 * cases.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param width
 *   The width of the image in pixels.
 * @param height
 *   The  height of the image in pixels.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_set_image_size(PdvDev pdv_p, int width, int height);

/**
 * @brief Returns the size of the image in bytes, absent any padding or header data.
 * Since padding and header data are usually absent, the value returned
 * from this is usually the same as that returned by @ref pdv_get_imghdr_size().
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The image size, in bytes.
 *
 * @see @ref pdv_get_imghdr_size()
 */
EDTAPI int pdv_get_image_size(PdvDev pdv_p);

/**
 * @brief Returns the size of the image buffer in bytes, based on its width,
 * height, and depth. The size returned includes allowance for buffer headers.
 *
 * Enabling a region of interest changes this value.  To obtain
 * the actual size of the image data without any optional header or other
 * padding, see @ref pdv_get_dma_size().
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The total number of bytes in the image, including buffer header
 * overhead.
 *
 * @see @ref pdv_set_roi()
 */
EDTAPI int pdv_get_imghdr_size(PdvDev pdv_p);

/**
 * @brief Returns the actual amount of image data for DMA.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The DMA size in bytes.
 *
 * @see @ref pdv_get_imghdr_size(), @ref pdv_get_header_position(),
 *      @ref pdv_extra_headersize().
 */
EDTAPI int pdv_get_dma_size(PdvDev pdv_p);

EDTAPI int pdv_get_fulldma_size(PdvDev pdv_p, int* extrasizep);

/**
 * @brief Returns the camera image width, in pixels, as set by the
 * configuration file directive \b width.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The image width in pixels.
 *
 * @see @ref pdv_get_width(), \b width directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_cam_width(PdvDev pdv_p);

/**
 * @brief Returns the camera image height, in pixels, as set by the
 * configuration file directive \b height, unaffected by changes made by
 * setting a region of interest. See @ref pdv_set_roi() for more
 * information.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The image height in pixels.
 *
 * @see @ref pdv_get_height(), \b height directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_cam_height(PdvDev pdv_p);

/**
 * @brief Checks for frame sync and frame count.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param image_p
 *   A pointer to previously acquired image (from @ref pdv_wait_images() for example)
 *   for which you want the framesync to be checked.
 * @param framecnt
 *   A pointer to location to put frame counter from this frame.
 *
 * @return 1 if out of sync is detected, 0 if no out of sync detected, -1 on error.
 *
 * @see @ref pdv_enable_framesync(), @ref pdv_framesync_mode()
 */
EDTAPI int pdv_check_framesync(PdvDev pdv_p, uint8_t* image_p, uint32_t* framecnt);

/**
 * @brief Enables frame sync footer and frame out-of-synch detection.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param mode
 *   The frame sync mode should be one of:
 *   @arg `PDV_FRAMESYNC_OFF`: Framesync functionality disabled.
 *   @arg `PDV_FRAMESYNC_ON`:  Framesync functionality enabled.
 *   @arg `PDV_FRAMESYNC_EMULATE_TIMEOUT`: Framesync errors will be reflected as timeouts.
 *
 * @return 0 on success, -1 if not supported by the device in use.
 *
 * @see @ref pdv_check_framesync(), @ref pdv_framesync_mode(), @ref pdv_timeouts()
 */
EDTAPI int pdv_enable_framesync(PdvDev pdv_p, int mode);

/**
 * @brief Returns the frame sync mode.
 *
 * Can be one of:
 *  - `PDV_FRAMESYNC_OFF`: Framesync functionality disabled.
 *  - `PDV_FRAMESYNC_ON`:  Framesync functionality enabled.
 *  - `PDV_FRAMESYNC_EMULATE_TIMEOUT`: Framesync functionality enabled, and framesync errors will be
 * reflected as timeouts.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if enabled, 0 if not enabled.
 *
 * @see @ref pdv_enable_framesync(), @ref pdv_check_framesync()
 */
EDTAPI int pdv_framesync_mode(PdvDev pdv_p);

/**
 * @brief Get the configuration file name.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The configuration file name.
 */
EDTAPI const char* pdv_get_cfgname(PdvDev pdv_p);

/**
 * @brief Set exposure, gain, and blacklevel to default values.
 *
 * @param pdv_p
 *   The open PDV device handle.
 */
EDTAPI void pdv_set_defaults(PdvDev pdv_p);

/**
 * @brief Enables external triggering.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param flag
 *   One of:
 *   @arg 0 -- turn off trigger
 *   @arg 1 -- turn on photo trigger
 *   @arg 2 -- turn on field ID trigger (through camera or cable). Does not apply
 *   to PCI C-Link.
 */
EDTAPI void pdv_enable_external_trigger(PdvDev pdv_p, int flag);

/**
 * @brief Set the frame period counter and enable/disable frame timing.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param period
 *   The frame period in microseconds-2, range 0-16777215
 * @param method
 *   One of:
 *   @arg  0                  -- disable frame counter
 *   @arg `PDV_FMRATE_ENABLE` -- continuous frame counter
 *   @arg `PDV_FVAL_ADJUST`   -- frame counter extends every frame valid
 *                               by 'period' microseconds
 *
 * @return -1 on error, 0 on success.
 *
 * @see @ref pdv_get_frame_period(),
 *      \b frame_period, \b method_frame_timing directives in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_set_frame_period(PdvDev pdv_p, int rate, int method);

/**
 * @brief Get the frame period.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The frame period in microseconds.
 *
 * @see @ref pdv_set_frame_period(), \b frame_period directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_frame_period(PdvDev pdv_p);

/**
 * @brief Get `htaps` and `vtaps` setting values.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] htaps
 *   On success, the current horizontal taps value is returned here. Pass NULL
 *   to ignore.
 * @param[out] vtaps
 *   On success, the current vertical taps value is returned here. Pass NULL
 *   to ignore.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_get_htaps_vtaps(PdvDev pdv_p, int* htaps, int* vtaps);

/**
 * @brief Get `hskip` and `vskip` setting values.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] hskip
 *   On success, the current horizontal skip value is returned here. Pass NULL
 *   to ignore.
 * @param[out] vskip
 *   On success, the current vertical skip value is returned here. Pass NULL
 *   to ignore.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_get_hskip_vskip(PdvDev pdv_p, int* hskip, int* vskip);

/**
 * @brief Get `hactv` and `vactv` setting values.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] hactv
 *   On success, the current horizontal active pixels value is returned here.
 *   Pass NULL to ignore.
 * @param[out] vactv
 *   On success, the current vertical active pixels value is returned here.
 *   Pass NULL to ignore.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_get_hactv_vactv(PdvDev pdv_p, int* hactv, int* vactv);

/**
 * @brief Get value of `fv_once` (frame valid once) flag.
 *
 * @brief pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if `fv_once` is true, 0 if false or on error.
 */
EDTAPI int pdv_get_fv_once(PdvDev pdv_p);

/** @} */ /* end settings */

/**
 * @addtogroup serial
 * @{
 */

/**
 * @brief Sets the baud rate on the serial lines; applies only to cameras with serial
 * control. Valid values are 9600, 19200, 38500, 57500, and 115200.
 *
 * @note The baud rate is ordinarily initialized using the value of the \b serial_baud
 * directive in the configuration file, and defaults to 9600 if the directive is not
 * present. Under most circumstances, applications do not need to set the baud
 * rate explicitly.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param baud
 *   The desired baud rate.
 *
 * @return 0 on success, -1 on error.
 *
 * @see \b serial_baud directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_serial_set_baud(PdvDev pdv_p, int baud);

/**
 * @brief Get the baud rate, typically initialized by the \b serial_baud directive in
 * the config file.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The baud rate in bits/sec, or 0 on error.
 *
 * @see \b serial_baud directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_serial_get_baud(PdvDev pdv_p);

/**
 * @brief Sends an ASCII serial command to the camera, with ASCII camera command
 * formatting. Applies only to cameras that use a serial control method for
 * camera-computer communications.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param cmd
 *   The command to send to the camera.
 *
 * @return 0 on success, -1 on failure
 *
 * @see @ref pdv_serial_term(), @ref pdv_serial_prefix(),
 *      @ref pdv_serial_set_delimiters(), @ref pdv_serial_write()
 */
EDTAPI int pdv_serial_command(PdvDev pdv_p, const char* cmd);

/**
 * @brief Sends a binary serial command to the camera. Applies only to cameras that
 * use a serial control method for for camera-computer communications.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param cmd
 *   A buffer containing the serial command.
 * @param len
 *   The number of bytes to send.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_serial_command(), @ref pdv_serial_read(), @ref pdv_serial_wait()
 */
EDTAPI int pdv_serial_binary_command(PdvDev pdv_p, const uint8_t* cmd, int len);

/**
 * @brief Get the number of bytes available in the driver's serial write buffer.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of bytes available in the driver's write buffer.
 */
EDTAPI int pdv_serial_write_available(PdvDev pdv_p);

/**
 * @brief Performs a serial write over the serial lines. Not recommended for
 * use in new code, consider using @ref pdv_serial_command() or
 * @ref pdv_serial_binary_command() instead.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param buf
 *  The  buffer containing the serial command.
 * @param size
 *  The number of bytes to send.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_serial_command(), @ref pdv_serial_binary_command()
 */
EDTAPI int pdv_serial_write(PdvDev pdv_p, const char* buf, int size);

/**
 * @brief Performs a serial read over the serial control lines.
 *
 * The serial data read will be stored in a user supplied buffer. That buffer
 * will be NULL-terminated. Use pdv_serial_read_nullterm(pdv_p, FALSE) if you
 * don't want that behavior.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] buf
 *   The buffer to receive the data. It must be at least @p size + 1 bytes
 *   (@p size bytes of data plus a one byte NULL terminator).
 * @param size
 *   The number of bytes to be read.
 *
 * @return The number of bytes read into the buffer.
 *
 * @see @ref pdv_serial_wait()
 */
EDTAPI int pdv_serial_read(PdvDev pdv_p, char* buf, int size);

/**
 * @brief Performs a serial read over the serial control lines, blocks until all
 * requested data is read.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] buf
 *   The buffer to receive the data. It must be at least @p size + 1 bytes
 *   (@p size bytes of data plus a one byte NULL terminator).
 * @param size
 *   The number of bytes to be read.
 *
 * @return The number of bytes read into the buffer.
 *
 * @see @ref pdv_serial_wait(), @ref pdv_serial_read()
 */
EDTAPI int pdv_serial_read_blocking(PdvDev pdv_p, char* buf, int size);

/**
 * @brief Performs a serial read over the serial control lines.
 * The buffer passed in will be NULL-terminated if nullterm is true.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] buf
 *   The buffer to receive the data. It must be at least @p size + 1 bytes
 *   (@p size bytes of data plus a one byte NULL terminator).
 * @param size
 *   The number of bytes to be read.
 * @param nullterm
 *   True to null terminate the buffer read in, false to disable that.
 *
 * @return The number of bytes read into buf.
 *
 * @see @ref pdv_serial_read()
 */
EDTAPI int pdv_serial_read_nullterm(PdvDev pdv_p, char* buf, int size, int nullterm);

EDTAPI int pdv_serial_read_enable(PdvDev pdv_p);
EDTAPI int pdv_serial_read_disable(PdvDev pdv_p);

/**
 * @brief Set the prefix and terminator for serial commands.
 *
 * The serial prefix and terminator are typically set through the
 * \b serial_term and \b serial_prefix config file directives.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param newprefix
 *   A string to prefix to serial commands, or NULL.
 * @param newterm
 *   A string to append to serial commands, or NULL.
 *
 * @see @ref pdv_serial_command(),
 *      \b serial_prefix and \b serial_term directives in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI void pdv_serial_set_delimiters(PdvDev pdv_p, const char* newprefix, const char* newterm);

/**
 * @brief Get the serial prefix.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The serial prefix string.
 *
 * @see @ref pdv_serial_command(), \b serial_prefix directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI const char* pdv_serial_prefix(PdvDev pdv_p);

/**
 * @brief Get the serial terminator.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The serial terminator string.
 *
 * @see @ref pdv_serial_command(), \b serial_term directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI const char* pdv_serial_term(PdvDev pdv_p);

/**
 * @brief Send a break condition on the serial connection.
 *
 * @param pdv_p
 *   The open PDV device handle.
 */
EDTAPI void pdv_serial_send_break(PdvDev pdv_p);

/**
 * @brief Resets the serial interface. Clears any outstanding reads and writes
 * and puts the serial interface in a known idle state.
 *
 * @param pdv_p
 *   The open PDV device handle.
 */
EDTAPI void pdv_serial_reset(PdvDev pdv_p);

/**
 * @brief Reads a Basler binary frame command.  Checks the framing and BCC.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] cmd
 *   The buffer to receive the data.
 * @param len
 *   The maximum number of bytes to be read.
 *
 * @return The number of characters read back, or 0 if none or failure.
 *
 * @see BASLER A202K Camera Manual Doc. ID number DA044003
 */
EDTAPI int pdv_serial_read_basler(PdvDev pdv_p, uint8_t* cmd, int len);

/**
 * @brief Send a Basler formatted serial frame. Adds the framing and BCC.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param cmd
 *   The serial data to send.
 * @param len
 *   The number of bytes to send.
 *
 * @return The number of characters read back, or 0 if none or failure.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see BASLER A202K Camera Manual Doc. ID number DA044003
 */
EDTAPI int pdv_serial_write_basler(PdvDev pdv_p, uint8_t* cmd, int len);

/**
 * @brief Read a binary serial response from a Duncantech MS and DT series
 * camera -- checks for STX and size, then waits for size+1 more bytes.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] frame
 *   The buffer to receive the data.
 *
 * @return The number of characters read back, or 0 if none or failure.
 *
 * @see @ref pdv_serial_write_duncan(),
 *      DuncanTech User Manual Doc # 9000-0001-05.
 */
EDTAPI int pdv_serial_read_duncan(PdvDev pdv_p, uint8_t* frame);

/**
 * @brief Send a Duncantech MS / DT series camera frame -- adds the framing and
 * checksum, then sends the command.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param cmdbuf
 *   A  buffer containing the command, minus framing information.
 * @param size
 *   The number of bytes in the cmdbuf.
 *
 * @return 0 on success, -1 on failure.
 *
 * @see @ref pdv_serial_read_duncan(),
 *      DuncanTech User Manual Doc # 9000-0001-05.
 */
EDTAPI int pdv_serial_write_duncan(PdvDev pdv_p, uint8_t* cmdbuf, int size);

/**
 * @brief Waits for a response from the camera as a result of a @ref pdv_serial_write() or
 * @ref pdv_serial_command().
 *
 * After calling this function, use @ref pdv_serial_read()
 * to get the data. For a detailed example of serial communications, see the
 * serial_cmd.c example program.
 *
 * If the timeout value is zero, the default from the \b serial_timeout directive in the
 * config file is used.  If no default timeout value was specified, the timeout will be
 * 1000 milliseconds.
 *
 * If two threads attempt to call pdv_serial_wait() at the same time, one will return
 * immediately with a result of -EBUSY.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param msecs
 *   Number of milliseconds to wait before timing out.
 * @param count
 *   The maximum number of bytes to wait for before returning.
 *
 * @return On success or timeout, the number of bytes of serial data returned from the camera.
 *   Negative error code on error.
 *
 * @see @ref pdv_serial_read(), \b serial_timeout directive in the
 *      <a href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_serial_wait(PdvDev pdv_p, int msecs, int count);

/**
 * @brief Wait for new serial data to come in. Ignore any previously received data.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param msecs
 *   Number of milliseconds to wait before timing out.
 * @param count
 *   The maximum number of bytes to wait for before returning.
 *
 * @return The number of bytes of serial data returned from the camera.
 */
EDTAPI int pdv_serial_wait_next(PdvDev pdv_p, int msecs, int count);

/**
 * @brief Set the serial wait character.
 *
 * Normally @ref pdv_serial_wait() will wait until the serial_timeout period expires before
 * returning (unless the max number of characters is seen). This is the most general
 * purpose and robust method since there's no other way of knowing all different camera
 * response formats. However if each response can be expected to be 1 line terminated
 * by the same character(such as a newline) every time, then setting the serial_waitchar
 * to that character can greatly shorten the time it takes for a @ref pdv_serial_wait() call to
 * return.
 *
 * This character can also be initialized in the
 * <a href="https://edt.com/downloads/camera-configuration-guide.pdf">camera configuration</A>
 * directive \b serial_waitchar.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param enable
 *   Nonzero to enable wait character detection.
 * @param wchar
 *   The end of message character to detect.
 *
 * @return 0 in success, -1 on failure
 *
 * @see @ref pdv_serial_get_waitchar() and \b serial_waitchar directive in the
 *      <a * href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_serial_set_waitchar(PdvDev pdv_p, int enable, unsigned char wchar);

/**
 * @brief Get serial wait character.
 *
 * If wait character detection is enabled, @ref pdv_serial_wait() will return
 * immediately after the wait character is detected.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param[out] wchar
 *   The character wait character detection will wait for.
 *
 * @return 1 if waitchar enabled, 0 if disabled.
 *
 * @see @ref pdv_serial_set_waitchar() and \b serial_waitchar directive in the
 *      <a * href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_serial_get_waitchar(PdvDev pdv_p, unsigned char* wchar);

/**
 * @brief Returns the number of bytes of unread data in the serial response buffer.
 *
 * Similar to @ref pdv_serial_wait() but doesn't wait for any timeout period,
 * nor does it have any minimum count parameter.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The number of bytes of serial data returned from the camera.
 */
EDTAPI int pdv_serial_get_numbytes(PdvDev pdv_p);

/**
 * @brief Returns the block size for serial writes.
 *
 * @return The serial block size in bytes.
 */
EDTAPI int pdv_serial_get_block_size(void);

/**
 * @brief Sets the block size for serial writes.
 *
 * The default size is 1024.
 *
 * @param newsize
 *   The new serial block size in bytes.
 */
EDTAPI void pdv_serial_set_block_size(int newsize);

/**
 * @brief Get the value of the serial timeout.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return The timeout in milliseconds.
 *
 * @see \b serial_timeout directive in the
 *      <a * href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_serial_get_timeout(PdvDev pdv_p);

/**
 * @brief Get the serial pause time, in milliseconds.
 *
 * Set by the @c pause_for_serial directive in a camera configuration file.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @returns The serial pause time in milliseconds.
 *
 * @see \b pause_for_serial directive in the
 *      <a * href="https://edt.com/downloads/camera-configuration-guide.pdf">
 *      Camera Configuration Guide</a>.
 */
EDTAPI int pdv_get_pause_for_serial(PdvDev pdv_p);
/** @} */ /* end serial */

/** @addtogroup utility
 * @{
 */

/**
 * @brief Convenience routine to free the memory allocated with @ref pdv_alloc().
 *
 * @param ptr
 *   Address of memory buffer to free.
 */
EDTAPI void pdv_free(void* ptr);

/**
 * @brief Convenience routine to allocate memory in a system-independent way.
 *
 * The buffer returned is page aligned. Page alignment is required for
 * some EDT image routines and always preferred.  This function uses
 * VirtualAlloc on Windows NT/2000/XP systems, or posix_memalign on
 * Linux/Unix systems.
 *
 * @param size
 *   The number of bytes of memory to allocate
 *
 * @return The address of the allocated memory, or NULL on error.  If NULL, use
 * @ref edt_msg_perror() to print the error.
 *
 * @see @ref pdv_free()
 */
EDT_DEALLOCATE_WITH(pdv_free)
EDTAPI void* pdv_alloc(int size);

/**
 * @brief Determines file access independent of operating system. This a convenience
 * routine that maps to \c access() on Unix/Linux systems, and \c _access on Windows
 * systems.
 *
 * @param fname
 *   The path name of the file to check access of.
 * @param perm
 *   The permission flag(s) to test for. See \c access() (Unix/Linux) or
 *   \c _access (Windows) for valid arguments.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_access(const char* fname, int perm);

/** @brief Draws the digits of a number into an image buffer.
 *
 * @param addr
 *   The address of the image to be marked.
 * @param n
 *   The number to be marked into the image.
 * @param width
 *   The width of the image.
 * @param height
 *   The height of the image.
 * @param x
 *   The position in the image where the number will be drawn.
 * @param y
 *   The position in the image where the number will be drawn.
 * @param depth
 *   The number of bits per pixel of the image.
 * @param fg
 *   The color value to use for drawing the digits.
 */
EDTAPI void
pdv_mark_ras_depth(void* addr, int n, int width, int height, int x, int y, int depth, int fg);

/**
 * @brief Returns bytes per line based on width and bit depth, including depth < 8.
 *
 * @param width
 *  Pixels per line.
 * @param depth
 *   Bits per pixel.
 *
 * @return Bytes per line.
 */
EDTAPI int pdv_bytes_per_line(int width, int depth);

/**
 * @brief Checks whether a camera is connected and turned on.
 *
 * Looks for an active (changing) pixel clock from the camera,
 * and returns 1 if detected.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if a pixel clock is detected, 0 if not detected.
 */
EDTAPI int pdv_cl_camera_connected(PdvDev pdv_p);

/**
 * @brief Infers that this device is connected to is a Camera Link camera
 * (as opposed to RS-422 or LVDS parallel), based on settings
 * from the loaded camera config file.
 *
 * Specifically for framegrabbers, will return false (0) for simulators.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if Camera Link framegrabber, 0 otherwise.
 */
EDTAPI int pdv_is_cameralink(PdvDev pdv_p);

/**
 * @brief Infers that this device is a simulator, e.g. VisionLink CLS.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if a simulator, 0 otherwise.
 */
EDTAPI int pdv_is_simulator(PdvDev pdv_p);
/** @} */ /* end utility */

#ifdef __cplusplus
}
#endif

#endif /* LIBPDV_H */
