/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2007 Engineering Design Team, Inc.
 * @brief EDT frame grabber initialization utilities.
 */
#ifndef EDT_PDV_INITCAM_H
#define EDT_PDV_INITCAM_H

#include "edtinc.h"
#include "pdv_dependent.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Enum for serial tagging options.
 */
enum serial_tag { SERIAL_TAG_BINARY, SERIAL_TAG_ASCII, SERIAL_TAG_BASLERF, SERIAL_TAG_DUNCANF };

/**
 * Contains information relating to serial nodes.
 */
typedef struct serial_init_node {
  enum serial_tag tag; /* What kind of serial the node is using. */
  char* data;          /* Serial initialization data. */
  struct serial_init_node* next;
} SerialInitNode;

/**
 * Contains important information not covered by the persistent dependent structure.
 */
typedef struct {
  uint32_t startdma;
  uint32_t enddma;
  uint32_t flushdma;
  uint32_t timeout;
  SerialInitNode* serial_init;
  int serial_init_delay;
  uint32_t cl_mgtspeed;
} Edtinfo;

/**
 * @brief Reads configuration file and fills in the dependent and edtinfo
 * structures based on the information in the file. These structures
 * can then be passed in to @ref pdv_initcam() to initialize the
 * board and camera.
 *
 * @param cfgfile
 *   The path name of configuration file to use.
 * @param dd_p
 *   Pointer to a persistent dependent structure; previously allocated with @ref
 * pdv_alloc_dependent(), but not yet initialized.
 * @param ei_p
 *   Structure holding non-persistent initialization strings and variables (information not in @p
 * dd_p).
 *
 * @return 0 on success, -1 on failure
 */
EDTAPI int pdv_readcfg(const char* cfgfile, PdvDependent dd_p, Edtinfo* ei_p);

/**
 * @brief Print the contents of the dependent structure.
 *
 * @param dd_p
 *   Pointer to a persistent dependent structure; previously allocated with @ref
 * pdv_alloc_dependent() and initialized with @ref pdv_readcfg().
 *
 * @return the size of the dependent structure provided.
 */
EDTAPI int pdv_printcfg(PdvDependent dd_p);

/**
 * @brief Initializes the frame grabber and sets the expected camera configuration.
 * This is the "guts" of the `initcam` CLI application that gets executed to
 * initialize when you choose a camera. The library subroutine is provided for
 * programmers who wish to incorporate the initialization procedure into their own
 * applications.
 *
 * @param pdv_p
 *   The open EDT device handle.
 * @param dd_p
 *   Pointer to a persistent dependent structure; previously allocated with @ref
 * pdv_alloc_dependent() and initialized with @ref pdv_readcfg().
 * @param unit
 *   Unit number of the device. Unit numbers start at 0.
 * @param ei_p
 *   Structure holding non-persistent initialization strings and variables (information not in @p
 * dd_p).
 * @param cfgfname
 *   The path name of the configuration file to use.
 * @param bitdir
 *   The directory path name for FPGA .bit files; if NULL, will search for bitfiles
 * under: ".", then "./camera_config/bitfiles".
 * @param pdv_debug
 *   Currently ignored, should be set to 0.
 *
 * @return 0 on success, -1 on failure.
 */
EDTAPI int pdv_initcam(PdvDev edt_p,
                       PdvDependent dd_p,
                       int unit,
                       Edtinfo* ei_p,
                       const char* cfgfname,
                       char* bitdir,
                       int pdv_debug);

#ifdef __cplusplus
} /* extern C */
#endif

#endif /* EDT_PDV_INITCAM_H */
