/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 * @brief PCI DV Library routines to reorder/interpolate image buffers from
 * cameras (header file).
 */

#ifndef PDV_INTERLACE_H
#define PDV_INTERLACE_H

#include "libpdv.h"
#include "pdv_dependent.h"
#include "edt_portability.h"

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PROCESS_ALL    0
#define PROCESS_SINGLE 1
#define PROCESS_NONE   2 /* disable */

typedef struct pdv_post_processing PdvPostProc;

typedef int (*post_process_f)(void* src, int width, int height, void* dest, PdvPostProc* pCtrl);

EDTAPI PdvPostProc* pdv_setup_postproc(PdvDev pdv_p, PdvDependent dd_p, const PdvPostProc* pInCtrl);

EDTAPI int pdv_set_postproc(PdvPostProc* pCtrl,
                            int depth,
                            int extdepth,
                            int frame_height,
                            int interlace,
                            int image_offset,
                            int order,
                            int n_intlv_taps,
                            PdvInterleaveTap* taps);

EDTAPI int pdv_set_postproc_order(PdvPostProc* pCtrl, int order);

EDTAPI const PdvPostProc* pdv_lookup_postproc(int interlace, int src_depth, int depth);

/* Loads external process library */

/**
 * @brief Unload post-processing library calls previously loaded with
 * @ref pdv_load_postproc_module()
 *
 * @param pCtrl
 *   The post-processing handle to unload.
 *   `dlclose()` (Linux) / `FreeLibrary()` (Windows) will be called on the
 *    PdvPostProc.dll_handle member.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_unload_postproc_module(PdvPostProc* pCtrl);

EDTAPI int pdv_load_postproc_module(PdvPostProc* pCtrl,
                                    const char* name,
                                    int srcdepth,
                                    int destdepth);

EDTAPI void pdv_set_full_bayer_parameters(int nSourceDepth,
                                          double scale[3],
                                          double gamma,
                                          int nBlackOffset,
                                          int bRedRowFirst,
                                          int bGreenPixelFirst,
                                          int quality,
                                          int bias,
                                          int gradientcolor);

EDTAPI int pdv_process_inplace(PdvDev pdv_p);

/**
 * @brief Check if de-interleaving post-processor has been initialized.
 *
 * @param pdv_p
 *   The open PDV device handle.
 *
 * @return 1 if true, 0 if false or error.
 */
EDTAPI int pdv_has_deinterleaver(PdvDev pdv_p);

/**
 * @brief Run de-interleave post-processing.
 *
 * @param pdv_p
 *   The open PDV device handle.
 * @param dd_p
 *   Frame grabber/camera settings. Can be NULL to use the internal PdvDependent
 *   struct from @p pdv_p.
 * @param dmabuf
 *   The frame buffer to process.
 * @param output_buf
 *   The destination buffer to store the post-processed data.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_deinterlace(PdvDev pdv_p, PdvDependent dd_p, uint8_t* dmabuf, uint8_t* output_buf);

#ifdef __cplusplus
}
#endif

#endif
