/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * @details
 * Configure or parse the IRIG time code information which may be added by the
 * frame grabber to a frame buffer's "footer" region.
 */
#ifndef PDV_IRIG_H
#define PDV_IRIG_H

#include "libpdv.h"
#include "libedt_timing.h"

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/** IRIG time code data is in the Unix format. */
#define IRIG2_TYPE_UNIX 3
/** IRIG time code data is in the "raw" format. */
#define IRIG2_TYPE_RAW 5

/** Expected value of @ref irig2_record::magic. Value is ASCII: 'E' 'D' 'T' 01 */
#define IRIG2_MAGIC 0x01544445

/**
 * Holds the IRIG info captured at the rising edge of frame valid.
 */
#pragma pack(push, 1)
typedef struct irig2_record {
  /** This should be @ref IRIG2_MAGIC. */
  uint32_t magic;
  /** This is the frame counter, which is reset when reset_intfc bit is cleared,
   * normally at the start of acquisition. */
  uint32_t framecnt;

  /**
   * There are two modes - seconds from 1970 (Unix time) or raw mode,
   * in which the raw values from IRIG are packed into 32 bits in the raw structure.
   */
  union {
    uint32_t seconds;
    ts_raw_t raw;
  } t;

  /** How many 40 MHz ticks in last second */
  uint32_t clocks;
  /** 40 MHz ticks since last second */
  uint32_t tickspps;

  /** Status bits. */
  struct {
    /** Footer type. Either @ref IRIG2_TYPE_UNIX or @ref IRIG2_TYPE_RAW */
    uint8_t type : 4;
    /** Has valid IRIG data */
    uint8_t irig_ok : 1;
    /** Is synched with PPS */
    uint8_t pps_ok : 1;
    /** Has seen IRIG error */
    uint8_t had_irig_error : 1;
    /** Has seen PPS error */
    uint8_t had_pps_error : 1;
  } status;

  uint8_t reserved_[3];

  /**
   * Monotonic Unix time with fractional seconds.
   * This value is computed by @ref pdv_irig_process_time() after DMA.
   * Fractional time is the 40 MHz count divided by 40 MHz max count.
   */
  double timestamp;

} Irig2Record;

#pragma pack(pop)

/**
 * @brief Reset the DMA frame counter in the frame grabber FPGA.
 *
 * The frame count value is written to @ref irig2_record::framecnt by the FPGA.
 *
 * @param pdv_p
 *   The open EDT device handle.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_reset_dma_framecount(PdvDev pdv_p);

/**
 * @brief Set the IRIG time code format to either "raw" or "Unix".
 *
 * @param pdv_p
 *   The open EDT device handle.
 * @param enable_raw
 *   @arg True (1) -- Use raw time code format.
 *   @arg False (0) -- Use Unix time code format.
 *
 * @return 0 on success.
 *
 * @see @ref edt_set_timecode_raw()
 */
EDTAPI int pdv_irig_set_raw(PdvDev pdv_p, int enable_raw);

/**
 * @brief Extract the IRIG time code data from an image footer.
 *
 * @ref pdv_get_header_offset() is used to determine the footer location.
 *
 * @param pdv_p
 *   The open EDT device handle.
 * @param imagedata
 *   The image frame buffer to inspect.
 *
 * @return The IRIG time code structure, or NULL on error.
 */
EDTAPI Irig2Record* pdv_irig_get_footer(PdvDev pdv_p, const uint8_t* imagedata);

/**
 * @brief Compute the monotonic Unix time and save it to the
 * @ref irig2_record::timestamp field.
 *
 * @param[in,out] footer
 *   Footer structure from an image frame. See @ref pdv_irig_get_footer().
 *
 * @return The computed timestamp value.
 */
EDTAPI double pdv_irig_process_time(Irig2Record* footer);

/**
 * @brief Reset IRIG time code decoding error flags.
 *
 * @param pdv_p
 *   The open EDT device handle.
 */
EDTAPI void pdv_irig_reset_errors(PdvDev pdv_p);

/**
 * @brief Set a positive time code offset to account for (1 PPS sync'd)
 * acquisition latencies.
 *
 * This is described in more detail by the @ref edt_set_timecode_seconds_offset()
 * function.
 *
 * @param pdv_p
 *   The open EDT device handle.
 * @param offset_sec
 *   Offset, in seconds, to add to the time code.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_irig_set_offset(PdvDev pdv_p, int offset_sec);

/**
 * @brief Read IRIG time code information directly from the FPGA.
 *
 * Reads @c seconds, @c clocks, @c tickpps and @c status data from the frame
 * grabber FPGA and stores them in the corresponding @ref irig2_record struct
 * fields.
 *
 * @param pdv_p
 *   The open EDT device handle.
 * @param[out] footer
 *   Time code information read from the FPGA will be returned here.
 *
 * @return 0 on success.
 */
EDTAPI int pdv_irig_get_current(PdvDev pdv_p, Irig2Record* footer);

#ifdef __cplusplus
}
#endif

#endif /* PDV_IRIG_H */
