#!/bin/bash
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (C) 2022 Engineering Design Team, Inc.

PATH="/sbin:$PATH"

#prefer to use TARGETDIR as set by installer (installp* scripts)
TARGETDIR=${TARGETDIR:-/opt/EDTpdv}
LOGFILE=$TARGETDIR/install.log


# shellcheck disable=SC2317
trap_handler () {
    errorCode=$? # save the exit code as the first thing done in the trap function
    echo "ERROR $errorCode"
    echo "  the command executing at the time of the error was"
    echo "  \"$BASH_COMMAND\""
    echo "  on line ${BASH_LINENO[0]}"

    echo ""
    echo '================[ !!! ]================'
    echo "ERROR encountered while running setup."
    echo "Contact your system adminstrator or EDT at <tech@edt.com>."
    echo "Include this log file ($LOGFILE) in any correspondence about this issue."
    echo "Setup aborted."
    echo '================[ !!! ]================'
    echo ""

    exit "$errorCode"
}

fn_setup_init ()
{
    echo "========= System Details ========="
    echo "Date: \"$(date --iso-8601=sec)\""
    (
        test -e /etc/os-release && os_release='/etc/os-release' || os_release='/usr/lib/os-release'
        . "$os_release"
        echo "OS Name: \"$NAME\" Version: \"$VERSION\" Like: \"$ID_LIKE\""
    )
    echo "Kernel: \"$(uname -rvm)\""
    echo "Motherboard Manufacturer: \"$(dmidecode -s baseboard-manufacturer | tr -d '\n')\" Name: \"$(dmidecode -s baseboard-product-name | tr -d '\n')\""
    echo "Processor Version: \"$(dmidecode -s processor-version | tr -d '\n')\""
    echo ""
    echo "EDT PCI devices:"
    lspci -d 123d: -nn
    echo "=================================="
    echo ""

    echo "$TARGETDIR/setup.sh"
    echo "Package version: $(cat "$TARGETDIR"/VERSION)"

    cd "$TARGETDIR"

    return 0
}


{
    set -e
    set -E
    # set -x

    trap trap_handler ERR

    if [ "$EUID" -ne "0" ]
        then echo "Please run as root"
        exit
    fi

    fn_setup_init

    kernel_release="$(uname -r)"

    # DKMS build & install
    dkms add "$TARGETDIR"/driver

    dkms_build_failed=0
    dkms build -m "edt_pci_pdv" -v "6.2.0" -k "$kernel_release" || dkms_build_failed=1

    # Copy DKMS make log(s)
    find /var/lib/dkms/edt_pci_pdv/6.2.0 -name 'make.log' -exec cp -v "{}" "$TARGETDIR"/install_dkms_make.log --backup=numbered \;

    if [ "$dkms_build_failed" -ne "0" ] ; then

        echo "ERROR: DKMS build failed. Kernel module not installed."

        kernel_build_dir="/lib/modules/${kernel_release}/build"

        if [ ! -d "$kernel_build_dir" ]; then
            echo ""
            echo "ERROR: Missing packages for compiling kernel modules."
            echo ""
            echo "Install kernel build packages for your system."
            echo "* CentOS / RHEL / AlmaLinux:"
            echo "  * kernel-devel"
            echo "    - e.g. \"kernel-devel-${kernel_release}\""
            echo ""
            echo "* Debian / Ubuntu:"
            echo "  * linux-headers"
            echo "    - e.g. \"linux-headers-${kernel_release}\""
            echo ""
            echo "You are currently running Linux: "
            echo "  * release: $(uname -r)"
            echo "  * version: $(uname -v)"
            echo "  * machine: $(uname -m)"
            echo ""
        fi

        echo "Re-run this script (${BASH_SOURCE[0]}) to retry building & installing the kernel driver module."
        exit 0
    fi

    # Install built module onto the kernel.
    dkms install --verbose -m "edt_pci_pdv" -v "6.2.0" -k "$kernel_release"
    # Load kernel module now.
    modprobe --verbose "edt_pci_pdv"
    # Load kernel module at boot time.
    # See "modules-load.d(5)" https://man7.org/linux/man-pages/man5/modules-load.d.5.html
    # This config file may not be necessary as the driver does export 'aliases'
    # so udev tooling should be able to detect that EDT's PCI devices are
    # supported by this driver and load it automatically. Leaving this here
    # until we understand it better.
    mkdir -p /usr/lib/modules-load.d
    echo "edt_pci_pdv" > /usr/lib/modules-load.d/"edt_pci_pdv".conf


    # TODO: Is this still necessary when "edtinit" service gets removed? (pci_driver#88)
    # If selinux in installed, add an selinux policy for driver.
    if command -v selinuxenabled &> /dev/null && selinuxenabled; then
        echo "Adding SELinux rule for driver..."
        echo ""
        sestatus || true
        if ! command -v audit2allow &> /dev/null; then
            echo "ERROR: Need 'audit2allow' to apply SELinux rule."
            echo "HINT: Try installing 'policycoreutils-python' OR 'policycoreutils-python-utils', depending on your platform."
            exit 1
        fi
        sleep 5
        ausearch -c 'modprobe' --raw --input-logs | audit2allow -M edt_pci_pdv-modprobe 1>/dev/null
        if [[ -f edt_pci_pdv-modprobe.pp ]]; then
            semodule --priority 300 --install edt_pci_pdv-modprobe.pp
        fi
    fi

    echo -n "Set environment variables... "
    if [ -d /etc/ld.so.conf.d ]; then
        echo "export EDT_PDV_DIR=\"${TARGETDIR}\"" > /etc/profile.d/edtpdv.sh
        chmod +x /etc/profile.d/edtpdv.sh

        echo "/etc/profile.d/edtpdv.sh"
    else
        echo "SKIPPED"
    fi

    echo -n "Add dynamic library references... "
    if [ -d /etc/ld.so.conf.d ]; then
        echo "${TARGETDIR}" >> /etc/ld.so.conf.d/EDTpdv-64bit.conf
        echo "done"
    else
        echo "SKIPPED"
    fi

    ldconfig


    echo "Setup complete"

} 2>&1 | tee --append "$LOGFILE"

exit $?
